#include "config.h"

#include <stdio.h>
#include <xgislib/coords.h>
#include <osg/ShapeDrawable>
#include <osg/Geode>
#include <osg/Group>
#include <osg/Geometry>
#include <osg/PolygonMode>
#include <osg/LineWidth>
#include <osg/Point>
#include <osgUtil/Tessellator>
#include <osgUtil/SmoothingVisitor>
#include <osgUtil/Optimizer>

#include <xgislib/relativealtitudegeometry.h>

#include <string.h>

#include "kmlloader.h"
/*
 *
 *  TODO:
 *			- Fix flattenGroup, add at end of ReadFile implementation
 *			- Implement a different "parseBoundaryIs" function for InnerBoundaryIs (Array of Arrays?)
 *			- Implement tessellation for Polygon, which can have several innerBoundaries
 *          - Solve bad possitioned StyleMaps
 *
 *  DONE:
 *			+ See where xmlTextReaderLocalName and where xmlTextReaderConstName are freed (with/without copy and need to free)
 *			+ Implement altutideMode as Group (absolute), or Transformation (clampToGround, relativeToGround)
 *			+ Implement tessellation for Polygon, which can be concave
 *			+ Implement 'Extrude' of polygons
 *          + add Name to, well, everything that has it.
 *
 *			WARNING: 'similar' colours mught be merged into one when overlapping polygons are read. This is a bug in OSG?
 *
*/

// Start tag
#define TAG_KML				"kml"

// Containers
#define TAG_DOCUMENT		"Document"
#define TAG_FOLDER			"Folder"
#define TAG_GEOMETRYCOLLECTION		"GeometryCollection"		/// Deprecated
#define TAG_MULTIGEOMETRY	"MultiGeometry"

// Features
#define TAG_PLACEMARK		"Placemark"

// Styles
#define TAG_STYLE				"Style"
#define TAG_STYLEMAP			"StyleMap"
#define TAG_STYLEMAPPAIR		"pair"
#define TAG_STYLEKEY			"key"
#define TAG_KEY_NORMAL			"normal"
#define TAG_STYLEURL			"styleUrl"      // Reference to a Style definition

// Geometry
#define TAG_POINT				"Point"
#define TAG_LINESTRING			"LineString"
#define TAG_POLYGON				"Polygon"
//#define TAG_MODEL				"Model"

//Elements for Point


//Elements for Polygon
#define TAG_OUTERBOUNDARYIS		"outerBoundaryIs"
#define TAG_INNERBOUNDARYIS		"innerBoundaryIs"
#define TAG_LINEARRING			"LinearRing"


// Values
#define TAG_NAME				"name"			// text
#define TAG_DESCRIPTION			"description"	// text
#define TAG_COORDINATES			"coordinates"	// lat,long(,alt)
#define TAG_EXTRUDE				"extrude"		// Boolean
#define TAG_TESSELLATE			"tessellate"	// Boolean
#define TAG_ALTITUDEMODE		"altitudeMode"	// clampToGround / relativeToGround / absolute
#define TAG_VISIBILITY			"visibility"    // Object is visible or not

// AltitudeModes
#define TAG_ALTITUDE_CAMPLED	"clampToGround"
#define TAG_ALTITUDE_RELATIVE	"relativeToGround"
#define TAG_ALTITUDE_ABSOLUTE	"absolute"

// Style Components
#define TAG_LINE_STYLE			"LineStyle"
#define TAG_POLY_STYLE			"PolyStyle"
#define TAG_LABEL_STYLE			"LabelStyle"
#define TAG_STYLE_COLOR			"color"
#define TAG_STYLE_WIDTH			"width"

/*
 * Auxuliary "safe" string functions
 */

#define MAX_STRING_VALUE 32766

const xmlChar * strnchr(const xmlChar* str, int car, size_t max_lengh=MAX_STRING_VALUE){
	if (str==NULL)
		return NULL;
	while( max_lengh>0 && *str!=car && *str!='\0' )
	{
			max_lengh--;
			str++;
	}
	return *str==car ? str : (const xmlChar *) NULL;
}

const xmlChar * skip_blanks(const xmlChar* str, size_t max_lengh=MAX_STRING_VALUE){
	if (str==NULL)
		return NULL;
	while( max_lengh>0 && (*str==' ' || *str=='\r' || *str=='\n' || *str=='\t') )
	{
			max_lengh--;
			str++;
	}
	return *str=='\0' ? NULL : str;
}

#define _Hexa_Val(h) ( ((h)>='0' && (h)<='9') ? ((h)-'0') :  ( (h>='A' && h<='F') ? (h-'A'+10) : ( (h>='a' && h<='f') ? (h-'a'+10) : 0) )  )
float parseHexa(const xmlChar* s)
{
	return (_Hexa_Val(s[0])*16.0 + _Hexa_Val(s[1]))/255.0f;
}

/*
 *
 *  KMLLoader class
 *
 */

/*
 * Parses a file and creates the Node (Group) containing the parsed objects,
 * and returns them inside a group.
 */
osg::Group* KMLLoader::readFile(const char* filename){
//	xmlTextReaderPtr reader;
	const xmlChar* tag;
	osg::ref_ptr<osg::Group> g;

	//printf("\n\nRead file %s\n", filename);
	xmlTextReaderPtr reader = xmlReaderForFile(filename, NULL, 0);

    if (reader == NULL)
		return NULL;

	KMLParsingContext* ctx = new KMLParsingContext(reader);
	if (ctx == NULL)
		return NULL;

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		// If kml tag opening
		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT  &&
			!strnicmp(TAG_KML, (char*) tag, sizeof(TAG_KML) ) )
		{
			g=parseGroup(ctx, tag);
			break;
		}
		xmlFree((void*)tag);
        ret = xmlTextReaderRead(ctx->reader);
    }

    xmlFreeTextReader(ctx->reader);
	emptyStyleMap(ctx);
	delete ctx;
	//printf("Read file %s, object address: %X\n", filename, g);

	return g.release();
	//return flattenGroup(g.get());
}

void KMLLoader::emptyStyleMap(KMLParsingContext* ctx)
{
/*	for(	std::map<const char*, KMLStyle, ltstr>::const_iterator it = ctx->styles.begin();
			it != ctx->styles.end() ; ++it)
	{
		printf("Style %s\n", it->first);
	}*/
	ctx->styles.clear();
}

/*
 * Parses an XML sub-tree containing an abstract group (container)
 * such as KML-root, Folder and Document, and returns it as a group
 */
osg::Group* KMLLoader::parseGroup(KMLParsingContext* ctx, const xmlChar* ParentTag){
	const xmlChar* tag;
    const xmlChar* name=NULL;
	osg::ref_ptr<osg::Group> g = new osg::Group();

    //printf("ParseGroup()\n");
    if (g == NULL || ctx->reader == NULL)
		return NULL;

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{

		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_NAME, (char*) tag, sizeof(TAG_NAME) ) )
			{
				name = getTextData(ctx);
				g->setName((char*) name);
				//printf("\nparseGroup()::name: %s\n", name);

			} else if(         !strnicmp(TAG_PLACEMARK, (char*) tag, sizeof(TAG_PLACEMARK) ) )
			{
				osg::Group* newGroup = parsePlacemark(ctx, tag);
				g->addChild(newGroup);
				// break; // there can be several placemarks in one Group (Folder/Document)
			} else if(  !strnicmp(TAG_FOLDER, (char*) tag, sizeof(TAG_FOLDER) ) ||
						!strnicmp(TAG_DOCUMENT, (char*) tag, sizeof(TAG_DOCUMENT) ) )
			{
				osg::Group* newGroup = parseGroup(ctx, tag);
				g->addChild(newGroup);

			} else if(	!strnicmp(TAG_STYLE, (char*) tag, sizeof(TAG_STYLE) ) )
			{
				parseStyle(ctx, tag);
			} else if(	!strnicmp(TAG_STYLEMAP, (char*) tag, sizeof(TAG_STYLEMAP) ) )
			{
				parseStyleMap(ctx, tag);
			} else
			{
				skipTag(ctx, tag);
			}
		}

		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

	return g.release();
}

/*
 * Parses a Placemark, and returns it as a group
 */
osg::Group* KMLLoader::parsePlacemark(KMLParsingContext* ctx, const xmlChar* ParentTag){
	const xmlChar* tag=NULL;
	const xmlChar* name=NULL;
	const xmlChar* description=NULL;
	const xmlChar* styleUrl=NULL;
	KMLStyle* localStyle = new KMLStyle();
	const KMLStyle* oldStyle=ctx->currentStyle;

    //printf("ParsePlaceMark()\n");
	osg::ref_ptr<osg::Group> g = new osg::Group();

    if (g == NULL || ctx->reader == NULL)
		return NULL;

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_NAME, (char*) tag, sizeof(TAG_NAME) ) )
			{
				name = getTextData(ctx);
				g->setName((char*) name);
				//printf("\nparsePlacemark()::name: %s\n", name);

			} else if(  !strnicmp(TAG_DESCRIPTION, (char*) tag, sizeof(TAG_DESCRIPTION) ) )
			{
				description = getTextData(ctx);
//				printf("\ndescription: %s", description);

			} else if(  !strnicmp(TAG_VISIBILITY, (char*) tag, sizeof(TAG_VISIBILITY) ) )
			{
				skipTag(ctx, ParentTag);  // If not visible, then ignore it completelly
				break;
			} else if(  !strnicmp(TAG_STYLE, (char*) tag, sizeof(TAG_STYLE) ) )
			{
			    //printf("parsePlacemark:<style>\n");
				modifyStyle(ctx, tag, localStyle);
				ctx->currentStyle = localStyle;
                   //skipTag(ctx, tag);
			}
            else if(  !strnicmp(TAG_STYLEURL, (char*) tag, sizeof(TAG_STYLEURL) ) )
			{
				if(styleUrl!=NULL)
					xmlFree((void*)styleUrl);
				styleUrl= getTextData(ctx);
				if(styleUrl!=NULL)
					styleUrl = xmlStrdup(styleUrl);
				if(styleUrl!=NULL && styleUrl[1]!='\0')
				{
//					printf("parseStyleMap::Found Style %s == %s\n", id, &(styleUrl[1]));
					std::map<const char*, KMLStyle, ltstr>::const_iterator it = ctx->styles.find((char*) &(styleUrl[1]));
					if ( it != ctx->styles.end( ) )
					{
//						printf("Placemark:currentStyle=%s\n", it->first);
						ctx->currentStyle = &(it->second);
						xmlFree((void*)styleUrl);
						styleUrl=NULL;
					}
				}

			} else if(  !strnicmp(TAG_POINT, (char*) tag, sizeof(TAG_POINT) ) )
			{
				osg::Group* newGroup = parsePoint(ctx, tag);
				g->addChild(newGroup);
			} else if(  !strnicmp(TAG_POLYGON, (char*) tag, sizeof(TAG_POLYGON) ) )
			{
				osg::Group* newGroup = parsePolygon(ctx, tag);
				g->addChild(newGroup);
			} else if(  !strnicmp(TAG_LINESTRING, (char*) tag, sizeof(TAG_LINESTRING) ) )
			{
				osg::Group* newGroup = parseLineString(ctx, tag);
				g->addChild(newGroup);
			} else if(  !strnicmp(TAG_GEOMETRYCOLLECTION, (char*) tag, sizeof(TAG_GEOMETRYCOLLECTION) ) ||  //Deprecated
						!strnicmp(TAG_MULTIGEOMETRY, (char*) tag, sizeof(TAG_MULTIGEOMETRY) )   )
			{
				osg::Group* newGroup = parsePlacemark(ctx, tag);
				g->addChild(newGroup);
			} else
			{
				skipTag(ctx, tag);
			}
		}
		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

    ctx->currentStyle = oldStyle;
    delete localStyle;
	return g.release();
}

/*
 * Parses a Point, and returns it as a group
 */
osg::Group* KMLLoader::parsePoint(KMLParsingContext* ctx, const xmlChar* ParentTag){
	const xmlChar* tag;
	const xmlChar* coordinates=NULL;
	const xmlChar* data;
	AltitudeMode altitude=ALTITUDE_ABSOLUTE;
	bool extrude=false;
	osg::ref_ptr<osg::Group> g = new osg::Group();

    if (g == NULL || ctx->reader == NULL)
		return NULL;

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		// TODO: ADD <altitudeMode> and <extrude>
		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_COORDINATES, (char*) tag, sizeof(TAG_COORDINATES) ) )
			{
				coordinates = getTextData(ctx);
				osg::ref_ptr<osg::Vec3Array> vertices = parseCoordinates(coordinates, altitude);

				if(vertices->size() > 0 ){
					osg::ref_ptr<osg::Geode> model=new osg::Geode;
					osg::ref_ptr<osg::Sphere> sphere = new osg::Sphere(vertices->operator[](0), 5.0);
					osg::ref_ptr<osg::ShapeDrawable> drawable = new osg::ShapeDrawable(sphere.get());

					if(ctx->currentStyle!=NULL)
					{
						drawable->setColor(ctx->currentStyle->getLabelColour());
						if(ctx->currentStyle->getLabelColour().a() < 1.0)
						{
							osg::StateSet* stateset = drawable->getOrCreateStateSet();
							stateset->setMode(GL_BLEND, osg::StateAttribute::ON);
							stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
						}
					}
					else
					{
						drawable->setColor(osg::Vec4(1.0f, .25f, .25f, 1.0f));
					}

					model->addDrawable(drawable.get());

					if(extrude && ctx->currentStyle!=NULL)
					{
						ExtrudeVertices(model.get(), vertices.get(),
							ctx->currentStyle->getPolyColour(),
							ctx->currentStyle->getLineColour(),
							altitude, EXTRUDE_POINT);
					}

					g->addChild(model.get());
				}
			} else if(  !strnicmp(TAG_ALTITUDEMODE, (char*) tag, sizeof(TAG_ALTITUDEMODE) ) )
			{
				data = getTextData(ctx);
				altitude = parseAltitudeMode(data);
//				printf("\naltitudeMode: %s", altitudeMode);
			} else if(  !strnicmp(TAG_EXTRUDE, (char*) tag, sizeof(TAG_EXTRUDE) ) )
			{
				data = getTextData(ctx);
				if(!strnicmp((const char*)data, "1", 1))
				{
					extrude = true;
				}
//				printf("\naltitudeMode: %s", altitudeMode);
			} else
			{
					skipTag(ctx, tag);
			}
		}
		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}
	return g.release();
}

/*
 * Parses a LineString (path), probably inside a Placemark, and returns it as a group
 */
osg::Group* KMLLoader::parseLineString(KMLParsingContext* ctx, const xmlChar* ParentTag){
	const xmlChar* tag;
	const xmlChar* coordinates;
	const xmlChar* data;
	AltitudeMode altitude=ALTITUDE_ABSOLUTE;
	bool extrude=false;
	osg::ref_ptr<osg::Group> g = new osg::Group();

    if (g == NULL || ctx->reader == NULL)
		return NULL;

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{

		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		//printf("\n%s::%s", ParentTag, tag);
		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_COORDINATES, (char*) tag, sizeof(TAG_COORDINATES) ) )
			{
				coordinates = getTextData(ctx);

				osg::ref_ptr<osg::Vec3Array> vertices = parseCoordinates(coordinates, altitude);

				if(vertices->size() > 0 ){
					osg::ref_ptr<osg::Geode> model=new osg::Geode;
					osg::ref_ptr<RelativeAltitudeGeometry> geom = new RelativeAltitudeGeometry();


					geom->setVertexArray(vertices.get());
					osg::ref_ptr<osg::Vec4Array> colours = new osg::Vec4Array;

					if(ctx->currentStyle!=NULL)
					{
						colours->push_back(ctx->currentStyle->getLineColour());
//						printf("parseLineString:LineColour:%f, %f, %f, %f\n",
//							ctx->currentStyle->getLineColour().r(), ctx->currentStyle->getLineColour().g(),
//							ctx->currentStyle->getLineColour().b(), ctx->currentStyle->getLineColour().a());
						if(ctx->currentStyle->getLineColour().a() < 1.0)
						{
							osg::StateSet* stateset = geom->getOrCreateStateSet();
							stateset->setMode(GL_BLEND, osg::StateAttribute::ON);
							stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
						}
					}
					else
						colours->push_back(osg::Vec4(0.75, 1.0f, 0.5f , 1.0f));

					geom->setColorArray(colours.get());
					geom->setColorBinding(osg::Geometry::BIND_OVERALL);
					geom->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::LINE_STRIP, 0, vertices->size()));

					osg::StateSet *state = geom->getOrCreateStateSet();
					if(ctx->currentStyle!=NULL)
					{
						state->setAttribute(new osg::LineWidth(ctx->currentStyle->getWidth()), osg::StateAttribute::ON);
					}
					else
						state->setAttribute(new osg::LineWidth(1.0f), osg::StateAttribute::ON);
					//state->setAttribute(new osg::LineWidth(1.0f), osg::StateAttribute::ON);

					osg::PolygonMode *polymode = new osg::PolygonMode;
					polymode->setMode(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::LINE);
					state->setAttributeAndModes(polymode, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);

			/*		osg::Point *point = new osg::Point;
					point->setSize(4.0f);
					state->setAttribute(point, osg::StateAttribute::ON);

					state->setMode(GL_LIGHTING, osg::StateAttribute::OFF);*/
					if(altitude!=ALTITUDE_ABSOLUTE)
					{
						geom->saveOriginalState(worldReference);
						geom->recalculateHeights();
						//model->setUpdateCallback(new RelativeAltitudeGeometryCB());
					}
					model->addDrawable(geom.get());


					if(extrude)
					{
						if(ctx->currentStyle != NULL)
							ExtrudeVertices(model.get(), vertices.get(),
								ctx->currentStyle->getPolyColour(),
								colours->operator[](0),
								altitude, EXTRUDE_LINE);
						else
							ExtrudeVertices(model.get(), vertices.get(),
								colours->operator[](0),
								colours->operator[](0),
								altitude, EXTRUDE_LINE);
					}

					g->addChild(model.get());

				}
			} else if(  !strnicmp(TAG_TESSELLATE, (char*) tag, sizeof(TAG_TESSELLATE) ) )
			{
				data = getTextData(ctx);
				if(!strnicmp((const char*)data, "1", 1))
					altitude = ALTITUDE_CLAMPED;
//				printf("\naltitudeMode: %s", altitudeMode);
			} else if(  !strnicmp(TAG_ALTITUDEMODE, (char*) tag, sizeof(TAG_ALTITUDEMODE) ) )
			{
				data = getTextData(ctx);
				altitude = parseAltitudeMode(data);
//				printf("\naltitudeMode: %s", altitudeMode);
			} else if(  !strnicmp(TAG_EXTRUDE, (char*) tag, sizeof(TAG_EXTRUDE) ) )
			{
				data = getTextData(ctx);
				if(!strnicmp((const char*)data, "1", 1))
				{
					extrude = true;
				}
//				printf("\naltitudeMode: %s", altitudeMode);
			}  else
			{
					skipTag(ctx, tag);
			}
		}

		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

	return g.release();
}

/*
 * Parses a Polygon, probably inside a Placemark, and returns it as a group
 */
osg::Group* KMLLoader::parsePolygon(KMLParsingContext* ctx, const xmlChar* ParentTag){
	const xmlChar* tag;
	const xmlChar* data;
	//const xmlChar* altitudeMode;
	AltitudeMode altitude=ALTITUDE_ABSOLUTE;
	bool extrude=false;
	osg::ref_ptr<osg::Group> g = new osg::Group();

    if (g == NULL || ctx->reader == NULL)
		return NULL;


    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{

		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";
		//printf("\n%s::%s", ParentTag, tag);
		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_OUTERBOUNDARYIS, (char*) tag, sizeof(TAG_OUTERBOUNDARYIS) ) )
			{
				osg::ref_ptr<osg::Vec3Array> vertices = parseBoundaryIs(ctx, tag, altitude, false); // single

				if(vertices->size() > 2 ){
					osg::ref_ptr<osg::Geode> model=new osg::Geode;
					osg::ref_ptr<osg::Geometry> geom = new RelativeAltitudeGeometry();
					//osg::ref_ptr<osg::Vec3Array> norms = new osg::Vec3Array();
					geom->setVertexArray(vertices.get());

					osg::ref_ptr<osg::Vec4Array> colours = new osg::Vec4Array;

					if(ctx->currentStyle!=NULL)
					{
						colours->push_back(ctx->currentStyle->getPolyColour());
/*						printf("parsePolygon:PolyColour:%f, %f, %f, %f\n",
							ctx->currentStyle->getPolyColour().r(), ctx->currentStyle->getPolyColour().g(),
							ctx->currentStyle->getPolyColour().b(), ctx->currentStyle->getPolyColour().a());
*/
						osg::StateSet* stateset = geom->getOrCreateStateSet();
						//stateset->setMode(GL_LIGHTING, osg::StateAttribute::OFF);
						if(ctx->currentStyle->getPolyColour().a() < 1.0)
						{
						//	printf("parsePolygon:PolyColour:TransparentBin\n");
							stateset->setMode(GL_BLEND, osg::StateAttribute::ON);
							stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
						}
					}
					else
					{
						colours->push_back(osg::Vec4(0.5f, 0.5f, 1.0f , 1.0f));
						printf("parsePolygon:PolyColour:using Default colour\n");
					}

					geom->setColorArray(colours.get());
					geom->setColorBinding(osg::Geometry::BIND_OVERALL);
					geom->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::POLYGON, 0, vertices->size()));

					model->addDrawable(geom.get());

					if(extrude)
					{
						if(ctx->currentStyle != NULL)
							ExtrudeVertices(model.get(), vertices.get(),
								colours->operator[](0),
								ctx->currentStyle->getLineColour(),
								altitude, EXTRUDE_POLYGON);
						else
							ExtrudeVertices(model.get(), vertices.get(),
								colours->operator[](0),
								colours->operator[](0),
								altitude, EXTRUDE_POLYGON);
					}


					osgUtil::Optimizer::TessellateVisitor tv;
					tv.apply(*model.get());
					osgUtil::SmoothingVisitor sm;
					sm.apply(*model.get());

					if(altitude!=ALTITUDE_ABSOLUTE)
					{
						((RelativeAltitudeGeometry*)geom.get())->saveOriginalState(worldReference);
						((RelativeAltitudeGeometry*)geom.get())->recalculateHeights();
						//model->setUpdateCallback(new RelativeAltitudeGeometryCB());
					}

					g->addChild(model.get());

/*
					// tessellation code: the number of vertexes will vary.
					osg::ref_ptr<osgUtil::Tessellator> tscx=new osgUtil::Tessellator;
 					tscx->setTessellationType(osgUtil::Tessellator::TESS_TYPE_GEOMETRY);
 					tscx->setBoundaryOnly(false);
 					tscx->setWindingType( osgUtil::Tessellator::TESS_WINDING_ODD);
 					tscx->retessellatePolygons(*(geom.get())); */
				}


			}
/*				else if(  !strnicmp(TAG_INNERBOUNDARYIS, (char*) tag, sizeof(TAG_INNERBOUNDARYIS) ) )
			{
				InnerLinearRingsFound=parseBoundaryIs(tag, true);  // multiple
			}
*/
			else if(  !strnicmp(TAG_TESSELLATE, (char*) tag, sizeof(TAG_TESSELLATE) ) )
			{
				data = getTextData(ctx);
				if(!strnicmp((const char*)data, "1", 1))
					altitude = ALTITUDE_CLAMPED;
//				printf("\naltitudeMode: %s", altitudeMode);
			} else if(  !strnicmp(TAG_ALTITUDEMODE, (char*) tag, sizeof(TAG_ALTITUDEMODE) ) )
			{
				data = getTextData(ctx);
				altitude = parseAltitudeMode(data);
//				printf("\naltitudeMode: %s", altitudeMode);
			} else if(  !strnicmp(TAG_EXTRUDE, (char*) tag, sizeof(TAG_EXTRUDE) ) )
			{
				data = getTextData(ctx);
				if(!strnicmp((const char*)data, "1", 1))
				{
					extrude = true;
				}
//				printf("\naltitudeMode: %s", altitudeMode);
			} else
			{
				skipTag(ctx, tag);
			}
		}

		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

	return g.release();
}


/*
 * Parses a Boundary (either outer or inner) of a Polygon, and returns its vertices
 */
osg::Vec3Array* KMLLoader::parseBoundaryIs(KMLParsingContext* ctx, const xmlChar* ParentTag, AltitudeMode altitude, bool multiple){
	const xmlChar* tag;
	int LinearRingsFound=0;
	osg::Vec3Array *vertices = NULL;

    if (ctx->reader == NULL)
		return 0;

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1 )
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";
		//printf("\n%s::%s", ParentTag, tag);
		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT )
		{
			if(         !strnicmp(TAG_LINEARRING, (char*) tag, sizeof(TAG_LINEARRING) ) )
			{
				if( multiple || LinearRingsFound<1)
				{
					LinearRingsFound++;
					vertices=parseLinearRing(ctx, tag, altitude);  // TODO: fix the hole thing for multiple linear rings.
				} else
				{
					LinearRingsFound++;
					vertices=parseLinearRing(ctx, tag, altitude);
				}
				//break;
			} else
			{
				skipTag(ctx, tag);
			}
		}

		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

	return vertices;
}


/*
 * Parses a LinearRing (closed path), probably as a boundary of a Polygon, and returns its vertices
 */
osg::Vec3Array* KMLLoader::parseLinearRing(KMLParsingContext* ctx, const xmlChar* ParentTag, AltitudeMode altitude){
	const xmlChar* tag;
	const xmlChar* coordinates;
	bool found=false;
	osg::Vec3Array *vertices = NULL;

    if (ctx->reader == NULL)
		return 0;

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{

		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		//printf("\n%s::%s", ParentTag, tag);
		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( !found && xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_COORDINATES, (char*) tag, sizeof(TAG_COORDINATES) ) )
			{
				coordinates = getTextData(ctx);
				//printf("\ncoordinates: %s", coordinates);
				vertices = parseCoordinates(coordinates, altitude);
				found=true;			// Should only be one!!
			}  else
			{
					skipTag(ctx, tag);
			}
		}

		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

	return vertices;
}

void KMLLoader::parseStyleMap(KMLParsingContext* ctx, const xmlChar* ParentTag)
{
	const xmlChar* tag=NULL;
	const xmlChar* id=NULL;
	const xmlChar* styleUrl=NULL;

    if (ctx->reader == NULL)
		return;

	id = xmlTextReaderGetAttribute(ctx->reader, (const xmlChar*)"id");
//	printf("KMLLoader::parseStyleMap::%s\n", id);

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_STYLEMAPPAIR, (char*) tag, sizeof(TAG_STYLEMAPPAIR) ) )
			{
				styleUrl = parseStyleMapPair(ctx, tag);

				if(styleUrl!=NULL && styleUrl[1]!='\0')
				{
//					printf("parseStyleMap::Found Style %s == %s\n", id, &(styleUrl[1]));
					std::map<const char*, KMLStyle, ltstr>::const_iterator it = ctx->styles.find((char*) &(styleUrl[1]));
					if ( it != ctx->styles.end( ) )
					{
//						printf("parseStyleMap::Added Style %s == %s\n", id, &(styleUrl[1]));
						ctx->styles[(char*)id] = it->second;
					}
					xmlFree((void*)styleUrl);
				}
			}else
			{
					skipTag(ctx, tag);
			}
		}
		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}


	return;
}

const xmlChar* KMLLoader::parseStyleMapPair(KMLParsingContext* ctx, const xmlChar* ParentTag)
{
	const xmlChar* tag=NULL;
	const xmlChar* styleUrl=NULL;
	const xmlChar* key=NULL;
	bool normal=false;

    if (ctx->reader == NULL)
		return NULL;

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_STYLEURL, (char*) tag, sizeof(TAG_STYLEURL) ) )
			{
				if(styleUrl!=NULL)
					xmlFree((void*)styleUrl);
				styleUrl= getTextData(ctx);
				if(styleUrl!=NULL)
					styleUrl = xmlStrdup(styleUrl);
				//printf("parseStyleMapPair::found %s\n", styleUrl);
				//printf("parseStyleMapPair::styleurl::key:%s, styleUrl: %s\n", key, styleUrl);
/*				if(styleUrl)
					styleUrl++;*/
			} else if(  !strnicmp(TAG_STYLEKEY, (char*) tag, sizeof(TAG_STYLEKEY) ) )
			{
				key = getTextData(ctx);
				//printf("parseStyleMapPair::key %s\n", key);
				//printf("parseStyleMapPair::stylekey::key:%s, styleUrl: %s\n", key, styleUrl);
				if( !strnicmp(TAG_KEY_NORMAL, (char*)key, sizeof(TAG_KEY_NORMAL)) )
					normal=true;
			} else
			{
					skipTag(ctx, tag);
			}
		}
		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

//	printf("parseStyleMapPair:: styleUrl: %s\n", normal ? styleUrl : NULL);
	return normal ? styleUrl : (const xmlChar* )NULL;
}

void KMLLoader::parseStyle(KMLParsingContext* ctx, const xmlChar* ParentTag)
{
	const xmlChar* tag;
	const xmlChar* id;
	KMLStyle newStyle;

    if (ctx->reader == NULL)
		return;

	id = xmlTextReaderGetAttribute(ctx->reader, (const xmlChar*)"id");
//	printf("KMLLoader::parseStyle::%s\n", id);

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		//printf("\n%s::%s", ParentTag, tag);
		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_POLY_STYLE, (char*) tag, sizeof(TAG_POLY_STYLE) ) )
			{
				osg::Vec4f colour;
				parseColorWidthStyle(ctx, tag, colour);
				newStyle.setPolyColour(colour);
			} else if(	!strnicmp(TAG_LABEL_STYLE, (char*) tag, sizeof(TAG_LABEL_STYLE) ) )
			{
				osg::Vec4f colour;
				parseColorWidthStyle(ctx, tag, colour);
				newStyle.setLabelColour(colour);
			} else if(	!strnicmp(TAG_LINE_STYLE, (char*) tag, sizeof(TAG_LINE_STYLE) ) )
			{
				osg::Vec4f colour;
				float width = parseColorWidthStyle(ctx, tag, colour);
				newStyle.setLineColour(colour);
				newStyle.setWidth(width);
			} else
			{
					skipTag(ctx, tag);
			}
		}

		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

	if (id != NULL && id[0]!='\0')
	{
//		printf("Saving Style %s\n", (const char*)id);
		ctx->styles[(const char*)id] = newStyle;
	}
	return;
}

void KMLLoader::modifyStyle(KMLParsingContext* ctx, const xmlChar* ParentTag, KMLStyle* style)
{
	const xmlChar* tag;

    if (ctx->reader == NULL)
		return;

    //printf("modifyStyle()\n");

    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		//printf("\n%s::%s", ParentTag, tag);
		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_POLY_STYLE, (char*) tag, sizeof(TAG_POLY_STYLE) ) )
			{
				osg::Vec4f colour;
				parseColorWidthStyle(ctx, tag, colour);
				style->setPolyColour(colour);
			} else if(	!strnicmp(TAG_LABEL_STYLE, (char*) tag, sizeof(TAG_LABEL_STYLE) ) )
			{
				osg::Vec4f colour;
				parseColorWidthStyle(ctx, tag, colour);
				style->setLabelColour(colour);
			} else if(	!strnicmp(TAG_LINE_STYLE, (char*) tag, sizeof(TAG_LINE_STYLE) ) )
			{
				osg::Vec4f colour;
				float width = parseColorWidthStyle(ctx, tag, colour);
				style->setLineColour(colour);
				style->setWidth(width);
			} else
			{
					skipTag(ctx, tag);
			}
		}

		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

	return;
}

float KMLLoader::parseColorWidthStyle(KMLParsingContext* ctx, const xmlChar* ParentTag, osg::Vec4f& colour)
{
	const xmlChar* tag;
	const xmlChar* value;
	float width=1.0;
	KMLStyle newStyle;

    if (ctx->reader == NULL)
		return 0;


    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
			tag = BAD_CAST "--";

		//printf("\n%s::%s", ParentTag, tag);
		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		if( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_ELEMENT)
		{
			if(         !strnicmp(TAG_STYLE_COLOR, (char*) tag, sizeof(TAG_STYLE_COLOR) ) )
			{
				value = getTextData(ctx);
				colour.set(parseHexa(&(value[6])), parseHexa(&(value[4])), parseHexa(&(value[2])), parseHexa(&(value[0])));
			} else if(  !strnicmp(TAG_STYLE_WIDTH, (char*) tag, sizeof(TAG_STYLE_WIDTH) ) )
			{
				value = getTextData(ctx);
				width = atof((const char*)value);

			} else
			{
					skipTag(ctx, tag);
			}
		}

		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

	return width;
}

/*
 * Parses and returns the data contained in the current XML node
 */
// WARNING: returns Const Pointer (make copy)
const xmlChar* KMLLoader::getTextData(KMLParsingContext* ctx){
		const xmlChar* data=NULL;
//		printf("\n%d\t%d\t%d\t%s", xmlTextReaderDepth(ctx->reader), xmlTextReaderNodeType(ctx->reader),
//			xmlTextReaderIsEmptyElement(ctx->reader), tag);

		if (ctx->reader == NULL)
			return 0;

		int ret = xmlTextReaderRead(ctx->reader);
		if(ret == 1 &&
			( xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_TEXT ||
			  xmlTextReaderNodeType(ctx->reader) == XML_READER_TYPE_CDATA )  )

		{
			if(xmlTextReaderHasValue(ctx->reader))
			{
				//data= xmlTextReaderConstValue(ctx->reader);
				data= xmlTextReaderConstValue(ctx->reader);
				//printf("\ndata: %s", data);
			}
		}
		return data;
}

/*
 * Skips all the content of a node-tree until after its closing tag
 */
void KMLLoader::skipTag(KMLParsingContext* ctx, const xmlChar* ParentTag){
	const xmlChar* tag;

    if (ctx->reader == NULL)
		return;
//	printf("\nSkipping %s", ParentTag);
    int ret = xmlTextReaderRead(ctx->reader);
    while (ret == 1)
	{
		tag = xmlTextReaderLocalName(ctx->reader);
		if (tag == NULL)
		tag = BAD_CAST "--";

		if( isClosingTag(ParentTag, tag, xmlTextReaderNodeType(ctx->reader) ) )
		{
			xmlFree((void*)tag);
			break;
		}

		xmlFree((void*)tag);
		ret = xmlTextReaderRead(ctx->reader);
	}

}

/*
 * Checks for ending tag property of the current inspected XML object
 */
inline bool KMLLoader::isClosingTag(const xmlChar* ParentName, const xmlChar* name, int type){
	return !stricmp((char*) ParentName, (char*) name) && type == XML_READER_TYPE_END_ELEMENT;
}

/*
 * Parses and returns the verteces defined inside a <coordinates> tag, being it one set or several
 */
osg::Vec3Array *KMLLoader::parseCoordinates(const xmlChar *str, AltitudeMode altitude){
	osg::Vec3Array* p=new osg::Vec3Array();
	double latitude, longitude, height;


	if(str==NULL)
		return p;

	const xmlChar *end =  strnchr(str, '\0');
	const xmlChar *next =  strnchr(str, ',');
	const xmlChar *current = skip_blanks(str);

	while(current < end && current!=NULL)
	{
		longitude = atof( (const char *) current );
		next =  strnchr(current, ',');
		if(current >= next || current==NULL) // error in format; ignore vector?
		{
			current = skip_blanks(current);
			break;
		}
		current = next+1;

		latitude = atof( (const char *) current );

		next =  strnchr(current, ',');
		if(current >= next || current==NULL) // error in format; ignore vector?
		{
			current = skip_blanks(current);
			break;
		}
		current = next+1;

		height = atof( (const char *) current );

		// have all 3, convert Long Lat GaussKr, pack them and save them
		Coords::WGS84 reprojection(longitude, latitude);
		Coords::GaussKr gauss(reprojection.toGaussKr());
		if(altitude == ALTITUDE_CLAMPED)
			p->push_back( osg::Vec3(gauss.x, gauss.y, 0.0f) );
			//p->push_back( osg::Vec3(gauss.x, gauss.y, worldReference->getTerrainHeight(gauss.x, gauss.y) ) );
		else
			p->push_back( osg::Vec3(gauss.x, gauss.y, height) );

		//printf("\n%16.16lf\t%16.16lf\t%16.16lf", longitude, latitude, height);
		//printf("\n%16.16lf\t%16.16lf\t%16.16lf - %s", gauss.x, gauss.y, height, altitude == ALTITUDE_CLAMPED ? "clamped" : "relative" );
		current = strnchr(current, ' ');
		current = skip_blanks(current);
	}
	return p;
}

AltitudeMode KMLLoader::parseAltitudeMode(const xmlChar *str)
{
	AltitudeMode altitude=ALTITUDE_ABSOLUTE;
	if(str==NULL)
		return altitude;
	if( !strncmp((const char*) str, TAG_ALTITUDE_CAMPLED, sizeof(TAG_ALTITUDE_CAMPLED) ) )
		altitude=ALTITUDE_CLAMPED;
	else if(!strncmp((const char*) str, TAG_ALTITUDE_RELATIVE, sizeof(TAG_ALTITUDE_RELATIVE) ) )
		altitude=ALTITUDE_RELATIVE;
//	else, default to Abosuloute

	return altitude;
}

void KMLLoader::ExtrudeVertices(osg::Geode* model,  osg::Vec3Array* vertices,
								const osg::Vec4 &polyColour, const osg::Vec4 &lineColour,
								AltitudeMode altitude, ExtrudeMode extrudeMode){

	int elementCount = vertices->size();
	int lastElement = elementCount-1;
	if(extrudeMode==EXTRUDE_LINE)
		lastElement--;
	//printf("ExtrudeVertices:elementCount=%d, lastElement=%d\n",elementCount, lastElement);
	for(int i=0; i<=lastElement; i++)
	{
		osg::ref_ptr<osg::Geometry> wallGeom = new RelativeAltitudeGeometry();

		osg::Vec3 curr = vertices->operator[](i);
		osg::Vec3 next = vertices->operator[]((i+1)%elementCount);
		osg::Vec3f nextUndergorund(next.x(), next.y(), -300.0f);
		osg::Vec3f currUndergorund(curr.x(), curr.y(), -300.0f);

		osg::ref_ptr<osg::Vec3Array> wallPoints = new osg::Vec3Array();

		wallPoints->push_back(next);
		wallPoints->push_back(curr);
		wallPoints->push_back(currUndergorund);
		wallPoints->push_back(nextUndergorund);

		wallGeom->setVertexArray(wallPoints.get());


		if(polyColour.a() < 1.0)
		{
			osg::StateSet* wallStateset = wallGeom->getOrCreateStateSet();
			wallStateset->setMode(GL_BLEND, osg::StateAttribute::ON);
			wallStateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
		}

		osg::ref_ptr<osg::Vec4Array> wallColour = new osg::Vec4Array;
		wallColour ->push_back(polyColour);

		wallGeom->setColorArray(wallColour.get());
		wallGeom->setColorBinding(osg::Geometry::BIND_OVERALL);
		wallGeom->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::POLYGON, 0, wallPoints->size()));
		if(altitude!=ALTITUDE_ABSOLUTE)
		{
			((RelativeAltitudeGeometry*)wallGeom.get())->saveOriginalState(worldReference);
			((RelativeAltitudeGeometry*)wallGeom.get())->recalculateHeights();
			//model->setUpdateCallback(new RelativeAltitudeGeometryCB());
		}
		model->addDrawable(wallGeom.get());

		osg::ref_ptr<osg::Geometry> linesGeom = new RelativeAltitudeGeometry();
		osg::ref_ptr<osg::Vec3Array> linesPoints = new osg::Vec3Array();
		linesPoints->push_back(curr);
		linesPoints->push_back(currUndergorund);
		if(extrudeMode!=EXTRUDE_POINT)
		{
			linesPoints->push_back(next);
			linesPoints->push_back(curr);
		}
		// If we are not closing the 'circle', then we have to draw the vertical line for the last point
		if(lastElement==i && lastElement < elementCount-1)
		{
			linesPoints->push_back(next);
			linesPoints->push_back(nextUndergorund);
		}

		linesGeom->setVertexArray(linesPoints.get());
		if(lineColour.a() < 1.0)
		{
			osg::StateSet* linesStateset = linesGeom->getOrCreateStateSet();
			linesStateset->setMode(GL_BLEND, osg::StateAttribute::ON);
			linesStateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
		}
		osg::ref_ptr<osg::Vec4Array> linesColour = new osg::Vec4Array;
		linesColour->push_back(lineColour);

		linesGeom->setColorArray(linesColour.get());
		linesGeom->setColorBinding(osg::Geometry::BIND_OVERALL);
		linesGeom->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::LINES, 0, linesPoints->size()));
		if(altitude!=ALTITUDE_ABSOLUTE)
		{
			((RelativeAltitudeGeometry*)linesGeom.get())->saveOriginalState(worldReference);
			((RelativeAltitudeGeometry*)linesGeom.get())->recalculateHeights();
		}
		model->addDrawable(linesGeom.get());
	} // for(int i=0; i<=lastElement; i++)

}

osg::Group* KMLLoader::flattenGroup(osg::Group* sparceGroup)
{
    if(sparceGroup==NULL)
        return NULL;

    osg::Geode* geode = dynamic_cast<osg::Geode*>(sparceGroup);

    if(geode==NULL)
        return sparceGroup;

    if( sparceGroup->getNumChildren() == 0)
    {
        if(geode==NULL)
            return NULL;
        if(geode->getNumDrawables() > 0)
            return sparceGroup;
        else
            return NULL;
    } else if( sparceGroup->getNumChildren() == 1)
    {
		osg::Group*g = dynamic_cast<osg::Group*>(sparceGroup->getChild(0));
		if(g!=NULL)
            return flattenGroup(g);
		else
			return sparceGroup;
    } else{
            return sparceGroup;
    }
	return sparceGroup;
}

/*
const char * strnchr(const char* str, int car, size_t max_lengh){
	while( max_lengh>0 && *str!=car && *str!='\0' )
	{
			max_lengh--;
			str++;
	}

	return *str==car ? str : (const char *) NULL;
}

void printDepth(int depth, const char *str=NULL){
	//putc('\n', stdout);
	while(depth--)
		putc(' ', stdout);
	if(str)
		puts(str);
}

void printGeode(osg::Geode* geod, int depth=0){
	for(unsigned int i=0; i < geod->getNumDrawables() ; i++)
	{
		osg::Drawable* draw = geod->getDrawable(i);
		if( ! stricmp(draw->className(), "Geometry") )
		{
			printDepth(depth, "-Geometry");
			osg::Geometry* geom = (osg::Geometry*) draw;
			osg::Vec3Array* vertices = (osg::Vec3Array*) geom->getVertexArray();
			for(unsigned int j=0; j<vertices->size(); j++)
			{
				printf("<%f, %f, %f>", vertices->operator[](i).x(), vertices->operator[](i).y(), vertices->operator[](i).z());
			}
		}
	}
}

void printGroup(osg::Group *group, int depth=0){
	for(unsigned int i=0; i < group->getNumChildren() ; i++)
	{
		osg::Node *node = group->getChild(i);
		if( ! stricmp(node->className(), "Group") )
		{
			printDepth(depth, "-Group");
			printGroup((osg::Group *)node, depth+1);
		}else if( ! stricmp(node->className(), "Geode") )
		{
			printDepth(depth, "-Geode");
			printGeode( (osg::Geode*) node, depth+1);
		} else
		{
			printDepth(depth);
			printf("%s", node->className() );
		}
	}
}

#include <conio.h>

int main(){
	//xgisKML::KMLLoader reader("tro.kml");
	//xgisKML::KMLLoader reader("tromostovje.kmz.kml");
	//xgisKML::KMLLoader reader("grad.kml");
	//xgisKML::KMLLoader reader("PotNaGrad.kml");

	xgisKML::KMLLoader reader("fe_fri.kml");
	osg::Group *g = reader.asGroup();

	//printGroup(g);

	printf("\n\nPress [ENTER] to continue\n");

//	delete n;

	getch();
}
*/

/*
normals
					int lastElement = vertices->size()-1;
					for(int i=0; i<=lastElement; i++)
					{
						osg::Vec3 prev = vertices->operator[](i%lastElement);
						osg::Vec3 curr = vertices->operator[](i);
						osg::Vec3 next = vertices->operator[]((i+1)%lastElement);

						osg::Vec3 v1 = curr - prev;
						//osg::Vec3 v2 = curr - next;
						osg::Vec3 v2 = next - curr;
						v1.normalize();
						v2.normalize();
						osg::Vec3 normal = osg::Vec3(
							v1.y()*v2.z() - v1.z()*v2.y(),
							v1.z()*v2.x() - v1.x()*v2.z(),
							v1.x()*v2.y() - v1.y()*v2.x()
						);
						norms->push_back(normal);
					}

					geom->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);
					geom->setNormalArray(norms.get());
*/
