#ifndef __XGIS_KML_OSG_KML_H__
#define __XGIS_KML_OSG_KML_H__

#include <osg/Group>
#include <libxml/xmlreader.h>
#include <xgislib/layerbuilder.h>
#include <xgislib/world.h>
#include <map>

typedef enum {ALTITUDE_ABSOLUTE=0, ALTITUDE_RELATIVE, ALTITUDE_CLAMPED} AltitudeMode;
typedef enum {EXTRUDE_POLYGON=0, EXTRUDE_LINE, EXTRUDE_POINT} ExtrudeMode;
class KMLStyle;
struct KMLParsingContext;
typedef std::pair<const char*, KMLStyle*> StylePair;

class KMLLoader: public LayerLoader{

public:
	KMLLoader() { mimeType = "application/vnd.google-earth.kml+xml"; };

	Layer *loadLayer(const std::string &file, LayerLoaderContext *ctx)	{
		worldReference = ctx->getWorld();
		osg::Node* node = readFile( file.c_str() );
		Layer* Root = new Layer(node);
		Root->setName("KML object");
		return Root;
	};

	osg::Group *loadNode(const std::string &file, World *world) {
		worldReference = world;
		return readFile(file.c_str());
	}

private:
	osg::Group* readFile(const char* filename);
	osg::Group* parseGroup(KMLParsingContext* ctx, const xmlChar* ParentName);
	osg::Group* parsePlacemark(KMLParsingContext* ctx, const xmlChar* ParentName);
	osg::Group* parsePoint(KMLParsingContext* ctx, const xmlChar* ParentTag);
	osg::Group* parsePolygon(KMLParsingContext* ctx, const xmlChar* ParentTag);
	osg::Group* parseLineString(KMLParsingContext* ctx, const xmlChar* ParentTag);
    osg::Group* flattenGroup(osg::Group* sparceGroup);
	osg::Vec3Array* parseBoundaryIs(KMLParsingContext* ctx, const xmlChar* ParentTag, AltitudeMode altitudeMode=ALTITUDE_ABSOLUTE, bool multiple=false);
	osg::Vec3Array* parseLinearRing(KMLParsingContext* ctx, const xmlChar* ParentTag, AltitudeMode altitudeMode=ALTITUDE_ABSOLUTE);

	void ExtrudeVertices(osg::Geode* model, osg::Vec3Array* vertices,
		const osg::Vec4 &polyColour, const osg::Vec4 &lineColour,
		AltitudeMode altitude=ALTITUDE_ABSOLUTE, ExtrudeMode extrudeMode=EXTRUDE_POINT);

	float parseColorWidthStyle(KMLParsingContext* ctx, const xmlChar* ParentTag, osg::Vec4f& colour);
	void parseStyle(KMLParsingContext* ctx, const xmlChar* ParentTag);
	void parseStyleMap(KMLParsingContext* ctx, const xmlChar* ParentTag);
	void modifyStyle(KMLParsingContext* ctx, const xmlChar* ParentTag, KMLStyle* style);
	const xmlChar* parseStyleMapPair(KMLParsingContext* ctx, const xmlChar* ParentTag);

	bool  isClosingTag(const xmlChar* ParentName, const xmlChar* name, int type);
	void  skipTag(KMLParsingContext* ctx, const xmlChar* ParentName);
	const xmlChar* getTextData(KMLParsingContext* ctx);
	osg::Vec3Array* parseCoordinates(const xmlChar *str, AltitudeMode altitudeMode=ALTITUDE_ABSOLUTE);
	AltitudeMode parseAltitudeMode(const xmlChar *str);
	void emptyStyleMap(KMLParsingContext* ctx);

	World* worldReference;
};


struct ltstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return stricmp(s1, s2) < 0;
  }
};

class KMLStyle
{

public:
	KMLStyle(): labelColour(1.0f, 0.5f, 0.5f, 1.0),
				lineColour(1.0f, 1.0f, 1.0f, 1.0),
				polyColour(0.5f, 0.5f, 1.0f, 1.0),
				lineWidth(1.0f)
				{};

	KMLStyle(osg::Vec4f poly, osg::Vec4f line, osg::Vec4f label, float w):
				polyColour(poly),
				lineColour(poly),
				labelColour(poly),
				lineWidth(w)
				{};
	KMLStyle(const KMLStyle& k):
				polyColour(k.getPolyColour()),
				lineColour(k.getLineColour()),
				labelColour(k.getLabelColour()),
				lineWidth(k.getWidth())
				{};

	void  setPolyColour(float r, float g, float b, float a)
		{	polyColour.set(r,g,b,a);		};
	void  setLineColour(float r, float g, float b, float a)
		{	lineColour.set(r,g,b,a);		};
	void  setLabelColour(float r, float g, float b, float a)
		{	labelColour.set(r,g,b,a);		};

	void  setPolyColour(osg::Vec4 v)	{	polyColour = v;		};
	void  setLineColour(osg::Vec4 v)	{	lineColour = v;		};
	void  setLabelColour(osg::Vec4 v)	{	labelColour = v;	};

	osg::Vec4f getPolyColour(void)  const		{	return polyColour;	};
	osg::Vec4f getLineColour(void)  const		{	return lineColour;	};
	osg::Vec4f getLabelColour(void)  const		{	return labelColour;	};

	void  setWidth(float w)			{	lineWidth=w;		};
	float getWidth(void) const			{	return lineWidth;	};

private:
	osg::Vec4f polyColour;
	osg::Vec4f lineColour;
	osg::Vec4f labelColour;
	float lineWidth;
};

struct KMLParsingContext
{
public:
	KMLParsingContext():reader(NULL), currentStyle(NULL){};
	KMLParsingContext(xmlTextReaderPtr reader):reader(reader), currentStyle(NULL){};

	std::map<const char*, KMLStyle, ltstr> styles;
	xmlTextReaderPtr reader;
	//const xmlChar* currentStyleUrl;
	const KMLStyle* currentStyle;
};
#endif /* !__XGIS_KML_OSG_KML_H__ */
