#include "config.h"

#include <osg/Geode>
#include <osg/Geometry>

#include <xgislib/osgutils.h>
#include <xgislib/xputils.h>

#include <xgislib/world.h>
#include <xgislib/worldloader.h>
#include <xgislib/relativealtitudegroup.h>

#include <osgGA/GUIEventAdapter>

#include <osgParticle/ExplosionEffect>
#include <osgParticle/ExplosionDebrisEffect>

#include <string.h>

#include "layer-building.h"

// BuildingsLayerPlugin

#define FLOAT_VAL_UNDEF 1e30

class BuildingDemolishCB : public osg::NodeCallback {
	osg::Node *toRemove;
	
public:
	BuildingDemolishCB(osg::Node *_toRemove) {
		toRemove = _toRemove;
	}

	virtual void operator()(osg::Node *node, osg::NodeVisitor *nv) {
		osg::Group *explosionGroup = new osg::Group;
		explosionGroup->addChild(new osgParticle::ExplosionDebrisEffect
			(toRemove->getBound().center(), 10.0f, 1.0f));
		explosionGroup->addChild(new osgParticle::ExplosionEffect
			(toRemove->getBound().center(), 20.0f, 1.0f));
		((osg::Group *)node)->addChild(explosionGroup);
		((osg::Group *)node)->removeChild(toRemove);
		node->setUpdateCallback(NULL);
	}

};

struct BuildingPickAction : public PickAction {
	virtual bool pick(osg::Node *picked, const osg::Vec3 &intersection,
		int button, bool doubleClick) {
		if(button == osgGA::GUIEventAdapter::RIGHT_MOUSE_BUTTON && doubleClick) {
			picked->getParents()[0]->setUpdateCallback(new BuildingDemolishCB(picked));
			return true;
		}
		return false;
	}
};

struct LayerBuildingPluginContext : public XGFLoaderPluginContext {
	std::vector<Coords::GaussKr> outline;
	float high, low;
	std::string id;
	std::string tmp;

	osgUtil::Tessellator tessellator;

	osg::Node *createBuildingGeode(XGFLoaderContext *ctx, const std::string &name,
		const std::vector<Coords::GaussKr> &outline, float high, float low) {
		osg::Geode *geode = new osg::Geode;
		float base = low;
		float height = high - base;
		
		base = 0;
		
		if(height > 500.0) {
			XGIS_LOG_DEBUG5("Building higher than 500m (base %f, top %f): "
				"this is probably an error.\n", low, high);
			height = 500;
		}
		else if(height <= 0.0) {
			XGIS_LOG_DEBUG5("Building lower than 0m (base %f, top %f): "
				"this is probably an error.\n", low, high);
			height = 2;
		}

		bool isRect = isRectOutline(outline);
		
		osg::ref_ptr<osg::Geometry> bodyGeometry =
			new osg::Geometry();
		osg::ref_ptr<osg::Vec3Array> bodyVertices = new osg::Vec3Array;
		osg::ref_ptr<osg::Vec3Array> bodyNormals = new osg::Vec3Array();

		unsigned int seed = 0;
		const char *nptr = (const char *)name.c_str();
		while(*nptr != '\0') {
			seed += (unsigned int)*nptr++;
		}
		srand(seed);

		osg::ref_ptr<osg::Vec4Array> wallColors = new osg::Vec4Array;
		float rf = (float)rand()/(float)RAND_MAX;
		wallColors->push_back(osg::Vec4(.4f + rf*0.4, .4f + rf*0.4, .4f + rf*0.4, 1.0f));	
		osg::ref_ptr<osg::Vec4Array> roofColors = new osg::Vec4Array;
		rf = (float)rand()/(float)RAND_MAX;
		roofColors->push_back(osg::Vec4(.2f + rf*0.5, .0f, .0f, 1.0f));
		
#ifdef XGIS_BUILDING_TEXTURES
		osg::ref_ptr<osg::Texture2D> wtex, rtex;
		osg::ref_ptr<osg::Vec2Array> wtexCoords, rtexCoords;
		osg::ref_ptr<osg::Vec4Array> white = new osg::Vec4Array;
		white->push_back(osg::Vec4(1.0f, 1.0f, 1.0f, 1.0f));	

		osg::ref_ptr<osg::Image> wimage = ctx->getResourceBag()->getWallTexture
			((WallTextureID)(name[0] % WALL_TEXTURE_LAST), ctx->getStorage());
		osg::ref_ptr<osg::Image> rimage = ctx->getResourceBag()->getRoofTexture
			((RoofTextureID)(name[0] % ROOF_TEXTURE_LAST), ctx->getStorage());
		osg::Vec2 wdtx, rdtx;
		bool useWallTex = (NULL != wimage && wimage->valid());
		bool useRoofTex = (NULL != rimage && rimage->valid());
		if(useWallTex) {
			wdtx = osg::Vec2(wimage->s()/50.0f, wimage->t()/50.0f);
			wtex = new osg::Texture2D;
			wtex->setWrap(osg::Texture::WRAP_S, osg::Texture::REPEAT);
			wtex->setWrap(osg::Texture::WRAP_T, osg::Texture::REPEAT);
			wtex->setImage(wimage.get());
		}
		if(useRoofTex) {
			rdtx = osg::Vec2(rimage->s()/50.0f, rimage->t()/50.0f);
			rtex = new osg::Texture2D;
			rtex->setWrap(osg::Texture::WRAP_S, osg::Texture::REPEAT);
			rtex->setWrap(osg::Texture::WRAP_T, osg::Texture::REPEAT);
			rtex->setImage(rimage.get());
		}
#endif /* XGIS_BUILDING_TEXTURES */

		int outlineVertexCount = outline.size();
		std::vector<Coords::GaussKr>::const_iterator curr = outline.begin();
		std::vector<Coords::GaussKr>::const_iterator end = outline.end();
#ifdef XGIS_BUILDING_TEXTURES
		wtexCoords = new osg::Vec2Array;
		osg::Vec2 lVertex((*curr).x, (*curr).y);
		float texx = 0.0f;
#endif /* XGIS_BUILDING_TEXTURES */
		osg::Vec3 previous;
		osg::Vec3 first = osg::Vec3((*curr).x, (*curr).y, base);
		osg::Vec3 firstTop = osg::Vec3((*curr).x, (*curr).y, base + height);

		while(curr != end) {
			osg::Vec3 bottom = osg::Vec3((*curr).x, (*curr).y, base);
			osg::Vec3 top = osg::Vec3((*curr).x, (*curr).y, base + height);
			
			/*calculate the normal*/
			if (curr != outline.begin()) {
				osg::Vec3 v1 = bottom - previous;
				osg::Vec3 v2 = top - bottom;
				v1.normalize();
				v2.normalize();
				/*vector product*/
				osg::Vec3 normal = osg::Vec3(
					v1.y()*v2.z() - v1.z()*v2.y(),
					v1.z()*v2.x() - v1.x()*v2.z(),
					v1.x()*v2.y() - v1.y()*v2.x());

				bodyNormals->push_back(normal);
			}
			previous = bottom;

			bodyVertices->push_back(bottom);
			bodyVertices->push_back(top);
#ifdef XGIS_BUILDING_TEXTURES
			osg::Vec2 cVertex((*curr).x, (*curr).y);
			texx += (cVertex - lVertex).length()/wdtx.x();
			if(useWallTex) {
				wtexCoords->push_back(osg::Vec2(texx, 0));
				wtexCoords->push_back(osg::Vec2(texx, height/wdtx.y()));
			}
			lVertex = cVertex;
#endif /* XGIS_BUILDING_TEXTURES */
			++curr;
		}
		
		osg::Vec3 v1 = previous - first;
		osg::Vec3 v2 = firstTop - first;

		v1.normalize();
		v2.normalize();
		osg::Vec3 normal = osg::Vec3(
			v1.y()*v2.z() - v1.z()*v2.y(),
			v1.z()*v2.x() - v1.x()*v2.z(),
			v1.x()*v2.y() - v1.y()*v2.x()
			);
		bodyNormals->push_back(normal);
		
		bodyGeometry->setVertexArray(bodyVertices.get());
		bodyGeometry->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);
		bodyGeometry->setNormalArray(bodyNormals.get());

		osg::ref_ptr<osg::DrawElementsUInt> bodyStrip =
			new osg::DrawElementsUInt(osg::PrimitiveSet::QUAD_STRIP);
		for(int i = 0; i < 2*outlineVertexCount - 1; i += 2) {
			bodyStrip->push_back(i);
			bodyStrip->push_back(i + 1);
		}
		bodyGeometry->addPrimitiveSet(bodyStrip.get());

#ifdef XGIS_BUILDING_TEXTURES
		if(useWallTex) {
			// if we have texture, all verts are white
			bodyGeometry->setColorArray(white.get());
			bodyGeometry->setColorBinding(osg::Geometry::BIND_OVERALL);
			bodyGeometry->setTexCoordArray(0, wtexCoords.get());
			// apply texture
			osg::StateSet *state = bodyGeometry->getOrCreateStateSet();
			state->setTextureAttribute(0, wtex.get());
			state->setTextureMode(0, GL_TEXTURE_2D, osg::StateAttribute::ON);
		} else {
#else /* XGIS_BUILDING_TEXTURES */
			bodyGeometry->setColorArray(wallColors.get());
			bodyGeometry->setColorBinding(osg::Geometry::BIND_OVERALL);
#endif
#ifdef XGIS_BUILDING_TEXTURES
		}
#endif /* XGIS_BUILDING_TEXTURES */

		osg::ref_ptr<osg::Vec3Array> sideNormals, topNormals;
		sideNormals = new osg::Vec3Array;
		topNormals = new osg::Vec3Array;
		
		if(isRect) {
			/* slant roof on rectangular buildings */
			float roofHeight = 2.0;
			float cx1 = (outline[0].x + outline[1].x)/2.0,
				cy1 = (outline[0].y + outline[1].y)/2.0,
				cx2 = (outline[2].x + outline[3].x)/2.0,
				cy2 = (outline[2].y + outline[3].y)/2.0;
			osg::ref_ptr<osg::Vec3Array> roofVertices = new osg::Vec3Array;
			roofVertices->push_back(osg::Vec3(outline[0].x, outline[0].y, base + height));
			roofVertices->push_back(osg::Vec3(cx1, cy1, base + height + roofHeight));
			roofVertices->push_back(osg::Vec3(outline[1].x, outline[1].y, base + height));
			roofVertices->push_back(osg::Vec3(outline[3].x, outline[3].y, base + height));
			roofVertices->push_back(osg::Vec3(cx2, cy2, base + height + roofHeight));
			roofVertices->push_back(osg::Vec3(outline[2].x, outline[2].y, base + height));
#ifdef XGIS_BUILDING_TEXTURES
			rtexCoords = new osg::Vec2Array;
			rtexCoords->push_back(osg::Vec2(0, 0));
			rtexCoords->push_back(osg::Vec2(0, sqrt((outline[0].x - cx1)*(outline[0].x - cx1) + (outline[0].y - cy1)*(outline[0].y - cy1))/rdtx.x()));
			rtexCoords->push_back(osg::Vec2(0, sqrt((outline[0].x - outline[1].x)*(outline[0].x - outline[1].x) + (outline[0].y - outline[1].y)*(outline[0].y - outline[1].y))/rdtx.x()));
			float ry = sqrt((outline[0].x - outline[3].x)*(outline[0].x - outline[3].x) + (outline[0].y - outline[3].y)*(outline[0].y - outline[3].y)); 
			rtexCoords->push_back(osg::Vec2(ry , 0));
			rtexCoords->push_back(osg::Vec2(ry, sqrt((outline[0].x - cx1)*(outline[0].x - cx1) + (outline[0].y - cy1)*(outline[0].y - cy1))/rdtx.x()));
			rtexCoords->push_back(osg::Vec2(ry, sqrt((outline[0].x - outline[1].x)*(outline[0].x - outline[1].x) + (outline[0].y - outline[1].y)*(outline[0].y - outline[1].y))/rdtx.x()));
#endif /* XGIS_BUILDING_TEXTURES */

			/* two triangles on roof sides */
			osg::ref_ptr<osg::Geometry> sideGeometry =
				new osg::Geometry();
			sideGeometry->setVertexArray(roofVertices.get());
			osg::ref_ptr<osg::DrawElementsUInt> sideTriangles =
				new osg::DrawElementsUInt(osg::PrimitiveSet::TRIANGLES);
			sideTriangles->push_back(0);
			sideTriangles->push_back(2);
			sideTriangles->push_back(1);
			sideTriangles->push_back(3);
			sideTriangles->push_back(5);
			sideTriangles->push_back(4);
			sideNormals->push_back((roofVertices->operator[](2) - roofVertices->operator[](0)) ^
				(roofVertices->operator[](1) - roofVertices->operator[](0)));
			sideNormals->push_back((roofVertices->operator[](5) - roofVertices->operator[](3)) ^
				(roofVertices->operator[](4) - roofVertices->operator[](3)));
			sideGeometry->setNormalArray(sideNormals.get());
			sideGeometry->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);
#ifdef XGIS_BUILDING_TEXTURES
			wtexCoords = new osg::Vec2Array;
			wtexCoords->push_back(osg::Vec2(0, 0));
			wtexCoords->push_back(osg::Vec2(0, sqrt((outline[0].x - outline[1].x)*(outline[0].x - outline[1].x) + (outline[0].y - outline[1].y)*(outline[0].y - outline[1].y))/rdtx.x()));
			wtexCoords->push_back(osg::Vec2(ry, sqrt((outline[0].x - cx1)*(outline[0].x - cx1) + (outline[0].y - cy1)*(outline[0].y - cy1))/rdtx.x()));
			wtexCoords->push_back(osg::Vec2(0, 0));
			wtexCoords->push_back(osg::Vec2(0, sqrt((outline[0].x - outline[1].x)*(outline[0].x - outline[1].x) + (outline[0].y - outline[1].y)*(outline[0].y - outline[1].y))/rdtx.x()));
			wtexCoords->push_back(osg::Vec2(ry, sqrt((outline[0].x - cx1)*(outline[0].x - cx1) + (outline[0].y - cy1)*(outline[0].y - cy1))/rdtx.x()));
			if(useWallTex) {
				// if we have texture, all verts are white
				sideGeometry->setColorArray(white.get());
				sideGeometry->setColorBinding(osg::Geometry::BIND_OVERALL);
				sideGeometry->setTexCoordArray(0, wtexCoords.get());
				// apply texture
				osg::StateSet *state = sideGeometry->getOrCreateStateSet();
				state->setTextureAttribute(0, wtex.get());
				state->setTextureMode(0, GL_TEXTURE_2D, osg::StateAttribute::ON);
			} else {
#else
				sideGeometry->setColorArray(wallColors.get());
				sideGeometry->setColorBinding(osg::Geometry::BIND_OVERALL);
#endif /* XGIS_BUILDING_TEXTURES */
#ifdef XGIS_BUILDING_TEXTURES
			}
#endif /* XGIS_BUILDING_TEXTURES */
			sideGeometry->addPrimitiveSet(sideTriangles.get());
			geode->addDrawable(sideGeometry.get());

			/* quad strip over the roof */
			osg::ref_ptr<osg::Geometry> topGeometry =
				new osg::Geometry();
			topGeometry->setVertexArray(roofVertices.get());
			osg::ref_ptr<osg::DrawElementsUInt> topQuads =
				new osg::DrawElementsUInt(osg::PrimitiveSet::QUAD_STRIP);
			topQuads->push_back(0);
			topQuads->push_back(3);
			topQuads->push_back(1);
			topQuads->push_back(4);
			topQuads->push_back(2);
			topQuads->push_back(5);
			topNormals->push_back((roofVertices->operator[](3) - roofVertices->operator[](0)) ^
				(roofVertices->operator[](1) - roofVertices->operator[](0)));
			topNormals->push_back((roofVertices->operator[](5) - roofVertices->operator[](2)) ^
				(roofVertices->operator[](1) - roofVertices->operator[](2)));
			topGeometry->setNormalArray(topNormals.get());
			topGeometry->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);
#ifdef XGIS_BUILDING_TEXTURES
			if(useRoofTex) {
				// if we have texture, all verts are white
				topGeometry->setColorArray(white.get());
				topGeometry->setColorBinding(osg::Geometry::BIND_OVERALL);
				topGeometry->setTexCoordArray(0, rtexCoords.get());
				// apply texture
				osg::StateSet *state = topGeometry->getOrCreateStateSet();
				state->setTextureAttribute(0, rtex.get());
				state->setTextureMode(0, GL_TEXTURE_2D, osg::StateAttribute::ON);
			} else {
#else
				topGeometry->setColorArray(roofColors.get());
				topGeometry->setColorBinding(osg::Geometry::BIND_OVERALL);
#endif /* XGIS_BUILDING_TEXTURES */
#ifdef XGIS_BUILDING_TEXTURES
			}
#endif /* XGIS_BUILDING_TEXTURES */

			topGeometry->addPrimitiveSet(topQuads.get());
			geode->addDrawable(topGeometry.get());

		}
		else {
			/* flat roof on non-rectangular buildings */
			osg::ref_ptr<osg::DrawElementsUInt> topSurface =
				new osg::DrawElementsUInt(osg::PrimitiveSet::POLYGON);
			for(int i = 0; i < 2*outlineVertexCount - 1; i += 2) {
				topSurface->push_back(i + 1);
			}
			osg::ref_ptr<osg::Geometry> topGeometry =
				new osg::Geometry();
			topGeometry->setVertexArray(bodyVertices.get());
			topGeometry->addPrimitiveSet(topSurface.get());
			topGeometry->setColorArray(roofColors.get());
			topGeometry->setColorBinding(osg::Geometry::BIND_OVERALL);

			topNormals->push_back(osg::Vec3(0, 0, 1.0));
			topGeometry->setNormalArray(topNormals.get());
			topGeometry->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);
			
			/* we need to retessellate the polygons as they are not necessarily
			   convex */
			tessellator.retessellatePolygons(*topGeometry);

			geode->addDrawable(topGeometry.get());
		}

#if 0
		// we don't really want text on every building; this is to be turned on just for
		// building debugging purposes
		DeferredText *text = new DeferredText((const char *)name,
				osgText::String::ENCODING_UTF8);
		text->setFont(resourceBag->getFont(FONT_TIMES));
		text->setAxisAlignment(osgText::Text::SCREEN);
		text->setColor(osg::Vec4f(0,0,0,1));
		text->setCharacterSize(10);
		text->setPosition(osg::Vec3(outline[0].x, outline[0].y, base + height + 20));
		text->setLayout(osgText::Text::LEFT_TO_RIGHT);
		text->setDrawMode(osgText::Text::TEXT);
		text->setCharacterSizeMode(osgText::Text::
			OBJECT_COORDS_WITH_MAXIMUM_SCREEN_SIZE_CAPPED_BY_FONT_HEIGHT);
		text->setAlignment(osgText::Text::CENTER_CENTER);
		text->setMaximumHeight(10);
		geode->addDrawable(text);
#endif /* 0/1 */

		geode->addDrawable(bodyGeometry.get());

		geode->setUserData(new BuildingPickAction());
		geode->setName((const char *)name.c_str());

		RelativeAltitudeGroup *group = new RelativeAltitudeGroup
			(osg::Vec2(outline[0].x, outline[0].y), 0.0, ctx->getWorld());
		group->addChild(geode);
		
		return group;
	}


	bool isRectOutline(const std::vector<Coords::GaussKr> &outline) {
		/* only outlines with 5 vertices (first and last repeat) can be considered rectangular */ 
		if(outline.size() != 5) {
			return false;
		}
#if 1
		/* simply: anything made out of 4 edges is rectangular */
		return true;
#else
		/* the shape is rectangular if pairs of subsequent edge vectors are
		 * orthogonal (ie edge(i)*edge(i+1) ~= 0)
		 * NOTE: however, do we really want to do these calculations for each
		 * building? */
		osg::Vec2f v1(outline[0].x - outline[1].x, outline[0].y - outline[1].y);
		v1.normalize();
		osg::Vec2f v2(outline[1].x - outline[2].x, outline[1].y - outline[2].y);
		v2.normalize();
		osg::Vec2f v3(outline[2].x - outline[3].x, outline[2].y - outline[3].y);
		v3.normalize();
			
		return (fabs(v1*v2) < 0.01 && fabs(v2*v3) < 0.01);
#endif /* 0/1 */
	}

};

LayerBuildingPlugin::LayerBuildingPlugin() : XGFLoaderPlugin() {
	sax.startElement = LayerBuildingPlugin::startElement;
	sax.endElement = LayerBuildingPlugin::endElement;
	sax.characters = LayerBuildingPlugin::characters;
}

void LayerBuildingPlugin::startElement(void *ctx, 
	const xmlChar *name, 
	const xmlChar **atts) {
	DECLARE_XGFLOADER_PLUGIN_CTX(lb, LayerBuildingPluginContext);

	if(0 == strcmp((const char *)name, "object")) {
		lb->high = lb->low = FLOAT_VAL_UNDEF;
		lb->id = "";
		lb->outline.clear();
	}
	lb->tmp = "";
}

void LayerBuildingPlugin::endElement(void *ctx, 
	const xmlChar *name) {
	DECLARE_XGFLOADER_CTX(b);
	DECLARE_XGFLOADER_PLUGIN_CTX(lb, LayerBuildingPluginContext);

	if(0 == strcmp((const char *)name, "object")) {
		XGIS_LOG_DEBUG5("Creating building: %s, %f->%f, %d", lb->id.c_str(),
			lb->low, lb->high, lb->outline.size());
		if(lb->id.length() > 0 && lb->outline.size() >= 3 &&
			lb->low != FLOAT_VAL_UNDEF && lb->high != FLOAT_VAL_UNDEF) {
			osg::Node *bld = lb->createBuildingGeode(b, lb->id, lb->outline, lb->high, lb->low);
			if(NULL != bld) {
				b->addResult(bld);
			}
		}
	}
	else if(0 == strcmp((const char *)name, "id")) {
		lb->id = lb->tmp;
		XGIS_LOG_DEBUG5("Got building ID: %s", lb->id.c_str());
	}
	else if(0 == strcmp((const char *)name, "high")) {
		if(1 != sscanf(lb->tmp.c_str(), "%f", &(lb->high))) {
			lb->high = FLOAT_VAL_UNDEF;
		}
		XGIS_LOG_DEBUG5("Got building high: %f", lb->high);
	}
	else if(0 == strcmp((const char *)name, "low")) {
		if(1 != sscanf(lb->tmp.c_str(), "%f", &(lb->low))) {
			lb->low = FLOAT_VAL_UNDEF;
		}
		XGIS_LOG_DEBUG5("Got building low: %f", lb->low);
	}
	else if(0 == strcmp((const char *)name, "point")) {
		float x = FLOAT_VAL_UNDEF, y = FLOAT_VAL_UNDEF;
		if(2 == sscanf(lb->tmp.c_str(), "%f,%f", &x, &y)) {
			lb->outline.push_back(Coords::GaussKr(x, y));
		}
		XGIS_LOG_DEBUG5("Got building point: %f,%f", x, y);
	}

}

void LayerBuildingPlugin::characters(void *ctx, 
	const xmlChar *ch, 
	int len) {
	DECLARE_XGFLOADER_PLUGIN_CTX(lb, LayerBuildingPluginContext);

	lb->tmp.insert(lb->tmp.length(), (const char *)ch, len);
}

const std::string &LayerBuildingPlugin::element() {
	static std::string e = "object";
	return e;
}

XGFLoaderPluginContext *LayerBuildingPlugin::createPluginContext() {
	return new LayerBuildingPluginContext();
}
