#include "config.h"

#include <osg/ShapeDrawable>
#include <osg/Switch>
#include <osg/Billboard>
#include <osg/PositionAttitudeTransform>
#include <osg/Geometry>

#include <osgGA/GUIEventAdapter>

#include <xgislib/osgutils.h>
#include <xgislib/xputils.h>
#include <xgislib/quad.h>
#include <xgislib/pickaction.h>

#include <string.h>

#include "layer-info.h"

#define FLOAT_VAL_UNDEF 1e30

struct InfoLocPickAction : public PickAction {
	virtual bool pick(osg::Node *picked, const osg::Vec3 &point, int button, bool doubleClick) {
		if(button == osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON) {
			osg::Switch *sw = (osg::Switch *)picked;
			sw->setValue(1, !sw->getValue(1));
			return true;
		}
		return false;
	}
};

static const float sphereRadius = 10.0;

struct LayerInfoPluginContext : public XGFLoaderPluginContext {
	float x, y;
	std::string name, symbol, url, info;
	std::string tmp;

	osg::Node *createInfoLocGeode(const std::string &name,
		float x, float y, const std::string &symbol, const std::string &url,
		const std::string &info, XGFLoaderContext *lb) {

		float z = lb->getWorld()->getTerrainHeight(x, y) + 100.0;//FIXME: make this height a parameter

		osg::PositionAttitudeTransform *rv = new osg::PositionAttitudeTransform;
		rv->setPosition(osg::Vec3(x, y, z));
		
		// clickable model that is always shown
		osg::Node *model = lb->getResourceBag()->getModel(symbol,
			std::vector<std::string>(), lb->getStorage());
		if(model == NULL) {
			osg::Sphere *sphere = new osg::Sphere(osg::Vec3(0, 0, 0), sphereRadius);		
			osg::ShapeDrawable *drawable = new osg::ShapeDrawable(sphere);
			drawable->setColor(osg::Vec4(1.0f, .25f, .25f, 1.0f));
			model = new osg::Geode;
			((osg::Geode *)model)->addDrawable(drawable);		
		}

		// switchable text panel
		std::string infoStr(info);
		for(int i = 0; i != -1; ) {
			i = infoStr.find(";;", i);
			if(i != -1) {
				infoStr.replace(i, 2, "\n");
			}
		}
		TextPanel *panel = new TextPanel(infoStr.c_str(), osgText::String::ENCODING_UTF8, 2);
		panel->setFont(lb->getResourceBag()->getFont(FONT_TIMES));

		osg::PositionAttitudeTransform *panelPos = new osg::PositionAttitudeTransform;
		panelPos->setPosition(osg::Vec3(-1.0, 0.0, -1.0) * sphereRadius);
		panelPos->addChild(panel);

		// switch
		osg::Switch *sw = new osg::Switch;
		sw->addChild(model, true);
		sw->addChild(panelPos, false);
		sw->setUserData(new InfoLocPickAction());

		rv->addChild(sw);
		rv->setName(name);
		return (osg::Node *)rv;
	}
};

LayerInfoPlugin::LayerInfoPlugin() : XGFLoaderPlugin() {
	sax.startElement = LayerInfoPlugin::startElement;
	sax.endElement = LayerInfoPlugin::endElement;
	sax.characters = LayerInfoPlugin::characters;
}

void LayerInfoPlugin::startElement(void *ctx, 
	const xmlChar *name, 
	const xmlChar **atts) {
	DECLARE_XGFLOADER_PLUGIN_CTX(ip, LayerInfoPluginContext);

	if(0 == strcmp((const char *)name, "infolocation")) {
		ip->x = ip->y = FLOAT_VAL_UNDEF;
		ip->name = "";
		ip->url = "";
		ip->symbol = "";
		ip->info = "";
	}
	ip->tmp = "";
}

void LayerInfoPlugin::endElement(void *ctx, 
	const xmlChar *name) {
	DECLARE_XGFLOADER_CTX(lb);
	DECLARE_XGFLOADER_PLUGIN_CTX(ip, LayerInfoPluginContext);

	if(0 == strcmp((const char *)name, "infolocation")) {
		XGIS_LOG_DEBUG5("Creating info location at (%f,%f)", ip->x, ip->y);
		if(ip->name.length() > 0 && 
			ip->x != FLOAT_VAL_UNDEF && ip->y != FLOAT_VAL_UNDEF) {
			osg::Node *il = ip->createInfoLocGeode(ip->name, ip->x, ip->y,
				ip->symbol, ip->url, ip->info, lb);
			if(NULL != il) {
				lb->addResult(il);
			}
		}
	}
	else if(0 == strcmp((const char *)name, "name")) {
		ip->name = ip->tmp;
	}
	else if(0 == strcmp((const char *)name, "url")) {
		ip->url = ip->tmp;
	}
	else if(0 == strcmp((const char *)name, "symbol")) {
		ip->symbol = ip->tmp;
	}
	else if(0 == strcmp((const char *)name, "info")) {
		ip->info = ip->tmp;
	}
	else if(0 == strcmp((const char *)name, "location")) {
		sscanf(ip->tmp.c_str(), "%f,%f", &(ip->x), &(ip->y));
	}

}

void LayerInfoPlugin::characters(void *ctx, 
	const xmlChar *ch, 
	int len) {
	DECLARE_XGFLOADER_PLUGIN_CTX(ip, LayerInfoPluginContext);

	ip->tmp.insert(ip->tmp.length(), (const char *)ch, len);
}

const std::string &LayerInfoPlugin::element() {
	static std::string e = "infolocation";
	return e;
}

XGFLoaderPluginContext *LayerInfoPlugin::createPluginContext() {
	return new LayerInfoPluginContext(); 
}
