#include "config.h"

#include <osg/PositionAttitudeTransform>

#include <xgislib/xputils.h>
#include <xgislib/quad.h>

#include <string.h>

#include "layer-model.h"

#define FLOAT_VAL_UNDEF 1e30

struct LayerModelPluginContext : public XGFLoaderPluginContext {
	float x, y;
	std::string name, model;
	std::vector<std::string> textures;
	std::string tmp;

	osg::Node *createModelGeode(const std::string &name,
		float x, float y, const std::string &model,
		const std::vector<std::string> &textures, XGFLoaderContext *lb) {

		osg::Node *modelNode = lb->getResourceBag()->getModel(model,
			textures, lb->getStorage());
		if(modelNode == NULL) {
			return NULL;
		}
		modelNode->setName(name);

		float z = lb->getWorld()->getTerrainHeight(x, y);
		osg::PositionAttitudeTransform *rv = new osg::PositionAttitudeTransform;
		rv->setPosition(osg::Vec3(x, y, z + 10));
		rv->setScale(osg::Vec3(0.05, 0.05, 0.05));
		rv->addChild(modelNode);
		
		return (osg::Node *)rv;
	}
};

LayerModelPlugin::LayerModelPlugin() : XGFLoaderPlugin() {
	sax.startElement = LayerModelPlugin::startElement;
	sax.endElement = LayerModelPlugin::endElement;
	sax.characters = LayerModelPlugin::characters;
}

void LayerModelPlugin::startElement(void *ctx, 
	const xmlChar *name, 
	const xmlChar **atts) {
	DECLARE_XGFLOADER_PLUGIN_CTX(mp, LayerModelPluginContext);

	if(0 == strcmp((const char *)name, "model")) {
		mp->x = mp->y = FLOAT_VAL_UNDEF;
		mp->name = "";
		mp->model = "";
		mp->textures.clear();
	}
	mp->tmp = "";
}

void LayerModelPlugin::endElement(void *ctx, 
	const xmlChar *name) {
	DECLARE_XGFLOADER_CTX(lb);
	DECLARE_XGFLOADER_PLUGIN_CTX(mp, LayerModelPluginContext);

	if(0 == strcmp((const char *)name, "model")) {
		XGIS_LOG_DEBUG5("Creating model at (%f,%f)", mp->x, mp->y);
		if(mp->model.length() > 0 &&
			mp->x != FLOAT_VAL_UNDEF && mp->y != FLOAT_VAL_UNDEF) {
			osg::Node *model = mp->createModelGeode(mp->name, mp->x, mp->y, mp->model, mp->textures, lb);
			if(NULL != model) {
				lb->addResult(model);
			}
		}
	}
	else if(0 == strcmp((const char *)name, "name")) {
		mp->name = mp->tmp;
	}
	else if(0 == strcmp((const char *)name, "object")) {
		mp->model = mp->tmp;
	}
	else if(0 == strcmp((const char *)name, "texture")) {
		mp->textures.push_back(mp->tmp);
	}
	else if(0 == strcmp((const char *)name, "location")) {
		sscanf(mp->tmp.c_str(), "%f,%f", &(mp->x), &(mp->y));
	}

}

void LayerModelPlugin::characters(void *ctx, 
	const xmlChar *ch, 
	int len) {
	DECLARE_XGFLOADER_PLUGIN_CTX(mp, LayerModelPluginContext);

	mp->tmp.insert(mp->tmp.length(), (const char *)ch, len);
}

const std::string &LayerModelPlugin::element() {
	static std::string e = "model";
	return e;
}

XGFLoaderPluginContext *LayerModelPlugin::createPluginContext() {
	return new LayerModelPluginContext(); 
}
