#include "config.h"

#include <osg/Node>
#include <osg/PolygonMode>
#include <osg/LineWidth>
#include <osg/Point>
#include <osg/Geometry>
#include <osg/Geode>

#include <xgislib/osgutils.h>
#include <xgislib/xputils.h>
#include <xgislib/quad.h>

#include <string.h>

#include "layer-vectors.h"

#define FLOAT_VAL_UNDEF 1e30

struct LayerVectorsPluginContext : public XGFLoaderPluginContext {
	std::string tmp;

	osg::ref_ptr<osg::Vec3Array> vertices;
	
	osg::Node *createVectorGeode(osg::Vec3Array *vertices, XGFLoaderContext *lb) {
		osg::Geode *rv = new osg::Geode;
		
		osg::Geometry *geom = new osg::Geometry();
		geom->setVertexArray(vertices);
		osg::DrawElementsUInt *lines = new osg::DrawElementsUInt(osg::PrimitiveSet::LINES);
		for(unsigned int i = 0; i < vertices->size(); i++) {
			lines->push_back(i);
		}
		geom->addPrimitiveSet(lines);
		osg::DrawElementsUInt *points = new osg::DrawElementsUInt(osg::PrimitiveSet::POINTS);
		for(unsigned int i = 1; i < vertices->size(); i += 2) {
			points->push_back(i);
		}
		geom->addPrimitiveSet(points);
		osg::Vec4Array *colors = new osg::Vec4Array;
		colors->push_back(osg::Vec4(1.0, 0.0, 0.0, 1.0));
		geom->setColorArray(colors);
		geom->setColorBinding(osg::Geometry::BIND_OVERALL);
		osg::StateSet *state = geom->getOrCreateStateSet();
		state->setAttribute(new osg::LineWidth(1.0f), osg::StateAttribute::ON);
		state->setAttribute(new osg::LineWidth(1.0f), osg::StateAttribute::ON);

		osg::PolygonMode *polymode = new osg::PolygonMode;
		polymode->setMode(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::LINE);
		state->setAttributeAndModes(polymode, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);

		osg::Point *point = new osg::Point;
		point->setSize(4.0f);
		state->setAttribute(point, osg::StateAttribute::ON);
		
		state->setMode(GL_LIGHTING, osg::StateAttribute::OFF);
		
		rv->addDrawable(geom);
		
		return rv;
	}
};

LayerVectorsPlugin::LayerVectorsPlugin() : XGFLoaderPlugin() {
	sax.startElement = LayerVectorsPlugin::startElement;
	sax.endElement = LayerVectorsPlugin::endElement;
	sax.characters = LayerVectorsPlugin::characters;
}

void LayerVectorsPlugin::startElement(void *ctx, 
	const xmlChar *name, 
	const xmlChar **atts) {
	DECLARE_XGFLOADER_PLUGIN_CTX(vp, LayerVectorsPluginContext);

	if(0 == strcmp((const char *)name, "vectors")) {
		vp->vertices = new osg::Vec3Array;
	}
	vp->tmp = "";
}

void LayerVectorsPlugin::endElement(void *ctx, 
	const xmlChar *name) {
	DECLARE_XGFLOADER_CTX(lb);
	DECLARE_XGFLOADER_PLUGIN_CTX(vp, LayerVectorsPluginContext);

	if(0 == strcmp((const char *)name, "vectors")) {
		if(vp->vertices->size() > 0) {
			osg::Node *vectors = vp->createVectorGeode(vp->vertices.get(), lb);
			if(NULL != vectors) {
				lb->addResult(vectors);
			}
		}
		vp->vertices = NULL;
	}
	else if(0 == strcmp((const char *)name, "v")) {
		float x0, y0, x1, y1;
		if(4 == sscanf(vp->tmp.c_str(), "%f,%f,%f,%f", &x0, &y0, &x1, &y1)) {
			vp->vertices->push_back(osg::Vec3(x0, y0,
				lb->getWorld()->getTerrainHeight(x0, y0)));
			vp->vertices->push_back(osg::Vec3(x1, y1,
				lb->getWorld()->getTerrainHeight(x1, y1)));
		}
	}
}

void LayerVectorsPlugin::characters(void *ctx, 
	const xmlChar *ch, 
	int len) {
	DECLARE_XGFLOADER_PLUGIN_CTX(vp, LayerVectorsPluginContext);

	vp->tmp.insert(vp->tmp.length(), (const char *)ch, len);
}

const std::string &LayerVectorsPlugin::element() {
	static std::string e = "vectors";
	return e;
}

XGFLoaderPluginContext *LayerVectorsPlugin::createPluginContext() {
	return new LayerVectorsPluginContext(); 
}
