#include "config.h"

#include <osg/Geode>
#include <osg/Geometry>

#include <xgislib/osgutils.h>
#include <xgislib/xputils.h>
#include <xgislib/quad.h>

#include <string.h>

#include "layer-water-surface.h"

#define FLOAT_VAL_UNDEF 1e30

struct LayerWaterSurfacePluginContext : public XGFLoaderPluginContext {
	float dx, dy;
	Coords::GaussKr ll, ur;
	osg::ref_ptr<osg::HeightField> hf;
	int hfData;

	std::string tmp;

	LayerWaterSurfacePluginContext() :
		ll(FLOAT_VAL_UNDEF, FLOAT_VAL_UNDEF), ur(FLOAT_VAL_UNDEF, FLOAT_VAL_UNDEF) {
	}
	
	static osg::Vec3 getVec3FromHeightField(const osg::HeightField *hf, int col, int row) {
		osg::Vec3 rv = hf->getVertex(col, row);

		if (isnan(rv.z())) {
			//if z is NaN, replace it with average of all non-NaN neighbours
			int num = 0;
			float sum = 0, z;
			if (col > 0) {
				z = hf->getVertex(col-1, row).z();
				if (!isnan(z)) {
					num++;
					sum += z;
				}
			}
			if (col < (int)hf->getNumColumns()-1) {
				z = hf->getVertex(col+1, row).z();
				if (!isnan(z)) {
					num++;
					sum += z;
				}
			}
			if (row > 0) {
				z = hf->getVertex(col, row-1).z();
				if (!isnan(z)) {
					num++;
					sum += z;
				}
			}
			if (row < (int)hf->getNumRows()-1) {
				z = hf->getVertex(col, row+1).z();
				if (!isnan(z)) {
					num++;
					sum += z;
				}
			}
			rv.z() = sum/num;
		}
		return rv;
	}

	osg::Node *createSurfaceGeode(osg::HeightField *hf,
		XGFLoaderContext *lb) {

		// replace huge values with NaNs
		int numRows = hf->getNumRows(),
			numCols = hf->getNumColumns();

		for (int col = 0; col < numCols; col++)
			for (int row = 0; row < numRows; row++)
				if (fabs(hf->getHeight(col, row)) > 1e+10)
					hf->setHeight(col, row, NAN);

		// create a geode and geometry (actual GL_QUADS go into the geometry)
		osg::Geode *geode = new osg::Geode;

		// load actual terrain
		osg::Geometry *terrain = new osg::Geometry;
	    osg::StateSet *statesetg = terrain->getOrCreateStateSet();
	    statesetg->setMode(GL_BLEND, osg::StateAttribute::ON);
	    statesetg->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	    geode->addDrawable(terrain);

		// read verts & calculate normals
		osg::Vec3Array *verts = new osg::Vec3Array;
		osg::Vec3Array *normals = new osg::Vec3Array;
		osg::Vec4Array *colors = new osg::Vec4Array;

		
#define DEPTH_NON_TRANSPARENT 2

		//inside verts
		for (int col = 0; col < numCols; col++) {
			for (int row = 0; row < numRows; row++) {	
				
				osg::Vec3 v = getVec3FromHeightField(hf, col, row);
				verts->push_back(v);
				osg::Vec3 left = getVec3FromHeightField(hf,
						col == 0 ? col : col-1, row),
					right = getVec3FromHeightField(hf,
						col == numCols-1 ? col : col+1, row),
					up = getVec3FromHeightField(hf,
						col, row == numRows-1 ? row : row+1),
					down = getVec3FromHeightField(hf,
						col, row == 0 ? row : row-1);
				normals->push_back( (right-left) ^ (up-down) );

				float alpha;
				alpha = 0.5 + (v.z() - lb->getWorld()->getTerrainHeight(v.x(), v.y()))/
					DEPTH_NON_TRANSPARENT;
				if(alpha < 0.0) {
					alpha = 0.0;
				}
				else if(alpha > 1.0) {
					alpha = 1.0;
				}
				
				colors->push_back(osg::Vec4f(0.0, 0.0, 0.75, alpha));
			}
		}

		terrain->setVertexArray(verts);
		terrain->setNormalArray(normals);
		terrain->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);
		terrain->setColorArray(colors);
		terrain->setColorBinding(osg::Geometry::BIND_PER_VERTEX);

		// build GL_QUAD strips for inside vertices
		for (int j = 0; j < numCols - 1; j++) {
			osg::DrawElementsUInt *strip = new osg::DrawElementsUInt(osg::PrimitiveSet::QUAD_STRIP);
			for (int i = 0; i < numRows; i++) {
				strip->push_back((j)*numRows + i);
				strip->push_back((j + 1)*numRows + i);
			}
			terrain->addPrimitiveSet(strip);
		}

		return geode;
	}
};

LayerWaterSurfacePlugin::LayerWaterSurfacePlugin() : XGFLoaderPlugin() {
	sax.startElement = LayerWaterSurfacePlugin::startElement;
	sax.endElement = LayerWaterSurfacePlugin::endElement;
	sax.characters = LayerWaterSurfacePlugin::characters;
}

void LayerWaterSurfacePlugin::startElement(void *ctx, 
	const xmlChar *name, 
	const xmlChar **atts) {
	DECLARE_XGFLOADER_PLUGIN_CTX(sp, LayerWaterSurfacePluginContext);

	if(0 == strcmp((const char *)name, "surface")) {
		sp->ll = Coords::GaussKr(FLOAT_VAL_UNDEF, FLOAT_VAL_UNDEF);
		sp->ur = Coords::GaussKr(FLOAT_VAL_UNDEF, FLOAT_VAL_UNDEF);
		sp->dx = sp->dy = FLOAT_VAL_UNDEF;
		sp->hf = NULL;
		sp->hfData = 0;
	}
	sp->tmp = "";
}

void LayerWaterSurfacePlugin::endElement(void *ctx, 
	const xmlChar *name) {
	DECLARE_XGFLOADER_CTX(lb);
	DECLARE_XGFLOADER_PLUGIN_CTX(sp, LayerWaterSurfacePluginContext);

	if(0 == strcmp((const char *)name, "surface")) {
		if(sp->hf.valid()) {
			XGIS_LOG_DEBUG5("Building surface...");
			osg::Node *srf = sp->createSurfaceGeode(sp->hf.get(), lb);
			if(NULL != srf) {
				lb->addResult(srf);
			}
			sp->hf = NULL;
		}
		else {
			XGIS_LOG_DEBUG5("Invalid height field.");
		}
	}
	else if(0 == strcmp((const char *)name, "step")) {
		sscanf(sp->tmp.c_str(), "%f,%f", &(sp->dx), &(sp->dy));
		sp->hfData |= (int)(1 << 0);
	}
	else if(0 == strcmp((const char *)name, "ll")) {
		sscanf(sp->tmp.c_str(), "%lf,%lf", &(sp->ll.x), &(sp->ll.y));
		sp->hfData |= (int)(1 << 1);
	}
	else if(0 == strcmp((const char *)name, "ur")) {
		sscanf(sp->tmp.c_str(), "%lf,%lf", &(sp->ur.x), &(sp->ur.y));
		sp->hfData |= (int)(1 << 2);
	}
	else if(0 == strcmp((const char *)name, "point")) {
		float x, y, z;
		if(sp->hf.valid() && 3 == sscanf(sp->tmp.c_str(), "%f,%f,%f", &x, &y, &z)) {
			int col = (x - sp->ll.x)/sp->dx;
			int row = (y - sp->ll.y)/sp->dy;
			XGIS_LOG_DEBUG5("Setting height (%d, %d, %f)", col, row, z);
			sp->hf->setHeight(col, row, z);
		}
	}
	
	if(!sp->hf.valid() &&
		((sp->hfData & ((int)(1 << 0) | (int)(1 << 1) | (int)(1 << 2))) == ((int)(1 << 0) | (int)(1 << 1) | (int)(1 << 2)))) {
		sp->hf = new osg::HeightField();
		XGIS_LOG_DEBUG5("LL (%lf,%lf), UR (%lf,%lf), step (%f, %f)",
			sp->ll.x, sp->ll.y, sp->ur.x, sp->ur.y, sp->dx, sp->dy);
		int colCount = (sp->ur.x - sp->ll.x)/sp->dx + 1;
		int rowCount = (sp->ur.y - sp->ll.y)/sp->dy + 1;
		sp->hf->allocate(colCount, rowCount);
		for(int col = 0; col < colCount; col++) {
			for(int row = 0; row < rowCount; row++) {
				sp->hf->setHeight(col, row, 0.0);
			}
		}
		sp->hf->setOrigin(osg::Vec3(sp->ll.x, sp->ll.y, 0));
		sp->hf->setXInterval(sp->dx);
		sp->hf->setYInterval(sp->dy);
		XGIS_LOG_DEBUG5("Created height field (%d x %d)@(%lf,%lf)",
			colCount, rowCount, sp->ll.x, sp->ll.y);
	}
}

void LayerWaterSurfacePlugin::characters(void *ctx, 
	const xmlChar *ch, 
	int len) {
	DECLARE_XGFLOADER_PLUGIN_CTX(sp, LayerWaterSurfacePluginContext);

	sp->tmp.insert(sp->tmp.length(), (const char *)ch, len);
}

const std::string &LayerWaterSurfacePlugin::element() {
	static std::string e = "surface";
	return e;
}

XGFLoaderPluginContext *LayerWaterSurfacePlugin::createPluginContext() {
	return new LayerWaterSurfacePluginContext(); 
}
