#include "config.h"

#include <xgislib/xputils.h>
#include <xgislib/quad.h>
#include <xgislib/cumulus.h>
#include <xgislib/cirrus.h>
#include <xgislib/stratus.h>
#include <xgislib/xgislog.h>
#include <xgislib/storage.h>
#include <xgislib/xgisconfig.h>

#include <osgParticle/PrecipitationEffect>
#include <osg/PositionAttitudeTransform>

#include <string.h>

#include "layer-weather.h"

#define FLOAT_VAL_UNDEF 1e30

struct LayerWeatherPluginContext : public XGFLoaderPluginContext {
	float rainDensity;
	float cloudDensity;
	osg::Vec4d rainColour;
	float size, x, y;

	osgParticle::PrecipitationEffect *precipitationEffect;
};

class WeatherLayerEventHandlers : public LayerEventHandlers {
	float rainDensity, cloudDensity;

public:
	WeatherLayerEventHandlers(float _rainDensity, float _cloudDensity) {
		rainDensity = _rainDensity;
		cloudDensity = _cloudDensity;
	}

	virtual void onEnter(LayerSet *set, Layer *layer) {
		// adding precipitation effect
		Quad *q = (Quad*)(set);
		if (rainDensity > 0) {
			osg::ref_ptr<osgParticle::PrecipitationEffect> precipitationEffect =
				new osgParticle::PrecipitationEffect();
			precipitationEffect->rain(rainDensity);
			precipitationEffect->setParticleSize(0.03);
			precipitationEffect->setParticleSpeed(10);
			precipitationEffect->setParticleColor(osg::Vec4(0.5, 0.64, 0.75, 1.0));
			precipitationEffect->setMaximumParticleDensity(30);
			precipitationEffect->setName("rain");
			//precipitationEffect->getOrCreateStateSet()->setRenderBinDetails(10,"DepthSortedBin");
			q->addChild(precipitationEffect.get());
			XGIS_LOG_DEBUG2("Added precipitation effect (%f,%f).",cloudDensity,rainDensity);
		} else {
			XGIS_LOG_DEBUG2("It doesn't seem to be raining here. Precipitation effect not added.");
		}
	}

	virtual void onLeave(LayerSet *set, Layer *layer) {			
		Quad *q = (Quad*)(set);
		unsigned int c = 0;
		for (unsigned int i=0; i < q->getNumChildren(); i++) {
			if (0 == strcmp(q->getChild(i)->getName().c_str(),"rain")) {
				c = i;
			}
		}

		if (q->getNumChildren() == c) {
			XGIS_LOG_DEBUG2("Can't find child rain! Precipitation effect not removed.");
		} else {
			q->removeChild(c);
			XGIS_LOG_DEBUG2("Precipitation effect removed.");
		}
	}
};

LayerWeatherPlugin::LayerWeatherPlugin() : XGFLoaderPlugin() {
	sax.startElement = LayerWeatherPlugin::startElement;
	sax.endElement = LayerWeatherPlugin::endElement;
}

void LayerWeatherPlugin::startElement(void *ctx, 
	const xmlChar *name, 
	const xmlChar **atts) {
	DECLARE_XGFLOADER_PLUGIN_CTX(wp, LayerWeatherPluginContext);

	if(0 == strcmp((const char *)name, "weather")) {
		while(atts != NULL && *atts != NULL) {
			if(0 == strcmp((const char *)*atts, "size")) {
				atts++;
				wp->size = 0;
				sscanf((const char *)*atts, "%f", &(wp->size));
				atts++;
			}
			else if(0 == strcmp((const char *)*atts, "x")) {
				atts++;
				wp->x = 0;
				sscanf((const char *)*atts, "%f", &(wp->x));
				atts++;
			}
			else if(0 == strcmp((const char *)*atts, "y")) {
				atts++;
				wp->y = 0;
				sscanf((const char *)*atts, "%f", &(wp->y));
				atts++;
			}
			else {
				atts += 2;
			}
		}
		wp->rainColour = osg::Vec4(0.5, 0.64, 0.75, 1.0);
	}
	else if(0 == strcmp((const char *)name, "rain")) {
		while(atts != NULL && *atts != NULL) {
			if(0 == strcmp((const char *)*atts, "density")) {
				atts++;
				if(1 != sscanf((const char *)*atts, "%f", &(wp->rainDensity))) {
					wp->rainDensity = FLOAT_VAL_UNDEF;
				}
				atts++;
			}
			else {
				atts += 2;
			}
		}
	}
	else if(0 == strcmp((const char *)name, "clouds")) {
		while(atts != NULL && *atts != NULL) {
			 if(0 == strcmp((const char *)*atts, "density")) {
				 atts++;
				 if(1 != sscanf((const char *)*atts, "%f", &(wp->cloudDensity))) {
					 wp->cloudDensity = FLOAT_VAL_UNDEF;
				 }
				 atts++;
			 }
			 else {
				 atts += 2; 
			 }
		}
	}  
}

static osg::Group *generateClouds(LayerWeatherPluginContext *wp,
	WorldStorage *s, XGisConfig *conf) {
	osg::Group *ret = new osg::Group();
	/* srand must have been called at application start! */
	/* probability that current layer has cloud */
	float p = 0.2;
	float p2 = 0.95;
	float r = (rand() % 100)/100.00;
	int numClouds = 0;
	float stratusHeight = 2500;

	if (r < p)
		numClouds = 1;

	osg::Vec3f base = osg::Vec3f(wp->x, wp->y, 0);

	/*stratus*/
	float sizeX = 3000.0;
	float border = 1200;

	//	cloudDensity = 0;

#if 1
	if (wp->cloudDensity > 0) {

		
		if (((int)(wp->x) % (int)sizeX == 0) 
			&& ((int)(wp->y) % (int)sizeX == 0)) {
			
			XGIS_LOG_DEBUG2("setting stratus clouds with density %f",wp->cloudDensity);

			float minD = (wp->cloudDensity > 0.85 ? 0.85 : (wp->cloudDensity));
			float maxD = minD + 0.2;
			maxD = MIN(maxD, 1.0);

			osg::ref_ptr<Stratus> stratus = new Stratus(s, 1,
				250,sizeX + border, sizeX + border,100,base, conf->getStartTime(), 
				minD, maxD);

			osg::ref_ptr<osg::PositionAttitudeTransform> stratusPos = 
						new osg::PositionAttitudeTransform();
			
			stratusPos->addChild(stratus.get());
			stratusPos->setPosition(osg::Vec3f(base.x(), base.y(), stratusHeight));
			ret->addChild(stratusPos.get());
		}
	} else {
#endif
#if 1
		float cirrusHeight = 4000;

		int qs = 7000;
		
		/*some cirrus, for decoration*/
		if (r > p2) {
			XGIS_LOG_DEBUG2("cirrus!");
			osg::ref_ptr<osg::PositionAttitudeTransform> cirrusPos = 
					new osg::PositionAttitudeTransform();
			
			osg::ref_ptr<Cirrus> cirrus = new Cirrus(s, 1, 1,
				qs, qs, 100.0, base, conf->getStartTime());
			cirrusPos->addChild(cirrus.get());
			cirrusPos->setPosition(osg::Vec3f(wp->x 
					+ qs/2, wp->y + qs/2, cirrusHeight));
			
			double sx = ((rand() % 150) / 100.0) + 0.7;
			double sy = ((rand() % 150) / 100.0) + 0.7;

			double att = ((rand() % 100) / 100)*2.0;

			cirrusPos->setScale(osg::Vec3d(sx,sy,1.0));
			cirrusPos->setAttitude(osg::Quat(att*osg::PI,osg::Vec3f(0,0,1)));

			ret->addChild(cirrusPos.get());
		}
#endif
#if 1
		/*cumulus*/
		int meanSizeX = 500;
		int partSize = 250;
		int maxOffsetOut = 1000;
		float height = 2500;

		for (int i=0; i< numClouds; i++) {
			
			osg::ref_ptr<osg::Group> cloudGroup = new osg::Group();
			osg::ref_ptr<osg::PositionAttitudeTransform> cloudPosOut = 
				new osg::PositionAttitudeTransform();
			ret->addChild(cloudPosOut.get());

			float sizeX = ((rand() % 100)/100.0 - 0.5)*meanSizeX/2.0 + meanSizeX;
			float ratioXY = 0.75 + ((rand() % 100)/100.0 - 0.5)*0.5;
			float ratioXZ = 0.6 + ((rand() % 100)/100.0)*0.5;
			float sizeY = sizeX*ratioXY;
			float sizeZ = sizeX*ratioXZ;

			sizeZ = MIN(sizeZ,400.0);

			XGIS_LOG_DEBUG2("cumulus height %f", sizeZ);

			osg::ref_ptr<Cumulus> cloud = new Cumulus(s, 1, partSize, sizeX, 
				sizeY, sizeZ, base, conf->getStartTime(), 1);

			float offX = (float)(rand() % maxOffsetOut);
			float offY = (float)(rand() % maxOffsetOut);

			cloudPosOut->addChild(cloud.get());

			cloudPosOut->setPosition(osg::Vec3f(wp->x + offX, 
				wp->y + offY, height));
			
		}
#endif
	}

	return ret;
}

void LayerWeatherPlugin::endElement(void *ctx, 
	const xmlChar *name) {
	DECLARE_XGFLOADER_CTX(b);
	DECLARE_XGFLOADER_PLUGIN_CTX(wp, LayerWeatherPluginContext);

	// TODO: fix this to work with new layer stuff
	WorldStorage *s = b->getStorage();
	XGisConfig *conf = XGisConfig::getInstance();

	if(0 == strcmp((const char *)name, "weather")) {
		if(wp->size > 0 && wp->x > 0 && wp->y > 0) {
			b->addResult(generateClouds(wp, s, conf));  
#define USE_PRECIPITATION
#ifdef USE_PRECIPITATION
			b->addResult(new WeatherLayerEventHandlers(wp->rainDensity, wp->cloudDensity));
#endif
		}
	}
}

const std::string &LayerWeatherPlugin::element() {
	static std::string e = "weather";
	return e;
}

XGFLoaderPluginContext *LayerWeatherPlugin::createPluginContext() {
	return new LayerWeatherPluginContext(); 
}
