#include "config.h"

#include <stdlib.h>

#include <osg/Geode>
#include <osg/Geometry>
#include <osg/Group>
#include <osg/Node>
#include <osg/PositionAttitudeTransform>

#include <osgText/Text>
#include <osgText/Font>

#include <xgislib/xputils.h>
#include <xgislib/world.h>
#include <xgislib/quad.h>

#include <string.h>

#include "xgfloader.h"

// #define XGIS_PROFILE_XGFLOADER to enable time profiling of parser
// and plugins
#undef XGIS_PROFILE_XGFLOADER

#ifdef XGIS_PROFILE_XGFLOADER

#include <osg/Timer>

static osg::Timer profilingTimer;
static double parsingTime = 0.0; /* in ms */

#endif /* XGIS_PROFILE_XGFLOADER */


// XGFLoaderPlugin

XGFLoaderPlugin::XGFLoaderPlugin() {
	memset(&sax, 0, sizeof(sax));
}

// XGFLoaderContext
XGFLoaderContext::XGFLoaderContext(const std::vector<XGFLoaderPlugin *> &_plugins,
	LayerLoaderContext *_ctx) : ctx(_ctx), plugins(_plugins) {
	layerNodes = new osg::Group;

	currentPlugin = NULL;
	pluginContext = NULL;
	
	depth = 0;
}

// XGFLoader

// SAX callbacks

void XGFLoader::startDocument(void *ctx) {
	XGIS_LOG_DEBUG5("Start document.");
}

void XGFLoader::endDocument(void *ctx) {
	XGIS_LOG_DEBUG5("End document.");
}

void XGFLoader::startElement(void *ctx, 
	const xmlChar *name, 
	const xmlChar **atts) {
	DECLARE_XGFLOADER_CTX(lb);

	lb->depth++;
	lb->elementStack.push_back((const char *)name);

	XGIS_LOG_DEBUG5("Start element %s, new depth %d.", name, lb->depth);

	if(lb->depth > 2) {
		// we are inside a plugin-specific element
		if(NULL != lb->currentPlugin) {
			if(NULL != lb->currentPlugin->sax.startElement) {
				// pass the whole thing to layer plugin 
				lb->currentPlugin->sax.startElement(ctx, name, atts);
			}
		}
	}
	else if(lb->depth == 2) {
		// plugin-specific element starts: find a suitable plugin
		std::vector<XGFLoaderPlugin *>::const_iterator i = lb->plugins.begin();
		while(i != lb->plugins.end()) {
			if((*i)->element() == std::string((const char *)name)) {
				lb->currentPlugin = *i;
				if(NULL != lb->pluginContext) {
					delete lb->pluginContext;
				}
				lb->pluginContext = lb->currentPlugin->createPluginContext();
				if(NULL != lb->currentPlugin->sax.startElement) {
					// pass the whole thing to layer plugin 
					lb->currentPlugin->sax.startElement(ctx, name, atts);
				}
				XGIS_LOG_DEBUG5("Element plugin found.");
				break;
			}
			i++;
		}
	}
	else if(lb->depth == 1) {
		// this should be the top-level layer element, objects
		if(0 != strcmp((const char *)name, "objects")) {
			XGIS_LOG_WARNING("Not a layer file: top-level element is not 'objects'.");
			// TODO: error!
		}
	}
}

void XGFLoader::endElement(void *ctx, 
	const xmlChar *name) {
	DECLARE_XGFLOADER_CTX(lb);

	XGIS_LOG_DEBUG5("End element %s.", name);
	
	if(lb->depth >= 2) {
		if(NULL != lb->currentPlugin) {
			if(NULL != lb->currentPlugin->sax.endElement) {
				// pass the whole thing to layer plugin 
				lb->currentPlugin->sax.endElement(ctx, name);
			}
			if(lb->depth == 2) {
				XGIS_LOG_DEBUG5("Leaving element plugin scope.");
				lb->currentPlugin = NULL;
				if(NULL != lb->pluginContext) {
					delete lb->pluginContext;
					lb->pluginContext = NULL;
				}
			}
		}
	}

	lb->depth--;
	// sanity check
	lb->elementStack.pop_back();
}

void XGFLoader::characters(void *ctx, 
	const xmlChar *ch, 
	int len) {
	DECLARE_XGFLOADER_CTX(lb);

#ifdef XGIS_ENABLE_DEBUG
	char *chars = new char[len + 1];
	strncpy(chars, (const char *)ch, len);
	chars[len] = '\0';
	XGIS_LOG_DEBUG5("Characters [%s].", chars);
	delete chars;
#endif /* XGIS_ENABLE_DEBUG */
	
	if(lb->depth >= 2) {
		if(NULL != lb->currentPlugin) {
			if(NULL != lb->currentPlugin->sax.characters) {
				// pass the whole thing to layer plugin 
				lb->currentPlugin->sax.characters(ctx, ch, len);
			}
		}
	}
}

void XGFLoader::ignorableWhitespace(void *ctx, 
	const xmlChar *ch, 
	int len) {
	DECLARE_XGFLOADER_CTX(lb);

#ifdef XGIS_ENABLE_DEBUG
	char *chars = new char[len + 1];
	strncpy(chars, (const char *)ch, len);
	chars[len] = '\0';
	XGIS_LOG_DEBUG5("Ignorable whitespace [%s].", chars);
	delete chars;
#endif /* XGIS_ENABLE_DEBUG */

	if(lb->depth >= 2) {
		if(NULL != lb->currentPlugin) {
			if(NULL != lb->currentPlugin->sax.ignorableWhitespace) {
				// pass the whole thing to layer plugin 
				lb->currentPlugin->sax.ignorableWhitespace(ctx, ch, len);
			}
		}
	}

}

void XGFLoader::warning(void *ctx, 
	const char *msg, 
	...) {
	XGIS_LOG_WARNING("Warning %s", msg);
	
}

void XGFLoader::error(void *ctx, 
	const char *msg, 
	...) {
	XGIS_LOG_ERROR("Error %s", msg);
}

void XGFLoader::fatalError(void *ctx, 
    const char * msg, 
    ...) {
	XGIS_LOG_FATAL("Fatal error %s", msg);
}

void XGFLoader::cdataBlock(void *ctx, 
	const xmlChar *value, 
	int len) {
	DECLARE_XGFLOADER_CTX(lb);

#ifdef XGIS_ENABLE_DEBUG
	char *chars = new char[len + 1];
	strncpy(chars, (const char *)value, len);
	chars[len] = '\0';
	XGIS_LOG_DEBUG5("CDATA [%s].", chars);
	delete chars;
#endif /* XGIS_ENABLE_DEBUG */

	if(lb->depth >= 2) {
		if(NULL != lb->currentPlugin) {
			if(NULL != lb->currentPlugin->sax.cdataBlock) {
				// pass the whole thing to layer plugin 
				lb->currentPlugin->sax.cdataBlock(ctx, value, len);
			}
		}
	}

}

void XGFLoader::startElementNs(void *ctx, 
	const xmlChar *localname, 
	const xmlChar *prefix, 
	const xmlChar *URI, 
	int nb_namespaces, 
	const xmlChar **namespaces, 
	int nb_attributes, 
	int nb_defaulted, 
	const xmlChar **attributes) {
	XGIS_LOG_DEBUG5("Start NS element {%s}%s.", localname, URI);
	
	// TODO: implement this when we move to namespaced layers
}

void XGFLoader::endElementNs(void *ctx, 
	const xmlChar *localname, 
	const xmlChar *prefix, 
	const xmlChar *URI) {
	XGIS_LOG_DEBUG5("End NS element {%s}%s.", localname, URI);

	// TODO: implement this when we move to namespaced layers
}

void XGFLoader::serror(void *ctx, 
	xmlErrorPtr error) {
	XGIS_LOG_ERROR("Structured error.");
	// TODO: more verbose output
}

Layer *XGFLoader::loadLayer(const std::string &file, LayerLoaderContext *ctx) {

	Layer *rv = new Layer();

	//if file is empty or does not exist, return an empty Group
	int fileSize = xpGetFileSize(file);
	if(fileSize <= 0)
		return rv;

#ifdef XGIS_PROFILE_XGFLOADER
	osg::Timer_t startTime = profilingTimer.tick();
#endif /* XGIS_PROFILE_XGFLOADER */

	XGFLoaderContext *xgfCtx = new XGFLoaderContext(plugins, ctx);

	if(0 != xmlSAXUserParseFile(&sax, xgfCtx, file.c_str())) {
		XGIS_LOG_DEBUG5("Parsing failed.");
	}
	else {
		XGIS_LOG_DEBUG5("Parsing succeeded.");
		rv->setNode(xgfCtx->layerNodes);
		std::vector<LayerEventHandlers *>::iterator i = xgfCtx->layerEventHandlers.begin();
		while(i != xgfCtx->layerEventHandlers.end()) {
			rv->addEventHandlers(*i);
			i++;
		}
	}

	delete xgfCtx;
	
#ifdef XGIS_PROFILE_XGFLOADER
	osg::Timer_t endTime = profilingTimer.tick();
	double timeSpent = profilingTimer.delta_m(startTime, endTime); 
	parsingTime += timeSpent;
	XGIS_LOG_INFO("Parsing time: current = %.2lfms, cumulative = %.2lfms", timeSpent, parsingTime);
#endif /* XGIS_PROFILE_XGFLOADER */

	return rv;

}

bool XGFLoader::registerPlugin(XGFLoaderPlugin *plugin) {
	std::vector<XGFLoaderPlugin *>::const_iterator i = plugins.begin();
	while(plugins.end() != i) {
		if((*i)->element() == plugin->element()) {
			return false;
		}
		i++;
	}
	plugins.push_back(plugin);
	return true;
}

XGFLoaderPlugin *XGFLoader::unregisterPlugin(const std::string &element) {
	std::vector<XGFLoaderPlugin *>::iterator i = plugins.begin();
	while(plugins.end() != i) {
		if((*i)->element() == element) {
			plugins.erase(i);
			return *i;
		}
		i++;
	}
	return NULL;
}

#include "layer-building.h"
#include "layer-text.h"
#include "layer-info.h"
#include "layer-model.h"
#include "layer-water-surface.h"
#include "layer-vectors.h"
#include "layer-weather.h"

XGFLoader::XGFLoader() {
	mimeType = "application/vnd.xlab-gis.xgf+xml";
	
	memset(&sax, 0, sizeof(sax));
	
	sax.startDocument = XGFLoader::startDocument;
	sax.endDocument = XGFLoader::endDocument;
	sax.startElement = XGFLoader::startElement;
	sax.endElement = XGFLoader::endElement;
	sax.characters = XGFLoader::characters;
	sax.ignorableWhitespace = NULL /* XGFLoader::ignorableWhitespace */;
	sax.warning = XGFLoader::warning;
	sax.error = XGFLoader::error;
	sax.fatalError = XGFLoader::fatalError;
	sax.cdataBlock = NULL /* XGFLoader::cdataBlock */;
	sax.startElementNs = XGFLoader::startElementNs;
	sax.endElementNs = XGFLoader::endElementNs;
	sax.serror = XGFLoader::serror;
	
	// register plugins
	registerPlugin(new LayerBuildingPlugin());
	registerPlugin(new LayerTextPlugin());
	registerPlugin(new LayerInfoPlugin());
	registerPlugin(new LayerModelPlugin());
	registerPlugin(new LayerWaterSurfacePlugin());
	registerPlugin(new LayerVectorsPlugin());
	registerPlugin(new LayerWeatherPlugin());	
}
