#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <xgislib/coords.h>
#include <xgislib/readerwriterquad.h>
#include <xgislib/xputils.h>
#include <xgislib/quadsingle.h>
#include <xgislib/layerbuilder.h>

#include <osg/Group>
#include <osg/Geode>
#include <osg/Geometry>
#include <osg/Texture2D>

#include <osgDB/ReadFile>
#include <osgDB/WriteFile>

static osgDB::RegisterReaderWriterProxy<ReaderWriterQuad> g_readerWriter_Quad_Proxy;
// static osgDB::RegisterReaderWriterProxy<ReaderWriterQcol> g_readerWriter_Qcol_Proxy;

// TODO: fix the tool to use storage (layer building needs it!)

static const char *usage =
	"Usage: %s [\n"
	"  terrain <world-dir> <quad-size> <quad-res> <txt-res> |\n"
	"  layer <layer-dir>\n"
	"]";

static int compileLayer(const char *layerFile, const char *iveFile) {
	fprintf(stdout, "Compiling layer XML file %s: %s\n", layerFile, iveFile);
	
	int rv = -42;

#if 0
	LayerBuilder builder;
	osg::ref_ptr<osg::Group> buildings = builder.parseLayerXml(layerFile, NULL);
	if(NULL != buildings) {
		if(osgDB::writeNodeFile(*buildings, iveFile)) {
			rv = 0;
		}
		else {
			rv = -101;
		}
	}
#endif /* 0/1 */
	
	return rv;
}

static int compileQuad(const char *quadFile, const char *textureFile, const char *iveFile) {
	int quadSize, quadResol, texResol, bx, by;

	fprintf(stdout, "Compiling quad file %s and texture file %s: %s\n",
		quadFile, textureFile, iveFile);
	
	osg::HeightField *heightField = osgDB::readHeightFieldFile(quadFile);
	if(NULL == heightField) {
		fprintf(stderr, "Failed to read quad file: %s\n", quadFile);
		return -1;
	}
	const char *ptr = strrchr(quadFile, dirSlash);
	if(NULL == ptr) {
		ptr = quadFile;
	}
	else {
		ptr++;
	}
	if(4 != sscanf(ptr, "q-%d-%d-%d-%d.quad", &quadSize, &quadResol, &bx, &by)) {
		fprintf(stderr, "Quad file is not properly named: %s\n", ptr);
		return -2;
	}
	const Coords::GaussKr base(bx, by);

	osg::Image *image = osgDB::readImageFile(textureFile);
	if(NULL == image || !image->valid()) {
		fprintf(stdout, "No texture file %s was found: "
			"will use height-based coloring...\n", textureFile);
	}
	else {
		ptr = strrchr(textureFile, dirSlash);
		if(NULL == ptr) {
			ptr = textureFile;
		}
		else {
			ptr++;
		}
		if(1 != sscanf(ptr, "t-%*d-%d-%*d-%*d.jpg", &texResol)) {
			fprintf(stderr, "Texture file is not properly named: %s\n", ptr);
			return -3;
		}
	}
	
	// create a geode and geometry (actual GL_QUADS go into the geometry)
	osg::ref_ptr<osg::Geode> geode = QuadSingle::compileTerrain(heightField, image,
			quadSize, quadResol, texResol, Coords::GaussKr(bx, by));

	int rv;
	
	if(osgDB::writeNodeFile(*geode, iveFile)) {
		rv = 0;
	}
	else {
		rv = -101;
	}

	return rv;
}

#define MAX_PATH_LEN 2048

static char *worldDir;
static int quadSize, quadRes, txtRes;

static bool compileQuadDirEntry(const char *dir, const char *entry, void *closure) {
	int size, res, x, y;
	char qpath[MAX_PATH_LEN + 1], tpath[MAX_PATH_LEN + 1], ipath[MAX_PATH_LEN + 1];
	
	if(4 != sscanf(entry, "q-%d-%d-%d-%d.quad", &size, &res, &x, &y)) {
		fprintf(stdout, "Ignoring improperly named file: %s\n", entry);
		return true;
	}
	
	if(size != quadSize || res != quadRes) {
		fprintf(stdout, "Ignoring quad with wrong size or resolution: %s", entry);
		return true;
	}

	snprintf(qpath, MAX_PATH_LEN, "%s%c%s", dir, dirSlash, entry);
	snprintf(tpath, MAX_PATH_LEN, "%s%ctexture%c%d%c%d%ct-%d-%d-%d-%d.jpg",
		worldDir, dirSlash, dirSlash, quadSize, dirSlash, txtRes,
		dirSlash, quadSize, txtRes, x, y);		
	snprintf(ipath, MAX_PATH_LEN, "%s%cive%c%d%c%d-%d%cc-%d-%d-%d-%d-%d.ive",
			worldDir, dirSlash, dirSlash, quadSize, dirSlash, quadRes, txtRes,
			dirSlash, quadSize, quadRes, txtRes, x, y);

	return (0 == compileQuad(qpath, tpath, ipath));
}

static bool compileLayerEntry(const char *dir, const char *entry, void *closure) {
	char bpath[MAX_PATH_LEN + 1], ipath[MAX_PATH_LEN + 1];
	int x, y;
	
	if(strlen(entry) < 4 || 0 != strcmp(entry + strlen(entry) - 4, ".xml")) {
		return true;
	}
	
	if(2 != sscanf(entry, "objects-%d-%d.xml", &x, &y)) {
		fprintf(stdout, "Ignoring improperly named file: %s\n", entry);
		return true;
	}
	
	snprintf(bpath, MAX_PATH_LEN, "%s%c%s", dir, dirSlash, entry);
	snprintf(ipath, MAX_PATH_LEN, "%s%cobjects-%d-%d.ive", dir, dirSlash, x, y);
	
	return (0 == compileLayer(bpath, ipath));
}

int main(int argc, char **argv) {
	if(argc < 2) {
		fprintf(stderr, usage, argv[0]);
	}
	else {
		if(0 == strcmp(argv[1], "terrain") && argc == 6) {
			if(!xpIsDirectory(argv[2])) {
				fprintf(stderr, "Second argument must be a directory.\n");
				return -83;
			}

			char *endptr;
			quadSize = strtol(argv[3], &endptr, 10);
			if(argv[3] == endptr || quadSize <= 0) {
				fprintf(stderr, "Quad size is not a valid number: %s", argv[3]);
				return -1;
			}
			quadRes = strtol(argv[4], &endptr, 10);
			if(argv[4] == endptr || quadRes <= 0) {
				fprintf(stderr, "Quad resolution is not a valid number: %s", argv[4]);
				return -1;
			}
			txtRes = strtol(argv[5], &endptr, 10);
			if(argv[5] == endptr || txtRes <= 0) {
				fprintf(stderr, "Texture resolution is not a valid number: %s", argv[5]);
				return -1;
			}

			worldDir = argv[2];
			
			char path[MAX_PATH_LEN + 1];
			snprintf(path, MAX_PATH_LEN, "%s%c3d%c%d%c%d",
				worldDir, dirSlash, dirSlash, quadSize, dirSlash, quadRes);		

			xpProcessDirectory(path, compileQuadDirEntry, NULL);
		}
		else if(0 == strcmp(argv[1], "buildings") && argc == 3) {
			
			if(!xpIsDirectory(argv[2])) {
				fprintf(stderr, "The second arguments must a building plan directory.\n");
				return -83;
			}
			
			xpProcessDirectory(argv[2], compileLayerEntry, NULL);
		}
	}
	
}
