#include "config.h"

#include <stdio.h>

#ifndef WIN32
#  include <errno.h>
#  include <string.h>
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <fcntl.h>
#else
#  define snprintf _snprintf
#  pragma warning(disable : 4996)
#endif /* !WIN32 */

#include <string>
#include <list>

#include "xgislib/pointdata.h"
#include "xgislib/xputils.h"

using namespace std;

static float quadSize = -1;
static int coarseFac = 1;
static size_t maxWrite = 100;
static string destDir;
static list<FILE*> fWrite;
static list<Vector2> locations;

static float startX = 0;
static float startY = 0;
static float endX = 1000000.0;
static float endY = 1000000.0;
static float xunit = 100;
static float yunit = 100;

static FILE* findFile(int locX, int locY) {
	if (locations.empty())
		return NULL;

	//try to pick last open file
	if ((int)locations.front().x == locX &&
		(int)locations.front().y == locY)
		return fWrite.front();
  
	//search if not last
	list<Vector2>::iterator l = locations.begin();
	list<FILE*>::iterator f = fWrite.begin();
	for (; l != locations.end(); ++l, ++f)
		if ((int)l->x == locX && (int)l->y == locY)
			return *f;
  
	return NULL;
}

static bool openFile(int locX, int locY, int qsize, float res, const char* mode) {
	//compose destination and filename
	int pointsPerKm = (int)(1000.0/res);
	char fname[256]; memset(fname, 0, 256);
	snprintf(fname, 256, "%sq-%i-%i-%i-%i.quad",
		destDir.c_str(), qsize, pointsPerKm, locX, locY);
  
	//try to open file for writing
	FILE *fnew = fopen(fname, mode);
	if (!fnew) {
		printf("\nFailed opening file '%s' for writing!\n", fname);
		return false;
	}
  
	// write header
	fwrite(&qsize, sizeof(int), 1, fnew);        // quad size
	fwrite(&pointsPerKm, sizeof(int), 1, fnew);  // quad resolution
	fwrite(&locX, sizeof(int), 1, fnew);         // LL X
	fwrite(&locY, sizeof(int), 1, fnew);         // LL Y

	//append to open files
	fWrite.push_front(fnew);
	locations.push_front(Vector2(locX,locY));
  
	//check if too many open
	if (fWrite.size() > maxWrite) {
		fclose(fWrite.back());
    	fWrite.pop_back();
		locations.pop_back();
	}
  
	return true;
}

static bool savePointBase() {
	//compose destination and filename
	char fname[256]; memset(fname, 0, 256);
	snprintf(fname, 256, "%spointbase.xy", destDir.c_str());
  
	//open point base file
	FILE *fBase = fopen(fname, "w");
	if (!fBase) {
		printf("\nFailed opening pointbase.xy file for writing!\n");
		return false;
	}
  
	//write base into file
	fprintf(fBase, "%lf\n", quadSize);
	fprintf(fBase, "%lf\n", xunit);

	fprintf(fBase, "%lf\n", startX);
	fprintf(fBase, "%lf\n", startY);
	fprintf(fBase, "%lf\n", endX);
	fprintf(fBase, "%lf\n", endY);

	printf("\n*** Saving settings into pointbase.xy:\n");
	printf("quadSize=%lf\n", quadSize);
	printf("xunit=%lf\n", xunit);

	fclose(fBase);
  
	return true;
}

static const char *usage =
	"Usage: xgis-makequads <quad-size> <dest-dir> <source-file> <unit> <coarse-fac>"
	" [<startx> <starty> <endx> <endy>]\n\n"
    "  quad-size - size of the quad that goes into one file (in metres)\n"
    "  dest-dir - destination directory for quad files\n"
    "  source-file - file containing original scanned points\n"
    "  unit - point grid resolution (in metres)\n"
    "  coarse-fac - the factor by which the resolution should be reduced for the coarse model\n"
    "  startx, starty, endx, endy - reference frame for beginning and end of coordinates (in metres, Gauss-Krueger coordinates)\n"
    "\n";

struct Vector3f {
	float x, y, z;
};

int main(int argc, char** argv) {
	if (argc < 6) {
		fprintf(stderr, usage);
		return -1;
	}

	sscanf(argv[1], "%f", &quadSize);

	destDir = argv[2];

	sscanf(argv[4], "%f", &xunit);
	yunit = xunit;
  
	sscanf(argv[5], "%d", &coarseFac);

	if (argc > 6) {
		sscanf(argv[6], "%f", &startX);
		sscanf(argv[7], "%f", &startY);
		sscanf(argv[8], "%f", &endX);
		sscanf(argv[9], "%f", &endY);
	}

	if (destDir[destDir.length()-1] != dirSlash)
    	destDir += dirSlash;

#ifndef WIN32
	// necessary to facilitate opening files >2GB
	int fdRead = open(argv[3], O_LARGEFILE|O_RDONLY);
	if(-1 == fdRead) {
		printf("Failed opening file %s for reading: %s.\n",
			argv[3], strerror(errno));
		return -1;
	}
	FILE *fRead = fdopen(fdRead, "rb");
#else
	FILE *fRead = fopen(argv[3], "rb");
	if (!fRead) {
		printf("Failed opening file %s for reading: %s.\n",
			argv[3], strerror(errno));
		return -1;
	}
#endif // !WIN32 

	int floatsRead = 0;
	Vector3f point;
	floatsRead = fscanf(fRead, "%f %f %f",
		&point.x, &point.y, &point.z);

	float minx = 1e+30f;
	float miny = 1e+30f;
	float maxx = -minx;
	float maxy = -miny;

	// get min and max coordinates 
	int counter = 0;
	printf("Scanning original points dimensions...\n");
	while (floatsRead == 3) {
		if(++counter % 10000 == 0) {
			printf("\rPoints scanned: %d           ", counter);
			fflush(stdout);
		}

		if (point.x < minx)
			minx = point.x;
		if (point.y < miny)
			miny = point.y;
		if (point.x > maxx)
			maxx = point.x;
		if (point.y > maxy)
			maxy = point.y;
      
		floatsRead = fscanf(fRead, "%f %f %f",
			&point.x, &point.y, &point.z);
	}

	printf("minX %lf minY %lf maxX %lf maxY %lf\n", minx, miny, maxx, maxy);

	if(startX < minx)
		startX = minx;
	if(startY < miny)
		startY = miny;
	if(endX > maxx)
		endX = maxx;
	if(endY > maxy)
		endY = maxy;

	int q = (int)floor(startX/quadSize);
	startX = q*quadSize;

	q = (int)floor(startY/quadSize);
	startY = q*quadSize;

	q = (int)ceil((endX - startX)/quadSize);
	endX = startX + q*quadSize;

	q = (int)ceil((endY - startY)/quadSize);
	endY = startY + q*quadSize;

	printf("SX %lf SY %lf EX %lf EY %lf\n", startX, startY, endX, endY);

	savePointBase();

	//stevilo tock v obeh smereh
	int diff = (int)ceil(endX - startX);
	int numX = (int)ceil(diff/xunit) + ((diff%(int)xunit == 0)?1:0);
	diff = (int)ceil(endY - startY);
	int numY = (int)ceil(diff/yunit) + ((diff%(int)yunit == 0)?1:0);
  
	float **lines; 
	lines = new float*[numX];

	//nan defined in a smart-ass way to avoid "possible division of zero" warning
	float nan = (float)endX; nan -= nan; nan /= nan;

	if(isnan(nan)) {
		printf("NaN is really not a number!\n");
	}
	else {
		printf("It seems NaN is a number... Aborting!\n");
		return -123;
	}

	printf("Allocating %d lines of %d floats...\n", numX, numY);
	for (int i = 0; i < numX; i++) {
		lines[i] = new float[numY];
		for (int j = 0; j < numY; j++)
			lines[i][j] = nan;
	}

	int readPoints = 0;
	rewind(fRead);
	floatsRead = fscanf(fRead, "%f %f %f", &point.x, &point.y, &point.z);

	// first pass - stuff it all into the line array
	// gee, I wonder how much memory this uses... do the maths!
	while (floatsRead == 3) {
		int yIndex = (int)((point.y - startY)/yunit);
		int xIndex = (int)((point.x - startX)/xunit);

		if (isnan(lines[xIndex][yIndex]))
			lines[xIndex][yIndex] = (float)point.z;

		if ((++readPoints) % 10000 == 0) {
			printf("\rPoints read: %d           ", readPoints);
			fflush(stdout);
		}
    
		//read next point
		floatsRead = fscanf(fRead, "%f %f %f",
			&point.x, &point.y, &point.z);
	}

	fclose(fRead);

	// parcel it out and write it... 
	int c = 0;

	int pointsPerQuadSide = (int)(quadSize/xunit);
	int xquadoffset = (int)floor(startX/quadSize);
	int yquadoffset = (int)floor(startY/quadSize);

	int quadsX = numX / pointsPerQuadSide;
	if(numX % pointsPerQuadSide != 0) quadsX++;
	int quadsY = numY / pointsPerQuadSide;
	if(numY % pointsPerQuadSide != 0) quadsY++;

	int files = 0, nanfiles = 0;
	
	for (int i = 0; i < quadsX; i++) {
		for (int j = 0; j < quadsY; j++) {
			int quadX = i + xquadoffset;
			int quadY = j + yquadoffset;

			int minX = i*pointsPerQuadSide;
			int minY = j*pointsPerQuadSide;
      
			// last already overwritten!!!!!!!!!!
			int maxX = minX + pointsPerQuadSide;
			int maxY = minY + pointsPerQuadSide;
      
			Point quad = Point(quadX, quadY);     

			// get file for writing
			FILE *f = findFile((int)(quad.x*quadSize), (int)(quad.y*quadSize));
			if (f == NULL) {
				if (!openFile((int)(quad.x*quadSize), (int)(quad.y*quadSize),
					(int)quadSize, xunit, (const char *)"a+b")) {
					fprintf(stderr, "Failed to create quad file.\n");
					return -4;
				}
				f = fWrite.front();
			}

			bool onlynans = true;
			int outnans = 0, innans = 0, total = 0;
			
			for (int k = minX; k <= maxX; k++) {
				for (int m = minY; m <= maxY; m++) {
					float data;

					if ((k >= numX) || (m >= numY)) {
						data = nan;
						outnans++;
					}
					else {
						data = lines[k][m];
						if(!isnan(data)) {
							onlynans = false;
						}
						else {
							innans++;
						}
					}

					if(1 != fwrite(&data, sizeof(float), 1, f)) {
						fprintf(stderr, "Failed to write quad file.\n");
						return -3;
					}
					
					total++;
				}
			}
			
			if(onlynans) {
				fprintf(stderr, "*** WARNING: only NANs (in %d, out %d) in file %d-%d-%d-%d\n",
					innans, outnans, (int)quadSize, (int)xunit,
					(int)(quad.x*quadSize), (int)(quad.y*quadSize));
				nanfiles++;
			}

			files++;
		}
	}

	printf("Total files written %d, only nans in %d\n", files, nanfiles);
	
	printf("%d\n",c);

	for (list<FILE*>::iterator fi=fWrite.begin();
		fi!=fWrite.end();
		++fi)
		fclose(*fi);
	fWrite.clear();

	// finally, create coarse-grained model using the specified reduction factor
	quadSize = endX - startX;
	if(quadSize < endY - startY)
		quadSize = endY - startY;

	if (openFile((int)startX, (int)startY, (int)quadSize,
		xunit*coarseFac, (const char *)"w+b")) {  	
		FILE *f = fWrite.front();
		for (int k = 0; k <= (int)quadSize/xunit; k += coarseFac) {
			for (int m = 0; m <= (int)quadSize/xunit; m += coarseFac) {
				float data;

				if ((k >= numX) || (m >= numY))
					data = nan;
				else
					data = lines[k][m];
				if(1 != fwrite(&data, sizeof(float), 1, f)) {
					fprintf(stderr, "Failed to write coarse model.\n");
					return -2;
				}   
			}
		}
		fclose(f);
	}
	else {
		fprintf(stderr, "Failed to open coarse model file...\n");
		return -5;
	}

	for (int i = 0; i < numX; i++)  {
		delete[] lines[i];
	}
	delete lines;

	printf("\n*** Done\n");

	return 0;
}
