#include <config.h>
#include "stdlib.h"

#include <algorithm>

#include <xgislib/xgislog.h>

#include <xgislib/cumulus.h>
#include <xgislib/pointdata.h>
#include <xgislib/xgisimgutils.h>

#include <osgSim/Impostor>
#include <osg/Depth>
#include <osg/LOD>
#include <osg/BlendFunc>
#include <osg/PointSprite>
#include <osg/Point>
#include <osgDB/ReadFile>
#include <osg/Texture2D>
#include <osg/PositionAttitudeTransform>

Cumulus::Cumulus(WorldStorage *_storage, int _numParticles, int _partSize,
	float _sizeX, float _sizeY, float _thick, const osg::Vec3f &_sunPos,
	time_t _startTime, int _subClouds) :
	Cloud(_storage, _numParticles, _partSize, _sizeX, _sizeY, _thick, _sunPos,
		_startTime) {
	this->subClouds = _subClouds;

	osg::StateSet *cloudState = this->getOrCreateStateSet();

	osg::PointSprite *sprite = new osg::PointSprite();
	cloudState->setTextureAttributeAndModes(0, sprite, osg::StateAttribute::ON);

	osg::Point *point = new osg::Point();
	point->setSize((float)(particleSize));

	cloudState->setAttribute(point);

	/// The texture for the sprites
	osg::Texture2D *tex = new osg::Texture2D();

	osg::Image *image = storage->getImage("smoke.rgb");
	if (image != NULL) {
		tex->setImage(image);
	} else {
		XGIS_LOG_DEBUG2("Cloud particle image not found!");
	}
	cloudState->setTextureAttributeAndModes(0, tex, osg::StateAttribute::ON);
	//cloudState->setRenderBinDetails(9,"RenderBin");

	cloudState->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);


	osg::Depth* depth = new osg::Depth();
	depth->setWriteMask(false);
	cloudState->setAttributeAndModes(depth, osg::StateAttribute::ON );

	cloudState->setMode(GL_LIGHTING, osg::StateAttribute::OFF);



	placeParticles();
}

void Cumulus::getCloudColour(float altitude, osg::Vec3f &cloudHighlight, 
							 osg::Vec3f &cloudShadow){
	
	int c1 = 5; int c2 = -5; int c3 = -7;
	float v2 = 0.5; float v3 = 0.3;

	if (altitude >= c1) {
		cloudHighlight = osg::Vec3f(1,1,1);
		cloudShadow = osg::Vec3f(0.75,0.75,0.75);
	} else if ((altitude < c1) && (altitude > c2)) {
		float val1 = (1 - (c1 - altitude)/(c1 - c2))*(1 - v2) + v2;
		val1 = MAX(val1, v2);
		float val2 = val1 - 0.25;

		cloudHighlight = osg::Vec3f(val1,val1,val2);
		cloudShadow = osg::Vec3f(val2,val2,val2);

	} else {
		float val1 = (1 - (c2 - altitude)/(c2 - c3))*(v2 - v3) + v3;
		val1 = MAX(val1, v3);
		float val2 = val1 - 0.2;

		cloudHighlight = osg::Vec3f(val1,val1,val2);
		cloudShadow = osg::Vec3f(val2,val2,val2);
	}
}



osg::Vec4Array *Cumulus::shadeCloud(const osg::Vec3Array *vertices, bool far,
	const osg::Vec3f &lightColour, double azimuth, double altitude){
	osg::Vec3Array::const_iterator it;
	osg::Vec4Array *colours = new osg::Vec4Array();

	double suZ = sin(osg::DegreesToRadians(altitude));
	double sproj = cos(osg::DegreesToRadians(altitude));//suZ/tan(osg::DegreesToRadians(altitude));//sqrt(1 - suZ*suZ);
	double xangle = osg::DegreesToRadians(270.0-azimuth);
	double suX = sproj*cos(xangle);
	double suY = sproj*sin(xangle);
	float mult = 10000;
	//printf("sun unit %f %f %f\n",suX,suY,suZ);
	osg::Vec3f sunUnit = osg::Vec3d(suX,suY,suZ);
	osg::Vec3f sunVec = osg::Vec3f(mult*suX, mult*suY, mult*suZ);

	XGIS_LOG_DEBUG3("sun unit %f %f %f ",sunUnit.x(), sunUnit.y(), sunUnit.z());
	float maxAngle = 0;
	float minAngle = osg::PI;
	std::vector<float> angles;
	osg::Vec2Array *distances = new osg::Vec2Array();
//	std::vector<float>::iterator ait;
	osg::Vec2Array::const_iterator ait;
	
	float minDistance = sunVec.length();
	float maxDistance = 0;

	float maxHeight = 0;

	osg::Vec3f cloudHi, cloudLow;

	getCloudColour(altitude, cloudHi, cloudLow);

	XGIS_LOG_DEBUG2("cloudHi %f cloudLow %f", cloudHi.x(), cloudLow.x());

	for (it = vertices->begin(); it != vertices->end(); it++) {
		osg::Vec3f v = (*it);
		osg::Vec3f vec = (*it);
		v.normalize();
		float angle = acos(v.x()*sunUnit.x() + v.y()*sunUnit.y() 
			+ v.z()*sunUnit.z());
		
		if (angle > osg::PI) angle -= osg::PI;

		if (angle > maxAngle) maxAngle = angle;
		if (angle < minAngle) minAngle = angle;

		angles.push_back(angle);

		float distance = (sunVec - vec).length();
		float height = vec.z();
		if (height > maxHeight) maxHeight = height;
		if (distance > maxDistance) maxDistance = distance;
		if (distance < minDistance) minDistance = distance;

		distances->push_back(osg::Vec2f(distance, height));
	}

	for (ait = distances->begin(); ait != distances->end(); ait++) {
		osg::Vec2f v = (*ait);
		float distance = v.x();
		float height = v.y();

		float closeness = cloudLow.x() + (
			(1 - (distance - minDistance)/(maxDistance - minDistance))
			*(cloudHi.x() - cloudLow.x()));

		float heightF;

		if (altitude > 0) {
			heightF = cloudLow.x();
			heightF += (height/maxHeight)*(cloudHi.x() - cloudLow.x());
		} else {
			heightF = cloudHi.x();
			heightF -= (height/maxHeight)*(cloudHi.x() - cloudLow.x());
		}

		float w1 = 2; float w2 = 1;
		float val = (w1*heightF + w2*closeness)/(w1 + w2);
		
		/*if (height < margin) {
			
			val *= 0.95;
		}*/

		osg::Vec4f colour;

		//if ((cloudHi.x() - val)/cloudHi.x() < 0.2){
			colour = osg::Vec4f(val*lightColour.x(),val*lightColour.y(),
				val*lightColour.z(),0.2);
	/*	} else {
			colour = osg::Vec4f(val,val,val,0.3);				
		}*/

		colours->push_back(colour);
	}

#if 0
	for (ait = angles.begin(); ait != angles.end(); ait++) {
		float val;

		float angle = (*ait);

		val = cloudLow.x() + (cloudHi.x() - cloudLow.x())*(1-(angle - minAngle)/(maxAngle - minAngle));


		colours->push_back(osg::Vec4f(val*lightColour.x(),val*lightColour.y(),val*lightColour.z(),0.4));

		//XGIS_LOG_DEBUG2("%f %f %f a %f val %f",v.x(), v.y(), v.z(), angle, val);
	}
#endif
	
	return colours;
}

void Cumulus::generateEllipsoid(float a, float b, float c,
	float deltaPhi, float deltaTheta, float offPercent, osg::Vec3Array *v) {

		for (float phi = 0; phi <= 2*osg::PI; phi += deltaPhi) {
	
			for (float theta = 0; theta <= osg::PI; theta += deltaTheta) {

				float x = a*sin(theta)*cos(phi);
				float y = b*sin(theta)*sin(phi);
		
				float z = ((rand() % 100) / 100.0)*c*fabs(cos(theta));

				float offX = ((rand() % 100) / 100.0 - 0.5)*a*offPercent;
				float offY = ((rand() % 100) / 100.0 - 0.5)*b*offPercent;
				float offZ = ((rand() % 100) / 100.0 - 0.5)*c*offPercent;

				XGIS_LOG_DEBUG3("%f %f %f",x + offX, y + offY, z + offZ);

				v->push_back(osg::Vec3f(x + offX,y + offY, z));

			}
		}
}

void Cumulus::generateFromFractal(float a, float b, float c,
								  float deltaPhi, float deltaTheta, 
								  float offPercent, osg::Image *image, 
								  bool isNear, osg::Vec3Array *v) {

	

	
	int width = image->s();
	int height = image->s();
	

	for (float phi = 0; phi <= 2*osg::PI; phi += deltaPhi) {

		for (float theta = 0; theta <= osg::PI; theta += deltaTheta) {

			float x = a*sin(theta)*cos(phi);
			float y = b*sin(theta)*sin(phi);

			int imgX = width/2 + osg::round((x/a)*(width)/2.0); 
			
			int imgY = height/2 + osg::round((y/b)*(height)/2.0);
	
			osg::Vec4f im = XGisImage::getImageValue(image, imgX, imgY,
				osg::Vec4f(0,0,0,0));

			float zHeight = im.r()*thickness;
			//if (!isNear) XGIS_LOG_DEBUG2("z %f",zHeight);
			
			/*fucking watch it for negative units here next time, m'kay?*/
			
			float unit; 
			if(isNear) {
				if (thickness > particleSize*2) {
					unit = zHeight*0.24;
				} else {
					unit = zHeight*0.33;
				}
			} else {
				unit = zHeight*0.49;
			}
			
			//if (!isNear) XGIS_LOG_DEBUG2("%f",unit);
			
			for (float z = 0; z < zHeight; z+=unit) {


				float offX = ((rand() % 100) / 100.0 - 0.5)*a*offPercent;
				float offY = ((rand() % 100) / 100.0 - 0.5)*b*offPercent;
				float offZ = ((rand() % 100) / 100.0 - 0.5)*zHeight*offPercent;

				if (((z + offZ) < 0) || ((z + offZ) > zHeight)) {
					z -= offZ;
				} else {
					z += offZ;
				}

				z = MAX(0, z);
				//z = MIN(zHeight, z);


				//if (!isNear) XGIS_LOG_DEBUG2("%f %f %f",x + offX, y + offY, z + offZ);

				v->push_back(osg::Vec3f(x + offX, y + offY, z));


			}

			

		}
	}

	
}

void Cumulus::placeParticles() {

	for (int cl=0; cl < subClouds; cl++) {

		/*make one "blob"*/
		osg::Geometry *subCloudNear = new osg::Geometry();
		osg::Vec3Array *verticesNear = new osg::Vec3Array();
		osg::Vec4Array *coloursNear;
		osg::ref_ptr<osg::Geode> subCloudGeodeNear = new osg::Geode();

		osg::Geometry *subCloud = new osg::Geometry();
		osg::Vec3Array *vertices = new osg::Vec3Array();
		osg::Vec4Array *colours;
		osg::ref_ptr<osg::Geode> subCloudGeode = new osg::Geode();

		osg::Geometry *subCloudFar = new osg::Geometry();
		osg::Vec3Array *verticesFar = new osg::Vec3Array();
		osg::Vec4Array *coloursFar;
		osg::ref_ptr<osg::Geode> subCloudGeodeFar = new osg::Geode();

		osg::Geometry *subCloudVeryFar = new osg::Geometry();
		osg::Vec3Array *verticesVeryFar = new osg::Vec3Array();
		osg::Vec4Array *coloursVeryFar;
		osg::ref_ptr<osg::Geode> subCloudGeodeVeryFar = new osg::Geode();
		
		float a = sizeX; float b = sizeY; float c = thickness;
		float offPercent = 0.4;

		float x1 = 600; float y1 = 4;
		float x2 = 400; float y2 = 5;
		float k = (y1 - y2)/(x1 - x2);
		float n = (y1 - k*x1);

		float deltaPhi = osg::DegreesToRadians(k*a + n);
		float deltaTheta = osg::DegreesToRadians(k*a + n);

		/*generating a random fractal image for our cloud height map*/
		osg::ref_ptr<osg::Image> image = new osg::Image;
		
		int numPixels = 129;
		image->allocateImage(numPixels,numPixels,1,GL_RGBA,GL_FLOAT);
		image->setInternalTextureFormat(GL_RGBA);
		
		XGisImage::initImage(image.get());
		XGisImage::generateFractal(image.get(), 0.8, false, true, true);
		//XGisImage::scaleImage(image.get(), numPixels-1, numPixels-1);
		XGisImage::setGrayRange(image.get(), 0.0, 1.0, true);

		/*cloud in the shape of a parametrized ellipsoid*/
		
		generateFromFractal(a,b,c,deltaPhi/2.0,deltaTheta/2.0,offPercent*0.6, 
			image.get(), true, verticesNear);
		
		generateFromFractal(a,b,c,deltaPhi,deltaTheta,offPercent*0.8, 
			image.get(), false, vertices);
		
		generateFromFractal(a,b,c,deltaPhi*2.5,deltaTheta*2.5,offPercent*1.5, 
			image.get(), false, verticesFar);
		
		generateFromFractal(a,b,c,deltaPhi*4.5,deltaTheta*4.5,offPercent*2.2, 
			image.get(), false, verticesVeryFar);
		

/*this makes a boxy blob. let's try with ellipsoids!*/
#if 0
		while (k < thickness) {

			i = 0;
			bool farX = true; 
			while (i < sizeX) {
				
				j = 0;
				c = 0;
				bool farY = true; 
				while (j < sizeY) {
					
					
					float p = ((rand() % 100)/100.0) - 0.5;
					float q = ((rand() % 100)/100.0) - 0.5;
					float r = ((rand() % 100)/100.0) - 0.5;

					float offX = i+p*unit*2;
					float offY = j+q*unit*2;
					float offZ = k + r*unit*2;

					//float opacity = (cos(osg::PI * (i/sizeX - 0.5))*cos(osg::PI * (j/sizeY - 0.5)) - 0.1)/2.0;

					XGIS_LOG_DEBUG3("%f %f %f %f",offX, offY,offZ, opacity);
					
					vertices->push_back(osg::Vec3f(offX,offY,offZ));
					//colors->push_back(osg::Vec4f(1.0,1.0,1.0,1.0));
					
					if (farX && farY && farZ) {
						XGIS_LOG_DEBUG3("far");
						verticesFar->push_back(osg::Vec3f(offX,offY,offZ));
					} else {
						XGIS_LOG_DEBUG3("not far");
					}

					farY = !(farY);
					

					j += unit;
				}
				i += unit;
				farX = !(farX);
			}
			k += unit;
			farZ = !(farZ);
		}
#endif
		double azimuth, altitude;
		Coords::GaussKr pos(sunPosition.x(), sunPosition.y());

		Pointdata::FeatureInfo::getSunPosition(pos, startTime, azimuth, altitude);

		osg::Vec3f lightColour = getLightColour(altitude);

		XGIS_LOG_DEBUG2("light %f %f %f", lightColour.x(), lightColour.y(), lightColour.z());

		coloursNear = shadeCloud(verticesNear, false, lightColour, azimuth, altitude);
		colours = shadeCloud(vertices, false, lightColour, azimuth, altitude);
		coloursFar = shadeCloud(verticesFar, false, lightColour, azimuth, altitude);
		coloursVeryFar = shadeCloud(verticesVeryFar, false, lightColour, azimuth, altitude);

		subCloudNear->setVertexArray(verticesNear);
		subCloudNear->setColorArray(coloursNear);
		subCloudNear->setColorBinding(osg::Geometry::BIND_PER_VERTEX);
		subCloudNear->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::POINTS, 0,verticesNear->size()));

		subCloud->setVertexArray(vertices);
		subCloud->setColorArray(colours);
		subCloud->setColorBinding(osg::Geometry::BIND_PER_VERTEX);
		subCloud->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::POINTS, 0,vertices->size()));

		subCloudFar->setVertexArray(verticesFar);
		subCloudFar->setColorArray(coloursFar);
		subCloudFar->setColorBinding(osg::Geometry::BIND_PER_VERTEX);
		subCloudFar->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::POINTS, 0,verticesFar->size()));

		subCloudVeryFar->setVertexArray(verticesVeryFar);
		subCloudVeryFar->setColorArray(coloursVeryFar);
		subCloudVeryFar->setColorBinding(osg::Geometry::BIND_PER_VERTEX);
		subCloudVeryFar->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::POINTS, 0,verticesVeryFar->size()));

		subCloudGeodeVeryFar->addDrawable(subCloudVeryFar);
		subCloudGeodeFar->addDrawable(subCloudFar);
		subCloudGeode->addDrawable(subCloud);
		subCloudGeodeNear->addDrawable(subCloudNear);

		osg::ref_ptr<osg::LOD> subCloudLOD = new osg::LOD();
		
		subCloudLOD->addChild(subCloudGeodeVeryFar.get(),9000.1f,1e10);
		subCloudLOD->addChild(subCloudGeodeFar.get(),3000.1f,9000.0);
		subCloudLOD->addChild(subCloudGeode.get(),1201.0f,3000.0);
		subCloudLOD->addChild(subCloudGeodeNear.get(),0.0f,1200.0);
		//subCloudLOD->addChild(imp,0.0f,1500.0);

		osg::PositionAttitudeTransform *subCloudPos = new osg::PositionAttitudeTransform();
		
		XGIS_LOG_DEBUG2("near particles: %d particles: %d far particles: %d very far particles: %d", verticesNear->size(), vertices->size(), verticesFar->size(), verticesVeryFar->size());

		//subCloudPos->addChild(subCloudGeode.get());
		subCloudPos->addChild(subCloudLOD.get());

		int maxOffset = sizeX*0.8;//osg::round(subClouds/5.0 * sizeX);

		float x = 0, y = 0;

		if (cl == 0) {
			x = 0; y = 0;
		} else if (cl == 1) {
			x = maxOffset; y = maxOffset/2.0;
		} else if (cl == 2) {
			x = 0; y = maxOffset;
		} else if (cl == 3) {
			x = -maxOffset; y = maxOffset/2.0;
		} else if (cl == 4) {
			x = -maxOffset; y = -maxOffset/2.0;
		} else if (cl == 5) {
			x = 0; y = -maxOffset;
		} else if (cl == 6) {
			x = maxOffset; y = -maxOffset/2.0;
		} 

		subCloudPos->setPosition(osg::Vec3f(x, y, 0));

		this->addChild(subCloudPos);

	}

}
