#include "config.h"

#include <xgislib/dataloader.h>
#include <xgislib/xputils.h>

#include <OpenThreads/Mutex>

LoaderTask::~LoaderTask() {
	// XGIS_LOG_INFO("Deleting loader task.");
}

DataLoaderQueue::DataLoaderQueue() {
	for(int i = 0; i < PRIORITY_LAST; i++) {
		qHead[i] = qTail[i] = NULL;
	}
	taskCount = 0;
}

void DataLoaderQueue::remove(LoaderTask *task) {
	qMut.lock();
	LoaderTask *ptr = qHead[task->staticPriority()], *pptr = NULL;
	while(NULL != ptr) {
		if(ptr == task) {
			if(NULL != pptr) {
				pptr->next = ptr->next;
			}
			else {
				qHead[task->staticPriority()] = ptr->next;
			}
			if(NULL == ptr->next) {
				qTail[task->staticPriority()] = pptr;
			}
			break;
		}
		pptr = ptr;
		ptr = ptr->next;
	}
	qMut.unlock();
}

void DataLoaderQueue::push(LoaderTask *task) {
	qMut.lock();
#ifdef XGIS_DYNAMIC_PRIORITIES
	// if dynamic priorities are enabled, the items in static priority queues
	// must be kept sorted according to dynamic priorities
	LoaderTask *ptr = qHead[task->staticPriority()], *pptr = NULL;
	while(NULL != ptr) {
		if(ptr->dynamicPriority() > task->dynamicPriority()) {
			break;
		}
		pptr = ptr;
		ptr = ptr->next;
	}
	if(NULL == pptr) {
		task->next = qHead[task->staticPriority()];
		qHead[task->staticPriority()] = task;
	}
	else {
		task->next = pptr->next;
		pptr->next = task;
	}
	if(NULL == ptr) {
		qTail[task->staticPriority()] = task;
	}
#else
	// otherwise (no dynamic priorities), we follow FIFO policy
	if(qTail[task->staticPriority()] == NULL) {
		qHead[task->staticPriority()] = task;
	}
	else {
		qTail[task->staticPriority()]->next = task;
	}
	qTail[task->staticPriority()] = task;
	task->next = NULL;
#endif /* XGIS_DYNAMIC_PRIORITIES */
	taskCount++;
	qSem.post();
	qMut.unlock();	
}

LoaderTask *DataLoaderQueue::pop() {
	LoaderTask *task = NULL;

	while(NULL == task) {
		qSem.wait();
		qMut.lock();
		for(int i = 0; i < PRIORITY_LAST; i++) {
			if(qHead[i] != NULL) {
				task = qHead[i];
				qHead[i] = qHead[i]->next;
				if(NULL == qHead[i]) {
					qTail[i] = NULL;
				}
				taskCount--;
				break;
			}
		}
		qMut.unlock();
	}

	return task;
}

LoaderTask *DataLoaderQueue::tryPop() {
	LoaderTask *task = NULL;

	if(0 == qSem.trywait()) {
		qMut.lock();
		for(int i = 0; i < PRIORITY_LAST; i++) {
			if(qHead[i] != NULL) {
				task = qHead[i];
				qHead[i] = qHead[i]->next;
				if(NULL == qHead[i]) {
					qTail[i] = NULL;
				}
				taskCount--;
				break;
			}
		}
		qMut.unlock();
	}
	return task;
}

void DataLoaderQueue::purge() {
	LoaderTask *task = NULL;
	
	qMut.lock();
	for(int i = 0; i < PRIORITY_LAST; i++) {
		while(qHead[i] != NULL) {
			task = qHead[i];
			qHead[i] = qHead[i]->next;
			taskCount--;
			delete task;
		}
	}
	qMut.unlock();
	
}

void DataLoaderThread::run() {
#	ifdef XGIS_ASYNC_LOADING
		LoaderTask *task;
		
		setSchedulePriority(
#			ifdef WIN32
				THREAD_PRIORITY_NOMINAL
#			else
				THREAD_PRIORITY_MIN
#			endif /* WIN32 */
		);
		live = true;
		XGIS_LOG_DEBUG1("Loader thread started\n");
		while(live) {
			task = source->pop();
			if(!live)
				continue;  // exit loader loop
			if(NULL != task) {
				doWork(task);
				sink->push(task);
			}
		}
		XGIS_LOG_DEBUG1("Loader thread stopped\n");
#	endif /* XGIS_ASYNC_LOADING */
}

void DataLoaderThread::start() {
#	ifdef XGIS_ASYNC_LOADING
		OpenThreads::Thread::start();
#	endif /* XGIS_ASYNC_LOADING */	
}

void DataLoaderThread::stop() {
#	ifdef XGIS_ASYNC_LOADING
		live = false;
		// TODO: need to wake up the thread waiting in source->pop()
#	endif /* XGIS_ASYNC_LOADING */
}

class LoadWorker : public DataLoaderThread {
public:
	LoadWorker(DataLoaderQueue *_source, DataLoaderQueue *_sink) :
		DataLoaderThread(_source, _sink) {
	}
	
	virtual void doWork(LoaderTask *task) {
		task->load();
	}
};

SerDataLoader::SerDataLoader(WorldStorage *_storage) : storage(_storage) {
	loadWorker = new LoadWorker(&loadQ, &updateQ);
}

void SerDataLoader::pushTask(LoaderTask *task) {
	task->setStorage(storage);

#ifdef XGIS_ASYNC_LOADING
		loadQ.push(task);
#else 
		// when using sync model, execute load() method at once
		task->load();

		// then push the task in the updateQueue (update can only
		// be performed in the update callback
		updateQ.push(task);
#endif /* XGIS_ASYNC_LOADING */
}

LoaderTask *SerDataLoader::popTask() {
	return updateQ.tryPop();
}

void SerDataLoader::start() {
	loadWorker->start();
}

void SerDataLoader::stop() {
	loadWorker->stop();
}

void SerDataLoader::purge() {
	loadQ.purge();
	updateQ.purge();
}

ParDataLoader::SingleDataLoader::SingleDataLoader(WorldStorage *_storage, DataLoaderQueue *_updateQ) 
	: updateQ(_updateQ), storage(_storage)  {
	loadWorker = new LoadWorker(&loadQ, updateQ);
}

void ParDataLoader::SingleDataLoader::pushTask(LoaderTask *task) {
	task->setStorage(storage.get());
#ifdef XGIS_ASYNC_LOADING
		loadQ.push(task);
#else 
		task->load();
		updateQ->push(task);
#endif /* XGIS_ASYNC_LOADING */
}

void ParDataLoader::SingleDataLoader::purge() {
	loadQ.purge();
}

ParDataLoader::ParDataLoader(WorldStorage *_storage) {
	addLoader(_storage);
	schedPolicy = _storage->getDefaultSchedPolicy();
	lastLoader = -1;
#	ifdef XGIS_MIPRO_GPX_FLIGHT
		startupDone = false;
#	endif /* XGIS_MIPRO_GPX_FLIGHT */

	//find alternative storages and create loaders for them
	const std::vector<WorldStorage*> &alternatives = _storage->getAlternativeStoragesList();
	for (size_t i = 0; i < alternatives.size(); i++) {
		addLoader(alternatives[i]);
	}
}

//pushes to an EXISTING loader, preferrably to i
//(so as to avoid crashes during testing of scheduling policies)
void ParDataLoader::pushClamp(int i, LoaderTask *task) {
	int iActual = osg::clampBetween(i, 0, (int)loaders.size()-1);
	if (iActual != i)
		XGIS_LOG_DEBUG3("pushing to loader %d instead of %d", iActual, i);
	loaders[iActual]->pushTask(task);
	lastLoader = iActual;
}

void ParDataLoader::pushTask(LoaderTask *task) {
#	ifdef XGIS_MIPRO_GPX_FLIGHT
		if (!startupDone) {
			pushClamp(0, task);
			return;
		}
#	endif /* XGIS_MIPRO_GPX_FLIGHT */

	switch (schedPolicy) { 
		case -1: //layers from 1, other stuff from 0
			if (task->staticPriority() == PRIORITY_LAYERS && loaders.size() > 1)
				pushClamp(1, task);
			else
				pushClamp(0, task);
			break;

		case -2: //round-robin
			if (loaders.size() == 1)
				pushClamp(0, task);
			else
				pushClamp( lastLoader%(loaders.size()-1)+1, task);
			break;

		case -3: //random
			pushClamp(rand()%(loaders.size()-1) + 1, task);
			break;
		
		case -4: { //least-busy (i.e. the one with shortest queue)
			/*XGIS_LOG_DEBUG4("q len %d %d, upd %d, all %d ",
				loaders[0]->getLoadTaskCount(), loaders[1]->getLoadTaskCount(),
				updateQ.getTaskCount(), getTaskCount());*/
			int minQL = loaders[1]->getLoadTaskCount(), minI = 1;
			for (size_t i = 2; i < loaders.size(); i++)
				if (loaders[i]->getLoadTaskCount() < minQL) {
					minQL = loaders[i]->getLoadTaskCount();
					minI = i;
				}
			pushClamp(minI, task);
			break;
		}

		default: //everything from loaders[policy]
			pushClamp(schedPolicy, task);
			break;
	}
}

void ParDataLoader::purge() {
	updateQ.purge();
	for(int i = 0; i < (int)loaders.size(); i++) {
		loaders[i]->purge();
	}
}
