#ifndef __OBSCENE_DATALOADER_H__
#  define __OBSCENE_DATALOADER_H__

#include <OpenThreads/Thread>

#include <vector>

#include <stdlib.h>

#include <xgislib/storage.h>
#include <xgislib/semaphore.h>

class LoaderTask {
	LoaderTask *next;

	WorldStorage *storage;

public:
	virtual ~LoaderTask();
	
	LoaderTask() {
		next = NULL;
		storage = NULL;
	}

	void setStorage(WorldStorage *_storage) {
		if (storage == NULL) {
			storage = _storage;
		}
		else { 
			XGIS_LOG_FATAL("LoaderTask::setStorage should only be called once!");
		}
	}

	WorldStorage *getStorage() {
		return storage;
	}

	virtual void load() = 0;
	virtual void updateWorld() = 0;

	virtual int staticPriority() = 0;
	virtual int dynamicPriority() = 0;
	
	friend class DataLoaderQueue;
};

typedef enum {
	PRIORITY_IMMEDIATE = 0,
	PRIORITY_TERRAIN_LO_RES = 1,
	PRIORITY_LAYERS = 2,
	PRIORITY_TERRAIN_HI_RES = 3,
	PRIORITY_UPDATE = 4,
	PRIORITY_LAST = 5
} DataLoaderTaskPriority;

class DataLoaderQueue {
protected:
	LoaderTask *qHead[PRIORITY_LAST], *qTail[PRIORITY_LAST];
	Semaphore qSem;
	OpenThreads::Mutex qMut;

	int taskCount;
	
public:
	DataLoaderQueue();

	void push(LoaderTask *task);
	LoaderTask *pop();
	LoaderTask *tryPop();
	
	void remove(LoaderTask *task);

	void purge();
	
	int getTaskCount() { return taskCount; }
};

class DataLoaderThread : public OpenThreads::Thread {
protected:
	bool live;
	
	virtual void doWork(LoaderTask *task) = 0;
	DataLoaderQueue *source, *sink;

public:
	DataLoaderThread(DataLoaderQueue *_source, DataLoaderQueue *_sink) :
		source(_source), sink(_sink) {
	}

	/* do the loading */
	virtual void run();
	virtual void start();
	void stop();
};

class DataLoader {
	/* abstract class that manages data loading */
public:
	virtual void pushTask(LoaderTask *task) = 0;
	virtual LoaderTask *popTask() = 0;

	virtual void start() = 0;
	virtual void stop() = 0;

	virtual void purge() = 0;
	
	virtual int getTaskCount() = 0;
};

class SerDataLoader : public DataLoader {
	/* manages data loading tasks that use a single storage */

protected:
	/* tasks waiting to load data (in data loader thread) */
	/* tasks waiting to update world (in world update callback) */ 
	DataLoaderQueue updateQ, loadQ;

	DataLoaderThread *loadWorker;
	
	WorldStorage *storage;
	
public:
	SerDataLoader(WorldStorage *_storage);

	virtual void pushTask(LoaderTask *task);
	virtual LoaderTask *popTask();
	
	virtual void start();
	virtual void stop();

	virtual void purge();
	
	virtual int getTaskCount() { return loadQ.getTaskCount() + updateQ.getTaskCount(); }	
};

class ParDataLoader : public DataLoader {
	/* distributes the incoming data loading tasts to multiple DataLoaderThreads,
	 * each using a different storage. */
	//TODO: take care of load dependencies
	//		(e.g. places should be loaded after quad, because we need to know how high to put the sign
	//TODO: prevent conflicts between DataLoaders
	//		(e.g. two DataLoaders, each using CacheStorage, starting to load the same thing
	//		 and then writing it to cache at the same time)

protected:
	class SingleDataLoader {
		/* aux class: loads data from a single storage but puts the loaded stuff
		 * into ParDataLoader's common updateQ */
		DataLoaderQueue loadQ, *updateQ;
		DataLoaderThread *loadWorker;
		osg::ref_ptr<WorldStorage> storage;

	public:
		SingleDataLoader(WorldStorage *_storage, DataLoaderQueue *_updateQ);

		virtual void pushTask(LoaderTask *task);
		
		virtual void start() { loadWorker->start(); }
		virtual void stop() { loadWorker->stop(); }
		
		virtual void purge();

		virtual int getLoadTaskCount() { return loadQ.getTaskCount(); }			
	};

	// one load queue, thread and storage per loader 
	std::vector<SingleDataLoader*> loaders;
	int schedPolicy;
	int lastLoader; //index of the last used loader
#	ifdef XGIS_MIPRO_GPX_FLIGHT
		bool startupDone;
		friend class ObsceneManipulator;
#	endif /* XGIS_MIPRO_GPX_FLIGHT */

	//a single update queue for all loaders
	DataLoaderQueue updateQ;

	void addLoader(WorldStorage *_storage) {
		SingleDataLoader *loader = new SingleDataLoader(_storage, &updateQ);
		loaders.push_back(loader);
	}
	
	void pushClamp(int i, LoaderTask *task);

public:
	ParDataLoader(WorldStorage *_storage);
	virtual ~ParDataLoader() {
		for (size_t i = 0; i < loaders.size(); i++) {
			delete loaders[i];
		}
	}

	virtual void pushTask(LoaderTask *task); //pushTask is where the scheduling will take place
	virtual LoaderTask *popTask() {
		return updateQ.tryPop();
	}
	
	virtual void start() {
		for (size_t i = 0; i < loaders.size(); i++)
			loaders[i]->start();
	}
	virtual void stop() {
		for (size_t i = 0; i < loaders.size(); i++)
			loaders[i]->stop();
	}

	virtual void purge();
	
	virtual int getTaskCount() {
		int rv = updateQ.getTaskCount();
		for (size_t i = 0; i < loaders.size(); i++)
			rv += loaders[i]->getLoadTaskCount();
		return rv;
	}
};

#endif /* !__OBSCENE_DATALOADER_H__ */
