#include "config.h"

#include "gpstrack.h"
#include "xputils.h"

#define E_GPX			(const xmlChar *)"gpx"
#define E_GPX_TRK		(const xmlChar *)"trk"
#define E_GPX_TRKSEG	(const xmlChar *)"trkseg"
#define E_GPX_TRKSPEED	(const xmlChar *)"xgisspeed"
#define E_GPX_TRKPT		(const xmlChar *)"trkpt"
#define E_GPX_LAT		(const xmlChar *)"lat"
#define E_GPX_LON		(const xmlChar *)"lon"
#define E_GPX_ELE		(const xmlChar *)"ele"

const char GpsTrack::typeName[] = "GpsTrack";

osg::Vec3 GpsTrack::parseTrkPt(xmlNodePtr node) {
	xmlChar
		*lat = xmlGetProp(node, E_GPX_LAT),
		*lon = xmlGetProp(node, E_GPX_LON),
		*ele = NULL;
	
	xmlNodePtr eleNode = node->xmlChildrenNode;
	
	while(eleNode != NULL) {
		if (0 == xmlStrcmp(eleNode->name, E_GPX_ELE)) {
			ele = xmlNodeListGetString(
					doc, eleNode->xmlChildrenNode, 1);
		}
		eleNode = eleNode->next;
	}
	//TODO: read time for each trackpoint

	Coords::WGS84 wgs84(0,0);
	float elev;
	char *endPtr;

	wgs84.latit = strtod((char *)lat, &endPtr);
	if (*endPtr != '\0') {
		wgs84.latit = NAN;
	}

	wgs84.longit = strtod((char *)lon, &endPtr);
	if (*endPtr != '\0') {
		wgs84.longit = NAN;
	}

	if (ele != NULL) {
		elev = (float)strtod((char *)ele, &endPtr);
		if (*endPtr != '\0') {
			elev = NAN;
		}
		xmlFree(ele);
	}
	else {
		elev = NAN;
	}

	xmlFree(lon);
	xmlFree(lat);

	Coords::GaussKr gk = wgs84.toGaussKr();
	
	return osg::Vec3((float)gk.x, (float)gk.y, elev);
}

bool GpsTrack::parseGpx(const char *uri) {
	speed = NAN;

	parse(uri);
	if (doc == NULL) {
		XGIS_LOG_ERROR("error parsing %s", uri);
		return false;
	}

	xmlNodePtr root = xmlDocGetRootElement(doc);
	if (0 != xmlStrcmp(root->name, E_GPX)) {
		XGIS_LOG_WARNING("root node is not %s", E_GPX);
		return false;
	}

	xmlNodePtr trk = root->xmlChildrenNode;
	while (trk != NULL) {
		if (0 == xmlStrcmp(trk->name, E_GPX_TRK)) {
			xmlNodePtr trkSeg = trk->xmlChildrenNode;
			while (trkSeg != NULL) {
				if (0 == xmlStrcmp(trkSeg->name, E_GPX_TRKSPEED)) {
					xmlChar *speedStr = xmlNodeListGetString(
						doc, trkSeg->xmlChildrenNode, 1);
					char *endPtr;
					speed = strtod((char*)speedStr, &endPtr);
					if (*endPtr != '\0') {
						XGIS_LOG_WARNING("Can't parse xgisspeed in GPX file");
						speed = NAN;
					}
					xmlFree(speedStr);
				} else if (0 == xmlStrcmp(trkSeg->name, E_GPX_TRKSEG)) {
					xmlNodePtr trkPt = trkSeg->xmlChildrenNode;
					while (trkPt != NULL) {
						if (0 == xmlStrcmp(trkPt->name, E_GPX_TRKPT)) {
							osg::Vec3 p = parseTrkPt(trkPt);
							if (isnan(p.x()) || isnan(p.y()))
								XGIS_LOG_WARNING("invalid track point: %f, %f, %f",
									p.x(), p.y(), p.z());
							points.push_back(p);
						}
						trkPt = trkPt->next;
					}
				}
				trkSeg = trkSeg->next;
			}
		}
		trk = trk->next;
	}
	return true;
}

void GpsTrack::addPoint(osg::Vec3 p) {
	points.push_back(p);
}

void GpsTrack::addPoint(Coords::GaussKr gk) {
	addPoint(osg::Vec3((float)gk.x, (float)gk.y, NAN));
}

