#include "config.h"

#include <string.h>

#include <xgislib/httpclient.h>
#include <xgislib/xputils.h>

#ifdef XGIS_HTTP_STREAMING

#	ifndef WIN32
#		include <netdb.h>
#		include <unistd.h>
#		include <netinet/in.h>
#		include <fcntl.h>
#		include <errno.h>
#		include <sys/types.h>
#		include <sys/socket.h>
#	endif /* !WIN32 */

#	define HEADER_MAX 4096
#	define HTTP_VERSION_STR "HTTP/1.1"
#	define USER_AGENT_STR "XGIS Http Client " VERSION
#	define REPLY_MAX 262144

#else

#	include <libxml/nanohttp.h>
#	include <libxml/xmlmemory.h>

#endif /* XGIS_HTTP_STREAMING */

HttpClient::HttpClient(const std::string &_server, unsigned short _port) :
	server(_server), port(_port) {

#ifdef XGIS_HTTP_STREAMING
	sSock = -1;
	buffer = new char[REPLY_MAX + 1];
	
#	ifdef WIN32
	WSADATA wsaData;
	WORD version;
	int error;
	
	version = MAKEWORD(2, 0);
	error = WSAStartup(version, &wsaData);
	/* naturally, we ignore error codes... ;) */
#	endif /* WIN32 */
	
#   ifdef XGIS_HTTP_CLIENT_THROTTLE		
	gettimeofday(&tvs, NULL);
	bytesInLastInterval = 0;
#   endif /* XGIS_HTTP_CLIENT_THROTTLE */

#endif /* XGIS_HTTP_STREAMING */
}

HttpClient::~HttpClient() {
#ifdef XGIS_HTTP_STREAMING
	delete[] buffer;
#endif /* XGIS_HTTP_STREAMING */
}

#ifdef XGIS_HTTP_STREAMING

// this is native http client (supporting streaming over keep-alive cnx)
// implementation: currently it does not build on win32, where we need to
// rewrite connect(), disconnect(), select() and read() for win32

int HttpClient::connect() {
	struct hostent *hp;
	struct sockaddr_in sAddr;
	
	/* get host info by name */
	if(NULL != (hp = gethostbyname(server.c_str()))) {
		memset((char *)&sAddr, 0, sizeof(sAddr));
		memmove((char *)&sAddr.sin_addr, hp->h_addr, hp->h_length);
		sAddr.sin_family = hp->h_addrtype;
		sAddr.sin_port = (unsigned short)htons(port);
	}
	else {
		XGIS_LOG_DEBUG5("gethostbyname() failed.\n");
		return -1; /* host not resolved */
	}

	/* create socket */
	if(!XP_SOCKET_OK((sSock = socket(AF_INET, SOCK_STREAM, 0)))) {
		XGIS_LOG_DEBUG5("socket() failed.\n");
		sSock = -1;
		return -2; /* no socket */
	}
	setsockopt(sSock, SOL_SOCKET, SO_KEEPALIVE, 0, 0);

	if(!XP_SOCKET_RV_OK(::connect(sSock, (struct sockaddr *)&sAddr, sizeof(sAddr)))) {
		xpSocketClose(sSock);
		sSock = -1;
		XGIS_LOG_DEBUG5("connect() failed.\n");
		return -3; /* connection failed */
	}

#ifdef WIN32
	u_long mode = 181;
	if(!XP_SOCKET_RV_OK(::ioctlsocket(sSock, FIONBIO, &mode))) {
#else
	int flags;
	flags = O_NONBLOCK;
	if(!XP_SOCKET_RV_OK(::fcntl(sSock, F_SETFL, flags))) {
#endif /* WIN32 */
		xpSocketClose(sSock);
		sSock = -1;
		XGIS_LOG_DEBUG5("fcntl(F_SETFL) failed.\n");
		return -5;
	}
	
	return 0;
}

bool HttpClient::isConnected() {
	return (sSock != -1);
}

int HttpClient::disconnect() {
	if(sSock != -1) {
		xpSocketClose(sSock);
		sSock = -1;
	}
	return 0;
}

static inline const char *skipWhiteSpace(const char *ptr) {
	while(*ptr == ' ' || *ptr == '\t') {
		ptr++;
	}
	return ptr;
}

static inline int hexToInt(const char *ptr) {
	int rv = 0;
	
	ptr = skipWhiteSpace(ptr);
	
	while(*ptr != '\0') {
		XGIS_LOG_DEBUG5("hextoint [%c]\n", *ptr);
		if(*ptr >= '0' && *ptr <= '9') {
			rv = 16*rv + (*ptr - '0');
		}
		else if(*ptr >= 'a' && *ptr <= 'f') {
			rv = 16*rv + 10 + (*ptr - 'a');		
		}
		else if(*ptr >= 'A' && *ptr <= 'F') {
			rv = 16*rv + 10 + (*ptr - 'A');		
		}
		else {
			break;
		}
		XGIS_LOG_DEBUG5("current = %d\n", rv);
		ptr++;
	}
	return rv;
}

static inline int decToInt(const char *ptr) {
	int rv = 0;
	
	ptr = skipWhiteSpace(ptr);
	
	while(*ptr != '\0') {
		if(*ptr >= '0' && *ptr <= '9') {
			rv = 10*rv + (*ptr - '0');
		}
		else {
			break;
		}
		ptr++;
	}
	return rv;
}

/*
 * t == SELECT_READ => read
 * t == SELECT_WRITE => write
 * returns: >0 - selected for reading
 *           0 - timeout
 *          <0 - error
 */ 
int HttpClient::select(SelectionType t) {
	fd_set fds;
	struct timeval tv;
	FD_ZERO(&fds);
	FD_SET(sSock, &fds);
	tv.tv_sec = 10; 
	tv.tv_usec = 0;

	XGIS_LOG_DEBUG5("Before select.\n");
	int n = xpSocketSelect(sSock,
		(t == SELECT_READ)?&fds:NULL,
		(t == SELECT_WRITE)?&fds:NULL,
		NULL,
		&tv);
	XGIS_LOG_DEBUG5("Selected (%d) %d.\n", t, n);
	
	return n;
}

int HttpClient::get(const std::string &_path, HttpGetCallback *cb) {
	char *header = buffer;
	const char *path = _path.c_str();
	int n, rv = 0;

	/* which parts of http response we still need to read */
	bool fStatus = true, fHeaders = true, fBody = true;
	/* buf points to the beginning of the yet-unprocessed buffer */
	char *buf = buffer;
	/* sptr points to the address where next read should start */
	/* lptr is a tmp pointer, pointing to line endings */
	/* tptr is another tmp pointer */
	char *sptr = buf, *lptr, *tptr;
	
	/* size stores size of the body (or size of the current body chunk in
	   case of chunked replies) */
	/* status stores HTTP response status (200, 404 etc.) */
	/* skip counts number of bytes that need to be skipped before reading
	   body further (CRLFs between chunks etc.) */
	/* csize stores currently written bytes */
	int size = -1, status = -1, skip = 0, csize = 0;
	/* closeAfter notifies that connection should be closed after get()
	   finishes (no-keep-alive header) */
	/* chunked denotes a chunked body (as per HTTP/1.1 rfc) */
	/* requiresRead denotes that data needs to be read from the socket
	   before body can be processed further */
	/* finally, finalChunk denotes that final chunk has been read and
	   after skipping final CRLF, get() should finish */
	bool closeAfter = false, chunked = false, requiresRead = false,
		finalChunk = false;
	
	/* assumption: buffer is large enough to hold complete headers (a few KB should do)! */
	
	mutex.lock();
	
	XGIS_LOG_DEBUG5("HttpClient.get(%s)\n", path);
	
	/* if not connected, connect */
	if(!isConnected()) {
		if(connect() < 0) {
			XGIS_LOG_DEBUG3("Connection failed.\n");
			rv = -1;
			goto l_get_done;
		}
	}

	/* format HTTP/1.1 request */
	if(HEADER_MAX < (n = snprintf(header, HEADER_MAX,
		"GET %s %s\r\n"
		"Host: %s:%d\r\n"
		"User-Agent: " USER_AGENT_STR "\r\n"
		"Connection: Keep-Alive\r\n"
		"Proxy-Connection: Keep-Alive\r\n"
		"\r\n",
		path, HTTP_VERSION_STR,
		server.c_str(), (int)port))) {
		rv = -1;
		goto l_get_done;
	}

	/* write request */
	while(n > 0) {
		if(select(SELECT_WRITE) <= 0) {
			XGIS_LOG_DEBUG3("Failed to select for writing header.");
			rv = -1;
			goto l_get_done;
		}
		int w;
		if(!XP_SOCKET_RV_IO_OK((w = xpSocketWrite(sSock, header, n)))) {
			if(XP_SOCKET_RV_AGAIN(w)) {
				continue;
			}
			XGIS_LOG_DEBUG3("Failed to write header.");
			rv = -1;
			goto l_get_done;
		}
		n -= w;
		header += w;
	}
	
	XGIS_LOG_DEBUG5("Header written.");	
	XGIS_LOG_DEBUG5("Reading status...\n");

	/* read status line */
	while(fStatus) {
		/* test connection */
		if(!isConnected()) {
			XGIS_LOG_DEBUG3("Not connected.");
			rv = -1;
			goto l_get_done;							
		}
		/* select socket for reading */
		if(select(SELECT_READ) <= 0) {
			XGIS_LOG_DEBUG3("Failed to select for reading status.");
			rv = -1;
			goto l_get_done;
		}
		/* read data while there is any */
		while(XP_SOCKET_RV_IO_OK((n = xpSocketRead(sSock, sptr, REPLY_MAX - (sptr - buffer))))) {
			XGIS_LOG_DEBUG5("Read %d.", n);
			/* null-terminate buffer for parsing */
			sptr += n;
			*sptr = '\0';
			/* determine end of status line */
			if(NULL != (lptr = strstr(buf, "\r\n"))) {
				/* found CRLF */
				*lptr = '\0';
				/* scan response status */
				if(1 != sscanf(buf, "HTTP/%*s %d", &status)) {
					/* malformed status line */
					rv = -1;
					goto l_get_done;
				}
				XGIS_LOG_DEBUG5("HTTP response status: %d", status);
				rv = status;
				/* handle response status */
				cb->handleStatus(status);
				fStatus = false;
				/* move current position after CRLF */
				buf = lptr + 2;
				break; /* out of read loop */
			}
		}
		if(XP_SOCKET_RV_CLOSE(n)) {
			/* if last read returned 0 (EOF), we have read the whole request
			   and keep-alive was not set - disconnect */
			XGIS_LOG_DEBUG5("Disconnecting.");
			disconnect();
		}
		else if(!XP_SOCKET_RV_IO_OK(n) && !XP_SOCKET_RV_AGAIN(n)) {
			/* an IO error */
			XGIS_LOG_DEBUG3("Failed to read status line.");
			rv = -1;
			goto l_get_done;
		}
	}

	/* inv: buf points at start of unprocessed data (start of first header line) */
	
	XGIS_LOG_DEBUG5("Reading headers...\n");
	assert(fStatus == false);	

	/* read headers */
	while(fHeaders) {
		/* while there are complete lines in the buffer */
		while(NULL != (lptr = strstr(buf, "\r\n"))) {
			/* found CRLF */
			if(lptr == buf) {
				/* empty line: body follows */
				fHeaders = false;
				buf += 2; /* move after CRLF */
				break; /* out of the line loop */
			}
			else {
				/* a header line */
				*lptr = '\0';
				XGIS_LOG_DEBUG5("HTTP header: '%s'\n", buf);
				if(NULL != (tptr = strstr(buf, ":"))) {
					*tptr = '\0';
					tptr++;
					if((0 == strcmp(buf, "Transfer-Encoding")) &&
						(NULL != strstr(tptr, "chunked"))) {
						chunked = true;
						size = -1;
						XGIS_LOG_DEBUG5("Got chunked body.");
					}
					else if(0 == strcmp(buf, "Content-Length")) {
						size = decToInt(tptr);
						XGIS_LOG_DEBUG5("Got body of size %d\n", size);
					}
					else if(0 == strcmp(buf, "Connection") &&
						(NULL != strstr(tptr, "close"))) {
						closeAfter = true;
						XGIS_LOG_DEBUG5("No keep-alive. Will close when done.\n");
					}
				}
				buf = lptr + 2; /* move after CRLF */
			}
		}
		if(fHeaders) {
			/* we need to read more data */
			/* test connection */
			if(!isConnected()) {
				XGIS_LOG_DEBUG3("Not connected.");
				rv = -1;
				goto l_get_done;							
			}
			/* select socket for reading */
			if(select(SELECT_READ) <= 0) {
				XGIS_LOG_DEBUG3("Failed to select for reading headers.");
				rv = -1;
				goto l_get_done;
			}
			/* read data while there is any */
			while(XP_SOCKET_RV_IO_OK(n = xpSocketRead(sSock, sptr, REPLY_MAX - (sptr - buffer)))) {
				sptr += n; /* advance end-of-data ptr */
			}
			if(XP_SOCKET_RV_CLOSE(n)) {
				/* if last read returned 0 (EOF), we have read the whole request
				   and keep-alive was not set - disconnect */
				XGIS_LOG_DEBUG5("Disconnecting.");
				disconnect();
			}
			else if(!XP_SOCKET_RV_IO_OK(n) && !XP_SOCKET_RV_AGAIN(n)) {
				/* an IO error */
				XGIS_LOG_DEBUG3("Failed to read header line: %d.", n);
				rv = -1;
				goto l_get_done;			
			}		
			*sptr = '\0'; /* null-terminate buffer for parsing */
		}
	}

	/* inv: buf points at start of unprocessed data (first byte of body) */

	XGIS_LOG_DEBUG5("Reading body...\n");
	assert(fHeaders == false);	

	/* read body */
	while(fBody) {
		requiresRead = false; /* assume we don't need to read any more data */
		/* process data while there is any in the buffer */
		while(!requiresRead) {
			if(skip > 0 && buf < sptr) {
				/* we have data that needs to be skipped */
				int rskip = buf - sptr;
				rskip = MIN(skip, rskip);
				buf += rskip;
				skip -= rskip;
			}
			if(skip == 0) {
				/* we skipped everything we needed to */
				if(finalChunk) {
					/* we skipped everything after final chunk: we're done */
					fBody = false;
					break;
				}
				if(size == -1) {
					/* it seems this is a chunked response: we need to
					   determine next chunk size */
					if(chunked) {
						/* indeed: a chunked response */
						if(NULL != (lptr = strstr(buf, "\r\n"))) {
							/* line with chunk size is in the buffer */ 
							size = hexToInt(buf); /* determine chunk size */
							buf = lptr + 2;       /* move current pos after CRLF */
							XGIS_LOG_DEBUG5("Chunk size %d\n", size);
						}
						else {
							/* not enough data in the buffer: read more */
							requiresRead = true;
						}
					}
					else {
						/* error: not a chunked response, but no content length
						   was given */
						XGIS_LOG_DEBUG3("Unable to handle non-chunked replies with unknown sizes.\n");
						rv = -1;
						goto l_get_done;			
					}			
				}
		
				if(size > 0) {
					/* there is body data waiting to be handled */ 
					csize = sptr - buf;
					if(size < csize) {
						/* determine max size of data to handle */
						csize = size;
					}
					if(csize > 0) {
						/* handle data */
						cb->handleData(buf, csize);
						buf += csize;  /* advance current pos */
						size -= csize; /* decrease size of body or chunk still
						                  remaining */
						XGIS_LOG_DEBUG5("Handled %d bytes, %d remaining.\n",
							csize, size);
						if(size == 0) {
							/* finished part of data, either chunk or complete
							   body */
							if(chunked) {
								/* chunk finished: wait for next one */
								size = -1;
								skip = 2;
							}
							else {
								/* body finished */
								fBody = false;
								break;
							}
						}
						if(buf >= sptr) {
							/* if current pos is at the end-of-data ptr,
							   we require reading from socket */
							requiresRead = true;
						}
					}
				}
				else if(size == 0) {
					/* size of body or chunk reached 0 */
					if(chunked) {
						/* last chunk: skip the trailing CRLF */
						XGIS_LOG_DEBUG5("Final chunk.\n");
						if(sptr - buf >= 2) {
							/* trailing CRLF is in the buffer: skip it */
							XGIS_LOG_DEBUG5("Done.\n");
							buf += 2;
							fBody = false;
							break;
						}
						else {
							/* we still need to read the trailing data */
							finalChunk = true;
							skip = 2;
							requiresRead = true;
						}
					}
					else {
						/* body finished */
						fBody = false;
						break;
					}
				}
			}
			else {
				/* we should have skipped some data, but there is none in
				   the buffer: require another read */
				requiresRead = true;
			}
		}
		/* we need to read more data */
		if(sptr >= buffer + REPLY_MAX) {
			/* end-of-data ptr points at the end of the complete buffer:
			   copy unprocessed data to beginning */
			lptr = buffer;
			while(buf < sptr) {
				*lptr++ = *buf++;
			}
			buf = buffer;
			sptr = lptr;
		}
		if(fBody) {
			/* test connection */
			if(!isConnected()) {
				XGIS_LOG_DEBUG3("Not connected.");
				rv = -1;
				goto l_get_done;							
			}
			/* select socket for reading */
			if(select(SELECT_READ) <= 0) {
				XGIS_LOG_DEBUG3("Failed to select for reading body.");
				rv = -1;
				goto l_get_done;
			}
			n = 1;
			/* read data while there is any */
			while((sptr < buffer + REPLY_MAX) &&
				XP_SOCKET_RV_IO_OK(n = xpSocketRead(sSock, sptr, REPLY_MAX - (sptr - buffer)))) {
				XGIS_LOG_DEBUG5("Read %d.", n);
				sptr += n; /* advance end-of-data ptr */
			}
			if(XP_SOCKET_RV_CLOSE(n)) {
				/* if last read returned 0 (EOF), we have read the whole request
				   and keep-alive was not set - disconnect */
				XGIS_LOG_DEBUG5("Disconnecting.");
				disconnect();
			}
			else if(!XP_SOCKET_RV_IO_OK(n) && !XP_SOCKET_RV_AGAIN(n)) {
				/* an IO error */
				XGIS_LOG_DEBUG3("Failed to read body: %d.", n);
				rv = -1;
				goto l_get_done;
			}	
			/* null-terminate buffer for parsing */
			*sptr = '\0';
		}
	}
		
l_get_done:
	/* we jump here at each error and reach this point after get() finishes
	   successfully */
	if(rv < 0) {
		XGIS_LOG_WARNING("Error occurred while fetching '%s'.", path);
	}

	if(rv != 200 || closeAfter) {
		/* close on any non-ok status or if keep-alive was not requested */
		disconnect();
	}
	
	mutex.unlock();

	/* return status code or <0 in case of IO error */
	return rv;
}

class HttpFileWriter : public HttpGetCallback {
	FILE *sink;
	std::string fname;
	
public:
	HttpFileWriter(const std::string &fname);
	~HttpFileWriter();
	virtual void handleStatus(int returnCode);
	virtual void handleData(const char *data, int dataSize);
};

HttpFileWriter::HttpFileWriter(const std::string &_fname) {
	sink = NULL;
	fname = _fname;
}

HttpFileWriter::~HttpFileWriter() {
	if(NULL != sink) {
		fclose(sink);
	}
}

void HttpFileWriter::handleStatus(int returnCode) {
	//FIXME: somehow return -2 from fetch() if opening sink file fails
	if(returnCode == 200) {
		sink = fopen(fname.c_str(), "w+b");
	}
}

void HttpFileWriter::handleData(const char *data, int dataSize) {
	if(NULL != sink) {
		XGIS_LOG_DEBUG5("Writing %d bytes to %s.", dataSize, fname.c_str());
		fwrite(data, dataSize, 1, sink);
	}
}

#else

#define MAX_URL 1024

int HttpClient::fetch(const std::string &path, const std::string &fname) {
/*
 *	we need this for two reasons:
 *	-xmlNanoHTTPFetch always saves the file in text mode,
 *		so binary files get corrupted on WIN32
 *	-xmlNanoHTTPFetch always saves the server's response to the file,
 *		even if the response is really an error message
*/
	// code modified from http://mail.gnome.org/archives/xml/2004-April/msg00138.html
	// Open HTTP connection
	char * pContentType = 0;
	void * ctxt = 0;
	char url[MAX_URL + 1];
	snprintf(url, MAX_URL, "http://%s:%d%s", server.c_str(), (int)port, path.c_str());
	ctxt = xmlNanoHTTPOpen(url, &pContentType);
	
	if (ctxt == 0)
		return -1;

	int rv = xmlNanoHTTPReturnCode(ctxt);

	if (rv != 200)
		return rv;

	// Open output file
	FILE *file = fopen(fname.c_str(), "wb");
	if (!file) {
		XGIS_LOG_DEBUG1("ERROR opening file %s\n", fname.c_str());
		return -2;
	}

	// Write to file
	const int SIZE = 2048;
	char buf[SIZE];
	int count;

	while ((count = xmlNanoHTTPRead(ctxt, buf, SIZE)) > 0)
		fwrite(buf, 1, count, file);
	fclose(file);

	if (count == -1)
		rv = -2;
	else
		rv = xmlNanoHTTPReturnCode(ctxt);

	if (rv != 200) {
		xpDeleteFile(std::string().append(fname));
	}

	// Free ressources
	xmlFree(pContentType);
	xmlNanoHTTPClose(ctxt);
	return rv;
}
#endif /* XGIS_HTTP_STREAMING */

int HttpClient::get(const std::string &path, const std::string &fname) {
#ifdef XGIS_HTTP_STREAMING
	HttpFileWriter w(fname);
	return get(path, &w);
#else
	return fetch(path, fname);
#endif /* XGIS_HTTP_STREAMING */
}
