#include "config.h"

#include <osg/Geode>
#include <osg/Geometry>
#include <osg/MatrixTransform>

#include <xgislib/osgutils.h>
#include <xgislib/xputils.h>
#include <xgislib/relativealtitudegeometry.h>
#include <xgislib/relativealtitudegroup.h>

#include <string.h>

// Orbit class

Orbit::Orbit()
{
	heading = M_PI/2.0;
}

void Orbit::operator()(osg::Node* node, osg::NodeVisitor* nv)
{ 
	osg::MatrixTransform *tx = dynamic_cast<osg::MatrixTransform *>(node);
	if( tx != NULL )
	{
		 heading += M_PI/180.0;
		 osg::Matrixd orbitRotation;
		 orbitRotation.makeRotate( 
				osg::DegreesToRadians(-10.0), osg::Vec3(0,1,0), // roll
				osg::DegreesToRadians(-20.0), osg::Vec3(1,0,0) , // pitch
				heading, osg::Vec3(0, 0, 1) ); // heading
		 osg::Matrixd orbitTranslation;
		 orbitTranslation.makeTranslate( 0,-40, 4 );
		 tx->setMatrix ( orbitTranslation * orbitRotation);
		 worldCoordMatrix = osg::computeLocalToWorld( nv->getNodePath() );
	}
	traverse(node, nv);
}


// misc helper functions

osg::Drawable* osgFindNamedDrawable(const std::string& searchName,
	osg::Node* currNode) {
   osg::Group* currGroup;
   osg::Drawable* foundNode;

   // check to see if we have a valid (non-NULL) node.
   // if we do have a null node, return NULL.
   if ( !currNode)
   {
      return NULL;
   }

   // We have a valid node, check to see if this is the node we 
   // are looking for. If so, return the current node.
   if (currNode->getName() == searchName)
   {
	   return dynamic_cast<osg::Drawable*>(currNode);
   }

   // We have a valid node, but not the one we are looking for.
   // Check to see if it has children (non-leaf node). If the node
   // has children, check each of the child nodes by recursive call.
   // If one of the recursive calls returns a non-null value we have
   // found the correct node, so return this node.
   // If we check all of the children and have not found the node,
   // return NULL
   
   currGroup = currNode->asGroup(); // returns NULL if not a group.
   if ( currGroup ) 
   {
      for (unsigned int i = 0 ; i < currGroup->getNumChildren(); i ++)
      { 
         foundNode = osgFindNamedDrawable(searchName, currGroup->getChild(i));
         if (foundNode)
            return foundNode; // found a match!
      }
      return NULL; // We have checked each child node - no match found.
   }
   else 
   {
	   osg::Geode* currGeode = dynamic_cast<osg::Geode*>(currNode);
	   if (currGeode)
	   {
		   for (unsigned int i=0;i<currGeode->getNumDrawables();i++)
		   {
				if (currGeode->getDrawable(i)->getName()==searchName)
				{
					foundNode = currGeode->getDrawable(i);
					return foundNode;
				}

		   }
		   return NULL;
	   } else {
		return NULL; // leaf node, no match 
	   }
   }
}

osg::Vec3 osgGetVec3FromHeightField(const osg::HeightField *hf, int col, int row) {

#	define MAX_ALLOWED_HEIGHT 100000.0 //heights above that are replaced by NaN

	osg::Vec3 rv = hf->getVertex(col, row);
	if (rv.z() > MAX_ALLOWED_HEIGHT)
		rv.z() = NAN;

	if (isnan(rv.z())) {
		//if z is NaN, replace it with average of all non-NaN neighbours
		int num = 0;
		float sum = 0, z;
		if (col > 0) {
			z = hf->getVertex(col-1, row).z();
			if (!isnan(z)) {
				num++;
				sum += z;
			}
		}
		if (col < (int)hf->getNumColumns()-1) {
			z = hf->getVertex(col+1, row).z();
			if (!isnan(z)) {
				num++;
				sum += z;
			}
		}
		if (row > 0) {
			z = hf->getVertex(col, row-1).z();
			if (!isnan(z)) {
				num++;
				sum += z;
			}
		}
		if (row < (int)hf->getNumRows()-1) {
			z = hf->getVertex(col, row+1).z();
			if (!isnan(z)) {
				num++;
				sum += z;
			}
		}
		rv.z() = sum/num;
	}
	return rv;
}

osg::Vec3 osgGetNormalFromHeightField(const osg::HeightField *hf, int col, int row) {
	int numRows = hf->getNumRows(),
		numCols = hf->getNumColumns();

	osg::Vec3 left = osgGetVec3FromHeightField(hf, col == 0 ? col : col-1, row),
		right = osgGetVec3FromHeightField(hf, col == numCols-1 ? col : col+1, row),
		up = osgGetVec3FromHeightField(hf, col, row == numRows-1 ? row : row+1),
		down = osgGetVec3FromHeightField(hf, col, row == 0 ? row : row-1);
	
	return (right-left) ^ (up-down);
}

#define SURF_HEIGHT_TOLERANCE 0.03 //meters; GLQuads differing less can be joined

osg::Node *osgCreateTransparentSurface(const osg::HeightField *hf, bool relative,
	const osg::Vec4 &baseCol, float depthNonTransparent, World *world) {

	// create a geode and set state
	osg::Geode *geode = new osg::Geode;
    osg::StateSet *statesetg = geode->getOrCreateStateSet();
    statesetg->setMode(GL_BLEND, osg::StateAttribute::ON);
    statesetg->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	int numRows = hf->getNumRows(),
		numCols = hf->getNumColumns();

	for (int col = 0; col < numCols-1; col++) {
		//build a drawable (a QUAD_STRIP) for each string of quads that contains no NANs
		int row = 0;
		osg::ref_ptr<osg::Vec3Array> verts = new osg::Vec3Array;
		osg::ref_ptr<osg::Vec3Array> normals = new osg::Vec3Array;
		osg::ref_ptr<osg::Vec4Array> colors = new osg::Vec4Array;
		while (row < numRows) {
			osg::Vec3 v1 = osgGetVec3FromHeightField(hf, col, row);
			osg::Vec3 v2 = osgGetVec3FromHeightField(hf, col+1, row);

#define OFFSET osg::Vec3(0,0,1.0)
			if (!relative) {
				v1 += OFFSET;
				v2 += OFFSET;
			}

			if (!isnan(v1.z()) && !isnan(v2.z())) {
				//valid pair of points; add them to arrays
				if (verts->size() >= 4) {
					//check whether some points can be left out 
					double z1p = (*verts)[verts->size()-2].z();
					double z1pp = (*verts)[verts->size()-4].z();
					double z2p = (*verts)[verts->size()-1].z();
					double z2pp = (*verts)[verts->size()-3].z();
					if (fabs( (z1pp + v1.z())/2 - z1p ) < SURF_HEIGHT_TOLERANCE
							&& fabs( (z2pp + v2.z())/2 - z2p ) < SURF_HEIGHT_TOLERANCE) {
						verts->pop_back();
						verts->pop_back();
						normals->pop_back();
						normals->pop_back();
						if (relative) {
							colors->pop_back();
							colors->pop_back();
						}
					}
				}

				verts->push_back(v1);
				verts->push_back(v2);
				normals->push_back( osgGetNormalFromHeightField(hf, col, row) );
				normals->push_back( osgGetNormalFromHeightField(hf, col+1, row) );

				if (relative) {
					colors->push_back(osg::Vec4(baseCol.x(), baseCol.y(), baseCol.z(),
						osg::clampBetween(baseCol.w()
							+ v1.z()/depthNonTransparent*(1.0f-baseCol.w()), 0.0f, 1.0f)));
					colors->push_back(osg::Vec4(baseCol.x(), baseCol.y(), baseCol.z(),
						osg::clampBetween(baseCol.w()
							+ v2.z()/depthNonTransparent*(1.0f-baseCol.w()), 0.0f, 1.0f)));
				}
			}
			if (row == numRows-1 || isnan(v1.z()) || isnan(v2.z())) {
				//end of column or invalid pair; make a quad strip (if nonempty), proceed to next one
				if (verts->size() >= 4) {
					osg::Geometry *terrain
						= relative ? new RelativeAltitudeGeometry : new osg::Geometry;
					terrain->setVertexArray(verts.get());
					terrain->setNormalArray(normals.get());
					terrain->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);
					terrain->setColorArray(colors.get());
					if (relative)
						terrain->setColorBinding(osg::Geometry::BIND_PER_VERTEX);
					else {
						terrain->setColorBinding(osg::Geometry::BIND_OVERALL);
						colors->push_back(baseCol);
					}
					
					osg::DrawArrays *strip = new osg::DrawArrays(osg::PrimitiveSet::QUAD_STRIP);
					strip->setCount(verts->size());
					terrain->addPrimitiveSet(strip);
					
					if (relative) {
						RelativeAltitudeGeometry *rg = dynamic_cast<RelativeAltitudeGeometry*>(terrain);
						rg->saveOriginalState(world);
						rg->recalculateHeights();
					}

					geode->addDrawable(terrain);

					verts = new osg::Vec3Array;
					normals = new osg::Vec3Array;
					colors = new osg::Vec4Array;
				} else {
					verts = new osg::Vec3Array;
					normals = new osg::Vec3Array;
					colors = new osg::Vec4Array;
				}
			}
			row++;
		}
	}

	return geode;
}

bool osgTestRectIntersection(const osg::Vec4f &b1, const osg::Vec4f &b2) {
	return /* (
			(	(b1.x() <= b2.x() && b1.x() + b1.z() >= b2.x()) ||
				(b1.x() <= b2.x() + b2.z() && b1.x() + b1.z() >= b2.x() + b2.z())	) &&
			(	(b1.y() <= b2.y() && b1.y() + b1.w() >= b2.y()) ||
				(b1.y() <= b2.y() + b2.w() && b1.y() + b1.w() >= b2.y() + b2.w())	)
			); */
		( ( (b1.x() + b1.z() >= b2.x()) && (b1.x() <= b2.x() + b2.z()) &&
		(b1.y() + b1.w() >= b2.y()) && (b1.y() <= b2.y() + b2.w()) ) );
	/* TODO: does this really cover all possible cases for intersection of two rects? */
}
	
// DeferredText

class DeferredTextUpdateCallback : public osg::Drawable::UpdateCallback {
public:
	virtual void update(osg::NodeVisitor *nv, osg::Drawable *drawable) {
		DeferredText *text = (DeferredText *)drawable;
		
		if (text->str) {
			text->setText(text->str, text->encoding);
			delete text->str;
			text->str = NULL;
		}
		text->setUpdateCallback(NULL);
	}
};

DeferredText::DeferredText(const char* _str, osgText::String::Encoding _encoding)
	: osgText::Text() {
	encoding = _encoding;
	if (_str) {
		int len = strlen(_str);
		str = new char[len+2];
		strncpy(str, _str, len+1);
		setUpdateCallback(new DeferredTextUpdateCallback());
	} else {
		str = NULL;
	}
}

DeferredText::~DeferredText() {
	if (str) {
		delete str;
	}
}


// TextPanel

class TextPanelUpdateCallback : public osg::NodeCallback {
public:
	virtual void operator()(osg::Node *node, osg::NodeVisitor *nv) {
		TextPanel *tp = (TextPanel *)node;
		DeferredText *text = tp->getText();
		
		text->update();
		osg::BoundingBox bbox = text->computeBound();
		if(bbox.valid() && ((bbox.xMax() - bbox.xMin()) > 0)) {
			osg::Vec3 panelPos = text->getPosition();
			panelPos.y() += 1.0;
			panelPos.z() += tp->border;
			panelPos.x() -= tp->border;
			float w = bbox.xMax() - bbox.xMin() + 2*tp->border;
			float h = bbox.zMax() - bbox.zMin() + 2*tp->border;
			tp->addDrawable(tp->createXZRect(panelPos, w, h, .5f));
			tp->setUpdateCallback(NULL);
		}
	}
};

TextPanel::TextPanel(const char *_str, osgText::String::Encoding _encoding, float _border) : osg::Billboard()
{
	setMode(osg::Billboard::POINT_ROT_WORLD);

	// the text
	osg::Vec4 textColor(0.0f, 0.0f, 0.0f, 1.0f);
	float textSize = 15.0f;		

	text = new DeferredText(_str, _encoding);
	text->setAxisAlignment(osgText::Text::XZ_PLANE);
	//text->setRotation(osg::Quat(osg::DegreesToRadians(90.0), osg::Vec3(1.0,0.0,0.0)));
	text->setColor(textColor);
	text->setCharacterSize(textSize);
	text->setLayout(osgText::Text::LEFT_TO_RIGHT);
	text->setAlignment(osgText::Text::LEFT_TOP);
	text->setDrawMode(osgText::Text::TEXT);
	text->setCharacterSizeMode(osgText::Text::OBJECT_COORDS /* _WITH_MAXIMUM_SCREEN_SIZE_CAPPED_BY_FONT_HEIGHT */);
	text->setMaximumHeight(textSize);

	addDrawable(text);
	
	border = _border;
	
	// render hints
	osg::StateSet *st = getOrCreateStateSet();
	st->setMode(GL_BLEND, osg::StateAttribute::ON);
    st->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
    
    setUpdateCallback(new TextPanelUpdateCallback());
}

osg::Geometry *TextPanel::createXZRect(const osg::Vec3 &pos, float w, float h, float alpha) {
	osg::Geometry *rv = new osg::Geometry();

	osg::Vec3Array *verts = new osg::Vec3Array;
	verts->push_back(pos + osg::Vec3(0.0, 0.0, 0.0));
	verts->push_back(pos + osg::Vec3(0.0, 0.0,  -h));
	verts->push_back(pos + osg::Vec3(  w, 0.0,  -h));
	verts->push_back(pos + osg::Vec3(  w, 0.0, 0.0));
	rv->setVertexArray(verts);

    osg::Vec3Array* normals = new osg::Vec3Array;
	normals->push_back(osg::Vec3(0.0,-1.0, 0.0));
	rv->setNormalArray(normals);
	rv->setNormalBinding(osg::Geometry::BIND_OVERALL);

    osg::Vec4Array* colors = new osg::Vec4Array;
    colors->push_back(osg::Vec4(1.0, 1.0, 1.0, alpha));
    rv->setColorArray(colors);
    rv->setColorBinding(osg::Geometry::BIND_OVERALL);

    rv->addPrimitiveSet(new osg::DrawArrays(GL_QUADS, 0, 4));

	return rv;
}

void TextPanel::setFont(osgText::Font *font) {
	text->setFont(font);
}
