#include "config.h"

#include "pointdata.h"

#include <sstream>
#include <time.h>
#include <string.h>

#include "xputils.h"
#include "xmlparser.h"
#include <xgislib/xgislog.h>

namespace Pointdata {
	
    static double startX; 
    static double startY; 
    static double endX;
    static double endY;
    static double xunit;
    static double yunit;
    static double quadSize;

	bool Elevations::getPointBaseData(const std::string &dirName) {
		// read map settings
		std::string baseName = dirName + "pointbase.xy";
  
		FILE *base = fopen(baseName.c_str(), "rb");
		if (base == NULL) {
			XGIS_LOG_ERROR("Failed opening '%s'!\n", baseName.c_str());
			return false;
		}
  
		fscanf(base, "%lf", &quadSize);
		fscanf(base, "%lf", &xunit);
		fscanf(base, "%lf", &startX);
		fscanf(base, "%lf", &startY);
		fscanf(base, "%lf", &endX);
		fscanf(base, "%lf", &endY);
		yunit = xunit;
		
		fclose(base);
		
		return true;
	}

	int Elevations::getQuadSize(const std::string &dirName) {
		int ret = -1;
		if (getPointBaseData(dirName+"1"+dirSlash))
		  ret = (int)quadSize;
		return ret;
	}

	double Elevations::getMaxResolution(const std::string &dirName) {
		double ret = -1.0f;

		if (getPointBaseData(dirName+"1"+dirSlash))
			ret = xunit;

		return ret;
	}

	int Elevations::getMinResolution(const std::string &dirName) {
		int ret = 1;
		
		std::string baseName = "pointbase.xy";
		std::string fullName = dirName+"8"+dirSlash+baseName;
		
		FILE *f = fopen(fullName.c_str(), "rb");
		if (f != NULL) {
			ret = 8;
		}
		else {
			fullName = dirName+"4"+dirSlash+baseName;
		
			f = fopen(fullName.c_str(), "rb");
			if (f != NULL) {
				ret = 4;
			}
		  	else {
				fullName = dirName+"2"+dirSlash+baseName;
				f = fopen(fullName.c_str(), "rb");
				if (f != NULL) {
					ret = 2;
				}
			}
		}
		
		return ret;
	}

	float *Elevations::getQuad(int u, int v, const std::string &baseDirName,
		int resolutionFactor) {
		std::ostringstream oss;
		oss<< resolutionFactor;

		std::string dirName = baseDirName+ oss.str()+dirSlash;

		getPointBaseData(dirName);

		size_t size = (size_t)(quadSize/xunit);
	    float *ret = NULL;
    
		// compose map filename
		char mapName[256]; memset(mapName,0,256);
		snprintf(mapName, 256, "%spointquad_%d_%d.xyz", dirName.c_str(),u,v);
      
		// try to open file
		FILE *map = fopen(mapName,"rb");
		if(map == NULL)
			XGIS_LOG_ERROR("Failed opening map file '%s'\n", mapName);
		else {
			ret = new float[size*size];
			XGIS_LOG_DEBUG2("Reading points from file: '%s'\n", mapName);
        
			size_t floatsRead = fread(ret, sizeof(float),size*size,map);

			if (floatsRead != size*size)
				XGIS_LOG_DEBUG2("read %d floats istead of %d\n",floatsRead,size*size);

			fclose(map);
		}

		return ret;
	}


	float *Elevations::getArea(Point ref, int width, int height, FILE* f) {
		int numX = (int)(width/xunit);
		int numY = (int)(height/yunit);
#if 0
		/* currently unused var */		
		int globalX = (int)((endX - startX)/xunit);
#endif /* 0/1 */		
		int globalY = (int)((endY - startY)/yunit);
		int count = 0;
		float *ret = new float[numX*numY];
      
		/* for each x row */
		for (int i = 0; i < numY; i++) {
			float *buffer = new float[numX];

			int column = (int)((ref.x - startX)/xunit);
			int row = (int)((ref.y - startY)/yunit + i);

			int pos = 0;
        
			pos += column*globalY*sizeof(float);
        
			pos += row*sizeof(float);

			XGIS_LOG_DEBUG3("pos %d\n",pos);
			
			fseek(f,pos,SEEK_SET);
			fread(buffer,sizeof(float),numX,f);

			for (int j = 0; j < numX; j++) {
				XGIS_LOG_DEBUG3("%f ",buffer[j]);
			}
			XGIS_LOG_DEBUG3("\n");

			int offset = count*numX;
			XGIS_LOG_DEBUG3("ret %p offset %d ret + offset %p \n", ret, offset, ret + offset);
			memcpy(ret + offset, buffer, numX*sizeof(float));
			count++;
		}

		return ret;
	}
 
	
	
	/****
	 * FeatureInfo
	 ****/

	int FeatureInfo::getMarkerPositions(const std::string &url,
		std::vector<osg::Vec3d> &list, std::vector<std::string> &list_labels) {
		std::vector<std::string> longitudes; 
		std::vector<std::string> latitudes; 
		list.empty();
	
		XmlParser parser;
		
		int ret = -1;
	  
		int dim = 0;
		
		if(parser.parse(url)) {
			dim = parser.getListNodeAttrVals("marker", "lng", longitudes);
			parser.getListNodeAttrVals("marker", "lat", latitudes);    
			parser.getListNodeAttrVals("marker", "label", list_labels);
		}
	
		if(dim > 0) {
			std::vector<std::string>::iterator itLong = longitudes.begin();
			std::vector<std::string>::iterator itLat = latitudes.begin();
			for(int i = 0; i < dim; i++) {
			//while ((itLong!=longitudes.end())&&(itLat!=latitudes.end()))
			//for (itLong=longitudes.begin(); itLong!=longitudes.end(); itLong++)

				XGIS_LOG_DEBUG3(longitudes.back().c_str());
				//std::string l = longitudes[i];
				float lng = (float)atof((*itLong).c_str());
				//longitudes.pop_back();
				float lt = (float)atof((*itLat).c_str());
				//latitudes.pop_back();

				Coords::WGS84 wgs = Coords::WGS84(lng,lt);
				Coords::GaussKr kr = wgs.toGaussKr();
				osg::Vec3d v = osg::Vec3d(kr.x, kr.y, 0.0f);
				list.push_back(v);
				itLat++;
				itLong++;
				
				
			}
			ret = dim;
		}
	
		return ret;
	}

	double toJulianDate(time_t rawtime)
	{
		tm* ptm;
		//time_t rawtime;
		//time (&rawtime);
		ptm = gmtime (&rawtime);
		
		double h = ptm->tm_hour;
		double min = ptm->tm_min;
		
		h += min/60.0;

		double d = ptm->tm_mday;

		double month = ptm->tm_mon+1;
		double j = ptm->tm_year+1900;

		double a = floor((14-month)/12.0f);
		double y = j + 4800 - a;
		double m = month + 12*a -3;

		double JDN = d + floor((153*m + 2)/5.0f) + 365*y + floor(y/4.0f) - floor(y/100.0) + floor(y/400.0) - 32045;
		double JD = JDN + (h - 12)/24.0;


		return JD;
	}

	void FeatureInfo::getMoonPosition(Coords::GaussKr place, time_t when, double& azimuth, double& altitude) {
		time (&when);
		tm* ptm;
		ptm = gmtime(&when);

		Coords::WGS84 w = place.toWGS84();

		double J = toJulianDate(when); //2354097.0;
		double phi = osg::DegreesToRadians(w.latit);//osg::DegreesToRadians(-33.92);
		/*Julian days since epoch*/
		double T = (J - 2451545.0);

		double L = 218.316 + 13.176396*T;
		double M = 134.963 + 13.064993*T;
		double F =  93.272 + 13.229350*T;

		/*ecliptic coordinates*/
		double lambda = L + 6.289*sin(M);			//longitude
		double beta = 5.128*sin(F);					//latitude
		double delta = 385001.0 + 20905.0*cos(M);	//distance
		
		double eps = osg::DegreesToRadians(23.4397);

		double declination = asin(sin(beta)*cos(eps) +cos(beta)*sin(eps)*sin(lambda));
		double ascension = atan2(sin(lambda)*cos(eps) - tan(beta)*sin(eps), cos(lambda));
		




	}

	/*AZIMUTH AND ALTITUDE ARE RETURNED IN DEGREES, NOT RADIANS!!*/
	void FeatureInfo::getSunPosition(Coords::GaussKr place, time_t when, double& azimuth, double& altitude)
	{
#if 0
		double th0 = 280.1600f;
		double th1 = 360.9856235f;
		int J2000 = 2451545;
#endif /* 0/1 */	
		tm* ptm;
		ptm = gmtime(&when);

		Coords::WGS84 w = place.toWGS84();

		double J = toJulianDate(when); //2354097.0;
		double phi = osg::DegreesToRadians(w.latit);//osg::DegreesToRadians(-33.92);
		double lw = w.longit;//-15.0;


		/*right ascension and declination in degrees*/
		double D = J - 2451545.0;
		double g = 357.529 + 0.98560028*D;
		double q = 280.459 + 0.98564736*D;
		double L = q + 1.915*sin(osg::DegreesToRadians(g)) + 0.020*sin(osg::DegreesToRadians(2*g));
		
		int q1 = (int)floor(L/360.0);
		L = L - q1*360;

#if 0
		double R = 1.00014 - 0.01671*cos(osg::DegreesToRadians(g)) - 0.00014*cos(osg::DegreesToRadians(2*g));
#endif /* 0/1 */
		double e = 23.439 - 0.00000036*D;
		double RA = atan(cos(osg::DegreesToRadians(e))*sin(osg::DegreesToRadians(L))/cos(osg::DegreesToRadians(L)));
		//double RA = atan(0.91764*tan(osg::DegreesToRadians(L)));
		
		double delta = osg::RadiansToDegrees(asin(sin(osg::DegreesToRadians(e))*sin(osg::DegreesToRadians(L))));
		double alpha = osg::RadiansToDegrees(RA);

		double lquadrant = floor(L/90.0)*90;
		double raquadrant = floor(alpha/90.0)*90;

		alpha = alpha + (lquadrant - raquadrant);

			
			//4.7585f;

		double HA = L - alpha + 180.0 + 15.0*(ptm->tm_hour+ptm->tm_min/60.0) + lw;
		double HArad = osg::DegreesToRadians(HA);
		double deltarad = osg::DegreesToRadians(delta);

		double h = asin(sin(phi)*sin(osg::DegreesToRadians(delta)) 
			+ cos(phi)*cos(osg::DegreesToRadians(delta))*cos(osg::DegreesToRadians(HA)));

		//XGIS_LOG_DEBUG2("HA %f phi %f delta %f", HArad, phi, deltarad);

		/*double A1 = sin(osg::DegreesToRadians(HA))/(cos(osg::DegreesToRadians(HA))*sin(phi) 
			- tan(osg::DegreesToRadians(delta)*cos(phi))); */
		//double A1 = sin(HArad) / ( cos(HArad)*sin(phi) - tan(deltarad)*cos(phi) );

		//HA 449.901758 phi 0.809288 delta 0.182170

		//double A = atan(A1);

		double A = atan2(sin(HArad),cos(HArad)*sin(phi) - tan(deltarad)*cos(phi));

		//XGIS_LOG_DEBUG2("A %f", A);
/*

		q1 = floor((th0 + th1*(J - J2000))/360.0f);
		
		double th = (th0 + th1*(J - J2000) - q1*360) - lw;

		
		double tau = th - alpha;

		

		double H = th - alpha;

		double A = atan2( sin(osg::DegreesToRadians(H)),
										cos(osg::DegreesToRadians(H))*sin(phi) - tan(osg::DegreesToRadians(delta))*cos(phi));

		double h = asin(sin(phi)*sin(osg::DegreesToRadians(delta)) + cos(phi)*cos(osg::DegreesToRadians(H))*cos(osg::DegreesToRadians(delta)));*/

		altitude = osg::RadiansToDegrees(h);

		/*west of south*/
		azimuth = osg::RadiansToDegrees(A);
/*		if (azimuth<0)
			azimuth += 360;
*/		
	}
	
}

