#include "config.h"

#include <osg/LineSegment>
#include <osgUtil/IntersectVisitor>

#include <xgislib/xputils.h>
#include <xgislib/osgutils.h>
#include <xgislib/quad.h>
#include <xgislib/quadsingle.h>

#ifdef XGIS_REFCOUNT_DEBUG
DebugObserver *dbgObserverQuad = NULL;
#endif /* XGIS_REFCOUNT_DEBUG */

float Quad::getTerrainHeight(float x, float y) {
	double ret = -1;

	osg::Vec3d v = osg::Vec3d(x, y, 0);
	osg::Vec3d v1 = v + osg::Vec3d(0, 0, 10000);

	osg::LineSegment* s1 = new osg::LineSegment();
	s1->set(v1, v);
	osgUtil::IntersectVisitor findHeight;
	findHeight.addLineSegment(s1);

	terrainGroup->accept(findHeight);

	osgUtil::IntersectVisitor::HitList elevLocHits;
	elevLocHits = findHeight.getHitList(s1);
	osgUtil::Hit heightTestResults;
	if (!elevLocHits.empty()) {
		heightTestResults = elevLocHits.front();
		osg::Vec3d terrainHeight =
			heightTestResults.getWorldIntersectPoint();
		return terrainHeight.z();
	}

	return ret;
}

Quad::Quad(const Coords::GaussKr &_base, int _size,
		   const osg::HeightField *cmodel, const osg::Image *ctexture)
		: osg::Group(), size(_size), base(_base) {
	// add the coarse quad
	coarse = new QuadSingle(base, size, cmodel, ctexture);
	
	terrainGroup = new osg::Group;
	
	terrainGroup->addChild(coarse.get());
	
	addChild(terrainGroup.get());
	addChild(getLayerGroup());
	
	fine = NULL;
#ifdef DEPRECATED
	loading = false;
	loadingLayers = false;
#endif /* DEPRECATED */
	showLayers = true;
	
#ifdef XGIS_REFCOUNT_DEBUG
	DebugObserver::createObserver(&dbgObserverQuad);
	dbgObserverQuad->objectCreated();
	addObserver(dbgObserverQuad);
#endif /* XGIS_REFCOUNT_DEBUG */
}

bool Quad::hasFine(int quadRes, int texRes, int subdivision) {
	// TODO: handle subdivisions

	if(fine.valid()) {
		// fine model already loaded.
		if(fine->getQuadResolution() == quadRes &&
			fine->getTextureResolution() == texRes) {
			// the same fine model already loaded: no-op
			return true;
		}
	}
	return false;
}

bool Quad::hasFine() {
	return (fine.valid());
}

void Quad::setFine(QuadSingle *single) {
	// XGIS_LOG_DEBUG4("Loading fine model: base %6f,%6f; size %d\n",
	//	base.x, base.y, storage->getQuadSize());

	if(NULL == single) {
		XGIS_LOG_WARNING("Attempting to set an invalid fine model.");
		return;
	}
	
	if(fine.valid()) {
		terrainGroup->removeChild(fine.get());
	}
	else {
		// no fine model loaded.
		terrainGroup->removeChild(coarse.get());
	}

	bool terrainChanged = (!fine.valid() || (fine->getQuadResolution() != single->getQuadResolution())); 
	fine = single;
	terrainGroup->addChild(fine.get());
	if(terrainChanged) {
		fireTerrainChangeEvent(terrainChangeHandlers);
	}
}

void Quad::clearFine() {
	// XGIS_LOG_DEBUG4("Unloading fine model: base %6f,%6f; size %d\n",
	//	base.x, base.y, storage->getQuadSize());
	
	if(!fine.valid()) {
		XGIS_LOG_DEBUG1("Fine model not loaded.\n");
		return;
	}

	terrainGroup->removeChild(fine.get());
	removeAllLayers();
	
	fine = NULL;
	terrainGroup->addChild(coarse.get());
}

void Quad::setShowLayers(bool show) {
	if(show != showLayers) {
		showLayers = show;
		if(!showLayers) {
			removeChild(getLayerGroup());
		}
		else {
			addChild(getLayerGroup());
		}
	}
}

Quad::~Quad(void) {
	removeAllLayers();
	if(fine.valid()) {
		fine = NULL;
	}
	if(coarse.valid()) {
		coarse = NULL;
	}
}

bool Quad::addLoaderTask(const std::string &taskId) {
	loaderTasks.push_back(taskId);
	return true;
}

void Quad::removeLoaderTask(const std::string &taskId) {
	std::vector<std::string>::iterator i = loaderTasks.begin();
	while(i != loaderTasks.end()) {
		if((*i) == taskId) {
			loaderTasks.erase(i);
			break;
		}
		i++;
	}
}

bool Quad::hasLoaderTask(const std::string &taskId) {
	std::vector<std::string>::iterator i = loaderTasks.begin();
	while(i != loaderTasks.end()) {
		if((*i) == taskId) {
			return true;
		}
		i++;
	}
	return false;
}

void Quad::fireTerrainChangeEvent(
	const std::vector<TerrainChangeHandler *> &handlerChain
	) {
	std::vector<TerrainChangeHandler *>::const_iterator i =
		handlerChain.begin();
	int n = 0, m = 0;
	while(i != handlerChain.end()) {
		if((*i)->handle(getTerrainBound())) {
			m++;
		}
		n++;
		i++;
	}
	XGIS_LOG_DEBUG5("Executed %d/%d terrain change handlers.", m, n);
}
