#ifndef OBSCENE_QUAD_H
#  define OBSCENE_QUAD_H 

#include <osg/LOD>

#include <xgislib/coords.h>
#include <xgislib/storage.h>
#include <xgislib/quadsingle.h>
#include <xgislib/layerset.h>

#ifdef XGIS_REFCOUNT_DEBUG
#	include <xgislib/debugobserver.h>

extern DebugObserver *dbgObserverQuad;
#endif /* XGIS_REFCOUNT_DEBUG */

class TerrainChangeHandler {
public:
	virtual osg::Vec4 getTerrainBound() = 0;
	virtual bool handle(const osg::Vec4f &bounds) = 0;
};

struct QuadC {
	int u, v;
	QuadC(int _u, int _v) : u(_u), v(_v) {}
	QuadC() : u(-1), v(-1) {}
	
	bool valid() {
		return (u != -1 && v != -1); 
	}
	
	bool operator!=(const QuadC &c) {
		return (u != c.u || v != c.v);
	}
	
	bool operator==(const QuadC &c) {
		return (u == c.u && v == c.v);
	}
};

class Quad : public osg::Group, public LayerSet {

protected:
	int size;
	bool showLayers;

	// QuadSingles
	osg::ref_ptr<QuadSingle> coarse; // coarse resolution model
	osg::ref_ptr<QuadSingle> fine;   // fine resolution model

	osg::ref_ptr<osg::Group> terrainGroup;
	
	Coords::GaussKr base;	// LL-corner

	OpenThreads::Mutex terrainChangeMutex;
	std::vector<TerrainChangeHandler *> terrainChangeHandlers;
	void fireTerrainChangeEvent(
		const std::vector<TerrainChangeHandler *> &handlerChain
		);

	std::vector<std::string> loaderTasks;
	
public:
	Quad(const Coords::GaussKr &_base,	int _size,
		const osg::HeightField *cmodel, const osg::Image *ctexture);
		// creates Quad with just the coarsest LOD (the one from coarse model)

	virtual ~Quad();

	bool hasFine(int quadRes, int texRes,     // test whether current fine
		int subdivision);                     // model is equal to given
	bool hasFine();                           // test whether quad has (any)
	                                          // fine model
	void clearFine();                         // removes fine model
	void setFine(QuadSingle *single);         // set fine model

	void setShowLayers(bool show);
	bool getShowLayers() {
		return showLayers;
	}
	
	osg::Group *getTerrainGroup() {
		return terrainGroup.get();
	}
	
	const Coords::GaussKr &getBase() { return base; }
	int getSize() { return size; }

	float getTerrainHeight(float x, float y);

	bool addLoaderTask(const std::string &taskId);
	void removeLoaderTask(const std::string &taskId);
	bool hasLoaderTask(const std::string &taskId);

	inline int getQuadResolution() {
		if(fine.valid()) {
			return fine->getQuadResolution();
		}
		else {
			return coarse->getQuadResolution();
		}
	}
	
	inline int getTextureResolution() {
		if(fine.valid()) {
			return fine->getTextureResolution();
		}
		else {
			return coarse->getTextureResolution();
		}
	}
	
	osg::Vec4 getTerrainBound() {
		return osg::Vec4f(base.x, base.y, (float)size, (float)size);
	}
	
	void addTerrainChangeHandler(TerrainChangeHandler *handler) {
		terrainChangeMutex.lock();
		terrainChangeHandlers.push_back(handler);
		terrainChangeMutex.unlock();
	}
	
	void removeTerrainChangeHandler(TerrainChangeHandler *handler) {
		terrainChangeMutex.lock();
		std::vector<TerrainChangeHandler *>::iterator i =
			terrainChangeHandlers.begin();
		while(i != terrainChangeHandlers.end()) {
			if(*i == handler) {
				terrainChangeHandlers.erase(i);
				break;
			}
			i++;
		}
		terrainChangeMutex.unlock();
	}

};

#endif /* OBSCENE_QUAD_H */
