#include "config.h"

#include <osg/Geode>
#include <osg/Geometry>
#include <osg/Texture2D>
#include <osg/ShapeDrawable>

#include <stdlib.h>
#include <string.h>

#include "quadsingle.h"
#include "xputils.h"

#ifdef XGIS_REFCOUNT_DEBUG
DebugObserver *dbgObserverQuadSingle = NULL;
#endif /* XGIS_REFCOUNT_DEBUG */

const std::string QuadSingle::terrainDescription = "XGIS::Terrain";

//extract a terrain vertex from HeightField
static osg::Vec3 getVec3(const osg::HeightField &hf, int col, int row,
	float dx, const Coords::GaussKr &base) {
	return osg::Vec3(base.x + col*dx, base.y + row*dx, hf.getVertex(col, row).z());
}

//calculate color-coding for elevation
static osg::Vec4 elev2color(double elev) {
	static const osg::Vec4 elevMap[] = {
					osg::Vec4(0.0, 0.2, 1.0, 1.0), //dark blue - sea
					osg::Vec4(0.05, 0.3, 0.0, 1.0),	//dark green
					osg::Vec4(0.15, 0.6, 0.2, 1.0),	//light green
					osg::Vec4(0.7, 0.7, 0.1, 1.0),	//yellow
					osg::Vec4(0.55, 0.55, 0.05, 1.0),	//brown
					osg::Vec4(0.6, 0.6, 0.6, 1.0),	//dark gray
					osg::Vec4(1.0, 1.0, 1.0, 1.0)};//white
	static const double elevs[] = {0, 0.5, 600, 1200, 1800, 2400, 2900};
	static const int numLevels = sizeof(elevs)/sizeof(double);

	if (elev >= elevs[numLevels-1])
		return elevMap[numLevels-1];
	if (elev <= elevs[0])
		return elevMap[0];

	int i = 0;
	while (elevs[i] < elev)
		i++;
	return elevMap[i-1] + (elevMap[i]-elevMap[i-1])
		* (elev-elevs[i-1] + (double)rand()/RAND_MAX*(elevs[i]-elevs[i-1])/500)/
		(elevs[i]-elevs[i-1]);
}

//extract a coarse terraine vertex from coarse HeightField
static osg::Vec3 getVec3fromCoarse(const osg::HeightField *coarse, int col, int row) {
	osg::Vec3 rv = coarse->getVertex(col, row);

	if (isnan(rv.z())) {
		//if z is NaN, replace it with average of all non-NaN neighbours
		int num = 0;
		float sum = 0, z;
		if (col > 0) {
			z = coarse->getVertex(col-1, row).z();
			if (!isnan(z)) {
				num++;
				sum += z;
			}
		}
		if (col < (int)coarse->getNumColumns()-1) {
			z = coarse->getVertex(col+1, row).z();
			if (!isnan(z)) {
				num++;
				sum += z;
			}
		}
		if (row > 0) {
			z = coarse->getVertex(col, row-1).z();
			if (!isnan(z)) {
				num++;
				sum += z;
			}
		}
		if (row < (int)coarse->getNumRows()-1) {
			z = coarse->getVertex(col, row+1).z();
			if (!isnan(z)) {
				num++;
				sum += z;
			}
		}
		rv.z() = sum/num;
	}
	return rv;
}

//extract color from coarse texture; if transparent, use color-coded height instead
static osg::Vec4 getColorForCoarse(const osg::Image *ctexture,
	const osg::HeightField *cmodel, int col, int row) {
	osg::Vec4 rv;

	if (NULL == ctexture) {
		rv = elev2color(cmodel->getHeight(col, row));
	} else {
		unsigned char color[4];
		memcpy(color, ctexture->data(row, col), 4);//FIXME upper and right row are missing
		if (color[3] == 0)
			rv = elev2color(cmodel->getHeight(col, row));
		else
			rv = osg::Vec4(color[0]/255.0, color[1]/255.0, color[2]/255.0, 1.0/*color[3]/255.0*/);
	}	
	return rv;
}

//push 2 coarse curtain vertices (one at terrain height, the other at 0)
static void pushCurtainCoarse(osg::Vec3Array *verts,
					   osg::Vec3Array *normals,
					   osg::Vec4Array *colors,
					   osg::Vec3 v, osg::Vec3 n, osg::Vec4 c) {
	verts->push_back(v);
	normals->push_back(osg::Vec3(0,0,1));
		/* the terrain height vertex should have the same normal as terrain
			(otherwise edge is visible), but we approximate it by "up" */
	colors->push_back(c);
	v.z() = 0;
	verts->push_back(v);
	normals->push_back(n);
		/* the 0-height vertex should have the normal facing away from the curtain plane */
	colors->push_back(c);
}


QuadSingle::QuadSingle(const Coords::GaussKr &_base,
	int quadSize, const osg::HeightField *cmodel, const osg::Image *ctexture)
		: osg::Group(), base(_base) {
	osg::Geode *geode = new osg::Geode;
	addChild(geode);
	geode->addDescription(getTerrainGeodeDescription());

	if (NULL != cmodel) {
		dx = cmodel->getXInterval();

		// load actual terrain
		osg::Geometry *terrain = new osg::Geometry;
		geode->addDrawable(terrain);

		// read verts & calculate normals
		osg::Vec3Array *verts = new osg::Vec3Array;
		osg::Vec3Array *normals = new osg::Vec3Array;
		osg::Vec4Array *colors = new osg::Vec4Array;

		float offX = base.x - cmodel->getOrigin().x(),
			offY = base.y - cmodel->getOrigin().y();
		int numRows = cmodel->getNumRows(),
			numCols = cmodel->getNumColumns();

		int lcol = (int)osg::round(offX/dx), //left col of this coarse quad
			rcol = (int)osg::round( (offX + quadSize)/dx ),
			brow = (int)osg::round(offY/dx),
			trow = (int)osg::round( (offY + quadSize)/dx ),
			numRowsQuad = (int)osg::round(quadSize/dx) + 1; //num of rows/cols in this coarse quad
		
		//inside verts
		for (int col = lcol; col <= rcol; col++) {
			for (int row = brow; row <= trow; row++) {
				verts->push_back(getVec3fromCoarse(cmodel, col, row));

				osg::Vec3 left = getVec3fromCoarse(cmodel, col == 0 ? col : col-1, row),
					right = getVec3fromCoarse(cmodel, col == numCols-1 ? col : col+1, row),
					up = getVec3fromCoarse(cmodel, col, row == numRows-1 ? row : row+1),
					down = getVec3fromCoarse(cmodel, col, row == 0 ? row : row-1);
				normals->push_back((right-left) ^ (up-down));
				colors->push_back(getColorForCoarse(ctexture, cmodel, col, row));
			}
		}
		//south curtain
		for (int col = lcol; col <= rcol; col++)
			pushCurtainCoarse(verts, normals, colors,
					getVec3fromCoarse(cmodel, col, brow), osg::Vec3(0, -1, 0),
					getColorForCoarse(ctexture, cmodel, col, brow));
		//north curtain
		for (int col = lcol; col <= rcol; col++)
			pushCurtainCoarse(verts, normals, colors,
					getVec3fromCoarse(cmodel, col, trow), osg::Vec3(0, 1, 0),
					getColorForCoarse(ctexture, cmodel, col, trow));
		//west curtain
		for (int row = brow; row <= trow; row++)
			pushCurtainCoarse(verts, normals, colors,
					getVec3fromCoarse(cmodel, lcol, row), osg::Vec3(-1, 0, 0),
					getColorForCoarse(ctexture, cmodel, lcol, row));
		//east curtain
		for (int row = brow; row <= trow; row++)
			pushCurtainCoarse(verts, normals, colors,
					getVec3fromCoarse(cmodel, rcol, row), osg::Vec3(1, 0, 0),
					getColorForCoarse(ctexture, cmodel, rcol, row));
		
		terrain->setVertexArray(verts);
		terrain->setNormalArray(normals);
		terrain->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);
		terrain->setColorArray(colors);
		terrain->setColorBinding(osg::Geometry::BIND_PER_VERTEX);

		// build GL_QUAD strips for inside vertices
		for (int i = 0; i < numRowsQuad - 1; i++) {
			osg::DrawElementsUInt *strip = new osg::DrawElementsUInt(osg::PrimitiveSet::QUAD_STRIP);
			for (int j = 0; j < numRowsQuad; j++) {
				strip->push_back(i*numRowsQuad + j);
				strip->push_back((i+1)*numRowsQuad + j);
			}
			terrain->addPrimitiveSet(strip);
		}

		//build one GL_QUAD strip for each curtain
		for (int curt = 0; curt < 4; curt++) {
			int vertsPerCurt = 2*numRowsQuad;
			osg::DrawElementsUInt *strip =
				new osg::DrawElementsUInt(osg::PrimitiveSet::QUAD_STRIP);
			for (int k = 0; k < vertsPerCurt; k++)
				strip->push_back(numRowsQuad*numRowsQuad + curt*vertsPerCurt + k);
			terrain->addPrimitiveSet(strip);
		}
	} else {
		//cmodel missing; just build a square at 0 height
		//(no curtains, so:
		//	if there is detailed terrain with negative height
		//	and cmodel missing, there will be cracks in terrain;
		//	SO WHAT!)
		dx = quadSize;
		osg::Geometry *terrain = new osg::Geometry;
		geode->addDrawable(terrain);

		osg::Vec3Array *verts = new osg::Vec3Array;
		osg::Vec3Array *normals = new osg::Vec3Array;
		osg::Vec4Array *colors = new osg::Vec4Array;

		verts->push_back(osg::Vec3(base.x, base.y, 0));
		verts->push_back(osg::Vec3(base.x+quadSize, base.y, 0));
		verts->push_back(osg::Vec3(base.x+quadSize, base.y+quadSize, 0));
		verts->push_back(osg::Vec3(base.x, base.y+quadSize, 0));
		normals->push_back(osg::Vec3(0, 0, 1));
		colors->push_back(osg::Vec4(0.2, 0.3, 0.1, 1.0));

		terrain->setVertexArray(verts);
		terrain->setNormalArray(normals);
		terrain->setNormalBinding(osg::Geometry::BIND_OVERALL);
		terrain->setColorArray(colors);
		terrain->setColorBinding(osg::Geometry::BIND_OVERALL);

		terrain->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::QUADS, 0, 4));
	}

#ifdef XGIS_REFCOUNT_DEBUG
	DebugObserver::createObserver(&dbgObserverQuadSingle);
	dbgObserverQuadSingle->objectCreated();
	addObserver(dbgObserverQuadSingle);
#endif /* XGIS_REFCOUNT_DEBUG */
}

QuadSingle::QuadSingle(const Coords::GaussKr &_base,
	int _size, int _quadResol, int _texResol,
	osg::HeightField *heightField, osg::Image *image)
		: osg::Group(), base(_base) {
	quadResol = _quadResol;
	texResol = _texResol;
	size = _size;
	dx = (double)size / (double)quadResol;

	osg::ref_ptr<osg::Geode> geode = QuadSingle::compileTerrain(heightField,
		image, size, quadResol, texResol, base); 

	
	addChild(geode.get());
	geode->addDescription(getTerrainGeodeDescription());

#ifdef XGIS_REFCOUNT_DEBUG
	DebugObserver::createObserver(&dbgObserverQuadSingle);
	dbgObserverQuadSingle->objectCreated();
	addObserver(dbgObserverQuadSingle);
#endif /* XGIS_REFCOUNT_DEBUG */
}

QuadSingle::~QuadSingle()
{
	// XGIS_LOG_INFO("Deleting QuadSingle (%lf, %lf)@%d\n", base.x, base.y, quadResol); 
}

//push 2 fine curtain vertices (one at terrain height, the other at 0)
static void pushCurtain(osg::Vec3Array *verts, osg::Vec3Array *normals,
				 osg::Vec3 v, osg::Vec3 n) {
	verts->push_back(v);
	normals->push_back(osg::Vec3(0,0,1));
	v.z() = 0;
	verts->push_back(v);
	normals->push_back(n);
}

osg::Geode *QuadSingle::compileTerrain(osg::HeightField *heightField,
	osg::Image *image,
	int quadSize, int quadResol, int texResol, const Coords::GaussKr &base) {
	float dx = (float)quadSize/(float)quadResol;

	// create a geode and geometry (actual GL_QUADS go into the geometry)
	osg::Geode *geode = new osg::Geode;

	// load actual terrain
	osg::Geometry *terrain = new osg::Geometry;
	geode->addDrawable(terrain);

	// read verts & calculate normals
	osg::Vec3Array *verts = new osg::Vec3Array;
	osg::Vec3Array *normals = new osg::Vec3Array;

	int numRows = heightField->getNumRows();
	assert(numRows = heightField->getNumColumns());
	assert(fabs(heightField->getOrigin().x() - base.x) < 0.1);
	assert(fabs(heightField->getOrigin().y() - base.y) < 0.1);
	assert(numRows == quadSize*quadResol / 1000 + 1);

	//inside vertices
	for (int col = 0; col < numRows; col++) {
		for (int row = 0; row < numRows; row++) {
			verts->push_back(getVec3(*heightField, col, row, dx, base));

			osg::Vec3 left = getVec3(*heightField, col == 0 ? col : col-1, row, dx, base),
				right = getVec3(*heightField, col == numRows-1 ? col : col+1, row, dx, base),
				up = getVec3(*heightField, col, row == numRows-1 ? row : row+1, dx, base),
				down = getVec3(*heightField, col, row == 0 ? row : row-1, dx, base);
			normals->push_back( (right-left) ^ (up-down) );
		}
	}
	//south curtain
	for (int col = 0; col < numRows; col++)
		pushCurtain(verts, normals,
				getVec3(*heightField, col, 0, dx, base), osg::Vec3(0, -1, 0));
	//north curtain
	for (int col = 0; col < numRows; col++)
		pushCurtain(verts, normals,
				getVec3(*heightField, col, numRows-1, dx, base), osg::Vec3(0, 1, 0));
	//west curtain
	for (int row = 0; row < numRows; row++)
		pushCurtain(verts, normals,
				getVec3(*heightField, 0, row, dx, base), osg::Vec3(-1, 0, 0));
	//east curtain (aka zelezna zavesa)
	for (int row = 0; row < numRows; row++)
		pushCurtain(verts, normals,
				getVec3(*heightField, numRows-1, row, dx, base), osg::Vec3(1, 0, 0));

	terrain->setVertexArray(verts);
	terrain->setNormalArray(normals);
	terrain->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);

	// calc colors or texture coords
	osg::Vec2Array *texCoords;
	osg::Vec4Array *colors = new osg::Vec4Array;
	if (image != NULL && image->valid()) {
		// if we have texture, all verts are white
		terrain->setColorBinding(osg::Geometry::BIND_OVERALL);
		colors->push_back(osg::Vec4(1,1,1,1));

		// we need texture coords
		osg::Vec2 texPixSize( //texture pixel width/height relative to the whole texture
			1.0/image->s(), 1.0/image->t());
		float //percentage of the texture's width/height that
				//corresponds to dx
			texWOfDx = 1.0 / (numRows - 1)  / (float)image->s() * (float)(image->s() - 2),
			texHOfDx = 1.0 / (numRows - 1)  / (float)image->t() * (float)(image->t() - 2);
		texCoords = new osg::Vec2Array;
		terrain->setTexCoordArray(0, texCoords);

		//texture coords of inside vertices
		for (int i = 0; i < numRows; i++) {
			for (int j = 0; j < numRows; j++) {
				texCoords->push_back(texPixSize
					+ osg::Vec2( (float)i * texWOfDx, (float)j * texHOfDx));
			}
		}
		//south curtain
		for (int i = 0; i < numRows; i++) {
			texCoords->push_back(texPixSize + osg::Vec2(
				(float)i * texWOfDx, 0 * texHOfDx));
			texCoords->push_back(texPixSize + osg::Vec2(
				(float)i * texWOfDx, 0 * texHOfDx));
		}
		//north curtain
		for (int i = 0; i < numRows; i++) {
			texCoords->push_back(texPixSize	+ osg::Vec2(
				(float)i * texWOfDx, (float)(numRows-1) * texHOfDx));
			texCoords->push_back(texPixSize	+ osg::Vec2(
				(float)i * texWOfDx, (float)(numRows-1) * texHOfDx));
		}
		//west curtain
		for (int j = 0; j < numRows; j++) {
			texCoords->push_back(texPixSize	+ osg::Vec2(
				0 * texWOfDx, (float)j * texHOfDx));
			texCoords->push_back(texPixSize	+ osg::Vec2(
				0 * texWOfDx, (float)j * texHOfDx));
		}
		//east curtain
		for (int j = 0; j < numRows; j++) {
			texCoords->push_back(texPixSize	+ osg::Vec2(
				(float)(numRows-1) * texWOfDx, (float)j * texHOfDx));
			texCoords->push_back(texPixSize	+ osg::Vec2(
				(float)(numRows-1) * texWOfDx, (float)j * texHOfDx));
		}

		// create & apply texture
		osg::Texture2D *tex = new osg::Texture2D;
		tex->setWrap(osg::Texture::WRAP_S, osg::Texture::CLAMP_TO_EDGE);
		tex->setWrap(osg::Texture::WRAP_T, osg::Texture::CLAMP_TO_EDGE);
		tex->setImage(image);
		tex->setUnRefImageDataAfterApply(true);
		osg::StateSet *state = terrain->getOrCreateStateSet();
		state->setTextureAttribute(0, tex);
		state->setTextureMode( 0, GL_TEXTURE_2D, osg::StateAttribute::ON );
	} else {
		// if no texture, just calculate colors from heights
		terrain->setColorBinding(osg::Geometry::BIND_PER_VERTEX);
		osg::Vec3Array::iterator it;
		for (it = verts->begin(); it != verts->end(); it++)
			colors->push_back(elev2color(it->z()));
	}
	terrain->setColorArray(colors);

	// build GL_QUAD strips from inside vertices
	for (int i = 0; i < numRows-1; i++) {
		osg::DrawElementsUInt *strip =
			new osg::DrawElementsUInt(osg::PrimitiveSet::QUAD_STRIP);
		for (int j = 0; j < numRows; j++) {
			strip->push_back(i*numRows + j);
			strip->push_back((i+1)*numRows + j);
		}
		terrain->addPrimitiveSet(strip);
	}

	//build one GL_QUAD strip for each curtain
	for (int curt = 0; curt < 4; curt++) {
		int vertsPerCurt = 2*numRows;
		osg::DrawElementsUInt *strip =
			new osg::DrawElementsUInt(osg::PrimitiveSet::QUAD_STRIP);
		for (int k = 0; k < vertsPerCurt; k++)
			strip->push_back(numRows*numRows + curt*vertsPerCurt + k);
		terrain->addPrimitiveSet(strip);
	}

	return geode;
}
