#include "readerwritergrd6.h"

/****************************************************************************
 *
 * OpenSceneGraph Reader/Writer plugin for "Surfer 6 Grid" format
 *
 * Written by Jaka Mocnik and Marjan Sterk, XLAB d.o.o., 2008
 *
 **********************************************************************/

struct Grd6Header {
	char id[4];
	short nx, ny;
	double xlo, xhi, ylo, yhi, zlo, zhi;

	static const char FormatID[4];
	void makeID() {
		for (int i = 0; i < sizeof(id); i++)
			id[i] = FormatID[i];
	}
	bool checkID() {
		for (int i = 0; i < sizeof(id); i++)
			if (id[i] != FormatID[i])
				return false;
		return true;
	}
};

const char Grd6Header::FormatID[] = {'D', 'S', 'B', 'B'};

ReaderWriterGrd6::ReadResult
ReaderWriterGrd6::readGrd6Stream(std::istream& fin) const {
	//read & process header
	Grd6Header header;
	fin.read((char *)(&header), sizeof(header) / sizeof(char));

	if (fin.rdstate() & std::ios::failbit) {
		osg::notify(osg::WARN) << "GRD file header cannot be read" << std::endl;
		return ReadResult::ERROR_IN_READING_FILE;
	}

	if (!header.checkID()) {
		osg::notify(osg::WARN) << "GRD file header does not begin with the correct ID" << std::endl;
		return ReadResult::FILE_NOT_HANDLED;
	}

	osg::HeightField *hf = new osg::HeightField();
	unsigned int
		colCount = header.nx,
		rowCount = header.ny;
	hf->allocate(colCount, rowCount);
	hf->setOrigin(osg::Vec3(header.xlo, header.ylo, 0));
	hf->setXInterval((header.xhi - header.xlo)/header.nx);
	hf->setYInterval((header.yhi - header.ylo)/header.ny);

	//read raw data
	float *data = new float[colCount * rowCount];
	fin.read( (char*)data, colCount * rowCount * sizeof(float) / sizeof(char) );
	if (fin.rdstate() & std::ios::failbit) {
		osg::notify(osg::WARN) << "GRD file cannot be read or ended prematurely" << std::endl;
		return ReadResult::ERROR_IN_READING_FILE;
	}

	//put data into heightField
	float *p = data;
	for (unsigned int row = 0; row < rowCount; row++) {
		for (unsigned int col = 0; col < colCount; col++) {
			hf->setHeight(col, row, *p);
			p++;
		}
	}

	delete[] data;
	return hf;
}

ReaderWriterGrd6::WriteResult
ReaderWriterGrd6::writeGrd6Stream(std::ostream &fout, const osg::HeightField &hf) const {
	//construct header 
	Grd6Header header;
	header.makeID();
	header.nx = hf.getNumColumns();
	header.ny = hf.getNumRows();
	header.xlo = hf.getOrigin().x();
	header.xhi = hf.getOrigin().x() + hf.getNumColumns()*hf.getXInterval();
	header.ylo = hf.getOrigin().y();
	header.yhi = hf.getOrigin().y() + hf.getNumRows()*hf.getYInterval();

	//copy data to raw table, and find min/max z
	header.zlo = 1e+308;
	header.zhi = -1e+308;
	float *data = new float[hf.getNumRows() * hf.getNumColumns()];

	for (unsigned int row = 0; row < hf.getNumRows(); row++)
		for (unsigned int col = 0; col < hf.getNumColumns(); col++) {
			double h = hf.getHeight(col, row);

			if (!osg::isNaN(h)) {
				if (h < header.zlo)
					header.zlo = h;
				if (h > header.zhi)
					header.zhi = h;
			}
			data[row*hf.getNumColumns() + col] = (float)h;
		}

	//write both to ostream
	fout.write((char *)(&header), sizeof(header) / sizeof(char));
	fout.write((char *)data, hf.getNumRows()*hf.getNumColumns()*sizeof(float)/sizeof(char));
	if (fout.rdstate() & std::ios::failbit) {
		osg::notify(osg::WARN) << "GRD file cannot be written" << std::endl;
		return WriteResult::ERROR_IN_WRITING_FILE;
	}

	delete[] data;
	return osgDB::ReaderWriter::WriteResult::FILE_SAVED;
}
