#include "config.h"

#include <xgislib/resourcebag.h>
#include <xgislib/storage.h>
#include <xgislib/xputils.h>

const std::string ResourceBag::fontNames[FONT_LAST] = {
	"times.ttf",
};

const std::string ResourceBag::roofTextureNames[ROOF_TEXTURE_LAST] = {
	"roof1.jpg",
	"roof2.jpg",
};

const std::string ResourceBag::wallTextureNames[WALL_TEXTURE_LAST] = {
	"brick1.jpg",
	"brick2.jpg",
	"wall1.jpg",
	"wall2.jpg",
};

ResourceBag::ResourceBag() {
	for(int i = 0; i < FONT_LAST; i++) {
		fonts[i] = NULL;
	}
	for(int i = 0; i < ROOF_TEXTURE_LAST; i++) {
		roofImages[i] = NULL;
	}
	for(int i = 0; i < WALL_TEXTURE_LAST; i++) {
		wallImages[i] = NULL;
	}
}

osgText::Font *ResourceBag::getFont(FontID id) {
	osgText::Font *rv = NULL;

	mutex.lock();

	if(id < FONT_LAST) {
		if(!fonts[id].valid()) {
			fonts[id] = osgText::readFontFile("fonts/times.ttf");			
		}
		if(fonts[id].valid()) {
			rv = fonts[id].get();
		}
	}

	mutex.unlock();

	return rv;
}

osg::Node *ResourceBag::getModel(const std::string &name,
	const std::vector<std::string> &textures, WorldStorage *storage) {
	osg::Node *model = NULL;
	std::vector<osg::ref_ptr<osg::Node> >::const_iterator i = models.begin();
	
	// first, check if model is loaded already
	while(i != models.end()) {
		if(name == (*i)->getName()) {
			model = (*i).get();
			break;
		}
		i++;
	}
	if(NULL == model) {
		// not loaded: check for textures, as we need to fetch them prior
		// to the model
		std::vector<std::string>::const_iterator ti = textures.begin();
		while(ti != textures.end()) {
			XGIS_LOG_DEBUG5("Checking texture: %s", (*ti).c_str());
			bool fetched = false;
			std::vector<std::string>::const_iterator fi = fetchedTextures.begin();
			while(fi != fetchedTextures.end()) {
				if(*fi == *ti) {
					fetched = true;
					break;
				}
				fi++;
			}
			if(!fetched) {
				XGIS_LOG_DEBUG5("Fetching texture: %s", (*ti).c_str());
				storage->getModelTexture(*ti);
				fetchedTextures.push_back(*ti);
			}
			ti++;
		}

		model = storage->getModel(name);
		if(NULL != model) {
			model->setName(name);
			models.push_back(model);
		}
	}
	return model;
}

osg::Image *ResourceBag::getRoofTexture(RoofTextureID id, WorldStorage *storage) {
	if(id < ROOF_TEXTURE_LAST) {
		if(!roofImages[id].valid()) {
			roofImages[id] = storage->getImage(roofTextureNames[id]);			
		}
		if(roofImages[id].valid()) {
			return roofImages[id].get();
		}
	}
	return NULL;
}

osg::Image *ResourceBag::getWallTexture(WallTextureID id, WorldStorage *storage) {
	if(id < WALL_TEXTURE_LAST) {
		if(!wallImages[id].valid()) {
			wallImages[id] = storage->getImage(wallTextureNames[id]);			
		}
		if(wallImages[id].valid()) {
			return wallImages[id].get();
		}
	}
	return NULL;
}
