#include "config.h"

#include <sstream>
#include <algorithm>

#include <osgDB/ReadFile>
#include <osgDB/FileUtils>

#include <xgislib/xputils.h>
#include <xgislib/storage.h>
#include <xgislib/quad.h>
#include <xgislib/xmlparser.h>

#define XGIS_MAX_CACHE_SIZE (128*1024*1024)

const char WorldStorage::typeName[] = "GpsTrack";

#define E_STORAGES		(const xmlChar *)"storages"
#define E_STORAGE		(const xmlChar *)"storage"
#define E_SCHED_POLICY	(const xmlChar *)"policy"
#define E_URL			(const xmlChar *)"url"

class StoragesXmlParser : protected XmlParser {
public:
	bool parseStoragesXml(std::string filename, std::vector<std::string> &urls, int &policy) {
		assert(urls.size() == 0);

		parse(filename);
		if (doc == NULL) {
			XGIS_LOG_FATAL("can't parse %s", filename.c_str());
			return false;
		}
		xmlNodePtr root = xmlDocGetRootElement(doc);
		if (0 != xmlStrcmp(root->name, E_STORAGES)) {
			XGIS_LOG_FATAL("root node is not %s", E_STORAGES);
			return false;
		}
		xmlChar	*polStr = xmlGetProp(root, E_SCHED_POLICY);
		if (1 != sscanf((char*)polStr, "%d", &policy))
			XGIS_LOG_ERROR("invalid policy: %s", polStr);

		xmlNodePtr stor = root->xmlChildrenNode;
		while (stor != NULL) {
			if (0 == xmlStrcmp(stor->name, E_STORAGE)) {
				xmlNodePtr url = stor->xmlChildrenNode;
				bool urlFound = false;
				while (url != NULL) {
					if (0 == xmlStrcmp(url->name, E_URL)) {
						if (urlFound)
							XGIS_LOG_WARNING("multiple URLs, only first used");
						else {
							urlFound = true;
							char *urlStr = (char*)xmlNodeListGetString(doc, url->xmlChildrenNode, 1);
							urls.push_back(std::string(urlStr));
							xmlFree(urlStr);
						}
					}
					url = url->next;
				}
			}
			stor = stor->next;
		}
		return urls.size() > 0;
	}
};	

WorldStorage *WorldStorage::createFromURL(std::string worldName, std::string urlBase) {
	WorldStorage *rv;
	if (urlBase.find("http://") == 0) {
		// net storage, possibly with cache
		rv = new HttpStorage(worldName, urlBase);
#if XGIS_MAX_CACHE_SIZE > 0
		rv = new CacheStorage((HttpStorage *)rv);
#endif /* XGIS_MAX_CACHE_SIZE > 0 */
	} else if (urlBase.find(".xml") == urlBase.length() - 4) {
		// storage list in xml file
		rv = NULL;
		std::vector<std::string> urls;
		int defPolicy;
		StoragesXmlParser parser;
		if (!parser.parseStoragesXml(urlBase, urls, defPolicy)) {
			XGIS_LOG_FATAL("No storages found!");
			return NULL;
		}
		for (size_t i = 0; i < urls.size(); i++) {
			WorldStorage *st = createFromURL(worldName, urls[i]);
			st->defaultSchedPolicy = defPolicy;
			if (rv == NULL) {
				XGIS_LOG_DEBUG1("primary storage %s", urls[i].c_str());
				rv = st;
			} else {
				XGIS_LOG_DEBUG1("alternative storage %s", urls[i].c_str());
				rv->altStorages.push_back(st);
			}
		}
	} else {
		// local storage
		rv = new LocalStorage(worldName, urlBase);
	}	
	return rv;
}

std::string WorldStorage::getImagePath(const std::string &id) const {
	std::stringstream rv;
	rv << basePath << slash() << "images" << slash() << id;
	return rv.str();
}

std::string WorldStorage::getModelPath(const std::string &id) const {
	std::stringstream rv;
	rv << basePath << slash() << "models" << slash() << id; 
	return rv.str();
}

std::string WorldStorage::getModelTexturePath(const std::string &id) const {
	std::stringstream rv;
	rv << basePath << slash() << "models" << slash() << id; 
	return rv.str();
}

std::string WorldStorage::getTexturePath(int resol, int size,
	const Coords::GaussKr &base) const {
	assert((int)base.x % size == 0);
	assert((int)base.y % size == 0);
	
	std::stringstream rv;
	rv << basePath << slash() << "worlds" << slash() <<
		worldName << slash() << "texture" << slash() << 
		size << slash() << resol << slash() << "t-" << size << "-" <<
		resol << "-" << (int)base.x << "-" << (int)base.y << ".jpg"; 
	return rv.str();
}

std::string WorldStorage::getQuadPath(int resol, int size, const Coords::GaussKr &base) const {
	assert((int)base.x % size == 0);
	assert((int)base.y % size == 0);

	std::stringstream rv;
	rv << basePath << slash() << "worlds" << slash() <<
		worldName << slash() << "3d" << slash() << 
		size << slash() << resol << slash() << "q-" << size << "-" <<
		resol << "-" << (int)base.x << "-" << (int)base.y << ".quad"; 
	return rv.str();
}

std::string WorldStorage::getLayerPath(const std::string &layerName, int size,
	const Coords::GaussKr &base) const {
	assert((int)base.x % size == 0);
	assert((int)base.y % size == 0);

	std::stringstream rv;
	rv << basePath << slash() << "worlds" << slash() <<
		worldName << slash() << "layers" << slash() << 
		layerName << slash() << size << slash() << "l-" <<
		(int)base.x << "-" << (int)base.y << ".xml"; 
	return rv.str();
}

std::string WorldStorage::getCoarseHeightsPath() const {
	std::stringstream rv;
	rv << basePath << slash() << "worlds" << slash() <<
		worldName << slash() << "coarse-model.quad"; 
	return rv.str();
}

std::string WorldStorage::getCoarseTexturePath() const {
	std::stringstream rv;
	rv << basePath << slash() << "worlds" << slash() <<
		worldName << slash() << "coarse-texture.qcol";
	return rv.str();
}

/*
* LocalStorage
* 
* WARNING: currently it lacks implementation of a number of required methods. this should be
* fixed though.
* 
* This storage expects the following directory layout (in contrast with the universe web server,
* which has the ability to indivdually configure all paths):
* 
* <dirBase>
*   +-- images
*   +-- models
*   +-- worlds
*         +-- <world-1-name>
*                  +-- 3d
*                       +-- <quad-size-1>
*                                +-- <quad-resolution-1>
*                                +-- ...
*                                +-- coarse-model.quad
*                       +-- ...
*                  +-- texture 
*                       +-- <quad-size-1>
*                                +-- <txt-resolution-1>
*                                +-- ...
*                                +-- coarse-model.cmap
*                       +-- ...
*                  +-- layers 
*                       +-- <layername>
*                                +-- <quad-size-1>
*                                +-- ...
*                       +-- .. 
*         +-- ..
* 
* directories <quad-resolution-*>, <txt-resolution-*> and layers/<layername>/<quad-size-*>
* must contain properly named quad, texture and layer files, respectively.
*/

void LocalStorage::getQuadIndex(const Coords::GaussKr &base, int size,
	int &u, int &v) const {
	u = (int)floor(base.x/size);
	v = (int)floor(base.y/size);
}

osg::Image* LocalStorage::getTexture(int resol, int size,
	const Coords::GaussKr &base) {
	return osgDB::readImageFile(getTexturePath(resol, size, base));
}

osg::HeightField* LocalStorage::getQuad(int resol, int size,
	const Coords::GaussKr &base) {
	return osgDB::readHeightFieldFile(getQuadPath(resol, size, base));
}

std::string LocalStorage::getLayer(const std::string &layer, int size,
	const Coords::GaussKr &base) {
	return getLayerPath(layer, size, base);
}

osg::HeightField *LocalStorage::getCoarseHeights() {
	return osgDB::readHeightFieldFile(getCoarseHeightsPath());
}

osg::Image *LocalStorage::getCoarseTexture() {
	return osgDB::readImageFile(getCoarseTexturePath());
}

osg::Image *LocalStorage::getImage(const std::string &id) {
	return osgDB::readImageFile(getImagePath(id));
}

osg::Node *LocalStorage::getModel(const std::string &id) {
	return osgDB::readNodeFile(getModelPath(id));
}

void LocalStorage::getModelTexture(const std::string &id) {
	// a no-op, since we will directly fetch texture from local
	// storage as model is loaded
}

/*
* HttpStorage
*/

HttpStorage::HttpStorage(const std::string &_worldName, 
	const std::string &_server, unsigned short _port,
	const std::string &_basePath) :
	WorldStorage(_worldName, _basePath), server(_server),
	port(_port) {
	std::stringstream rv;
	rv << "http://" << server << ":" << port;
	urlBase = rv.str();
}

HttpStorage::HttpStorage(const std::string &_worldName, const std::string &_urlBase) :
	WorldStorage(_worldName, "/"),
	urlBase(_urlBase), server("localhost"), port(80)  {
	char *str = strdup(urlBase.c_str()), *p1, *p2, *p3;

	if(0 != strncmp(str, "http://", strlen("http://"))) {
		assert("Only http:// URLs are accepted by HttpStorage." == NULL);
	}

	p1 = str + strlen("http://");
	p2 = strchr(p1, ':');
	if(NULL != p2) {
		*p2 = '\0';
		p2++;
		p3 = strchr(p2, '/');
		if(NULL != p3) {
			basePath = p3;
			p3 = '\0';
		}
		else {
			basePath = "/";
		}
		port = (unsigned short)::atoi(p2);
	}
	else {
		port = 80;
		p3 = strchr(p1, '/');
		if(NULL != p3) {
			basePath = p3;
			*p3 = '\0';
		}
		else {
			basePath = "/";
		}			
	}
	server = p1;

	std::stringstream rv;
	rv << "http://" << server << ":" << port;
	urlBase = rv.str();

	XGIS_LOG_DEBUG2("Server %s, port %d, base path %s\n",
		server.c_str(), (int)port, basePath.c_str());
}

std::string HttpStorage::getTextureURL(int resol, int size,
	const Coords::GaussKr &base) const {
	assert((int)base.x % size == 0);
	assert((int)base.y % size == 0);
	
	std::stringstream rv;
	rv << urlBase << getTexturePath(resol, size, base);
	return rv.str();
}

osg::Image* HttpStorage::getTexture(int resol, int size,
	const Coords::GaussKr &base) {
	return osgDB::readImageFile(getTextureURL(resol, size, base));
}

std::string HttpStorage::getQuadURL(int resol, int size,
	const Coords::GaussKr &base) const {
	assert((int)base.x % size == 0);
	assert((int)base.y % size == 0);

	std::stringstream rv;
	rv << urlBase << getQuadPath(resol, size, base);
	return rv.str();
}

osg::HeightField* HttpStorage::getQuad(int resol, int size,
	const Coords::GaussKr &base) {
	return osgDB::readHeightFieldFile(getQuadURL(resol, size, base));
}

std::string HttpStorage::getLayer(const std::string &layer, int size,
	const Coords::GaussKr &base) {
	return getLayerURL(layer, size, base);
}

std::string HttpStorage::getLayerURL(const std::string &layer, int size,
	const Coords::GaussKr &base) const {
	assert((int)base.x % size == 0);
	assert((int)base.y % size == 0);

	std::stringstream rv;
	rv << urlBase << getLayerPath(layer, size, base);
	return rv.str();
}

std::string HttpStorage::getCoarseHeightsURL() const {
	std::stringstream rv;
	rv << urlBase << getCoarseHeightsPath();
	return rv.str();
}

osg::HeightField *HttpStorage::getCoarseHeights() {
	return osgDB::readHeightFieldFile(getCoarseHeightsURL());
}

std::string HttpStorage::getCoarseTextureURL() const {
	std::stringstream rv;
	rv << urlBase << getCoarseTexturePath();
	return rv.str();
}

osg::Image *HttpStorage::getCoarseTexture() {
	return osgDB::readImageFile(getCoarseTextureURL());
}

std::string HttpStorage::getImageURL(const std::string &id) const {
	std::stringstream rv;
	rv << urlBase << getImagePath(id);
	return rv.str();
}

osg::Image *HttpStorage::getImage(const std::string &id) {
	return osgDB::readImageFile(getImageURL(id));
}

std::string HttpStorage::getModelURL(const std::string &id) const {
	std::stringstream rv;
	rv << urlBase << getModelPath(id);
	return rv.str();
}

osg::Node *HttpStorage::getModel(const std::string &id) {
	return osgDB::readNodeFile(getModelURL(id));
}

void HttpStorage::getModelTexture(const std::string &id) {
	// a no-op, since we will directly fetch texture from HTTP
	// as model is loaded (using osgDB logic)
}

/****
 * CacheStorage
 ****/

static const std::string
	TMP_DIR_NAME = "xgis-cache",

	CS_TXT_PREFIX = "txt",
	CS_QUAD_PREFIX = "quad",
	CS_COMPILED_QUAD_PREFIX = "cquad",
	CS_CHEIGHTS_PREFIX = "cheights",
	CS_CTEX_PREFIX = "ctex",
	CS_LAYER_PREFIX = "layer",

	CS_TXT_EXT = "jpg",
	CS_QUAD_EXT = "quad",
	CS_CHEIGHTS_EXT = "quad",
	CS_CTEX_EXT = "qcol",
	CS_COMPILED_EXT = "ive",
	CS_LAYER_EXT = "xml";

bool CacheStorage::initTempDirectory() {
	std::string tmpPath = xpGetTempDirectory();
	XGIS_LOG_DEBUG4("System temp directory: %s\n", tmpPath.c_str());
	
	tmpPath = tmpPath + dirSlash + TMP_DIR_NAME;
#	ifdef XGIS_UNIQUE_CACHE_DIR
		if(!xpPathExists(tmpPath)) {
			XGIS_LOG_WARNING("Tmp dir does not exist.\n");
			if(!xpMakeDir(tmpPath)) {
				XGIS_LOG_ERROR("Failed to create tmp dir.");
				return false;
			}
		}
		else if(!xpIsDirectory(tmpPath)) {
			XGIS_LOG_ERROR("Tmp dir path exists but is not a directory.");
			return false;
		}
		//find a non-existant cache subdirectory to use
		for (int i = 0; true; i++) {
			char buf[40];
			sprintf(buf, "%d", i);
			if (!xpPathExists(tmpPath + dirSlash + buf)) {
				if (xpMakeDir(tmpPath + dirSlash + buf)) {
					tmpPath = tmpPath + dirSlash + buf;
					break;
				}
			}
		}
#	endif /* XGIS_UNIQUE_CACHE_DIR */
	XGIS_LOG_DEBUG4("Tmp dir: %s\n", tmpPath.c_str());
	
	if(!xpPathExists(tmpPath)) {
		XGIS_LOG_WARNING("Tmp dir does not exist.\n");
		if(!xpMakeDir(tmpPath)) {
			XGIS_LOG_ERROR("Failed to create tmp dir.");
			return false;
		}
	}
	else if(!xpIsDirectory(tmpPath)) {
		XGIS_LOG_ERROR("Tmp dir path exists but is not a directory.");
		return false;
	}

	tempDirPath = tmpPath;

	// set osgDB data file path to cache dir
	osgDB::FilePathList dbPaths;
	dbPaths.push_back(tempDirPath);
	osgDB::setDataFilePathList(dbPaths);
	
	/* TODO: read current contents of the temp directory and fill the
	 * cache data with the filenames! */

	return true;
}

void CacheStorage::addEntry(const std::string &fname) {
	int size = xpGetFileSize(fname);
	if(-1 == size)
		return;
	CacheEntry *entry = new CacheEntry(fname, size, time(NULL));
	entriesByTime.push_back(entry);
	totalSize += size;
	XGIS_LOG_DEBUG4("Added entry %s @ %d\n",
		entry->fname.c_str(), (int)entry->lastAccess);
	purgeCache();
}

void CacheStorage::purgeCache() {
	XGIS_LOG_DEBUG4("Current cache size %d, max %d\n", totalSize, maxSize);
	while(totalSize > maxSize && !entriesByTime.empty()) {
		CacheEntry *entry = *entriesByTime.begin();
		entriesByTime.erase(entriesByTime.begin());
		totalSize -= entry->size;
		xpDeleteFile(entry->fname);
		XGIS_LOG_DEBUG4("Purging entry %s @ %d\n",
			entry->fname.c_str(), (int)entry->lastAccess);  
		delete entry;
		XGIS_LOG_DEBUG4("Current cache size %d, max %d\n", totalSize, maxSize);
	}
}

void CacheStorage::markAccess(const std::string &fname) {
	// TODO: ignore until we have a hashtable impl. long vectors take way too much time.
	/*
	std::vector<CacheEntry *>::iterator i = entriesByTime.begin();
	while(i != entriesByTime.end()) {
		CacheEntry *entry = *i;
		if(entry->fname == fname) {
			entry->lastAccess = time(NULL);
			XGIS_LOG_DEBUG3("Marked entry %s @ %d\n",
				entry->fname.c_str(), (int)entry->lastAccess);
			entriesByTime.erase(i);
			entriesByTime.push_back(entry);		
			break;
		}
		i++;
	}
	*/
}

std::string CacheStorage::buildCFNameTexture(int resol, int size,
	const Coords::GaussKr &base) const {
	assert((int)base.x % size == 0);
	assert((int)base.y % size == 0);

	std::stringstream rv;
	rv << tempDirPath << dirSlash << storage->getWorldName()
		<< '-' << CS_TXT_PREFIX
		<< '-' << size
		<< '-' << resol
		<< '-' << (int)base.x
		<< '-' << (int)base.y
		<< '.' << CS_TXT_EXT;
	return rv.str();
}

std::string CacheStorage::buildCFNameQuad(int resol, int size,
	const Coords::GaussKr &base) const {
	assert((int)base.x % size == 0);
	assert((int)base.y % size == 0);

	std::stringstream rv;
	rv << tempDirPath << dirSlash << storage->getWorldName()
		<< '-' << CS_QUAD_PREFIX
		<< '-' << size
		<< '-' << resol
		<< '-' << (int)base.x
		<< '-' << (int)base.y
		<< '.' << CS_QUAD_EXT;
	return rv.str();
}


std::string CacheStorage::buildCFNameLayer(const std::string &layer, int size,
	const Coords::GaussKr &base) const {
	assert((int)base.x % size == 0);
	assert((int)base.y % size == 0);

	std::stringstream rv;
	rv << tempDirPath << dirSlash << storage->getWorldName()
		<< '-' << CS_LAYER_PREFIX
		<< '-' << size
		<< '-' << layer
		<< '-' << (int)base.x
		<< '-' << (int)base.y
		<< '.' << CS_LAYER_EXT;
	return rv.str();
}

std::string CacheStorage::buildCFNameCoarseHeights() const {
	std::stringstream rv;
	rv << tempDirPath << dirSlash << storage->getWorldName()
		<< '-' << CS_CHEIGHTS_PREFIX
		<< '.' << CS_CHEIGHTS_EXT;
	return rv.str();
}

std::string CacheStorage::buildCFNameCoarseTexture() const {
	std::stringstream rv;
	rv << tempDirPath << dirSlash << storage->getWorldName()
		<< '-' << CS_CTEX_PREFIX
		<< '.' << CS_CTEX_EXT;
	return rv.str();
}

std::string CacheStorage::buildCFNameData(const std::string &fname) const {
	std::stringstream rv;
	rv << tempDirPath << dirSlash << fname;
	return rv.str();
}

osg::Image* CacheStorage::getTexture(int resol,	int size,
	const Coords::GaussKr &base) {
	std::string cfname = buildCFNameTexture(resol, size, base);

	osg::Image *rv = NULL;
	
	if(!xpIsFile(cfname)) {
		std::string URL = storage->getTexturePath(resol, size, base);
		if(200 != httpClient.get(URL, cfname)) {
			XGIS_LOG_DEBUG3("Error fetching file: %s from %s\n",
				URL.c_str(), storage->getServer().c_str());
			return NULL;
		}
	}

	if(xpIsFile(cfname)) {
		rv = osgDB::readImageFile(cfname);
		markAccess(cfname);
	}
	return rv;
}

osg::HeightField* CacheStorage::getQuad(int resol, int size,
	const Coords::GaussKr &base) {
	std::string cfname = buildCFNameQuad(resol, size, base);

	osg::HeightField *rv = NULL;

	if(!xpIsFile(cfname)) {
		std::string URL = storage->getQuadPath(resol, size, base); 
		if(200 != httpClient.get(URL, cfname)) {
			XGIS_LOG_DEBUG3("Error fetching file: %s from %s\n",
				URL.c_str(), storage->getServer().c_str());
			return NULL;
		}
	}

	if(xpIsFile(cfname)) {
		rv = osgDB::readHeightFieldFile(cfname);
		markAccess(cfname);
	}
	return rv;
}

std::string CacheStorage::getLayer(const std::string &layer, int size,
	const Coords::GaussKr &base) {
	std::string cfname = buildCFNameLayer(layer, size, base);

	if (!xpIsFile(cfname)) {
		std::string URL = storage->getLayerPath(layer, size, base);
		int httpRes = httpClient.get(URL, cfname);
		if(200 != httpRes) {
			if (404 == httpRes) {
				XGIS_LOG_DEBUG5("Can't fetch file: %s, supposing (and caching) "
								"that this layer has no data in this quad\n", URL.c_str());
				FILE *cf = fopen(cfname.c_str(), "w");
				fclose(cf);
			} else {
				XGIS_LOG_DEBUG3("Error fetching file: %s from %s\n",
					URL.c_str(), storage->getServer().c_str());
				return "";
			}
		}
	}

	if(xpIsFile(cfname)) {
		markAccess(cfname);
	}

	return cfname;
}

osg::HeightField* CacheStorage::getCoarseHeights() {
	std::string cfname = buildCFNameCoarseHeights();

	osg::HeightField *rv = NULL;

	if(!xpIsFile(cfname)) {
		std::string URL = storage->getCoarseHeightsPath(); 
		if(200 != httpClient.get(URL, cfname)) {
			XGIS_LOG_DEBUG3("Error fetching file: %s from %s\n",
				URL.c_str(), storage->getServer().c_str());
			return NULL;
		}
	}

	if(xpIsFile(cfname)) {
		rv = osgDB::readHeightFieldFile(cfname);
		markAccess(cfname);
	}
	return rv;
}

osg::Image* CacheStorage::getCoarseTexture() {
	std::string cfname = buildCFNameCoarseTexture();

	osg::Image *rv = NULL;
	
	if(!xpIsFile(cfname)) {
		std::string URL = storage->getCoarseTexturePath();
		if(200 != httpClient.get(URL, cfname)) {
			XGIS_LOG_DEBUG3("Error fetching file: %s from %s\n",
				URL.c_str(), storage->getServer().c_str());
			return NULL;
		}
	}

	if(xpIsFile(cfname)) {
		rv = osgDB::readImageFile(cfname);
		markAccess(cfname);
	}
	return rv;
}

osg::Image *CacheStorage::getImage(const std::string &id) {
	std::string cfname = buildCFNameData(id);

	osg::Image *rv = NULL;

	if(!xpIsFile(cfname)) {
		std::string URL = storage->getImagePath(id);
		if(200 != httpClient.get(URL, cfname)) {
			XGIS_LOG_DEBUG3("Error fetching file: %s from %s\n",
				URL.c_str(), storage->getServer().c_str());
			return NULL;
		}
	}
	
	if(xpIsFile(cfname)) {
		rv = osgDB::readImageFile(cfname);
		markAccess(cfname);
	}
	return rv;
}

osg::Node *CacheStorage::getModel(const std::string &id) {
	std::string cfname = buildCFNameData(id);

	osg::Node *rv = NULL;
	
	if(!xpIsFile(cfname)) {
		std::string URL = storage->getModelPath(id); 
		if(200 != httpClient.get(URL, cfname)) {
			XGIS_LOG_DEBUG3("Error fetching file: %s from %s\n",
				URL.c_str(), storage->getServer().c_str());
			return NULL;
		}
	}

	if(xpIsFile(cfname)) {
		rv = osgDB::readNodeFile(cfname);
		markAccess(cfname);
	}
	return rv;
}

void CacheStorage::getModelTexture(const std::string &id) {
	std::string cfname = buildCFNameData(id);
	
	if(!xpIsFile(cfname)) {
		std::string URL = storage->getModelTexturePath(id); 
		if(200 != httpClient.get(URL, cfname)) {
			XGIS_LOG_DEBUG3("Error fetching file: %s from %s\n",
				URL.c_str(), storage->getServer().c_str());
			return;
		}
	}

	if(xpIsFile(cfname)) {
		markAccess(cfname);
	}
}
