#ifndef __XGISLIB_STORAGE_H__
#  define __XGISLIB_STORAGE_H__

#include <vector>
#include <string>
#include <sstream>

#include <assert.h>

#include <osg/Node>
#include <osg/Shape>
#include <osg/Image>
#include <osg/Geode>

#include <xgislib/coords.h>
#include <xgislib/xputils.h>
#include <xgislib/httpclient.h>

// abstract class whose descendants describe where the world is stored
// and can provide us the world data
class WorldStorage : public osg::Referenced {
private:
	static const char typeName[];
	std::vector<WorldStorage*> altStorages; // list of alternative storages
	int defaultSchedPolicy;                 // default policy for scheduling
	                                        // among storages (irrelevant in
	                                        // case of only one storage)

protected:
	std::string worldName;
	std::string basePath;

	virtual char slash() const = 0; //should return appropriate dir-slash, i.e. / for Http, platfrom-dep. for Local

	WorldStorage(const std::string &_worldName, const std::string &_basePath) :
		defaultSchedPolicy(0), worldName(_worldName), basePath(_basePath) {
	}

public:
	virtual const char *getTypeName() { return typeName; }

	//creates a 'most appropriate' storage:
	//-if urlBase starts with http://, CacheStorage
	//-elsif urlBase ends with .xml, reads storage URL and list of alternatives from the file
	//-else LocalStorage
	static WorldStorage *createFromURL(std::string worldName, std::string urlBase);

	virtual ~WorldStorage() {}

	inline std::string getWorldName() const { return worldName; }
	inline std::string getBasePath() const { return basePath; }

	const std::vector<WorldStorage*> getAlternativeStoragesList() const {
		return altStorages;
	}

	int getDefaultSchedPolicy() const {
		return defaultSchedPolicy;
	}

	// quad fine data: heights, texture, and buildings
	virtual osg::HeightField *getQuad(int resol, int size, const Coords::GaussKr &base) = 0;
	virtual osg::Image *getTexture(int resol, int size, const Coords::GaussKr &base) = 0;

	virtual std::string getLayer(const std::string &layer, int size, const Coords::GaussKr &base) = 0;

	// quad coarse data
	virtual osg::HeightField *getCoarseHeights() = 0;	// coarse model of the whole world, aka "hen"
	virtual osg::Image *getCoarseTexture() = 0;

	virtual osg::Image *getImage(const std::string &id) = 0; // images like sky texture etc
	virtual osg::Node *getModel(const std::string &id) = 0;  // 3ds models
	virtual void getModelTexture(const std::string &id) = 0; // model texture
	
	std::string getTexturePath(int resol, int size, const Coords::GaussKr &base) const;
	std::string getQuadPath(int resol, int size, const Coords::GaussKr &base) const;
	std::string getLayerPath(const std::string &layer, int size,
		const Coords::GaussKr &base) const;
	std::string getCoarseHeightsPath() const;
	std::string getCoarseTexturePath() const;
	std::string getImagePath(const std::string &id) const;
	std::string getModelPath(const std::string &id) const;
	std::string getModelTexturePath(const std::string &id) const;

};

class LocalStorage : public WorldStorage {
private:
	void getQuadIndex(const Coords::GaussKr &base, int size, int &u, int &v) const;

protected:
	char slash() const { return dirSlash; }

public:
	LocalStorage(const std::string &_worldName, const std::string &_basePath) :
		WorldStorage(_worldName, _basePath) {
	}
	virtual ~LocalStorage() {}

	virtual osg::HeightField* getQuad(int resol, int size, const Coords::GaussKr &base);	
	virtual osg::Image* getTexture(int resol, int size, const Coords::GaussKr &base);

	virtual std::string getLayer(const std::string &layer, int size,
		const Coords::GaussKr &base);

	virtual osg::HeightField *getCoarseHeights();
	virtual osg::Image *getCoarseTexture();

	virtual osg::Image *getImage(const std::string &id);
	virtual osg::Node *getModel(const std::string &id);
	virtual void getModelTexture(const std::string &id);

};

class HttpStorage : public WorldStorage {
protected:
	char slash() const { return '/'; }

private:
	std::string urlBase, server;
	unsigned short port;

public:
	HttpStorage(const std::string &_worldName,
		const std::string &_server, unsigned short _port,
		const std::string &_basePath);
	HttpStorage(const std::string &_worldName, const std::string &_urlBase);
	virtual ~HttpStorage() {}
	
	virtual int getAlternativeStoragesList(std::vector<WorldStorage*> &list) {
		// TODO: get a real list from somewhere, e.g. server-supplied XML
		return 0;
	}

	virtual osg::HeightField* getQuad(int resol, int size, const Coords::GaussKr &base);
	virtual osg::Image* getTexture(int resol, int size, const Coords::GaussKr &base);

	virtual std::string getLayer(const std::string &layer, int size,
		const Coords::GaussKr &base);

	virtual osg::HeightField *getCoarseHeights();
	virtual osg::Image *getCoarseTexture();

	virtual osg::Image *getImage(const std::string &id);
	virtual osg::Node *getModel(const std::string &id);
	virtual void getModelTexture(const std::string &id);

	// just get URLs; used by CacheStorage
	std::string getQuadURL(int resol, int size, const Coords::GaussKr &base) const;
	std::string getTextureURL(int resol, int size, const Coords::GaussKr &base) const;
	std::string getLayerURL(const std::string &layer, int size,
		const Coords::GaussKr &base) const;
	std::string getCoarseHeightsURL() const;
	std::string getCoarseTextureURL() const;
	std::string getImageURL(const std::string &id) const;
	std::string getModelURL(const std::string &id) const;
	std::string getSurfaceURL(const std::string &id) const;
	
	std::string getServer() { return server; }
	unsigned short getPort() { return port; }

};

class CacheStorage : public WorldStorage {
protected:
	char slash() const { 
		assert("slash() should never be called on CacheStorage" == NULL);
		return '\n'; //return some nonsense
	}
private:
	struct CacheEntry {
		std::string fname;
		int size;
		time_t lastAccess;
		
		CacheEntry(const std::string &_fname, int _size, time_t _now) :
			fname(_fname), size(_size), lastAccess(_now) {}
	};
	
	HttpStorage *storage;
	std::vector<int> quadResolutionList, textureResolutionList; 
	std::string tempDirPath;
	std::vector<CacheEntry *> entriesByTime;
	int totalSize, maxSize;

	HttpClient httpClient;

	bool initTempDirectory(void);
	
	void markAccess(const std::string &fname);
	void addEntry(const std::string &fname);
	void purgeCache(void);

	// the next functions build the name of the cache file, including path
	std::string buildCFNameTexture(int resol, int size,
		const Coords::GaussKr &base) const;
	std::string buildCFNameQuad(int resol, int size,
		const Coords::GaussKr &base) const;
	std::string buildCFNameLayer(const std::string &layer, int size,
		const Coords::GaussKr &base) const;
	std::string buildCFNameCoarseHeights() const;
	std::string buildCFNameCoarseTexture() const;
	std::string buildCFNameData(const std::string &fname) const;

public:
	CacheStorage(HttpStorage *_storage, int _maxSize = 100000) :
		WorldStorage(_storage->getWorldName(), _storage->getBasePath()),
			quadResolutionList(), textureResolutionList(), entriesByTime(),
		totalSize(0), maxSize(_maxSize),
		httpClient(_storage->getServer(), _storage->getPort()) {
		storage = _storage;
		initTempDirectory();
	}
	virtual ~CacheStorage() {
		std::vector<CacheEntry*>::iterator i = entriesByTime.begin();
		std::vector<CacheEntry*>::iterator g = entriesByTime.end();
		while (i != g)
			if (*i != NULL)
				delete *i;
		delete storage;
	}

	virtual int getAlternativeStoragesList(std::vector<WorldStorage*> &list) {
		return storage->getAlternativeStoragesList(list);
	}

	virtual osg::Image* getTexture(int resol, int size, const Coords::GaussKr &base);
	virtual osg::HeightField* getQuad(int resol, int size, const Coords::GaussKr &base);

	virtual std::string getLayer(const std::string &layer, int size,
		const Coords::GaussKr &base);

	virtual osg::HeightField *getCoarseHeights();
	virtual osg::Image *getCoarseTexture();

	virtual osg::Image *getImage(const std::string &id);
	virtual osg::Node *getModel(const std::string &id);
	virtual void getModelTexture(const std::string &id);

};

#endif /* __XGISLIB_STORAGE_H__ */
