#include <config.h>
#include "stdlib.h"

#include <xgislib/sun.h>
#include <xgislib/xgislog.h>
#include <xgislib/pointdata.h>
#include <xgislib/xputils.h>
#include <xgislib/xgisimgutils.h>

#include <osg/PositionAttitudeTransform>
#include <osg/Billboard>
#include <osg/Texture2D>
#include <osg/BlendFunc>

#define SUN_DISTANCE 150000.0

Sun::Sun(time_t rawtime, osg::Vec3f _detailRefPoint) {
	detailRefPoint = _detailRefPoint;
	this->sunPosition = setupSun(rawtime);
	this->addChild(sunPosition.get());
	setupLight();
	//this->sunPosition->addChild(setupLight());
}

void Sun::setSunPosition(time_t rawtime) {
	osg::Vec3d pos = getSunPosition(rawtime);
	sunPosition->setPosition(pos);
	sunLight->setPosition(osg::Vec4f(pos.x(),pos.y(),pos.z(),1.0));
}

osg::Vec4d Sun::getSunColour(time_t rawtime) {
/*let's just assume azimuth and altitude are the right ones...*/
	osg::Vec4d morningSun = osg::Vec4d(0.878,0.847,0.26,1.0);
	/*0.75 is the target value for midday*/
	float add = MIN(0.5, 0.5*altitude/20.0);
	osg::Vec4d ret = morningSun + osg::Vec4d(0,0,add,0);

	return ret;
}

osg::Vec3d Sun::getSunPosition(time_t rawtime) {

	Coords::GaussKr center(detailRefPoint.x(), detailRefPoint.y());
	
	Pointdata::FeatureInfo::getSunPosition
		(center, rawtime, azimuth, altitude);
	
	if (azimuth < 0) azimuth += 360;
	double azimuthX = 270 - azimuth;	
	double sunX = 0;
	
	double sunY = 0;

	double sunZ = sin(osg::DegreesToRadians(altitude))*SUN_DISTANCE;
	double p1 = cos(osg::DegreesToRadians(altitude));
	double proj = fabs(p1)*SUN_DISTANCE;
	
	sunX = cos(osg::DegreesToRadians(azimuthX))*proj+center.x;


	sunY = sin(osg::DegreesToRadians(azimuthX))*proj+center.y;

	osg::Vec3d sunLoc = osg::Vec3d(sunX, sunY, sunZ);
	return sunLoc;
}

//osg::ref_ptr<osg::Geode> World::setupSun(time_t rawtime)
osg::PositionAttitudeTransform *Sun::setupSun(time_t rawtime)
{
	Coords::GaussKr center(detailRefPoint.x(), detailRefPoint.y()); 

	osg::Vec3d sunLoc = getSunPosition(rawtime);
	
	osg::ref_ptr<osg::StateSet> sunState = new osg::StateSet;

	sunState->setRenderBinDetails(-1, "RenderBin");

	sunState->setMode( GL_LIGHTING, osg::StateAttribute::OFF);
	sunState->setMode(GL_CULL_FACE, osg::StateAttribute::OFF);
	sunState->setMode(GL_DEPTH_TEST, osg::StateAttribute::ON);

	float radius = SUN_DISTANCE*0.03;

	osg::Billboard *sunBillBoard = new osg::Billboard();

	float width = 4*radius;
	float height = width;

	osg::Geometry* sunQuad = new osg::Geometry;

	osg::Vec3Array* sunVerts = new osg::Vec3Array(4);
	(*sunVerts)[0] = osg::Vec3(-width/2.0f, 0, 0);
	(*sunVerts)[1] = osg::Vec3( width/2.0f, 0, 0);
	(*sunVerts)[2] = osg::Vec3( width/2.0f, 0, height);
	(*sunVerts)[3] = osg::Vec3(-width/2.0f, 0, height);
	
	sunQuad->setVertexArray(sunVerts);
	
	osg::Vec2Array* sunTexCoords = new osg::Vec2Array(4);
	(*sunTexCoords)[0].set(0.0f,0.0f);
	(*sunTexCoords)[1].set(1.0f,0.0f);
	(*sunTexCoords)[2].set(1.0f,1.0f);
	(*sunTexCoords)[3].set(0.0f,1.0f);
	sunQuad->setTexCoordArray(0,sunTexCoords);
	
	sunQuad->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::QUADS,0,4));
	
	sunQuad->setColorBinding(osg::Geometry::BIND_OFF);
	
	osg::Texture2D *sunTexture = new osg::Texture2D;

	osg::ref_ptr<osg::Image> image = new osg::Image;
		
	int numPixels = 64;
	image->allocateImage(numPixels,numPixels,1,GL_RGBA,GL_FLOAT);
	image->setInternalTextureFormat(GL_RGBA);
		
	XGisImage::initImage(image.get(), 1.0, 1.0);

	XGisImage::makeSunCircle(image.get());
	sunTexture->setImage(image.get());

	sunTexture->setWrap(osg::Texture2D::WRAP_S, osg::Texture::CLAMP_TO_EDGE);
	sunTexture->setWrap(osg::Texture2D::WRAP_T, osg::Texture::CLAMP_TO_EDGE);
	sunState->setTextureAttributeAndModes(0, sunTexture, osg::StateAttribute::ON);

	sunState->setMode(GL_BLEND, osg::StateAttribute::ON);
	osg::BlendFunc *fn = new osg::BlendFunc();
	fn->setFunction(osg::BlendFunc::SRC_ALPHA, osg::BlendFunc::ONE_MINUS_SRC_ALPHA);
	sunState->setAttributeAndModes(fn, osg::StateAttribute::ON);

	sunBillBoard->setStateSet(sunState.get());
	sunBillBoard->setMode(osg::Billboard::POINT_ROT_EYE);

	sunBillBoard->addDrawable(sunQuad);

	osg::PositionAttitudeTransform *sunPos = new osg::PositionAttitudeTransform();

	sunPos->addChild(sunBillBoard);

	sunPos->setPosition(sunLoc);
	sunPos->setName("sunPos");

	return sunPos;
}

void Sun::setLightBrightness(){
	osg::Vec4 lightMagnitudes = osg::Vec4(0.3, 0.3, 0.3, 1.0);
	float add = 0;
	
	if (altitude > 0) {
		add = MIN(0.6,0.7*altitude/20.0);
	}

	sunLight->setAmbient(lightMagnitudes+osg::Vec4(add,add,add,0));	
}

void Sun::setupLight() {
	sunLight = new osg::Light();

	//sunLight->setDirection(osg::Vec3(0.0, 0.0, 0.0));		// has no effect
	sunLight->setPosition(osg::Vec4(detailRefPoint.x(),detailRefPoint.y(),
		detailRefPoint.z(),1.0));

	/*this makes houses in the buildings layer ugly*/
	sunLight->setSpecular(osg::Vec4(0.0,0.0,0.0,0.0));		// remove default value that reflects too much
	sunLight->setDiffuse(osg::Vec4(0.05,0.05,0.05,0.50));	// leave enough to give sence of lighting without deforming colour too much
	setLightBrightness();
	
	sunLightSource = new osg::LightSource();
	sunLightSource->setLight(sunLight.get());
}

