#include "config.h"

#include <math.h>

#include "vectors.h"

const double PI = 3.141592653;

double norm(const Vector2 &v) {
	return sqrt(v.x*v.x + v.y*v.y);
}

double norm(const Vector3 &v) {
	return sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
}

Vector2 normalize(const Vector2 &v) {
	double N = norm(v);
	if (N == 0.0f) return v;
	else return v / N;
}

Vector3 normalize(const Vector3 &v) {
	double N = norm(v);
	if (N == 0.0f) return v;
	else return v / N;
}

double cross(const Vector2 &v1, const Vector2 &v2) {
	return v1.x*v2.y - v2.x*v1.y;
}

Vector3 cross(const Vector3 &v1, const Vector3 &v2) {
	return Vector3(v1.y * v2.z - v1.z * v2.y,
		 v1.z * v2.x - v1.x * v2.z,
		 v1.x * v2.y - v1.y * v2.x);
}

double dot(const Vector2 &v1, const Vector2 &v2) {
	return v1.x*v2.x + v1.y*v2.y;
}

double dot(const Vector3 &v1, const Vector3 &v2) {
	return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z;
}

void offset(Vector2 *v, const Vector2 &vOff, double d) {
	//assume vOff normalized
	v->x += vOff.x*d;
	v->y += vOff.y*d;
}

void offset(Vector3 *v, const Vector3 &vOff, double d) {
	//assume vOff normalized
	v->x += vOff.x*d;
	v->y += vOff.y*d;
	v->z += vOff.z*d;
}

void rotate(Vector3 *v, double radAngle) {
	//rotation around Z axis - formerly Y, thence the inconsistency
	double cosa = cos(radAngle);
	double sina = sin(radAngle);
	double newx = cosa * v->x - sina * v->y;
	double newz = sina * v->x + cosa * v->y;
	v->x = newx; v->y = newz;
}

void rotate(Vector3 *v, const Vector3 &vRot, double radAngle) {
	//project v onto vRot (assume vRot normalized)
	Vector3 vProj = vRot * dot(*v, vRot);
	//X unit of rotation plane
	Vector3 unitX = (*v) - vProj;
	//its length (equals length of v in rotation space)
	double vNorm = norm(unitX);
	//normalize unitX;
	unitX /= vNorm;
	//Y unit of rotation plane
	Vector3 unitY = cross(vRot, unitX);
	//offset pivot by rotation angle in rotation space
	offset(&vProj, unitX, cos(radAngle)*vNorm);
	offset(&vProj, unitY, sin(radAngle)*vNorm);
	*v = vProj;
}

double area(const Vector2 &v1, const Vector2 &v2, const Vector2 &v3) {
	Vector2 edge1 = v2 - v1;
	Vector2 edge2 = v3 - v1;
	double a = cross(edge1, edge2) / 2;
	return (a < 0.0f ? -a : a);
}

double area(const Vector3 &v1, const Vector3 &v2, const Vector3 &v3) {
	Vector3 edge1 = v2 - v1;
	Vector3 edge2 = v3 - v1;
	return norm(cross(edge1,edge2)) / 2;
}

double angle(const Vector2 &v1, const Vector2 &v2) {
	//assume both vectors normalized
	return acos(dot(v1,v2));
}

double angle(const Vector3 &v1, const Vector3 &v2) {
	//assume both vectors normalized
	return acos(dot(v1,v2));
}

double intersectionCoeff(const Vector2 &o1, const Vector2 &v1,
	const Vector2 &o2, const Vector2 &v2) {	
	double rightU = o2.x - o1.x;
	double rightD = o2.y - o1.y;
	
	double D =	v1.x	 * (-v2.y) - v1.y	 * (-v2.x);
	double DX = rightU * (-v2.y) - rightD * (-v2.x);

	return DX/D;
}

Vector2 intersection(const Vector2 &o1, const Vector2 &v1,
	const Vector2 &o2, const Vector2 &v2) {
	double d = intersectionCoeff(o1,v1, o2,v2);
	return Vector2(o1.x + d*v1.x, o1.y + d*v1.y);
}

bool linesIntersect(const Vector2 &p11, const Vector2 &p12,
	const Vector2 &p21, const Vector2 &p22) {
	Vector2 v1(p12.x-p11.x, p12.y-p11.y);
	Vector2 v2(p22.x-p21.x, p22.y-p21.y);
	double d1 = intersectionCoeff(p11,v1, p21,v2);
	double d2 = intersectionCoeff(p21,v2, p11,v1);
	
	return
		d1 >= 0.0f && d1 <= 1.0f &&
		d2 >= 0.0f && d2 <= 1.0f;
}

double distPtPt(const Vector2 &v1, const Vector2 &v2) {
	double dx = v2.x - v1.x;
	double dy = v2.y - v1.y;
	return sqrt(dx*dx + dy*dy);
}

double distPtPt(const Vector3 &v1, const Vector3 &v2) {
	double dx = v2.x - v1.x;
	double dy = v2.y - v1.y;
	double dz = v2.z - v1.z;
	return sqrt(dx*dx + dy*dy + dz*dz);
}

void evalArcForY(double y, double sweepY, const Vector2 &v,
	double *x1, double *x2) {
	double x0 = v.x;
	double y0 = v.y;
	double Y = sweepY;

	double A = 1;
	double B = -2*x0;
	double C = x0*x0 - 2*y*y0 + y0*y0 - Y*Y + 2*Y*y;
	
	double D = B*B - 4*A*C;
	double Ds = 0.0f;
	if (D > 0.0f) Ds = sqrt(D);
	if (B < 0.0f) Ds = -Ds;
	double q = -0.5f * (B + Ds);
	
	*x1 = q / A;
	*x2 = C / q;

	if (*x2 < *x1) {
		double tmp = *x1;
		*x1 = *x2;
		*x2 = tmp;
	} 
}

double evalArcForX(double x, double sweepY, const Vector2 &v) {
	return
		(x*x - 2*x*v.x + v.x*v.x + v.y*v.y - sweepY*sweepY) /
		(2 * (v.y - sweepY));
}

void intersectionArcArc(double sweepY, const Vector2 &v1, const Vector2 &v2,
	Vector2 *i1, Vector2 *i2) {
	double dy1 = v1.y - sweepY;
	double dy2 = v2.y - sweepY;
	//printf("dy1:%f dy2:%f\n", dy1,dy2);
	
	if (v1.y == v2.y) {
		
		i1->x = i2->x = (v1.x + v2.x)/2;
		
	}
	else{
		
		if (dy1 == 0.0f)
			
			i1->x = i2->x = v1.x;
		
		else if (dy2 == 0.0f)
			
			i1->x = i2->x = v2.x;
		
		else {
			
			double A = 1/(2*dy1) - 1/(2*dy2);
			double B = v2.x/dy2 - v1.x/dy1;
			double C =
				(v1.x*v1.x + v1.y*v1.y - sweepY*sweepY) / (2*dy1) -
				(v2.x*v2.x + v2.y*v2.y - sweepY*sweepY) / (2*dy2);
			
			double D = B*B - 4*A*C;
			double Ds = 0.0f;
			if (D > 0.0f) Ds = sqrt(D);
			if (B < 0.0f) Ds = -Ds;
			double q = -0.5f * (B + Ds);
			//printf("A:%f B:%f C:%f D:%f Ds:%f\n",A,B,C,D,Ds);
			
			i1->x = q / A;
			i2->x = C / q;
			
			if (i2->x < i1->x) {
				double tx = i1->x;
				i1->x = i2->x;
				i2->x = tx;
			}
		}
	}
	
	//printf("i1->x:%f i2->x:%f\n", i1->x, i2->x);
	
	if (dy1 == 0.0f) {
		i1->y = evalArcForX(i1->x, sweepY, v2);
		i2->y = evalArcForX(i2->x, sweepY, v2);
	}
	else {
		i1->y = evalArcForX(i1->x, sweepY, v1);
		i2->y = evalArcForX(i2->x, sweepY, v1);
	}
	
	//printf("i1->y:%f i2->y:%f\n", i1->y, i2->y);
}
