#ifndef __SOCALIB_VECTORS_H__
#	define __SOCALIB_VECTORS_H__

//////////////////////////////////////////////////////
//
// Discreet 2D point used for (U,V) quad coordinates
//
//////////////////////////////////////////////////////

class Point {
public:
	int x;
	int y;

	Point() {
		x=0; y=0;
	}
	Point(int x, int y) {
		this->x = x; this->y=y;
	}
	Point(const Point &p) {
		this->x = p.x; this->y = p.y;
	}
	void set(int x, int y) {
		this->x = x; this->y = y;
	}
	Point& operator=(const Point &p) {
		this->x = p.x; this->y = p.y;
		return *this;
	}
	bool operator==(const Point &p) {
		return (this->x == p.x && this->y == p.y);
	}
	bool operator!=(const Point &p) {
		return (this->x != p.x || this->y != p.y);
	}
};

//////////////////////////////////////////////////////
//
// Floating-point 2d, 3d and 4d vectors
//
//////////////////////////////////////////////////////


class Vector2 {
public:
	double x,y;
	Vector2() : x(0.0f), y(0.0f) {
	}
	Vector2(double x, double y) {
		this->x=x; this->y=y;
	}
	Vector2(const Vector2 &v) {
		x=v.x; y=v.y;
	}
	void set(double x, double y) {
		this->x=x; this->y=y;
	}
	bool operator==(const Vector2 &v) const {
		return (x==v.x && y==v.y);
	}
	Vector2 operator+(const Vector2 &v) const {
		return Vector2(x+v.x, y+v.y);
	}
	Vector2 operator-(const Vector2 &v) const {
		return Vector2(x-v.x, y-v.y);
	}
	Vector2 operator*(double s) const {
		return Vector2(x*s, y*s);
	}
	Vector2 operator/(double s) const{
		return Vector2(x/s, y/s);
	}
	void operator+=(const Vector2 &v) {
		x+=v.x; y+= v.y;
	}
	void operator-=(const Vector2 &v) {
		x-=v.x; y-=v.y;
	}
	void operator*=(double s) {
		x*=s; y*=s;
	}
	void operator/=(double s) {
		x/=s; y/=s;
	}
};

class Vector3 {
public:
	double x,y,z;
	Vector3() : x(0.0f), y(0.0f), z(0.0f) {
	}
	Vector3(double x, double y, double z) {
		this->x=x; this->y=y; this->z=z;
	}
	Vector3(const Vector3 &v) {
		x=v.x; y=v.y; z=v.z;
	}
	Vector3(const Vector2 &v) {
		x=v.x; y=v.y; z=0.0f;
	}
	void set(double x, double y, double z) {
		this->x=x; this->y=y; this->z=z;
	}
	bool operator==(const Vector3 &v) const {
		return (x==v.x && y==v.y && z==v.z);
	}
	Vector3 operator+(const Vector3 &v) const {
		return Vector3(x+v.x, y+v.y, z+v.z);
	}
	Vector3 operator-(const Vector3 &v) const {
		return Vector3(x-v.x, y-v.y, z-v.z);
	}
	Vector3 operator*(double s) const {
		return Vector3(x*s, y*s, z*s);
	}
	Vector3 operator/(double s) const{
		return Vector3(x/s, y/s, z/s);
	}
	void operator+=(const Vector3 &v) {
		x+=v.x; y+= v.y; z+=v.z;
	}
	void operator-=(const Vector3 &v) {
		x-=v.x; y-=v.y; z-=v.z;
	}
	void operator*=(double s) {
		x*=s; y*=s; z*=s;
	}
	void operator/=(double s) {
		x/=s; y/=s; z/=s;
	}
};

class Vector4 {
public:
	double x,y,z,w;
	Vector4() : x(0.0f), y(0.0f), z(0.0f), w(0.0f) {
	}
	Vector4(double x, double y, double z, double w) {
		this->x=x; this->y=y; this->z=z, this->w=w;
	}
	Vector4(const Vector4 &v) {
		x=v.x; y=v.y; z=v.z; w=v.w;
	}
	Vector4(const Vector3 &v) {
		x=v.x; y=v.y; z=v.z; w=0.0f;
	}
	Vector4(const Vector2 &v) {
		x=v.x; y=v.y; z=0.0f; w=0.0f;
	}
	void set(double x, double y, double z, double w) {
		this->x=x; this->y=y; this->z=z; this->w=w;
	}
	bool operator==(const Vector4 &v) const {
		return (x==v.x && y==v.y && z==v.z && w==v.w);
	}
	Vector4 operator+(const Vector4 &v) const {
		return Vector4(x+v.x, y+v.y, z+v.z, w+v.w);
	}
	Vector4 operator-(const Vector4 &v) const {
		return Vector4(x-v.x, y-v.y, z-v.z, w-v.w);
	}
	Vector4 operator*(double s) const {
		return Vector4(x*s, y*s, z*s, w*s);
	}
	Vector4 operator/(double s) const{
		return Vector4(x/s, y/s, z/s, w/s);
	}
	void operator+=(const Vector4 &v) {
		x+=v.x; y+= v.y; z+=v.z; w+=v.w;
	}
	void operator-=(const Vector4 &v) {
		x-=v.x; y-=v.y; z-=v.z; w-=v.w;
	}
	void operator*=(double s) {
		x*=s; y*=s; z*=s; w*=s;
	}
	void operator/=(double s) {
		x/=s; y/=s; z/=s; w/=s;
	}
};

double norm(const Vector2 &v);
double norm(const Vector3 &v);
Vector2 normalize(const Vector2 &v);
Vector3 normalize(const Vector3 &v);
double cross(const Vector2 &v1, const Vector2 &v2);
Vector3 cross(const Vector3 &v1, const Vector3 &v2);
double dot(const Vector2 &v1, const Vector2 &v2);
double dot(const Vector3 &v1, const Vector3 &v2);
void offset(Vector2 *v, const Vector2 &vOff, double d);
void offset(Vector3 *v, const Vector3 &vOff, double d);
void rotate(Vector3 *v, double radAngle);
void rotate(Vector3 *v, const Vector3 &vRot, double radAngle);
double area(const Vector2 &v1, const Vector2 &v2, const Vector2 &v3);
double area(const Vector3 &v1, const Vector3 &v2, const Vector3 &v3);
double angle(const Vector2 &v1, const Vector2 &v2);
double angle(const Vector3 &v1, const Vector3 &v2);
Vector2 intersection(const Vector2 &o1, const Vector2 &v1,
	const Vector2 &o2, const Vector2 &v2);
bool linesIntersect(const Vector2 &p11, const Vector2 &p12,
	const Vector2 &p21, const Vector2 &p22);

double distPtPt(const Vector2 &v1, const Vector2 &v2);
double distPtPt(const Vector3 &v1, const Vector3 &v2);
	
double evalArcForX(double x, double sweepY, const Vector2 &v);

void evalArcForY(double y, double sweepY, const Vector2 &v,
	double *x1, double *x2);

void intersectionArcArc(double sweepY, const Vector2 &v1, const Vector2 &v2,
	Vector2 *i1, Vector2 *i2);

extern const double PI;

#endif /* __SOCALIB_VECTORS_H__ */
