#include "config.h"

#include <xgislib/pointdata.h>
#include <xgislib/xputils.h>
#include <xgislib/xgisimgutils.h>
#include <xgislib/world.h>
#include <xgislib/worldloader.h>
#include <xgislib/xgisconfig.h>
#include <xgislib/layerbuilder.h>

#include <osg/BlendFunc>
#include <osg/Geometry>
#include <osgDB/ReadFile>
#include <osg/Billboard>
#include <osg/ShapeDrawable>
#include <osg/LineSegment>
#include <osg/Texture2D>
#include <osg/LightSource>

#include <osgUtil/IntersectVisitor>

#include <math.h>

// setup of detail levels
// TODO: detail levels should be loaded from the server!
static DetailLevel detailLevelsXDSL[] = {
	{ 500, 1, 80, 256, PRIORITY_TERRAIN_HI_RES, true },
	{ 1000, 1, 40, 128, PRIORITY_TERRAIN_HI_RES, true },
	{ 2000, 1, 20, 64, PRIORITY_TERRAIN_HI_RES, true },
	{ 4000, 1, 10, 32, PRIORITY_TERRAIN_LO_RES, false },
	{ 8000, 1, 10, 32, PRIORITY_TERRAIN_LO_RES, false },
	{ 0, 0, 0, 0, PRIORITY_TERRAIN_LO_RES, false }		
};

static DetailLevel detailLevelsLAN[] = {
	{ 500, 1, 80, 1024, PRIORITY_TERRAIN_HI_RES, true },
	{ 1000, 1, 80, 512, PRIORITY_TERRAIN_HI_RES, true },
	{ 2000, 1, 40, 256, PRIORITY_TERRAIN_HI_RES, true },
	{ 4000, 1, 20, 128, PRIORITY_TERRAIN_HI_RES, true },
	{ 12000, 1, 10, 32, PRIORITY_TERRAIN_LO_RES, false },
#if 0
	{ 16000, 1, 5, 32, PRIORITY_TERRAIN_LO_RES, false },
#endif /* 0/1 */
	{ 0, 0, 0, 0, PRIORITY_TERRAIN_LO_RES, false }
};

static DetailLevel detailLevelsMipro[] = {
	{ 1000, 1, 80, 256, PRIORITY_TERRAIN_HI_RES, true },
	{ 2000, 1, 40, 64, PRIORITY_TERRAIN_HI_RES, true },
	{ 6000, 1, 10, 32, PRIORITY_TERRAIN_LO_RES, false },
	{ 0, 0, 0, 0, PRIORITY_TERRAIN_LO_RES, false }
};

static DetailLevel detailLevelsMiproCoarse[] = {
	{ 600, 1, 40, 256, PRIORITY_TERRAIN_HI_RES, true },
	{ 3000, 1, 10, 64, PRIORITY_TERRAIN_LO_RES, false },
	{ 0, 0, 0, 0, PRIORITY_TERRAIN_LO_RES, false }
};

static DetailLevel detailLevelsHi[] = {
        { 1000, 1, 80, 1024, PRIORITY_TERRAIN_LO_RES, true },
        { 2000, 1, 80, 512, PRIORITY_TERRAIN_LO_RES, true },
        { 4000, 1, 40, 256, PRIORITY_TERRAIN_LO_RES, true },
        { 8000, 1, 20, 128, PRIORITY_TERRAIN_LO_RES, true },
#if 0
		{ 16000, 1, 10, 64, PRIORITY_TERRAIN_LO_RES, true },
#endif
        { 0, 0, 0, 0, PRIORITY_TERRAIN_LO_RES, false }
};

static DetailLevel detailLevelsUltraHi[] = {
        { 2000, 1, 80, 1024, PRIORITY_TERRAIN_LO_RES, true },
        { 4000, 1, 80, 512, PRIORITY_TERRAIN_LO_RES, true },
        { 8000, 1, 40, 256, PRIORITY_TERRAIN_LO_RES, true },
        { 16000, 1, 20, 128, PRIORITY_TERRAIN_LO_RES, true },
        { 32000, 1, 10, 64, PRIORITY_TERRAIN_LO_RES, true },
        { 0, 0, 0, 0, PRIORITY_TERRAIN_LO_RES, false }
};

static DetailLevel detailLevelsInsane[] = {
        { 4000, 1, 80, 1024, PRIORITY_TERRAIN_LO_RES, true },
        { 8000, 1, 80, 512, PRIORITY_TERRAIN_LO_RES, true },
        { 16000, 1, 40, 256, PRIORITY_TERRAIN_LO_RES, true },
        { 32000, 1, 20, 128, PRIORITY_TERRAIN_LO_RES, true },
        { 64000, 1, 10, 64, PRIORITY_TERRAIN_LO_RES, true },
        { 0, 0, 0, 0, PRIORITY_TERRAIN_LO_RES, false }
};

static DetailLevelSet defaultDetailLevelSets[] = {
	{
		"Poor detail level (xDSL).",
		detailLevelsXDSL
	},
	{
		"Rich detail level (10Mbps+).",
		detailLevelsLAN
	},
#if 0
	{
		"MIPRO testing detail level.",
		detailLevelsMipro
	},
	{
		"MIPRO-coarse testing detail level.",
		detailLevelsMiproCoarse
	},
#endif
	{
		"High detail level (off-line movie rendering).",
		detailLevelsHi
	},
#if 0
	{
		"Ultra high detail level (off-line movie rendering).",
		detailLevelsUltraHi
	},
	{
		"Insane detail level (off-line movie rendering).",
		detailLevelsInsane
	},
#endif 
	{
		"", NULL
	}
};					

#define START_ALTITUDE 50

class WorldConfigChangeHandlers : public XGisConfigChangeHandlers {
	friend class World;
	
private:
	World *world;
		
public:
	WorldConfigChangeHandlers(World *_world) : world(_world) {}
	
	virtual void onAddLayer(const std::string &layer) {
		world->getUpdateCallback()->triggerLayerAdd(layer);
	}
	virtual void onRemoveLayer(const std::string &layer) {
		world->getUpdateCallback()->triggerLayerRemove(layer);
	}
};

World::World(const std::string &name, int _quadSize, DataLoader *_loader) : osg::Group() {
	
	quadSize = _quadSize;
	dataLoader = _loader;
	worldName = name;

	doAutomaticUpdates = true;
	quads = NULL;

	cfg = XGisConfig::getInstance();
	cfg->addChangeHandlers(new WorldConfigChangeHandlers(this));
	
	// set update callback
	this->setUpdateCallback(new WorldUpdateCallback());

	// set state to support lighting
	osg::ref_ptr<osg::StateSet> state = getOrCreateStateSet();
	state->setMode( GL_LIGHTING, osg::StateAttribute::ON );
	state->setMode( GL_LIGHT0, osg::StateAttribute::ON );
	state->setMode( GL_NORMALIZE, osg::StateAttribute::ON );

}

void World::setup(const osg::Vec3 &_center, LayerBuilder *lb,
	osg::HeightField *cmodel, osg::Image *ctex) {

	detailRefPoint = osg::Vec3f(_center.x(), _center.y(), 100000);
	layerBuilder = lb;
	
	// read coarse model to find out size of the world
	float coarseW, coarseH;

	coarseW = (cmodel->getNumColumns()-1)*cmodel->getXInterval();
	coarseH = (cmodel->getNumRows()-1)*cmodel->getYInterval();
	llLoadedQ = getQuadFromPoint(cmodel->getOrigin().x(), cmodel->getOrigin().y());
	urLoadedQ = QuadC((int)(llLoadedQ.u + coarseW / quadSize),
			(int)(llLoadedQ.v + coarseH / quadSize));

	// create the whole world, load coarse model into all quads
	quads = new osg::ref_ptr<Quad>*[urLoadedQ.u - llLoadedQ.u + 1];
	for (int u = llLoadedQ.u; u <= urLoadedQ.u; u++) {
		quads[u - llLoadedQ.u] =
			new osg::ref_ptr<Quad>[urLoadedQ.v - llLoadedQ.v + 1];
		for (int v = llLoadedQ.v; v <= urLoadedQ.v; v++) {
			QuadC qc(u, v);
			if (NULL == cmodel || !isQuadEmpty(qc, *cmodel)) {
				quads[u - llLoadedQ.u][v - llLoadedQ.v] = new Quad(getLLPointOfQuad(qc), quadSize,
					cmodel, ctex);
				addChild(getQuad(qc));
			}
		}
	}

	detailLevelSets = defaultDetailLevelSets;
	for(detailLevelSetCount = 0;
		detailLevelSets[detailLevelSetCount].levels != NULL;
		detailLevelSetCount++) ;
	detailLevelMax = 0;
	detailLevelSetCurrent = 0;
	setCurrentDetailLevelSet(cfg->getDefaultDetailLevelSet());

	// move detail reference point to START_ALTITUDE m above terrain: this
	// inititates an update to quad detail levels.
	osg::Vec3f realCenter(_center);
	QuadC qc = getQuadFromPoint(Coords::GaussKr
		(realCenter.x(), realCenter.y()));
	if(isnan(realCenter.z())) {
		realCenter.z() = getQuad(qc)->getTerrainHeight
			(realCenter.x(), realCenter.y()) + START_ALTITUDE;
	}
	setDetailRefPoint(realCenter);

	localGroup = new osg::Group;
	addChild(localGroup.get());
	
}

// returns squared distance from the closest point on the quad
float World::getDistanceFromQuad(const osg::ref_ptr<Quad> &q,
	const osg::Vec3 &p) const {
	float dx, dy, dz, cx, cy, dist;
	float qx = q->getBase().x, qy = q->getBase().y, qs = q->getSize();
	
	if(p.x() < qx) {
		dx = qx - p.x();
		cx = qx;
	}
	else if(p.x() > qx + qs) {
		dx = p.x() - qs - qx;
		cx = qx + qs;
	}
	else {
		dx = 0;
		cx = p.x();
	}
	if(p.y() < qy) {
		dy = qy - p.y();
		cy = qy;
	}
	else if(p.y() > qy + qs) {
		dy = p.y() - qs - qy;
		cy = qy + qs;
	}
	else {
		dy = 0;
		cy = p.y();
	}
	dz = p.z() - q->getTerrainHeight(cx, cy);
	dist = dx*dx + dy*dy + dz*dz;
	
	return dist;
}

#define DETAIL_REF_POINT_DELTA_THRESHOLD (200*200)

void World::setDetailRefPoint(const osg::Vec3 &eye) {
	// only move detail ref point when it is significantly distant
	// from previous ref point
	float dist = (eye.x() - detailRefPoint.x())*(eye.x() - detailRefPoint.x()) +
		(eye.y() - detailRefPoint.y())*(eye.y() - detailRefPoint.y()) +
		(eye.z() - detailRefPoint.z())*(eye.z() - detailRefPoint.z());
	
	if(dist < DETAIL_REF_POINT_DELTA_THRESHOLD) {
		return;
	}

	if(getDoAutomaticUpdates()) {
		QuadC llQ, urQ;
		getUpdateFrame(eye, llQ, urQ);
		detailRefPoint = eye;
		updateDetailLevel(llQ, urQ);
	}
	else {
		detailRefPoint = eye;
	}

	QuadC qc = getQuadFromPoint(eye.x(), eye.y());
	if(qc != currentQuad) {
		if(currentQuad.valid()) {
			Quad *q = getQuad(currentQuad);
			if(NULL != q) {
				q->onLeave();
			}
		}
		currentQuad = qc;
		if(currentQuad.valid()) {
			Quad *q = getQuad(currentQuad);
			if(NULL != q) {
				q->onEnter();
			}
		}
	}
}

void World::updateDetailLevel() {
	QuadC llQ, urQ;
	getUpdateFrame(getDetailRefPoint(), llQ, urQ);
	updateDetailLevel(llQ, urQ);	
}

void World::updateDetailLevel(const QuadC &llQ, const QuadC &urQ) {
	// TODO: this gets called *a lot*, so it could *really* benefit from
	// optimisation.
	
	int u, v;
	Quad *q;
	
	updateMutex.lock();
	
	for(u = llQ.u; u <= urQ.u; u++) {
		for(v = llQ.v; v <= urQ.v; v++) {
			q = getQuad(QuadC(u, v));
			if(NULL == q) {
				// no quad there
				continue;
			}
			if(q->hasLoaderTask(LoadDetails::getTaskId()) &&
				q->hasLoaderTask(LoadDetails::getTaskId())) {
				continue;
			}
			float dist = getDistanceFromQuad(q, getDetailRefPoint());
			DetailLevel *d = getDetailLevel(q, dist);
			// calculate detail level: if changed, create a loader
			// task
			if(!q->hasLoaderTask(LoadDetails::getTaskId())) { 
				if(d->maxDistance > 1.0) {
					// in case we have *no* detail level at all, load the lowest one first (which is fast)
					if(!q->hasFine()) {
						LoadDetails *task = new LoadDetails(q, this, getLowestDetailLevel(), PRIORITY_TERRAIN_LO_RES, dist);
						getDataLoader()->pushTask(task);
						// q->setLoading(true);
					}
					else if(!q->hasFine(d->hfResolution,
						d->txtResolution,
						d->subdivision)) {
						LoadDetails *task = new LoadDetails(q, this, NULL, d->pri, dist);
						getDataLoader()->pushTask(task);
						// q->setLoading(true);
					}
				}
				else if(q->hasFine()) {
					// no detail level found: use coarse model if not already set
					LoadDetails *task = new LoadDetails(q, this, d, d->pri, -dist);
					getDataLoader()->pushTask(task);
					// q->setLoading(true);
				}
			}

			if(!q->hasLoaderTask(LoadLayers::getTaskId()) && d->layers && !q->hasLayers()) {
				getDataLoader()->pushTask(new LoadLayers(q, this, dist, cfg->getVisibleLayers()));
				// q->setLoadingLayers(true);
			}
		}
	}

	updateMutex.unlock();
	
}

void World::clearLayers() {
	QuadC llQ, urQ;
	getUpdateFrame(getDetailRefPoint(), llQ, urQ);

	int u, v;
	Quad *q;

	XGIS_LOG_INFO("Clearing layers.");
	
	updateMutex.lock();
	
	for(u = llQ.u; u <= urQ.u; u++) {
		for(v = llQ.v; v <= urQ.v; v++) {
		}
		q = getQuad(QuadC(u, v));
		if(NULL == q) {
			// no quad there
			continue;
		}
		q->removeAllLayers();
	}

	updateMutex.unlock();

	XGIS_LOG_INFO("Clearing layers END.");
	
}

void World::updateLayers() {
	QuadC llQ, urQ;
	getUpdateFrame(getDetailRefPoint(), llQ, urQ);
	updateLayers(llQ, urQ);	
}

void World::updateLayers(const QuadC &llQ, const QuadC &urQ) {
	int u, v;
	Quad *q;
	
	updateMutex.lock();
	
	for(u = llQ.u; u <= urQ.u; u++) {
		for(v = llQ.v; v <= urQ.v; v++) {
			q = getQuad(QuadC(u, v));
			if(NULL == q) {
				// no quad there
				continue;
			}
			float dist = getDistanceFromQuad(q, getDetailRefPoint());
			DetailLevel *d = getDetailLevel(q, dist);
			if(!q->hasLoaderTask(LoadLayers::getTaskId()) && d->layers && !q->hasLayers()) {
				getDataLoader()->pushTask(new LoadLayers(q, this, dist, cfg->getVisibleLayers()));
				// q->setLoadingLayers(true);
			}
		}
	}

	updateMutex.unlock();

}


void World::getUpdateFrame(const osg::Vec3f &nRefPoint, float maxDistance,
	QuadC &llQ, QuadC &urQ) const {

	float llx1, lly1, urx1, ury1;
	float llx2, lly2, urx2, ury2;

	QuadC rqc = getQuadFromPoint(nRefPoint.x(), nRefPoint.y());
	QuadC orqc = getQuadFromPoint(detailRefPoint.x(), detailRefPoint.y());

	llx1 = detailRefPoint.x() - maxDistance;
	lly1 = detailRefPoint.y() - maxDistance;
	urx1 = detailRefPoint.x() + maxDistance;
	ury1 = detailRefPoint.y() + maxDistance;
	llx2 = nRefPoint.x() - maxDistance;
	lly2 = nRefPoint.y() - maxDistance;
	urx2 = nRefPoint.x() + maxDistance;
	ury2 = nRefPoint.y() + maxDistance;
	
	QuadC tllQ = getQuadFromPoint(MIN(llx1, llx2), MIN(lly1, lly2));
	QuadC turQ = getQuadFromPoint(MAX(urx1, urx2), MAX(ury1, ury2));
	
	// better safe than sorry ;)
	llQ.u = tllQ.u - 1; llQ.v = tllQ.v - 1;
	urQ.u = turQ.u + 1; urQ.v = turQ.v + 1;

}

void World::getUpdateFrame(const osg::Vec3f &nRefPoint,
	QuadC &llQ, QuadC &urQ) const {

	getUpdateFrame(nRefPoint,
		getCurrentDetailLevelSet()->levels[detailLevelMax].maxDistance,
		llQ, urQ);
}

bool World::isQuadEmpty(const QuadC &qc, const osg::HeightField &coarse) {
	float offX = getLLPointOfQuad(qc).x - getLLPointOfWorld().x,
		offY = getLLPointOfQuad(qc).y - getLLPointOfWorld().y;

	for (int col = (int)osg::round(offX / coarse.getXInterval());
			col <= (int)osg::round((offX + quadSize) / coarse.getXInterval()); col++)
		for (int row = (int)osg::round(offY / coarse.getYInterval());
				row <= (int)osg::round((offY + quadSize) / coarse.getYInterval()); row++) {
			if (isnan(coarse.getHeight(col, row)))
				return true;
		}
	return false;
}

double World::getTerrainHeight(float x, float y) const {
	Quad *quad = getQuad(getQuadFromPoint(x, y));
	if (NULL != quad) {
		return quad->getTerrainHeight(x, y);
	}
	else {
		return 0;
	}
}


Quad *World::getQuad(const QuadC &q) const {
	if ((llLoadedQ.u <= q.u && q.u <= urLoadedQ.u) &&
		(llLoadedQ.v <= q.v && q.v <= urLoadedQ.v)) {
		return quads[q.u - llLoadedQ.u][q.v - llLoadedQ.v].get();
	}
	else {
		return NULL;
	}
}

World::~World(void) {

	dataLoader->stop();
	
	for (int u = llLoadedQ.u; u <= urLoadedQ.u; u++)
		delete quads[u - llLoadedQ.u];
	delete quads;

}

QuadC World::getQuadFromPoint(float x, float y) const {
	//XGIS_LOG_DEBUG4("get quad %f %f\n",x,y);
	return QuadC((int)floor(x / (double)quadSize),
		(int)floor(y / (double)quadSize));
}

Coords::GaussKr World::getLLPointOfQuad(const QuadC &quad) const {
	return Coords::GaussKr(quad.u*(double)quadSize,
		quad.v*(double)quadSize);
}

Coords::GaussKr World::getURPointOfQuad(const QuadC &quad) const {
	return Coords::GaussKr((quad.u + 1)*(double)quadSize,
		(quad.v + 1)*(double)quadSize);
}

Coords::GaussKr World::getLLPointOfWorld() const {
	QuadC quad = llLoadedQ;
	return getLLPointOfQuad(quad);
}

Coords::GaussKr World::getURPointOfWorld() const {
	QuadC quad = urLoadedQ;
	return getURPointOfQuad(quad);
}

osg::Matrixd World::getDefaultCameraPos() {
	osg::Matrixd rv;
	rv *= osg::Matrixd::rotate(1.5, osg::Vec3(1, 0, 0));
	rv *= osg::Matrixd::translate(detailRefPoint.x(), detailRefPoint.y(), detailRefPoint.z());
	return rv;
}

DetailLevel *World::getDetailLevel(Quad *q, float dist) const {
	int d;
	DetailLevel *levels = getCurrentDetailLevelSet()->levels; 
	for(d = 0; levels[d].maxDistance != 0; d++) {
		if(levels[d].subdivision > 1) {
			// TODO: implement detail levels with subdivision > 1
			// currently, we ignore them
			continue;
		}
		if(dist <= levels[d].maxDistance*levels[d].maxDistance) {
			break;  // this is the detail level we want
		}
	}
	return &(levels[d]);
}

DetailLevel *World::getDetailLevel(Quad *q) const {
	float dist = getDistanceFromQuad(q, getDetailRefPoint());
	return getDetailLevel(q, dist);
}

DetailLevel *World::getLowestDetailLevel() const {
	return &(getCurrentDetailLevelSet()->levels[detailLevelMax]);
}

const DetailLevelSet *World::getDetailLevelSets() const{
	return detailLevelSets;
}

const DetailLevelSet *World::getCurrentDetailLevelSet() const {
	return &(detailLevelSets[detailLevelSetCurrent]);
}

void World::nextDetailLevelSet() {
	int n = detailLevelSetCurrent + 1;
	if(n >= detailLevelSetCount) {
		n = 0;
	}
	setCurrentDetailLevelSet(n);
}

void World::prevDetailLevelSet() {
	int n = detailLevelSetCurrent - 1;
	if(n < 0) {
		n = detailLevelSetCount - 1;
	}
	setCurrentDetailLevelSet(n);
}

void World::setCurrentDetailLevelSet(int n) {
	int oldMaxDistance;
	if(detailLevelSetCurrent != -1) {
		oldMaxDistance = getCurrentDetailLevelSet()->levels[detailLevelMax].maxDistance;
	}
	else {
		oldMaxDistance = 0;
	}
	detailLevelSetCurrent = (n % detailLevelSetCount);
	
	int i;
	for(i = 0;
		detailLevelSets[detailLevelSetCurrent].levels[i].maxDistance != 0;
		i++) ;	
	detailLevelMax = i - 1;
	
	if(oldMaxDistance <	detailLevelSets[detailLevelSetCurrent].
		levels[detailLevelMax].maxDistance) {
		oldMaxDistance = detailLevelSets[detailLevelSetCurrent].
			levels[detailLevelMax].maxDistance;
	}

	XGIS_LOG_INFO("Using detail level set %d", n);
	
	// clear all pending tasks...
	dataLoader->purge();
	
	// update complete detailed terrain
	QuadC llQ, urQ;
	getUpdateFrame(detailRefPoint, oldMaxDistance, llQ, urQ);
	updateDetailLevel(llQ, urQ);

}

void World::addTerrainChangeHandler(TerrainChangeHandler *handler) {
	osg::Vec4 b = handler->getTerrainBound();
	float i, j;
	for(i = b.x(); i <= b.x() + b.z(); i += quadSize) {
		for(j = b.y(); j <= b.y() + b.w(); j += quadSize) {
			Quad *q = getQuad(i, j);
			if(NULL != q) {
				q->addTerrainChangeHandler(handler);
			}
		}
	}
}

void World::removeTerrainChangeHandler(TerrainChangeHandler *handler) {
	osg::Vec4 b = handler->getTerrainBound();
	float i, j;
	for(i = b.x(); i <= b.x() + b.z(); i += quadSize) {
		for(j = b.y(); j <= b.y() + b.w(); j += quadSize) {
			Quad *q = getQuad(i, j);
			if(NULL != q) {
				q->removeTerrainChangeHandler(handler);
			}
		}
	}
}
