#ifndef __OBSCENE_WORLD_H__
#  define __OBSCENE_WORLD_H__

#include <OpenThreads/Mutex>

#include <osg/Node>
#include <osg/PolygonMode>
#include <osg/PositionAttitudeTransform>

#include <xgislib/quad.h>
#include <xgislib/xgisconfig.h>
#include <xgislib/worldupdatecallback.h>
#include <xgislib/dataloader.h>

class LayerBuilder;

struct DetailLevel {
	int maxDistance;   // maximum distance from the observer to apply this level
	int subdivision;   // number of subdivisions of each quad at this level
	int hfResolution;  // height field resolution (points-per-km)
	int txtResolution; // texture resolution (pixels-per-km)
	DataLoaderTaskPriority pri; // priority of the data loader task
	bool layers;       // load layers?
};

struct DetailLevelSet {
	const std::string description;
	struct DetailLevel *levels;
};

class World : public osg::Group {
	friend class WorldUpdateCallback;

protected:
	int quadSize;
	QuadC llLoadedQ, urLoadedQ;        // lower-left and upper-right indices
	osg::ref_ptr<Quad> **quads;        // table of all loaded quads

	osg::Vec3f detailRefPoint;         // reference point for detail level
	                                   // determination
	QuadC currentQuad;    			   // current quad
	
	osg::ref_ptr<osg::Group> localGroup;
	
	// available resolutions of quads and textures
	std::vector<int> quadResList, texResList;
	
	std::string worldName;

	DataLoader *dataLoader;
	LayerBuilder *layerBuilder;
	XGisConfig *cfg;

	// locked when world updates are being processed!
	OpenThreads::Mutex updateMutex;
	
	// detail level information
	DetailLevelSet *detailLevelSets;
	int detailLevelSetCurrent, detailLevelSetCount, detailLevelMax;
	
	bool doAutomaticUpdates;

	bool isQuadEmpty(const QuadC &q, const osg::HeightField &coarse);
	
public:
	const std::string& getName() { return worldName; }

	World(const std::string &name, int _quadSize, DataLoader *_loader); 
	virtual ~World();

	void setup(const osg::Vec3 &_center, LayerBuilder *lb,
		osg::HeightField *cmodel, osg::Image *ctex);

	osg::Matrixd getDefaultCameraPos();

	Quad *getQuad(const QuadC &q) const;
	Quad *getQuad(const Coords::GaussKr &point) const {
		return getQuad(getQuadFromPoint(point.x, point.y));
	}
	Quad *getQuad(float x, float y) const {
		return getQuad(getQuadFromPoint(x, y));
	}

	// coordinate stuff
	QuadC getQuadFromPoint(float x, float y) const;
	QuadC getQuadFromPoint(const Coords::GaussKr &point) const {
		return getQuadFromPoint(point.x, point.y);
	}
	
	Coords::GaussKr getLLPointOfQuad(const QuadC &quad) const;
	Coords::GaussKr getURPointOfQuad(const QuadC &quad) const;

	Coords::GaussKr getLLPointOfWorld() const;
	Coords::GaussKr getURPointOfWorld() const;
	
	QuadC getLLQuadC() const { return llLoadedQ; }
	QuadC getURQuadC() const { return urLoadedQ; }	
	
	void setDetailRefPoint(const osg::Vec3 &eye);
	const osg::Vec3 &getDetailRefPoint() const {
		return detailRefPoint;
	}

	float getDistanceFromQuad(const osg::ref_ptr<Quad> &q,
		const osg::Vec3 &p) const;

	void getUpdateFrame(const osg::Vec3f &nRefPoint, float maxDistance,
		QuadC &llQ, QuadC &urq) const;
	void getUpdateFrame(const osg::Vec3 &nRefPoint,
		QuadC &llQ, QuadC &urQ) const;
	
	double getQuadSize() const { return quadSize; }

	double getTerrainHeight(float x, float y) const;

	void updateDetailLevel(const QuadC &llQ, const QuadC &urQ);
	void updateDetailLevel();
	void updateLayers(const QuadC &llQ, const QuadC &urQ);
	void updateLayers();
	void clearLayers();

	class WorldUpdateCallback *getUpdateCallback() { 
		return dynamic_cast<class WorldUpdateCallback *>
			(osg::Group::getUpdateCallback());
	}

	DataLoader *getDataLoader() {
		return dataLoader;
	}

	LayerBuilder *getLayerBuilder() {
		return layerBuilder;
	}
	
	XGisConfig *getConfiguration() {
		return cfg;
	}
	
	const DetailLevelSet *getDetailLevelSets() const;
	const DetailLevelSet *getCurrentDetailLevelSet() const;
	void setCurrentDetailLevelSet(int n);
	void nextDetailLevelSet();
	void prevDetailLevelSet();

	DetailLevel *getDetailLevel(Quad *q) const;
	DetailLevel *getDetailLevel(Quad *q, float dist) const;
	DetailLevel *getLowestDetailLevel() const;
	
	bool getDoAutomaticUpdates() { return doAutomaticUpdates; }
	void setDoAutomaticUpdates(bool v) { doAutomaticUpdates = v; } 

	void addLocalNode(osg::Node *node) {
		localGroup->addChild(node);
	}
	
	void removeLocalNode(osg::Node *node) {
		localGroup->removeChild(node);
	}
	
	void addTerrainChangeHandler(TerrainChangeHandler *handler);
	void removeTerrainChangeHandler(TerrainChangeHandler *handler);
};

#endif /* __OBSCENE_WORLD_H__ */
