#include "config.h"

#include <xgislib/world.h>
#include <xgislib/worldupdatecallback.h>
#include <xgislib/worldloader.h>
#include <xgislib/xputils.h>

// movement distance threshold that triggers detail level updates, in meters
// actually the square of the distance in order to save one sqrt() operation,
// i.e. 10000 signifies 100m
#define DETAIL_UPDATE_THRESHOLD_2 10000 

#define XGIS_TIME_ADAPTIVE_UPDATE_PROCESSING 1

// how many tasks to process in one update callback
#ifndef XGIS_TIME_ADAPTIVE_UPDATE_PROCESSING
#	define TASKS_TO_PROCESS 5
#endif /* !XGIS_TIME_ADAPTIVE_UPDATE_PROCESSING */

void WorldUpdateCallback::fireUpdateEvent(World *world,
	const std::vector<WorldUpdateHandler *> &handlerChain, const osg::FrameStamp *stamp) {
	std::vector<WorldUpdateHandler *>::const_iterator i =
		handlerChain.begin();
	while(i != handlerChain.end()) {
		(*i)->handle(world, stamp);
		i++;
	}
}

void WorldUpdateCallback::operator()(osg::Node *node, osg::NodeVisitor *nv) {
	World* w = dynamic_cast<World*>(node);
	assert(w);

	// in case of async loading, we need to check if we have any finished
	// loading tasks
	osg::Timer_t updateStart, updateEnd;
	int nTasks = 0;
	LoaderTask *task;
	updateEnd = updateStart = updateTimer.tick();
#ifdef XGIS_TIME_ADAPTIVE_UPDATE_PROCESSING
	while(updateTimer.delta_s(updateStart, updateEnd) < updateMaxTime) {
#else
	while(nTasks < TASKS_TO_PROCESS) {
#endif /* XGIS_TIME_ADAPTIVE_UPDATE_PROCESSING */
		task = w->dataLoader->popTask();
		if(NULL != task) {
			XGIS_LOG_DEBUG5("Task processed, %d remaining.", nTasks);
			task->updateWorld();
			delete task;
			nTasks++;
#ifdef XGIS_TIME_ADAPTIVE_UPDATE_PROCESSING
			updateEnd = updateTimer.tick();
#endif /* XGIS_TIME_ADAPTIVE_UPDATE_PROCESSING */
		}
		else {
			break; /* no more tasks... */
		}
	}
	
	// if neccessary, toggle showing of details on all quads
	if(toggleDetailsFlag) {
		for (int u = w->llLoadedQ.u; u <= w->urLoadedQ.u; u++) {
			for (int v = w->llLoadedQ.v; v <= w->urLoadedQ.v; v++) {
				osg::ref_ptr<Quad> &q = 
					w->quads[u - w->llLoadedQ.u][v - w->llLoadedQ.v];
				if(q.get() != NULL) {
					q->setShowLayers(!q->getShowLayers());
				}
			}
		}
		toggleDetailsFlag = false;
	}

	if(!layersToRemove.empty()) {
		QuadC llQ, urQ;
		w->getUpdateFrame(w->getDetailRefPoint(), llQ, urQ);
		while(!layersToRemove.empty()) {		
			std::string layer = *(layersToRemove.begin());
			layersToRemove.erase(layersToRemove.begin());

			int u, v;
			Quad *q;
			for(u = llQ.u; u <= urQ.u; u++) {
				for(v = llQ.v; v <= urQ.v; v++) {
					q = w->getQuad(QuadC(u, v));
					if(NULL == q) {
						// no quad there
						continue;
					}
					if(q->hasLayer(layer)) {
						q->removeLayer(layer);
					}
				}
			}
		}
	}
	if(!layersToAdd.empty()) {
		QuadC llQ, urQ;
		w->getUpdateFrame(w->getDetailRefPoint(), llQ, urQ);
		
		int u, v;
		Quad *q;
		for(u = llQ.u; u <= urQ.u; u++) {
			for(v = llQ.v; v <= urQ.v; v++) {
				q = w->getQuad(QuadC(u, v));
				if(NULL == q) {
					// no quad there
					continue;
				}
				float dist = w->getDistanceFromQuad(q, w->getDetailRefPoint());
				DetailLevel *d = w->getDetailLevel(q, dist);
				if(d->layers) {
					w->getDataLoader()->pushTask(new LoadLayers(q, w, dist, layersToAdd));
					q->addLoaderTask(LoadLayers::getTaskId());
				}
			}
		}
		layersToAdd.clear();
	}
	
	// set polygon shading mode
	if (newPolyMode != NULL) {
		osg::ref_ptr<osg::StateSet> state = w->getOrCreateStateSet();
		state->setAttributeAndModes(newPolyMode,
			osg::StateAttribute::ON | osg::StateAttribute::OVERRIDE );
		newPolyMode = NULL;
	}

	// call update handlers
	fireUpdateEvent(w, updateHandlers, nv->getFrameStamp());

	traverse(node, nv);
}

void WorldUpdateCallback::triggerLayerAdd(const std::string &layer) {
	std::vector<std::string>::iterator i = layersToRemove.begin();
	while(i != layersToRemove.end()) {
		if((*i) == layer) {
			layersToRemove.erase(i);
		}
		i++;
	}
	i = layersToAdd.begin();
	while(i != layersToAdd.end()) {
		if((*i) == layer) {
			return;
		}
		i++;
	}
	layersToAdd.push_back(layer);
}

void WorldUpdateCallback::triggerLayerRemove(const std::string &layer) {
	std::vector<std::string>::iterator i = layersToAdd.begin();
	while(i != layersToAdd.end()) {
		if((*i) == layer) {
			layersToAdd.erase(i);
		}
		i++;
	}
	i = layersToRemove.begin();
	while(i != layersToRemove.end()) {
		if((*i) == layer) {
			return;
		}
		i++;
	}
	layersToRemove.push_back(layer);
}

/*
void WorldUpdateCallback::signalTerrainChange(const osg::Vec4f &changeBound) {
	terrainChangeMutex.lock();
	if(terrainChangeBound.z() <= 0.001) {
		terrainChangeBound = changeBound;
	}
	else {
		if(changeBound.x() < terrainChangeBound.x()) {
			terrainChangeBound.z() += terrainChangeBound.x() - changeBound.x();
			terrainChangeBound.x() = changeBound.x();
		}
		if(changeBound.x() + changeBound.z() > terrainChangeBound.x() + terrainChangeBound.z()) {
			terrainChangeBound.z() = changeBound.x() + changeBound.z() - terrainChangeBound.x();
		}
		if(changeBound.y() < terrainChangeBound.y()) {
			terrainChangeBound.w() += terrainChangeBound.y() - changeBound.y();
			terrainChangeBound.y() = changeBound.y();
		}
		if(changeBound.y() + changeBound.w() > terrainChangeBound.y() + terrainChangeBound.w()) {
			terrainChangeBound.w() = changeBound.y() + changeBound.w() - terrainChangeBound.y();
		}
	}
	terrainChangeMutex.unlock();
}
*/

WorldUpdateCallback::WorldUpdateCallback() {
	newPolyMode = NULL;
	toggleDetailsFlag = false;
	statusText = NULL;
	statusTimeout = 0;
	updateMaxTime = 1.0/12.0; /* 1/12 s */
}

