#include "config.h"
#include "stdlib.h"

#include <xgislib/xgislog.h>
#include <xgislib/xgisimgutils.h>
#include "xputils.h"

#include <vector>
#include <osg/GraphicsContext>

class MyGraphicsContext {
    public:
        MyGraphicsContext()
        {
            osg::ref_ptr<osg::GraphicsContext::Traits> traits = new osg::GraphicsContext::Traits;
            traits->x = 0;
            traits->y = 0;
            traits->width = 1;
            traits->height = 1;
            traits->windowDecoration = false;
            traits->doubleBuffer = false;
            traits->sharedContext = 0;
            traits->pbuffer = true;

            _gc = osg::GraphicsContext::createGraphicsContext(traits.get());

            if (!_gc)
            {
                traits->pbuffer = false;
                _gc = osg::GraphicsContext::createGraphicsContext(traits.get());
            }

            if (_gc.valid())
            {
                _gc->realize();
                _gc->makeCurrent();
            }
        }

        bool valid() const { return _gc.valid() && _gc->isRealized(); }

    private:
        osg::ref_ptr<osg::GraphicsContext> _gc;
};

void XGisImage::scaleImage(osg::Image *image, int newSizeX, int newSizeY) {
	MyGraphicsContext gc;
    if (!gc.valid())
    {
		XGIS_LOG_ERROR("Invalid context. Unable to scale image.");
	} else {
		XGIS_LOG_DEBUG5("image scale %d %d to %d %d", image->s(), image->t(),
			newSizeX, newSizeY);
		image->scaleImage(newSizeX, newSizeY, 1);
	}
}

osg::Vec4f XGisImage::getImageValue(osg::Image *image, int i,
									int j, osg::Vec4f empty) {
	return getImageValue(image, i,j,empty,false);
}

osg::Vec4f XGisImage::getImageValue(osg::Image *image, int i,
									int j, osg::Vec4f empty, bool wrap) {
	int width = image->s();
	int height = image->t();

	osg::Vec4f ret;

	if ((i < 0) || (j < 0) || (i >= width) || (j >= height)) {
		if (!wrap) {
			ret = empty;
		} else {
			/*make it wrap, for seamless textures*/
			XGIS_LOG_DEBUG5("seeking wrap for %d %d", i,j);
			if (i < 0) {
				/*falling off the left side, continue on the right*/
				/*i is negative here, and equal to half the unit, so we can do
				this*/
				int indxI = (width - 1) + i;
				ret = (*((osg::Vec4f*)image->data(indxI,j)));

			} else if (i > width) {
				/*falling off right side*/
				int indxI = i - (width - 1);
				ret = (*((osg::Vec4f*)image->data(indxI,j)));

			} else if (j < 0) {
				/*falling off bottom*/
				int indxJ = (height - 1) + j;
				ret = (*((osg::Vec4f*)image->data(i,indxJ)));
			} else if (j > height) {
				/*falling off top*/
				int indxJ = j - (height - 1);
				ret = (*((osg::Vec4f*)image->data(i,indxJ)));
			}
		}
	} else {

		ret = (*((osg::Vec4f*)image->data(i,j)));
		if(ret.a() == 0)
			ret = empty;
	}

	return ret;
}

void XGisImage::setGrayRange(osg::Image *image, float rangeMin, float rangeMax,
							 bool checkRange){
	/*assumming 0 is low and 1 is high for now*/
	/*maybe determine actual low and high values? But I suspect they would be
	1 and 0 anyway*/
	int width = image->s();
	int height = image->t();

	float valLow = 0.0;
	float valHigh = 1.0;

	if (checkRange) {
		valLow = 1.0;
		valHigh = 0.0;



		/*this makes it slow, but boy are clouds awful without it!*/
		for (int i = 0; i < width; i++) {
			for (int j = 0; j < height; j++) {
				osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);
				if (data->r() > valHigh) valHigh = data->r();
				if (data->r() < valLow) valLow = data->r();

			}
		}

	}

	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);

			float col = (data->r()/(valHigh - valLow))*(rangeMax - rangeMin)
						+ rangeMin;

			col = MAX(col, rangeMin);
			col = MIN(col, rangeMax);

			osg::Vec4f colour = osg::Vec4f(col,col,col,1.0);
			*data = colour;
		}
	}


}

void XGisImage::clampImage(osg::Image *image, float value){
	int width = image->s();
	int height = image->t();

	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);

			float opacity = value - data->a();

			opacity = MAX(opacity, 0);

			osg::Vec4f colour = osg::Vec4f(1,1,1,opacity);
			*data = colour;

		}
	}
}

void XGisImage::fixBorder(osg::Image *image) {
	int width = image->s();
	int height = image->t();

	for (int i = 1; i < width-1; i++) {
		osg::Vec4f colour = *((osg::Vec4f*)image->data(i,1));
		osg::Vec4f *data = (osg::Vec4f*)image->data(i,0);

		*data = colour;

		colour = *((osg::Vec4f*)image->data(i,height-2));
		data = (osg::Vec4f*)image->data(i,height-1);

		*data = colour;

	}

}


void XGisImage::squareRootImage(osg::Image *image){

	int width = image->s();
	int height = image->t();

	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);

			float opacity = data->a();



			osg::Vec4f colour = osg::Vec4f(1,1,1,sqrt(opacity));
			*data = colour;

		}
	}
}

void XGisImage::initImage(osg::Image *image, float f, float opacity){
	int width = image->s();
	int height = image->t();

	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);
			osg::Vec4f colour = osg::Vec4f(f,f,f,opacity);
			*data = colour;
		}
	}
}

void XGisImage::initImage(osg::Image *image){
	XGisImage::initImage(image, -1, 0);
}

void XGisImage::squareImage(osg::Image *image){
	int width = image->s();
	int height = image->t();

	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);

			float opacity = data->a();



			osg::Vec4f colour = osg::Vec4f(1,1,1,opacity*opacity);
			*data = colour;

		}
	}
}

void XGisImage::addPercentNoise(osg::Image *image, float amount){
	int width = image->s();
	int height = image->t();
	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);
			float r = ((rand() % 100) / 100.0 - 0.5)*amount + data->r();

			r = MIN(1,r);
			r = MAX(0,r);

			osg::Vec4f colour = osg::Vec4f(r,r,r,data->a());

			*data = colour;

		}
	}
}

void XGisImage::makeSunCircle(osg::Image *image){
	int width = image->s();
	int height = image->t();


	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);



			float x = osg::PI + osg::PI*0.5*(i - (width/2.0))/(float)(width/2.0);
			float y = osg::PI + osg::PI*0.5*(j - (height/2.0))/(float)(height/2.0);

			float opacity = 1.5*cos(x)*cos(y);
			opacity -= 0.2;

			opacity = MIN(opacity, 1);
			opacity = MAX(opacity,0);

			opacity *= opacity;
			opacity *= opacity;

			osg::Vec4f colour = osg::Vec4f(1.0, 1.0, 0.9, opacity);


			*data = colour;

		}
	}


}


void XGisImage::makeCircle(osg::Image *image){
	int width = image->s();
	int height = image->t();

	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);



			float x = osg::PI + osg::PI*0.5*(i - (width/2.0))/(float)(width/2.0);
			float y = osg::PI + osg::PI*0.5*(j - (height/2.0))/(float)(height/2.0);

			float opacity = 4*cos(x)*cos(y);
			//opacity -= 0.1;

			opacity = MAX(opacity,0);
			opacity = MIN(opacity,1);
			opacity *= opacity;

			opacity = MIN(opacity, data->a());

			osg::Vec4f colour = osg::Vec4f(data->r(), data->g(), data->b(),
				opacity);

			*data = colour;

		}
	}
}

void XGisImage::smoothBorder(osg::Image *image, int margin,
							 float opacity, float goal){
	int width = image->s();
	int height = image->t();

	float diff = 1.0 - goal;

	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {

			if ((i < margin) ||( i > width - margin - 1)) {

				osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);
				float c = (i > margin ? width - i : i + 1);

				osg::Vec4f colour;

				float opacity = data->a() - (1.0 - c/(float)margin);
				opacity = MAX(opacity, 0);

				if (0 == i) XGIS_LOG_DEBUG5("i 0 %f %f", data->r(), opacity);
				if (width-1 == i) XGIS_LOG_DEBUG5("i %d %f %f", width-1, data->r(), opacity);
				float col;

				if (opacity) {
					col = MAX(0,data->r());
					colour = osg::Vec4f(col,col,col,opacity);
				} else {

					col = data->r() - (diff - diff*c/float(margin));
					//col = MAX(col, goal);
					colour = osg::Vec4f(col, col, col, opacity);
				}

				*data = colour;


			}

			if ((j < margin) || (j > height - margin)) {
				osg::Vec4f *data = (osg::Vec4f*)image->data(i,j);

				float c = (j > margin ? height - j : j + 1);
				osg::Vec4f colour;
				float opacity = data->a() - (1.0 - c/(float)margin);
				opacity = MAX(opacity, 0);

				float col;

				if (0 == c) XGIS_LOG_DEBUG2("j %f", data->r());

				if (opacity) {
					col = MAX(0,data->r());
					colour = osg::Vec4f(col,col,col,opacity);
				} else {

					col = data->r() - diff*(1.0 - c/float(margin));
					//col = MAX(col, goal);
					colour = osg::Vec4f(col, col, col, opacity);
				}
				*data = colour;


			}

		}
	}

}

void XGisImage::convolve(osg::Image *image, int type){
	int width = image->s();
	int height = image->t();

	int size = 1;

	osg::Vec4f empty = osg::Vec4f(0,0,0,0);

	/*let's say for now that 0 == box blur*/
	if (0 == type) {
		for (int i=size; i < width - size; i++) {
			for (int j=size; j < height - size; j++) {
				osg::Vec4f *dataPtr = (osg::Vec4f*)image->data(i,j);
				osg::Vec4f v00, v01, v02;
				osg::Vec4f v10, v12;
				osg::Vec4f v20, v21, v22;

				v00 = getImageValue(image, i - 1, j - 1, empty);
				v01 = getImageValue(image, i, j - 1, empty);
				v02 = getImageValue(image, i + 1, j - 1, empty);

				v10 = getImageValue(image, i, j - 1, empty);
				v12 = getImageValue(image, i, j + 1, empty);

				v20 = getImageValue(image, i + 1, j - 1, empty);
				v21 = getImageValue(image, i, j + 1, empty);
				v22 = getImageValue(image, i + 1, j + 1, empty);

				float nonZero =   ceilf(v00.a()) + ceilf (v10.a())
								+ ceilf(v20.a())
								+ ceilf(v01.a()) + ceilf (dataPtr->a())
								+ ceilf(v21.a())
								+ ceilf(v02.a()) + ceilf (v12.a())
								+ ceilf(v22.a());

				float valR =  (*dataPtr).r();
				valR +=   v00.r() + v10.r() + v20.r()
						+ v01.r() + v21.r()
						+ v02.r() + v12.r() + v22.r();
				valR /= 9.0;

				float valG =  (*dataPtr).g();
				valG +=   v00.g() + v10.g() + v20.g()
						+ v01.g() + v21.g()
						+ v02.g() + v12.g() + v22.g();
				valG /= 9.0;


				float valB =  (*dataPtr).r();
				valB +=   v00.b() + v10.b() + v20.b()
						+ v01.b() + v21.b()
						+ v02.b() + v12.b() + v22.b();
				valB /= 9.0;


				float valA =  (*dataPtr).r();
				valA +=   v00.a() + v10.a() + v20.a()
						+ v01.a() + v21.a()
						+ v02.a() + v12.a() + v22.a();
				valA /= 9.0;

				//XGIS_LOG_DEBUG2("%f %f %f %f", valR, valG, valB, valA);


				osg::Vec4f colour;

				if (nonZero != 0) {
					colour = osg::Vec4f(valR,valG,valB,valA);
					XGIS_LOG_DEBUG5("%f %f %f %f", valR, valG, valB, valA);
				} else {
					colour = osg::Vec4f(0,0,0,0);
				}

				*dataPtr = colour;
			}
		}
	}
}


/*jesus. continue this some other day.
DONE!*/
void XGisImage::generateFractal(osg::Image* image, float randRange,
								bool opacity, bool wrap, bool low) {
	int width = image->s();
	int height = image->t();

	osg::Vec4f empty = osg::Vec4f(0,0,0,0);

	int unit = width-1;

	/*beginning seed*/
	float s = (rand() % 100)/100.0;
	osg::Vec4f colour1;
	if (low) {
		colour1 = (opacity ? osg::Vec4f(1,1,1,0) : osg::Vec4f(0,0,0,1));
	} else {
		colour1 = (opacity ? osg::Vec4f(1,1,1,s) : osg::Vec4f(s,s,s,1));
	}

	osg::Vec4f* dataPtr = (osg::Vec4f*)image->data(0,0);

	*dataPtr = colour1;

	if (wrap) {

		/*make it wrap around nicely, so seed all corners with the same value*/
		dataPtr = (osg::Vec4f*)image->data(0,unit);
		*dataPtr = colour1;

		dataPtr = (osg::Vec4f*)image->data(unit,unit);
		*dataPtr = colour1;

		dataPtr = (osg::Vec4f*)image->data(unit, 0);
		*dataPtr = colour1;


	} else {
		s = (rand() % 100)/100.0;
		osg::Vec4f colour2;
		colour2 = (opacity ? osg::Vec4f(1,1,1,s) : osg::Vec4f(s,s,s,1));
		dataPtr = (osg::Vec4f*)image->data(0,unit);
		*dataPtr = colour2;

		s = (rand() % 100)/100.0;
		osg::Vec4f colour3;
		colour3 = (opacity ? osg::Vec4f(1,1,1,s) : osg::Vec4f(s,s,s,1));
		dataPtr = (osg::Vec4f*)image->data(unit,unit);
		*dataPtr = colour3;

		s = (rand() % 100)/100.0;
		osg::Vec4f colour4;
		colour4 = (opacity ? osg::Vec4f(1,1,1,s) : osg::Vec4f(s,s,s,1));
		dataPtr = (osg::Vec4f*)image->data(unit, 0);
		*dataPtr = colour4;
	}


	while (unit > 1) {
		XGIS_LOG_DEBUG5("unit %d",unit);
		XGIS_LOG_DEBUG5("rr %f",randRange);

		int subUnit = unit/2;

		if (0 == subUnit) { break; }



		XGIS_LOG_DEBUG5("diamond step");
		for (int i=0; i < width; i+= unit) {
			for (int j=0; j < height; j+=unit) {
				XGIS_LOG_DEBUG5("ds %d %d",i,j);
				/*do the diamond step*/
				if ((i + unit < width) && (j + unit < height)) {

					osg::Vec4f ll = getImageValue(image,i,j, empty, wrap);
					osg::Vec4f ul = getImageValue(image, i,j + unit, empty, wrap);
					osg::Vec4f ur = getImageValue(image, i + unit,j, empty, wrap);
					osg::Vec4f lr = getImageValue(image, i + unit,j + unit,
						empty, wrap);

					osg::Vec4f* resultPtr =
						(osg::Vec4f*)image->data(i + subUnit, j + subUnit);

					float r = (((rand() % 100)/100.0) - 0.5)*randRange;

					float val;
					val = (opacity ? (ul.a() + ur.a() + ll.a() + lr.a())/4.0
						: (ul.r() + ur.r() + ll.r() + lr.r())/4.0);

					val += r;
					val = MIN(1.0, val);
					val = MAX(0.0, val);


					osg::Vec4f colour;
					colour = (opacity ? osg::Vec4f(1, 1, 1, val)
						: osg::Vec4f(val, val, val, 1));
					//osg::Vec4f colour = osg::Vec4f(val, val, val, val);

					if (resultPtr->r() < 0) *resultPtr = colour;

				}

			} /*for j*/
		} /*for i*/


		XGIS_LOG_DEBUG5("square step");
		for (int i=0; i < width; i+= unit) {
			for (int j=0; j < height; j+=unit) {
				/*do the square step*/
				XGIS_LOG_DEBUG5("ss %d %d",i,j);

				osg::Vec4f up;
				osg::Vec4f left;
				osg::Vec4f right;
				osg::Vec4f down;

				float r = (((rand() % 100)/100.0) - 0.5)*randRange;

				osg::Vec4f* resultPtr = NULL;
				float val;
				osg::Vec4f colour;



				/*presume our corner is left*/
				if (i + unit < width) {

					XGIS_LOG_DEBUG5("%d %d left", i, j);

					XGIS_LOG_DEBUG5("getting %d %d", i+subUnit, j+subUnit);
					up = getImageValue(image, i + subUnit,
													j + subUnit, empty, wrap);

					left = getImageValue(image, i, j, empty, wrap);

					right = getImageValue(image, i + unit, j, empty, wrap);

					down = getImageValue(image, i + subUnit,
													j - subUnit, empty, wrap);

					resultPtr =
						(osg::Vec4f*)image->data(i + subUnit,j);


					val = (opacity ?
						(up.a() + down.a() + left.a() + right.a())/4.0
						: (up.r() + down.r() + left.r() + right.r())/4.0);

					val += r;
					val = MIN(1.0, val);
					val = MAX(0.0, val);



					colour = (opacity ? osg::Vec4f(1, 1, 1, val)
						: osg::Vec4f(val, val, val, 1));

					if (resultPtr->r() < 0) *resultPtr = colour;

					if (wrap) {

						/*if you want a seamless texture, write the new values
						on the other edge, too.*/

						if (width-1 == i + subUnit) {
							int newI = 0;
							resultPtr = (osg::Vec4f*)image->data(newI,j);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}

						if (0 == j) {
							int newJ = height - 1;
							int newI = i + subUnit;
							resultPtr = (osg::Vec4f*)image->data(newI, newJ);
							XGIS_LOG_DEBUG5("writing to %d %d", newI, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}

						if (height-1 == j) {
							int newJ = 0;
							int newI = i + subUnit;
							resultPtr = (osg::Vec4f*)image->data(newI, newJ);
							XGIS_LOG_DEBUG5("writing to %d %d", newI, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}


					}

				}

				/*right*/
				if (i - unit > 0) {

					XGIS_LOG_DEBUG5("%d %d right", i, j);

					XGIS_LOG_DEBUG5("getting %d %d", i - subUnit, j - subUnit);
					up = getImageValue(image, i - subUnit,
													j - subUnit, empty, wrap);

					XGIS_LOG_DEBUG5("getting %d %d", i - unit, j);
					left = getImageValue(image, i - unit, j, empty, wrap);

					XGIS_LOG_DEBUG5("getting %d %d", i, j);
					right = getImageValue(image, i, j, empty, wrap);

					XGIS_LOG_DEBUG5("getting %d %d", i-subUnit, j-subUnit);
					down = getImageValue(image, i - subUnit,
													j - subUnit, empty, wrap);

					resultPtr =
						(osg::Vec4f*)image->data(i - subUnit,j);


					val = (opacity ?
						(up.a() + down.a() + left.a() + right.a())/4.0
						: (up.r() + down.r() + left.r() + right.r())/4.0);

					val += r;
					val = MIN(1.0, val);
					val = MAX(0.0, val);


					colour = (opacity ? osg::Vec4f(1, 1, 1, val)
						: osg::Vec4f(val, val, val, 1));

					if (resultPtr->r() < 0) *resultPtr = colour;

					if (wrap) {

						if (0 == i - subUnit) {
							int newI = width - 1;
							resultPtr = (osg::Vec4f*)image->data(newI,j);
							XGIS_LOG_DEBUG5("writing to %d %d", newI, j);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}

						if (0 == j) {
							int newI = i - subUnit;
							int newJ = height - 1;
							resultPtr = (osg::Vec4f*)image->data(newI,newJ);
							XGIS_LOG_DEBUG5("writing to %d %d", newI, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}

						if (height-1 == j) {
							int newI = i - subUnit;
							int newJ = 0;
							resultPtr = (osg::Vec4f*)image->data(newI,newJ);
							XGIS_LOG_DEBUG5("writing to %d %d", newI, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}
					}

				}

				/*presume our corner is up*/
				if (j - unit > 0) {

					XGIS_LOG_DEBUG5("%d %d up", i, j);

					XGIS_LOG_DEBUG5("getting %d %d", i, j);
					up = getImageValue(image, i, j, empty, wrap);

					XGIS_LOG_DEBUG5("getting %d %d", i-subUnit, j-subUnit);
					left = getImageValue(image, i - subUnit,
						j - subUnit, empty, wrap);

					XGIS_LOG_DEBUG5("getting %d %d", i+subUnit, j-subUnit);
					right = getImageValue(image, i + subUnit,
						j - subUnit, empty, wrap);

					XGIS_LOG_DEBUG5("getting %d %d", i, j-unit);
					down = getImageValue(image, i, j - unit, empty, wrap);



					resultPtr =
						(osg::Vec4f*)image->data(i,j - subUnit);


					val = (opacity ?
						(up.a() + down.a() + left.a() + right.a())/4.0
						: (up.r() + down.r() + left.r() + right.r())/4.0);

					val += r;
					val = MIN(1.0, val);
					val = MAX(0.0, val);



					colour = (opacity ? osg::Vec4f(1, 1, 1, val)
						: osg::Vec4f(val, val, val, 1));
					if (resultPtr->r() < 0) *resultPtr = colour;

					if (wrap) {


						if (0 == i) {
							int newI = width - 1;
							int newJ = j - subUnit;
							resultPtr = (osg::Vec4f*)image->data(newI,newJ);
							XGIS_LOG_DEBUG5("writing to %d %d", newI, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}


						if (width-1 == i) {

							int newI = 0;
							resultPtr = (osg::Vec4f*)image->data(newI, j);
							XGIS_LOG_DEBUG5("writing to %d %d", newI, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}


						if (0 == j - subUnit) {
							int newJ = height - 1;
							resultPtr = (osg::Vec4f*)image->data(i,newJ);
							XGIS_LOG_DEBUG5("writing to %d %d", i, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}


					}

				}


				/*presume our corner is down*/
				if (j + unit < height) {

					XGIS_LOG_DEBUG5("%d %d down", i, j);

					XGIS_LOG_DEBUG5("getting %d %d down", i, j+unit);
					up = getImageValue(image, i, j + unit, empty, wrap);

					XGIS_LOG_DEBUG5("getting %d %d down", i-subUnit, j+subUnit);
					left = getImageValue(image, i - subUnit,
						j + subUnit, empty, wrap);

					XGIS_LOG_DEBUG5("getting %d %d down", i+subUnit, j+subUnit);
					right = getImageValue(image, i + subUnit,
						j + subUnit, empty, wrap);

					XGIS_LOG_DEBUG5("getting %d %d down", i, j);
					down = getImageValue(image, i, j, empty, wrap);

					resultPtr =
						(osg::Vec4f*)image->data(i,j + subUnit);

					val = (opacity ?
						(up.a() + down.a() + left.a() + right.a())/4.0
						: (up.r() + down.r() + left.r() + right.r())/4.0);

					val += r;
					val = MIN(1.0, val);
					val = MAX(0.0, val);


					colour = (opacity ? osg::Vec4f(1, 1, 1, val)
						: osg::Vec4f(val, val, val, 1));
					if (resultPtr->r() < 0) *resultPtr = colour;

					if (wrap) {

						if (0 == i) {
							int newI = width -1;
							int newJ = j + subUnit;
							resultPtr = (osg::Vec4f*)image->data(newI, newJ);
							XGIS_LOG_DEBUG5("%f writing to %d %d", resultPtr->r(), newI, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}


						if (width-1 == i) {
							int newI = 0;
							int newJ = j + subUnit;
							resultPtr = (osg::Vec4f*)image->data(newI, newJ);
							XGIS_LOG_DEBUG5("writing to %d %d", newI, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}


						if (height-1 == j) {
							int newJ = 0;
							resultPtr = (osg::Vec4f*)image->data(i,newJ);
							XGIS_LOG_DEBUG5("writing to %d %d", i, newJ);
							if (resultPtr->r() < 0) *resultPtr = colour;
						}
					}

				}


			}/*for j*/
		}/*for i*/


		randRange /= 2;
		unit /= 2;



	} /*while unit*/

}

