#include "config.h"

#include <osgDB/WriteFile>

#include <xgislib/xputils.h>

#include "consoleui.h"

#define RENDER_IMG_SUFFIX "png"

struct SaveImageCallback : public osg::Camera::DrawCallback {
private:
	osg::ref_ptr<osg::Image> image;
	int *n;
	const std::string &dir;
	
public:
	SaveImageCallback(osg::Image *_image, const std::string &_dir, int sf) :
		image(_image), dir(_dir) {
		n = new int; // need to do this via pointer as operator[]
		             // that we need to override is const!
		*n = sf;
	}
	
	virtual void operator() (const osg::Camera &camera) const {
		char fname[1024 + 1];
		snprintf(fname, 1024, "%s%cxgis-render-%06d." RENDER_IMG_SUFFIX,
			dir.c_str(), dirSlash, *n);
		if(osgDB::writeImageFile(*(image.get()), fname)) {
			XGIS_LOG_INFO("Saved image %d to: %s", *n, fname);			
		}
		else {
			XGIS_LOG_WARNING("Failed to save image %d to: %s", *n, fname);						
		}
		(*n)++;
	}
};

void ConsoleUI::setupViewer() {
    setupUI();
}

void ConsoleUI::run() {
    getViewer();
    osg::ref_ptr<osg::Viewport> v = new osg::Viewport(0, 0, width, height);
    viewer->getCamera()->setViewport(v.get());
    if(saveRenderedImages()) {
    	// render to off-screen pixel buffer
	    viewer->getCamera()->setRenderTargetImplementation
	    	(osg::Camera::PIXEL_BUFFER);
	    // create image for rendering to and attach it to main camera
	    osg::ref_ptr<osg::Image> image = new osg::Image;
	    image->allocateImage(width, height, 1, GL_RGB, GL_UNSIGNED_BYTE);
	    viewer->getCamera()->attach(osg::Camera::COLOR_BUFFER, image.get());
	    // set post-draw callback that will save image to file
	    viewer->getCamera()->setPostDrawCallback
	    	(new SaveImageCallback(image.get(), imageDir, startFrame));
	    // create gfx context
	    osg::ref_ptr<osg::GraphicsContext::Traits> traits =
	    	new osg::GraphicsContext::Traits;
	    traits->x = traits->y = 0;
	    traits->width = width;
	    traits->height = height;
	    traits->pbuffer = true;
	    traits->doubleBuffer = false;
	    traits->windowDecoration = false;
	    traits->sharedContext = 0;
	    osg::ref_ptr<osg::GraphicsContext> ctx =
	    	osg::GraphicsContext::createGraphicsContext(traits.get());
	    if(ctx.valid()) {
	    	ctx->realize();
	    	ctx->makeCurrent();
	    }
	    else {
	    	// failed to create gfx context: exit
	    	XGIS_LOG_FATAL("Failed to create graphics context for image rendering.");
	    	return;
	    }
	    viewer->getCamera()->setGraphicsContext(ctx.get());
    }
    // we have initialized UI
    postInitialized();
    // run startup procedure: load world
    startup();
    // wait for startup to finish
    startupSemaphore.wait();
    // now update and render the scene graph until we reach the end of track
    double stime = (float)startFrame/(float)viewer->getMaxFPS();
    while(!viewer->done()) {
    	if(saveRenderedImages()) {
    		/* we want to render images */
    		viewer->advance(stime);
    		viewer->eventTraversal();
			int i = 0;
    		do {
    			viewer->updateTraversal();
				if(i++ % 100 == 0) {
	    			XGIS_LOG_INFO("Pending tasks: %d",
	    				world->getDataLoader()->getTaskCount());
				}
    		} while(world->getDataLoader()->getTaskCount() > 0);
    		viewer->renderingTraversals();
			stime += 1.0/viewer->getMaxFPS();
    	}
    	else {
    		/* we don't really want to render images;
			 * we thus use real time to enable load performance measurments */
    		viewer->frameLimiter();
    		viewer->advance();
    		viewer->eventTraversal();
    		viewer->updateTraversal();
			stime = viewer->getFrameStamp()->getSimulationTime();
    	}
    }
    XGIS_LOG_INFO("Console viewer done.");
}
