#include "config.h"

#ifdef WIN32
#	define _WINSOCKAPI_
#  include <windows.h>
#endif

#include <iostream>
#include <sstream>

#include "obsviewer.h"
#include "obscenemanipulator.h"
#include "hud.h"
#include "skydome.h"

#include <xgislib/sun.h>
#include <xgislib/world.h>
#include <xgislib/xputils.h>
#include <xgislib/coords.h>
#include <xgislib/pointdata.h>

#include <osg/Transform>
#include <osg/ShapeDrawable>
#include <osg/PositionAttitudeTransform>


// PeriodicUpdateHandler requests a terrain update every period
class PeriodicUpdateHandler : public WorldUpdateHandler {
	int period;
	double last;
	osg::Timer timer;
	
public:
	PeriodicUpdateHandler(int _period) : period(_period) {
		last = timer.tick();
	};
	
	virtual void handle(World *world, const osg::FrameStamp *stamp) {
		double now = timer.tick();

		if(timer.delta_m(last, now) > period &&
			world->getDoAutomaticUpdates() &&
			world->getDataLoader()->getTaskCount() == 0) {
			
			XGIS_LOG_DEBUG5("Invoking periodic detail update.");
			QuadC llQ, urQ;
			world->getUpdateFrame(world->getDetailRefPoint(), llQ, urQ);
			world->updateDetailLevel(llQ, urQ);
		}
		last = now;		
	};
};

// WorldLoader takes care of loading terrain as the camera moves
class WorldLoader : public WorldUpdateHandler {
	World *world;
	ObsceneManipulator *manipulator;

public:
	
	WorldLoader(World *_world, ObsceneManipulator *_manipulator) :
		world(_world), manipulator(_manipulator) {}
	
	void handle(class World *world, const osg::FrameStamp *stamp) {
		world->setDetailRefPoint(manipulator->getEye());
	}
};

class SunUpdater : public WorldUpdateHandler {
	Sun *sun;
	
	int period;
	double last;
	osg::Timer timer;
public:	
	SunUpdater(Sun *_sun, int _period) :
	  sun(_sun), period(_period) {
		last = timer.tick();
	  }

	virtual void handle(class World *world, const osg::FrameStamp *stamp) {


		double now = timer.tick();

		if(world->getDoAutomaticUpdates() &&
			timer.delta_m(last, now) > period) {
			
			XGIS_LOG_DEBUG5("Invoking periodic sun update.");
			time_t rawtime = (time_t)(world->getConfiguration()->getStartTime()
					+ stamp->getSimulationTime());
			sun->setSunPosition(rawtime);

			sun->setLightBrightness();

			last = now;
		}
		
	}
};

class SkyUpdater : public WorldUpdateHandler {
	SkyDome *skyDome;
	
	int period;
	double last;
	osg::Timer timer;

public:	
	SkyUpdater(SkyDome* _skyDome, int _period) :
	  skyDome(_skyDome), period(_period) {
		last = timer.tick();
	  }

	virtual void handle(class World *world, const osg::FrameStamp *stamp) {

		double now = timer.tick();

		// skyDome->setCenter(world->getDetailRefPoint());

		if(world->getDoAutomaticUpdates() &&
			timer.delta_m(last, now) > period) {
			
			XGIS_LOG_DEBUG5("Invoking periodic sky update.");
			time_t rawtime = (time_t)(world->getConfiguration()->getStartTime()
				+ stamp->getSimulationTime());
			skyDome->setSkyColour(rawtime);
			last = now;	
		}
		
	}
};

class HUDUpdater : public WorldUpdateHandler {
	HUD *hud;
	ObsceneManipulator *manipulator;

public:
	HUDUpdater(HUD *_hud, ObsceneManipulator *_manipulator) :
		hud(_hud), manipulator(_manipulator) {}

	virtual void handle(class World *world, const osg::FrameStamp *stamp) {
		// this one should be done only on resize!
		//hud->updateTextPosition();

		hud->updateTasks(world->getDataLoader()->getTaskCount());

		osg::Vec3 fwd, north(0, 1, 0);
		fwd = manipulator->getForward();
		fwd[2] = 0;
		fwd.normalize();

		double angle = fwd*north;		
		angle = acos(angle);
		if(fwd[0] < 0) angle = 2*M_PI - angle;
		
		hud->updateCompass(angle);

		hud->updatePosition(manipulator->getSpeed(), manipulator->getEye());

		hud->updateStatus();
	}	
};
