#include "config.h"

#include "hud.h"

#include <sstream>

#include <osg/Geometry>
#include <osg/PositionAttitudeTransform>
#include <osg/MatrixTransform>

#include <osgUtil/Tessellator>

#include "xgislib/osgutils.h"
#include "xgislib/xputils.h"

/* world edge detector */
struct Edge {
	Edge(int _u0, int _v0, int _u1, int _v1) :
		u0(_u0), v0(_v0), u1(_u1), v1(_v1) {
	}
	
	int u0, v0, u1, v1;
};

int HUD::getWorldEdgeVertices(World *world, float mapQuadSize,
	osg::Vec3Array *vertices) {
	QuadC llQ, urQ;
	std::vector<Edge> edges;

	XGIS_LOG_DEBUG5("Starting edge detection.");
	
	llQ = world->getLLQuadC();
	urQ = world->getURQuadC();
	
	for(int u = llQ.u; u <= urQ.u; u++) {
		for(int v = llQ.v; v <= urQ.v; v++) {
			/* for each quad: detect edges */
			Quad *q = world->getQuad(QuadC(u, v));
			if(NULL != q) {
				if(NULL == world->getQuad(QuadC(u + 1, v))) {
					/* nothing to the east... */
					edges.push_back(Edge(u + 1, v, u + 1, v + 1));
				}
				if(NULL == world->getQuad(QuadC(u - 1, v))) {
					/* nothing to the west... */
					edges.push_back(Edge(u, v + 1, u, v));
				}
				if(NULL == world->getQuad(QuadC(u, v + 1))) {
					/* nothing to the north... */
					edges.push_back(Edge(u + 1, v + 1, u, v + 1));
				}
				if(NULL == world->getQuad(QuadC(u, v - 1))) {
					/* nothing to the south... */
					edges.push_back(Edge(u, v, u + 1, v));
				}
			}
		}
	}
	
	/* at this point (and after O(U*V) time) we have all the edges in world in no
	 * particular order. now we need to get a sequence of vertices from the edges
	 * that will describe a closed polygon. we will use O((U*V)^2) more time. */
	int nu, nv, su, sv, n = 0;
	Edge e = *(edges.begin());
	edges.erase(edges.begin());
	vertices->push_back
		(osg::Vec3((e.u0 - llQ.u)*mapQuadSize, (e.v0 - llQ.v)*mapQuadSize, 0));
	nu = e.u1;
	nv = e.v1;
	su = e.u0;
	sv = e.v0;
	n++;
	XGIS_LOG_DEBUG5("Inserted (%d,%d), next (%d,%d)", su, sv, nu, nv);
	while(!edges.empty()) {
		std::vector<Edge>::iterator i = edges.begin();
		while(i != edges.end()) {
			e = *i;
			if(e.u0 == nu && e.v0 == nv) {
				vertices->push_back
					(osg::Vec3((nu - llQ.u)*mapQuadSize, (nv - llQ.v)*mapQuadSize, 0));
				nu = e.u1;
				nv = e.v1;
				edges.erase(i);
				n++;
				XGIS_LOG_DEBUG5("Inserted (%d,%d), next (%d,%d)", e.u0, e.v0, nu, nv);
				break;
			}
			else if(e.u1 == nu && e.v1 == nv) {
				vertices->push_back
					(osg::Vec3((nu - llQ.u)*mapQuadSize, (nv - llQ.v)*mapQuadSize, 0));
				nu = e.u0;
				nv = e.v0;
				edges.erase(i);
				n++;
				XGIS_LOG_DEBUG5("Inserted (%d,%d), next (%d,%d)", e.u1, e.v1, nu, nv);
				break;
			}
			i++;
		}
	}
	
	/* supposedly, we should have ended where we started! */
	XGIS_LOG_DEBUG5("S(%d,%d), E(%d,%d)", su, sv, nu, nv);
	if(su != nu || sv != nv) {
		XGIS_LOG_DEBUG5("Polygon does not close. Strange... ;)");
	}
	
	return n;
}

HUD::HUD(osg::Viewport *_viewport, World *_world, ResourceBag *resourceBag) :
	osg::Group(), viewport(_viewport), world(_world), mapDim(0) {

	createStatusBar(world.get(), resourceBag);
	createMap(world.get());
	createInfoDisplay(world.get(), resourceBag);
	createCompass(world.get());
}

void HUD::updateTasks(int tasks) {
	char s[1024 + 1];
	
	snprintf(s, 1024, "Pending tasks: %5d", tasks);
	textTasks->setText(s);
}

void HUD::updatePosition(double speed, const osg::Vec3 &position) {
	char s[1024 + 1];
	
	Coords::GaussKr gkr = Coords::GaussKr(position.x(), position.y());
	snprintf(s, 1024, "Gauss-Krueger: (%8.1lf,%8.1lf)", gkr.x, gkr.y);
	textGaussKr->setText(s);

	Coords::WGS84 wgs = Coords::WGS84(gkr.toWGS84());
	snprintf(s, 1024, "WGS84: (%8.4lf,%8.4lf)", wgs.latit, wgs.longit);
	textLatLon->setText(s);
	
	QuadC qc = world->getQuadFromPoint(gkr);
	int x = (qc.u - world->getLLQuadC().u);
	int y = (qc.v - world->getLLQuadC().v);
	mapCursorTransform->setPosition(osg::Vec3d(x*mapQuadSize, y*mapQuadSize, 0.0));

	snprintf(s, 1024, "Speed: %6.1lf m/s, alt: %6.1lf m", speed, position.z());
	textSpeedAlt->setText(s);
}

void HUD::updateCompass(double angle) {
	osg::Vec3d pos = osg::Vec3d(30, 30, 0);
	osg::Matrix m = orientationTransform->getMatrix();
	orientationTransform->setMatrix(osg::Matrix::translate(-pos)*
		osg::Matrix::rotate(angle, osg::Vec3f(0,0,1))*
		osg::Matrix::translate(pos));
}

void HUD::updateTextPosition() {
	osg::BoundingBox bb;
    for(unsigned int i=0;i<textGeode->getNumDrawables();++i)
    {
        bb.expandBy(textGeode->getDrawable(i)->getBound());
    }
   
    float depth = bb.zMin()-0.1;
	float margin = 15;
	float compassSide = 60;
	
	textPosition->setPosition(osg::Vec3d(margin, 2*statusBarFontSize, depth));
	
	if (mapDim >0) {
		double minW = viewport->width() - mapDim - margin;
		double minH = 2*statusBarFontSize;
		mapPosition->setPosition(osg::Vec3d(minW, minH, -0.2));
	}		
	
	compassPosition->setPosition(osg::Vec3d(viewport->width() - compassSide - margin,
											viewport->height() - compassSide - margin,
											depth));

}

void HUD::createMap(World *world) {	
	/* map goes in its own group */
	osg::Group *mapGroup = new osg::Group();
	
	/* a transform to position the map display on screen */
	osg::PositionAttitudeTransform *positionTransform = 
		mapPosition = new osg::PositionAttitudeTransform();
	
	positionTransform->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
	mapGroup->addChild(positionTransform);

	/* another transform to move the eye position indicator relative to
	 * map position */
	mapCursorTransform = new osg::PositionAttitudeTransform();
	positionTransform->addChild(mapCursorTransform);

	/* add a map geode to position transform */
	osg::Geode *mapGeode = new osg::Geode();
	positionTransform->addChild(mapGeode);

	/* no lighting for map */
	osg::StateSet *stateset = mapGeode->getOrCreateStateSet();
    stateset->setMode(GL_LIGHTING, osg::StateAttribute::OFF);
	
    /* map geometry */
	osg::Geometry *geom = new osg::Geometry();
    mapGeode->addDrawable(geom);

	osg::Vec3Array *vertices = new osg::Vec3Array;
    osg::Vec4Array *colors = new osg::Vec4Array;
    osg::Vec3Array *normals = new osg::Vec3Array;
    normals->push_back(osg::Vec3(0.0f,0.0f,1.0f));
	
    QuadC llQ = world->getLLQuadC(), urQ = world->getURQuadC();
    int mw = urQ.u - llQ.u + 1; 
    int mh = urQ.v - llQ.v + 1; 
    
	mapQuadSize = .5;
	/* make the map a single polygon. simpler scene graph. */
	int nv = getWorldEdgeVertices(world, mapQuadSize, vertices);

	osg::DrawElementsUInt* mapPoly =
				new osg::DrawElementsUInt(osg::PrimitiveSet::POLYGON);
	for(int i = 0; i < nv; i ++) {
		mapPoly->push_back(i);
	}
	
	normals->push_back(osg::Vec3(0.0, 0.0, 1.0));
	colors->push_back(osg::Vec4(.75, .75, .75, .6));
    geom->setVertexArray(vertices);
    geom->setNormalArray(normals);
	geom->setNormalBinding(osg::Geometry::BIND_OVERALL);
    geom->setColorArray(colors);
	geom->setColorBinding(osg::Geometry::BIND_OVERALL);
	geom->addPrimitiveSet(mapPoly);

	// we need to retessellate the polygons as they are not necessarily
	// convex
	osgUtil::Tessellator tsl;
	tsl.retessellatePolygons(*geom);
	
	/* adjust geometry state set */
    stateset = geom->getOrCreateStateSet();
    stateset->setMode(GL_BLEND, osg::StateAttribute::ON);
    stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

    /* position the map relative to the screen */
    mapDim = mapQuadSize*mw;

    double border = 10;
	double minW = viewport->width() - mapQuadSize*mw - border;
	double minH = 2*statusBarFontSize;
	positionTransform->setPosition(osg::Vec3d(minW, minH, -0.2));

	/* create another geode for eye position indicator */
	osg::Geode *indicatorGeode = new osg::Geode();
	mapCursorTransform->addChild(indicatorGeode);
	mapCursorTransform->setPosition
		(osg::Vec3d(mw*mapQuadSize/2, mh*mapQuadSize/2, 0.0));
	
	/* adjust indicator geode state set */
	stateset = indicatorGeode->getOrCreateStateSet();
    stateset->setMode(GL_LIGHTING, osg::StateAttribute::OFF);
	
    /* create indicator geometry */
	geom = new osg::Geometry();
    indicatorGeode->addDrawable(geom);
	
	vertices = new osg::Vec3Array;
    colors = new osg::Vec4Array;
    
    /* a rectangle for indicator (with each side 8 quads long) */
	vertices->push_back(osg::Vec3(4*mapQuadSize, -4*mapQuadSize, 0));
	vertices->push_back(osg::Vec3(4*mapQuadSize, 4*mapQuadSize, 0));
	vertices->push_back(osg::Vec3(-4*mapQuadSize, 4*mapQuadSize, 0));
	vertices->push_back(osg::Vec3(-4*mapQuadSize, -4*mapQuadSize, 0));
	
	/* indicator will be bright red */
	colors->push_back(osg::Vec4(1.0, 0.0, 0.0, 1.0));

    geom->setVertexArray(vertices);
    geom->setNormalArray(normals);
	geom->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);
    geom->setColorArray(colors);
	geom->setColorBinding(osg::Geometry::BIND_PER_PRIMITIVE);
	geom->addPrimitiveSet(new osg::DrawArrays(GL_QUADS, 0, 4));

	/* adjust state set of the indicator geometry */
    stateset = geom->getOrCreateStateSet();
    stateset->setMode(GL_BLEND, osg::StateAttribute::ON);
    stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	addChild(mapGroup);
}

void HUD::createCompass(World *world) {
	/* compass goes in another geode! */
	osg::Group* compassGroup = new osg::Group();

	osg::Geode* compassGeode = new osg::Geode();
	osg::StateSet* stateset = compassGeode->getOrCreateStateSet();
    stateset->setMode(GL_LIGHTING,osg::StateAttribute::OFF);
	
	osg::Geometry* geom = new osg::Geometry();
	osg::PositionAttitudeTransform* positionTransform = 
	compassPosition = new osg::PositionAttitudeTransform();
	//osg::PositionAttitudeTransform* orientationTransform = new osg::PositionAttitudeTransform();
	positionTransform->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
	orientationTransform = new osg::MatrixTransform();

	osg::Vec3Array* vertices = new osg::Vec3Array;

    float depth = -0.1;
	float side = 60;
	float border = 15;

	double minW = viewport->width() - side-border;
	double minH = viewport->height() - side-border;

    vertices->push_back(osg::Vec3(0,0,0));
    vertices->push_back(osg::Vec3(0,side,0));
    vertices->push_back(osg::Vec3(side,side,0));
    vertices->push_back(osg::Vec3(side,0,0));
    geom->setVertexArray(vertices);

    osg::Vec3Array* normals = new osg::Vec3Array;
    normals->push_back(osg::Vec3(0.0f,0.0f,1.0f));
    geom->setNormalArray(normals);
	geom->setNormalBinding(osg::Geometry::BIND_PER_PRIMITIVE);

    osg::Vec4Array* colors = new osg::Vec4Array;
    
	colors->push_back(osg::Vec4(0.8f,0.8f,0.8f,0.5f));
    geom->setColorArray(colors);
	geom->setColorBinding(osg::Geometry::BIND_OVERALL);

	geom->addPrimitiveSet(new osg::DrawArrays(GL_QUADS,0,4));

    osg::StateSet* statesetg = geom->getOrCreateStateSet();
    statesetg->setMode(GL_BLEND,osg::StateAttribute::ON);
    statesetg->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

    compassGeode->addDrawable(geom);

	positionTransform->setPosition(osg::Vec3d(minW, minH, 2*depth));
	
	/*now for the real compass*/
	osg::Geode* dirGeode = new osg::Geode();
	
	orientationTransform->addChild(dirGeode);
	osg::Geometry* geomDir = new osg::Geometry();
	osg::Vec3Array* vertDir = new osg::Vec3Array;
		
	/*south*/
	vertDir->push_back(osg::Vec3(side/3.0,side/2.0,depth));
	vertDir->push_back(osg::Vec3(2*side/3.0,side/2.0,depth));
	vertDir->push_back(osg::Vec3(side/2.0,0,depth));

	/*west*/
	vertDir->push_back(osg::Vec3(0,side/2.0,depth));
	vertDir->push_back(osg::Vec3(side/2.0,2*side/3.0,depth));
	vertDir->push_back(osg::Vec3(side/2.0,side/3.0,depth));

	/*east*/
	vertDir->push_back(osg::Vec3(side/2.0,2*side/3.0,depth));
	vertDir->push_back(osg::Vec3(side,side/2.0,depth));
	vertDir->push_back(osg::Vec3(side/2.0,side/3.0,depth));

	/*north*/
	vertDir->push_back(osg::Vec3(side/3.0,side/2.0,depth));
	vertDir->push_back(osg::Vec3(side/2.0,side,depth));
	vertDir->push_back(osg::Vec3(2*side/3.0,side/2.0,depth));

	geomDir->setVertexArray(vertDir);

    osg::StateSet* ss = geomDir->getOrCreateStateSet();
    ss->setMode(GL_BLEND,osg::StateAttribute::ON);
    ss->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
	ss->setMode(GL_LIGHTING,osg::StateAttribute::OFF);

	osg::Vec4Array* colors2 = new osg::Vec4Array;
	
    colors2->push_back(osg::Vec4(0.8f,0.0f,0.0f,0.8f));
	colors2->push_back(osg::Vec4(0.8f,0.0f,0.0f,0.8f));
	colors2->push_back(osg::Vec4(0.8f,0.0f,0.0f,0.8f));
	colors2->push_back(osg::Vec4(0.0f,0.0f,0.0f,0.8f));				

    geomDir->setColorArray(colors2);
	geomDir->setColorBinding(osg::Geometry::BIND_PER_PRIMITIVE);
	
	geomDir->setNormalArray(normals);
	geomDir->setNormalBinding(osg::Geometry::BIND_OVERALL);

	geomDir->addPrimitiveSet(new osg::DrawArrays(GL_TRIANGLES,0,12));
	dirGeode->addDrawable(geomDir);
	osg::Vec3d pos = osg::Vec3d(side/2.0,side/2.0,0);
	osg::Matrix m = orientationTransform->getMatrix();
	
	positionTransform->addChild(orientationTransform);
	positionTransform->addChild(compassGeode);
	compassGroup->addChild(positionTransform);

	addChild(compassGroup);
}

void HUD::createStatusBar(World *world, ResourceBag *resourceBag) {
	osg::Geode *geode = statusGeode = new osg::Geode;
    osg::StateSet* stateset = geode->getOrCreateStateSet();
    stateset->setMode(GL_LIGHTING,osg::StateAttribute::OFF);

    osg::PositionAttitudeTransform* positionTransform =
		new osg::PositionAttitudeTransform();
	positionTransform->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
	positionTransform->addChild(geode);
	
	newStatusTimeout = 0;
	statusTimeout = 0;
	statusVisible = false;
	
	statusText = new osgText::Text;
	
	statusBarFontSize = 10;
	
	osgText::Font *timesFont = resourceBag->getFont(FONT_TIMES);

	statusText->setCharacterSizeMode(osgText::Text::OBJECT_COORDS); 
	statusText->setFont(timesFont);
	statusText->setPosition(osg::Vec3(statusBarFontSize/2, statusBarFontSize/2, -0.1f));
	statusText->setText("");
	statusText->setColor(osg::Vec4(0, 0, 0, 1));
	statusText->setCharacterSize(statusBarFontSize);
	
	positionTransform->setPosition(osg::Vec3(0, 0, 0));
	
	addChild(positionTransform);
}

void HUD::createInfoDisplay(World *world, ResourceBag *resourceBag) {
    /* finally, text in its own geode */
    osg::Geode* geode = textGeode = new osg::Geode;
	osg::PositionAttitudeTransform* positionTransform =
		textPosition = new osg::PositionAttitudeTransform();
	positionTransform->addChild(geode);
	positionTransform->setReferenceFrame(osg::Transform::ABSOLUTE_RF);

	osgText::Font* timesFont = resourceBag->getFont(FONT_TIMES);

    // turn lighting off for the text and disable depth test
    // to ensure its always ontop.
    osg::StateSet* stateset = geode->getOrCreateStateSet();
    stateset->setMode(GL_LIGHTING,osg::StateAttribute::OFF);
    stateset->setMode(GL_BLEND,osg::StateAttribute::ON);
	stateset->setMode(GL_DEPTH_TEST,osg::StateAttribute::OFF);
    stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	double offset = 10;
	double fontSize = 15;

    osg::Vec3 position(offset,offset,0.0f);
    osg::Vec3 delta(0.0f,20.0f,0.0f);

    osgText::Text *text;
    {
        text = textTasks = new osgText::Text;
        text->setCharacterSizeMode(osgText::Text::OBJECT_COORDS); 
        geode->addDrawable( text );

        text->setFont(timesFont);
        text->setPosition(position);
        text->setText("Pending tasks:     0");
		text->setColor(osg::Vec4(0,0,0,0.6));
		text->setCharacterSize(fontSize);

        position += delta;
    } 

    {
        text = textGaussKr = new osgText::Text;
        text->setCharacterSizeMode(osgText::Text::OBJECT_COORDS);
        geode->addDrawable( text );

        text->setFont(timesFont);
        text->setPosition(position);
        text->setText("Gauss-Krueger: (000000.0,000000.0)");
		text->setColor(osg::Vec4(0,0,0,0.6));
		text->setCharacterSize(fontSize);

        position += delta;
    } 

    {
        text = textLatLon = new  osgText::Text;
        text->setCharacterSizeMode(osgText::Text::OBJECT_COORDS);
        geode->addDrawable( text );

        text->setFont(timesFont);
        text->setPosition(position);
        text->setText("WGS84: (000.0000,000.0000)");
		text->setColor(osg::Vec4(0,0,0,0.6));
		text->setCharacterSize(fontSize);

        position += delta;
    }

    {
        text = textSpeedAlt = new  osgText::Text;
        text->setCharacterSizeMode(osgText::Text::OBJECT_COORDS);
        geode->addDrawable( text );

        text->setFont(timesFont);
        text->setPosition(position);
		text->setText("Speed: 0.0 m/s, alt: 0.0 m");
		text->setColor(osg::Vec4(0,0,0,0.6));
		text->setCharacterSize(fontSize);

        position += delta;
    } 

    {
        text = textWorld = new  osgText::Text;
        text->setCharacterSizeMode(osgText::Text::OBJECT_COORDS);
        geode->addDrawable( text );

        text->setFont(timesFont);
        text->setPosition(position);
        text->setText(world->getName());
		text->setColor(osg::Vec4(0,0,0,0.6));
		text->setCharacterSize(fontSize);

        position += delta;
    } 

    {
        osg::BoundingBox bb;
        for(unsigned int i = 0; i < geode->getNumDrawables(); ++i) {
            bb.expandBy(geode->getDrawable(i)->getBound());
        }

        osg::Geometry* geom = new osg::Geometry();

        osg::Vec3Array* vertices = new osg::Vec3Array;
        float depth = bb.zMin();
		float margin = 10;

		double xSide = (bb.xMax() - bb.xMin()) + 2*margin;
		double ySide = (bb.yMax() - bb.yMin()) + 2*margin;

        vertices->push_back(osg::Vec3(0, ySide, depth));
        vertices->push_back(osg::Vec3(0, 0, depth));
        vertices->push_back(osg::Vec3(xSide, 0, depth));
        vertices->push_back(osg::Vec3(xSide, ySide, depth));

		positionTransform->setPosition(osg::Vec3(bb.xMin(), bb.yMin() + 2*statusBarFontSize, depth));

        geom->setVertexArray(vertices);

        osg::Vec3Array* normals = new osg::Vec3Array;
        normals->push_back(osg::Vec3(0.0f, 0.0f, 1.0f));
        geom->setNormalArray(normals);
        geom->setNormalBinding(osg::Geometry::BIND_OVERALL);

        osg::Vec4Array* colors = new osg::Vec4Array;
        colors->push_back(osg::Vec4(0.8f, 0.8f, 0.8f, 0.3f));
        geom->setColorArray(colors);
        geom->setColorBinding(osg::Geometry::BIND_OVERALL);

        geom->addPrimitiveSet(new osg::DrawArrays(GL_QUADS, 0, 4));

        geode->addDrawable(geom);
    }

	addChild(positionTransform);
}

void HUD::setStatus(const std::string &text, unsigned int timeout) {
#ifdef XGIS_UI_CONSOLE
	XGIS_LOG_INFO(text.c_str());
#endif
	statusMutex.lock();
	
	newStatusMessage = text;
	newStatusTimeout = timeout;
	
	statusMutex.unlock();
}

void HUD::updateStatus() {
	statusMutex.lock();
	
	if(newStatusTimeout > 0) {
		statusTimeout = timer.time_m() + newStatusTimeout*1000;
		
		statusText->setText(newStatusMessage);
		if(!statusVisible) {
			statusGeode->addDrawable(statusText.get());
			statusVisible = true;
		}
		
		newStatusTimeout = 0;
	}
	else if(statusVisible && statusTimeout > 0) {
		if(timer.time_m() > statusTimeout) {
			statusGeode->removeDrawable(statusText.get());
			statusTimeout = 0;
			statusVisible = false;
		}
	}
	
	statusMutex.unlock();
}

Coords::GaussKr HUD::map2world(float mapX, float mapY) {
	Coords::GaussKr rv(world->getLLPointOfWorld());
	XGIS_LOG_DEBUG2("llpointofW = %f,%f", rv.x, rv.y);
	rv.x += (mapX - mapPosition->getPosition().x()) / mapQuadSize * 1000;
	rv.y += (mapY - mapPosition->getPosition().y()) / mapQuadSize * 1000;
	XGIS_LOG_DEBUG2("map2world(%f,%f)= %f,%f", mapX, mapY, rv.x, rv.y);
	return rv;
}

void HUD::setViewport(osg::Viewport *_viewport) {
	viewport = _viewport;
	updateTextPosition();
}
