#include "config.h"

#include <osg/PolygonMode>

#include <osgUtil/IntersectionVisitor>

#include <osgParticle/ExplosionEffect>
#include <osgParticle/ExplosionDebrisEffect>

#include <xgislib/xputils.h>
#include <xgislib/pickaction.h>
#include <xgislib/worldloader.h>

#include <string.h>

#include "interactivemanipulator.h"

InteractiveManipulator::InteractiveManipulator(World *_world,
	ObsViewer *_viewer, HUD *_hud, const osg::Matrixd &_home) :
	ObsceneManipulator(_world, _viewer, _home), hud(_hud) {
	_usePolytopeIntersector = false;
	_useWindowCoordinates = false;
	_mX = _mY = 0;
}

InteractiveManipulator::~InteractiveManipulator() {
	
}

bool InteractiveManipulator::handleKeyDown(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter& us)
{
	assert(ea.getEventType() == osgGA::GUIEventAdapter::KEYDOWN);
	
	float f = (ea.getModKeyMask() & osgGA::GUIEventAdapter::MODKEY_SHIFT) ?
		shiftFactor : 1;

#if 0
	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 4; j++)
			XGIS_LOG_DEBUG4("%4.2f ", matrix(i, j));
		XGIS_LOG_DEBUG4("\n");
	}
#endif /* 0/1 */

	float dlt;

	dlt = 1.0;
	int key = ea.getKey();
	
	switch(key) {
	// E.T. go home
	case 'h': 
	case 'H':
		home();
		break;

	//view/hide details
	case 'v':
	case 'V':
		world->getUpdateCallback()->toggleDetails();
		hud->setStatus("Switching details visibility...", 5);
		break;

	// set draw mode	
	case 'p':
		world->getUpdateCallback()->setPolygonMode
			(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::LINE);
		hud->setStatus("Setting wireframe display mode...", 5);
		break;
	case 'P':
		world->getUpdateCallback()->setPolygonMode
			(osg::PolygonMode::FRONT_AND_BACK, osg::PolygonMode::FILL);
		hud->setStatus("Setting filled polygon display mode...", 5);
		break;

	// moving (in world orientation)
	case 'a': 
	case 'A':
		translate(-f*moveStep*dlt, 0, 0);
		break;
	case 'd':
	case 'D':
		translate(f*moveStep*dlt, 0, 0);
		break;
	case 's':
	case 'S':
		translate(0, -f*moveStep*dlt, 0);
		break;
	case 'w':
	case 'W':
		translate(0, f*moveStep*dlt, 0);
		break;
	
	// moving (in camera orientation)
	case 'j': 
	case 'J':
		left(f*moveStep*dlt);
		break;
	case 'l':
	case 'L':
		right(f*moveStep*dlt);
		break;
	case 'k':
	case 'K':
		backward(f*moveStep*dlt);
		break;
	case 'i':
	case 'I':
		forward(f*moveStep*dlt);
		break;

	// change altitude
	case osgGA::GUIEventAdapter::KEY_Page_Up:
	case 'q':
	case 'Q':
		if (isHeightLocked)
			lockedHeight += f*moveStep*dlt;
		translate(0, 0, f*moveStep*dlt);
		break;
	case osgGA::GUIEventAdapter::KEY_Page_Down:
	case 'e':
	case 'E':
		if (isHeightLocked)
			lockedHeight -= f*moveStep*dlt;
		if (lockedHeight < minEyeHeight)
			lockedHeight = minEyeHeight;
		translate(0, 0, -f*moveStep*dlt);
		break;

	// rotating view
	case osgGA::GUIEventAdapter::KEY_Left: 
		yaw(f*rotStep);
		break;
	case osgGA::GUIEventAdapter::KEY_Right: 
		yaw(-f*rotStep);
		break;
	case osgGA::GUIEventAdapter::KEY_Down:
		pitch(-f*rotStep);
		break;
	case osgGA::GUIEventAdapter::KEY_Up: 
		pitch(f*rotStep);
		break;

	// rolling (rotation around direction vector)
	case osgGA::GUIEventAdapter::KEY_Insert:
	case 'u':
	case 'U':
		roll(-f*rotStep);
		break;
	case osgGA::GUIEventAdapter::KEY_Home:
	case 'o':
	case 'O':
		roll(f*rotStep);
		break;

	// changing current detail level set
	case '.':
		world->nextDetailLevelSet();
		hud->setStatus("New detail level set: " +
			world->getCurrentDetailLevelSet()->description, 5);
		break;		
	case ',':
		world->prevDetailLevelSet();
		hud->setStatus("New detail level set: " +
			world->getCurrentDetailLevelSet()->description, 5);
		break;
		
	// switch automatic detail level updates
	case 'm':
	case 'M':
		world->setDoAutomaticUpdates(!world->getDoAutomaticUpdates());
		hud->setStatus("Automatic updates set to " + std::string(((world->getDoAutomaticUpdates())?"true":"false")), 5);
		break;
		
	// changing move/rotate speed
	case '+':
		if (moveStep < 100000)
			moveStep *= plusFactor;
		if (speedStep < 10000)
			speedStep *= plusFactor;
		break;
	case '-':
		if (moveStep/plusFactor > 0)
			moveStep /= plusFactor;
		if (speedStep/plusFactor > 0)
			speedStep /= plusFactor;
		break;

	//lock/unlock height
	case 'y':
	case 'Y':
		if(!isHeightLocked) {
			osg::Vec3d eye = getEye();
			lockedHeight = world->getTerrainHeight(eye[0], eye[1]);
			if(lockedHeight != -1) {
				lockedHeight = eye[2] - lockedHeight;
				XGIS_LOG_DEBUG2("Locking height to %f\n", lockedHeight);
				isHeightLocked = true;
			}
			hud->setStatus("Locking altitude...", 5);
		}
		else if(isHeightLocked) {
			XGIS_LOG_DEBUG2("Unlocking height.\n");
			hud->setStatus("Unlocking altitude...", 5);
			isHeightLocked = false;
		}
		break;

#ifdef XGIS_REFCOUNT_DEBUG
	// refcount debug report
	case 't':
	case 'T': {
		XGIS_LOG_INFO("Quad refcount report:");
		dbgObserverQuad->report();
		XGIS_LOG_INFO("\n");
		XGIS_LOG_INFO("QuadSingle refcount report:");
		dbgObserverQuadSingle->report();
		XGIS_LOG_INFO("\n");
		XGIS_LOG_INFO("Layer refcount report:");
		dbgObserverLayer->report();
		XGIS_LOG_INFO("\n");
	}
#endif /* XGIS_REFCOUNT_DEBUG */
		
	default:
		return false;
	}
	
	return true;
}

bool InteractiveManipulator::handleMouseButton(const osgGA::GUIEventAdapter &ea,
		osgGA::GUIActionAdapter &aa, bool doubleClick) {

	if(doubleClick && isAnimating()) {
		// a gouble click will immediately cancel each and every on-going
		// motion!
		stop();
		return true;
	}
	
	osg::ref_ptr<osgUtil::IntersectionVisitor> iv = createIV(ea);
	
	if(pickOnHUDMap(iv.get(), ea, doubleClick)) {
		return true;
	}

	if(pickOnScene(iv.get(), ea, doubleClick)) {
		return true;
	}

	return false;
}

osgUtil::IntersectionVisitor *InteractiveManipulator::createIV(const osgGA::GUIEventAdapter &ea) {
	// setup the Intersector object
	osgUtil::LineSegmentIntersector* picker;
	if (!_useWindowCoordinates) {
		// use non dimensional coordinates - in projection/clip space
		picker = new osgUtil::LineSegmentIntersector
			(osgUtil::Intersector::PROJECTION,
			ea.getXnormalized(), ea.getYnormalized());
	}
	else {
		// use window coordinates
		// remap the mouse x,y into viewport coordinates.
		osg::Viewport* viewport = viewer->getCamera()->getViewport();
		float mx = viewport->x() +
			(int)((float)viewport->width()*(ea.getXnormalized()*0.5f + 
			0.5f));
		float my = viewport->y() +
			(int)((float)viewport->height()*(ea.getYnormalized()*0.5f +
			0.5f));
		picker = new osgUtil::LineSegmentIntersector
			(osgUtil::Intersector::WINDOW, mx, my);
    }
	return new osgUtil::IntersectionVisitor(picker);
}

bool InteractiveManipulator::pickOnHUDMap(osgUtil::IntersectionVisitor *iv,
	const osgGA::GUIEventAdapter &ea, bool doubleClick)
	//check if the pick was on the overlay map;
	//if yes, process it & return true
	//otherwise return false
{
	viewer->getHUDCamera()->accept(*iv);
	if (!iv->getIntersector()->containsIntersections()) {
		return false;
	}
	
	if (doubleClick) {
		osgUtil::LineSegmentIntersector::Intersection intersection =
			dynamic_cast<osgUtil::LineSegmentIntersector*>(iv->getIntersector())
			->getFirstIntersection();
		Coords::GaussKr clickLoc = hud->map2world(
				intersection.getWorldIntersectPoint().x(),
				intersection.getWorldIntersectPoint().y());
		switch (ea.getButton()) {
		case osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON:
			jumpTo(osg::Vec3(clickLoc.x, clickLoc.y,
				getEye().z()));
			break;
		case osgGA::GUIEventAdapter::RIGHT_MOUSE_BUTTON:
			flyTo(osg::Vec3(clickLoc.x, clickLoc.y,
				getEye().z()), 5.0);
			break;
		default:
			break;
		}
	}
	return true;
}

void InteractiveManipulator::moveToTerrainAtIntersection
	(const osgUtil::LineSegmentIntersector::Intersection &intersection,
		const osgGA::GUIEventAdapter &ea) {
	const osg::NodePath &nodePath = intersection.nodePath;
	for (unsigned int i = 1; i < nodePath.size(); i++) {
		osg::Node::DescriptionList dscs = nodePath[i]->getDescriptions();
		XGIS_LOG_DEBUG4("  object %s (%s), descriptions: %d",
			nodePath[i]->className(),
			nodePath[i]->getName().c_str(), dscs.size());
		osg::Node::DescriptionList::const_iterator it = dscs.begin();
		while(it != dscs.end()) {
			XGIS_LOG_DEBUG4("Found description '%s' on %p",
				(*it).c_str(), nodePath[i]);
			if(*it == QuadSingle::getTerrainGeodeDescription()) {
				// if this is a terrain geode (identified by an appropriate description)
				osg::Vec3 point = intersection.localIntersectionPoint;
				point.z() += (getEye().z() - world->getTerrainHeight(getEye().x(), getEye().y()));
				if(ea.getModKeyMask() &
					osgGA::GUIEventAdapter::MODKEY_SHIFT) {
					// slow, straight flight to location
					flyTo(point);
				}
				else {
					// 5 second jump over a parabole to location
					flyTo(point, 5.0);
				}
				return;
			}
			it++;
		}
	}
}

void InteractiveManipulator::displayIntersectionLocation
	(const osgUtil::LineSegmentIntersector::Intersection &intersection) {
	char locmsg[128 + 1];
	snprintf(locmsg, 128, "Clicked on location (%f,%f,%f)",
		intersection.localIntersectionPoint.x(),
		intersection.localIntersectionPoint.y(),
		intersection.localIntersectionPoint.z());
	hud->setStatus(locmsg);	
}

int InteractiveManipulator::executePickActions
	(const osgUtil::LineSegmentIntersector::Intersection &intersection,
		const osgGA::GUIEventAdapter &ea, bool doubleClick) {
	const osg::NodePath &nodePath = intersection.nodePath;
	int actionsExecuted = 0;
	XGIS_LOG_DEBUG5("Picking object: ");
	for (unsigned int i = 0; i < nodePath.size(); i++) {
		SceneUserData *udata = (SceneUserData *)nodePath[i]->getUserData();
		XGIS_LOG_DEBUG5("  object %s (%s), user data: %p",
			nodePath[i]->className(), nodePath[i]->getName().c_str(),
			udata);
		if(NULL != udata) {
			if(0 == strcmp(udata->getTypeName(), PickAction::TYPE_NAME)) {
				XGIS_LOG_DEBUG5("  executing pick action.");
				((PickAction *)udata)->pick(nodePath[i], intersection.localIntersectionPoint, ea.getButton(), doubleClick);
				actionsExecuted++;
			}
		}
	}
	return actionsExecuted;
}

bool InteractiveManipulator::pickOnScene(osgUtil::IntersectionVisitor *iv,
	const osgGA::GUIEventAdapter &ea, bool doubleClick) {

	viewer->getCamera()->accept(*iv);
	
	osgUtil::LineSegmentIntersector *picker =
		(osgUtil::LineSegmentIntersector *)iv->getIntersector();
	
	if (picker->containsIntersections()) {

		osgUtil::LineSegmentIntersector::Intersection intersection =
			picker->getFirstIntersection();

		XGIS_LOG_DEBUG5("Picked %f %f %f\n",
			intersection.localIntersectionPoint.x(),
			intersection.localIntersectionPoint.y(),
			intersection.localIntersectionPoint.z());

		if(executePickActions(intersection, ea, doubleClick) > 0) {
			return true;
		}
		
		if(doubleClick) {
			if(ea.getButton() == osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON) {
				// double-click: animate move to location, if a terrain geode
				// is found among the intersections
				moveToTerrainAtIntersection(intersection, ea);
			}
			else if(ea.getButton() == osgGA::GUIEventAdapter::MIDDLE_MOUSE_BUTTON) {
				displayIntersectionLocation(intersection);				
			}
		}
	}
	
	return true;
}

bool InteractiveManipulator::handle(const osgGA::GUIEventAdapter& ea,
	osgGA::GUIActionAdapter& us) {

	ctrl = (ea.getModKeyMask() & osgGA::GUIEventAdapter::MODKEY_CTRL) ? 1  : 0;
	shift = (ea.getModKeyMask() & osgGA::GUIEventAdapter::MODKEY_SHIFT) ? 1  : 0;

	if (ea.getHandled())
		return true;

	switch(ea.getEventType()) {

	case(osgGA::GUIEventAdapter::KEYDOWN): {
		if(handleKeyDown(ea, us)) {
			return true;
		}
		break;
	}
	
	case(osgGA::GUIEventAdapter::PUSH): {
		_mX = ea.getX();
		_mY = ea.getY();
		handleMouseButton(ea, us, false);
		return true;
	}
	
	case(osgGA::GUIEventAdapter::DOUBLECLICK): {
		_mX = ea.getX();
		_mY = ea.getY();
		handleMouseButton(ea, us, true);
		return true;
	}
	
	case(osgGA::GUIEventAdapter::DRAG): {
		float dx = _mX - ea.getX();
		_mX = ea.getX();
		float dy = _mY - ea.getY();
		_mY = ea.getY();
		if(ea.getButtonMask() == osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON) {
			if(dx != 0) {
				rotate(0.01*dx, osg::Vec3f(0, 0, 1));
			}
			if(dy != 0) {
				rotate(0.01*dy, getStrafe());
			}
		}
		else if(ea.getButtonMask() == osgGA::GUIEventAdapter::RIGHT_MOUSE_BUTTON) {
			if(dx != 0) {
				translate(getStrafe()*dx);
			}
			if(dy != 0) {
				osg::Vec3 horizFwd = getForward();
				horizFwd.z() = 0.0;
				horizFwd *= 4.0;
				translate(horizFwd*dy);
			}				
		}
		return true;
	}

	case (osgGA::GUIEventAdapter::SCROLL): {
		double shiftFactor = 1;
		if (ea.getModKeyMask() & osgGA::GUIEventAdapter::MODKEY_SHIFT) {
			shiftFactor *= getShiftFactor();
		}
		if(ea.getScrollingMotion() == osgGA::GUIEventAdapter::SCROLL_DOWN) {
			translate(getForward()*shiftFactor*-20.0);
		}
		else if(ea.getScrollingMotion() == osgGA::GUIEventAdapter::SCROLL_UP) {
			translate(getForward()*shiftFactor*20.0);
		}
		return true;
	}									   

	default:
		break;
	}
	
	return ObsceneManipulator::handle(ea, us);
}
