#include "config.h"

#ifdef WIN32
#	define _WINSOCKAPI_
#	include <windows.h>
#endif

#include <iostream>

#include <osg/Fog>
#include <osg/MatrixTransform>
#include <osg/Timer>

#include <OpenThreads/Thread>

#include "xgisui.h"
#include "skydome.h"
#include "handlers.h"

#include <xgislib/world.h>
#include <xgislib/sun.h>
#include <xgislib/xgisconfig.h>

#include <xgislib/cirrus.h>
#include <xgislib/cumulus.h>
#include <xgislib/stratus.h>

#include <xgislib/xputils.h>
#include <xgislib/gpstrack.h>
#include <xgislib/readerwriterquad.h>
#include <xgislib/readerwriterqcol.h>
#include <xgislib/readerwritergrd6.h>

#include <xgislib/layerbuilder.h>

#include <loaders/xgf/xgfloader.h>
#include <loaders/kml/kmlloader.h>

// register the Quad/Qcol readers with Registry
osgDB::RegisterReaderWriterProxy<ReaderWriterQuad> g_readerWriter_Quad_Proxy;
osgDB::RegisterReaderWriterProxy<ReaderWriterQcol> g_readerWriter_Qcol_Proxy;
osgDB::RegisterReaderWriterProxy<ReaderWriterGrd6> g_readerWriter_Grd6_Proxy;

using namespace std;

/* interval for sky colour updates (in ms) */
#define SKY_UPDATE_INTERVAL 20000
/* interval for sun position updates (in ms) */
#define SUN_UPDATE_INTERVAL 20000

#define XGIS_DEFAULT_QUAD_SIZE 1000

// we need a separate thread to do the initial loading, as we need to refresh
// UI during that time (which can be quite long, especially on slow xDSL
// connections)
class StartupThread : public OpenThreads::Thread {
	std::string worldName;
	const osg::Vec3 base;
	WorldStorage *storage;
	XGisUI *ui;

public:
	StartupThread(const std::string &_worldName, const osg::Vec3 &_base,
		WorldStorage *_storage, XGisUI *_ui) :
		worldName(_worldName), base(_base), storage(_storage), ui(_ui)
	{}

	/* do the loading */
	virtual void run() {
		// create root of our scene graph
		osg::ref_ptr<osg::Group> root = new osg::Group();
		osg::ref_ptr<osg::StateSet> state = root->getOrCreateStateSet();
		state->setMode( GL_LIGHTING, osg::StateAttribute::ON );
		state->setMode( GL_LIGHT0, osg::StateAttribute::ON );
		state->setMode( GL_NORMALIZE, osg::StateAttribute::ON );

		// create a DataLoader
		DataLoader *dataLoader = new SerDataLoader(storage);
		dataLoader->start();
		// create world
		osg::ref_ptr<World> world =
			new World(worldName, XGIS_DEFAULT_QUAD_SIZE, dataLoader);

		// create layer builder
		ResourceBag *rb = new ResourceBag();
		LayerBuilder *lb = new LayerBuilder(world.get(), rb, storage);
		lb->registerLoader(new XGFLoader());

		// setup coarse model
		osg::ref_ptr<osg::HeightField> cmodel = storage->getCoarseHeights();
		osg::ref_ptr<osg::Image> ctex = storage->getCoarseTexture();
		world->setup(base, lb, cmodel.get(), ctex.get());

#if 0
		//MARIANO

		KMLLoader kml;

		osg::ref_ptr<osg::Node> kml_object;

		const char *kmlTestFiles[] = {
				"egrad.kml",
				"exlab.kml",
				"pot_absolute.kml",
				"epotnagrad.kml",
				"fe_fri.kml",
				"pot_claped.kml",
				"epresernovplac.kml",
				"grad.kml",
				"pot_relative.kml",
		};

		for(unsigned int i = 0;
			i < sizeof(kmlTestFiles)/sizeof(const char *);
			i++) {
			XGIS_LOG_INFO("Loading file %d: %s", i, kmlTestFiles[i]);
			kml_object = kml.loadNode(kmlTestFiles[i], world.get());
			world->addChild(kml_object.get());
		}

		//MARIANO
#endif

		Coords::GaussKr ll = world->getLLPointOfWorld(),
						ur = world->getURPointOfWorld();

		if(world->getConfiguration()->getFogSize() > 1.0) {
			osg::StateSet* worldStateSet = world->getOrCreateStateSet();
			osg::Fog *fog = new osg::Fog();
			fog->setMode(osg::Fog::LINEAR);
			fog->setDensity(0.3);
			fog->setColor(osg::Vec4(0.8,0.8,0.8,1.0));
			fog->setStart(world->getConfiguration()->getFarPlane() - world->getConfiguration()->getFogSize());
			fog->setEnd(world->getConfiguration()->getFarPlane());

			worldStateSet->setAttribute(fog, osg::StateAttribute::ON);
			worldStateSet->setMode(GL_FOG, osg::StateAttribute::ON);
		}

		WorldUpdateCallback *wu = world->getUpdateCallback();

		osg::Vec3 c(base.x(), base.y(), 0 );

		// create sun (FIAT LUX!)
		Sun *sunGroup =
			new Sun(world->getConfiguration()->getStartTime(), c);

		wu->addUpdateHandler(new SunUpdater(sunGroup,
			SUN_UPDATE_INTERVAL));
		root->addChild(sunGroup);
		world->addChild(sunGroup->getLightSource());

		float r = world->getConfiguration()->getFarPlane();
		// add sky dome to scene graph
		osg::ref_ptr<SkyDome> skyDome = new SkyDome(c, r, storage,
			world.get(), 2.0, world->getConfiguration()->getStartTime());
		root->addChild(skyDome.get());
		wu->addUpdateHandler(new SkyUpdater(skyDome.get(), SKY_UPDATE_INTERVAL));

		srand((unsigned)time(0));

		// add HUD camera
		osg::ref_ptr<HUD> hud =
			new HUD(new osg::Viewport(0, 0, 800, 600), world.get(), rb);

//#define USE_CLOUDS
#ifdef USE_CLOUDS

		XGIS_LOG_DEBUG2("start cloud creation");

		float height = 2500;

		//int numCloudsOut = 200;
		int numCloudsOut = 1;

		int numParticles = 10000;
		int partSize = 100;
		int meanSizeX = 10000;

		int maxOffsetOut = 25000;



		for (int j=0; j < numCloudsOut; j++) {
			osg::ref_ptr<osg::Group> cloudGroup = new osg::Group();
			osg::ref_ptr<osg::PositionAttitudeTransform> cloudPosOut = new osg::PositionAttitudeTransform();
			root->addChild(cloudPosOut.get());

			/*float sizeX = ((rand() % 100)/100.0 - 0.5)*meanSizeX/2.0 + meanSizeX;
			float ratioXY = 0.75 + ((rand() % 100)/100.0 - 0.5)*0.5;
			float sizeY = sizeX*ratioXY;*/

			float sizeX = 1000.0;

			//osg::ref_ptr<Stratus> cloud = new Stratus(storage, numParticles,250,
			//	sizeX,sizeX,125.0,base, world->getConfiguration()->getStartTime());
			osg::ref_ptr<Cumulus> cloud = new Cumulus(storage, numParticles,partSize,
				650,650,250.0,base,world->getConfiguration()->getStartTime(),1);
			//osg::ref_ptr<Cirrus> cloud = new Cirrus(storage, numParticles,
			//	partSize,sizeX,sizeX,100.0,base,world->getConfiguration()->getStartTime());

			float offX = 0;//(float)(rand() % maxOffsetOut) - maxOffsetOut/2.0;
			float offY = 0;//(float)(rand() % maxOffsetOut) - maxOffsetOut/2.0;


			cloudPosOut->addChild(cloud.get());

			cloudPosOut->setPosition(osg::Vec3f(base.x()+offX, base.y()+offY, height));


		}
		XGIS_LOG_DEBUG2("end cloud creation");
#endif

		ui->setScene(root.get(), world.get(), hud.get());
		ui->postStartedUp();
	}
};

int main(int argc, char **argv) {
	// first let's deal with application usage string and command line
	// parameters
	std::string cliUsage = "xgis-viewer [-h | --help] "
		"[--storage <storage-url>] [--start-track <filename>] "
		"[--fps <fps>] [--view-size <width> <height>]"
		"[--rotation-speed <speed>] [--translation-speed <speed>]"
		"[--far-plane <dist>] [--fog-size <dist>]";
	osg::ApplicationUsage *usage = new osg::ApplicationUsage;
	usage->setApplicationName("xgis-viewer");
	usage->addCommandLineOption("-h or --help", "Displays this message.");
	usage->addCommandLineOption("--storage <storage-url>", "Set storage URL.", "http://app.gmajna.net/uniserver-0.2.0/raw");
	usage->addCommandLineOption("--start-date <YYYYMMDD-hhmm>", "Starting time of visualisation (for sun and sky purposes).", "");
	usage->addCommandLineOption("--start-track <filename>", "Follow track in GPX file <filename> after starting (overrides --start-*).");
	usage->addCommandLineOption("--fps <fps>", "Maximum number of frames per second.", "30");
	usage->addCommandLineOption("--view-size <width> <height>", "Size of viewport.", "800 600");
	usage->addCommandLineOption("--rotation-speed <speed>", "Rotation speed in radians per second.", "0.314");
	usage->addCommandLineOption("--translation-speed <speed>", "Translation speed in metres per second.", "10");
	usage->addCommandLineOption("--far-plane <dist>", "Distance of far plane (determines visibility range) in metres.", "50000");
	usage->addCommandLineOption("--fog-size <dist>", "Distance of fog layer start from far plane in metres.", "0");
	usage->addCommandLineOption("--layers <layer1,layer2...>", "Set of layers to show.", "");
	usage->addCommandLineOption("--detail-level-set <n>", "Default detail level set to use.", "");
#ifdef XGIS_UI_CONSOLE
	usage->addCommandLineOption("--image-dir <path>", "Path to directory for rendered images.");
	usage->addCommandLineOption("--start-frame <n>", "Start rendering a movie at frame n.");
	usage->setDescription("A batch XGIS visualization client: follows a given GPX track after startup, "
		"fetching data from the storage as required. "
		"Optionally renders each frame to a file for latter processing.");
	cliUsage += " [--image-dir <path>] [--start-frame <n>]";
#else
	usage->addCommandLineOption("--start-gkr <x> <y>", "Set startup location in Gauss-Krueger coordinates.");
	usage->addCommandLineOption("--start-wgs84 <latitude> <longitude>", "Set startup location in WGS-84 coordinates.");
	usage->addCommandLineOption("--start-altitude <z>", "Set startup altitude in metres.");
	usage->addCommandLineOption("--flood-path <path>", "Path where flood surfaces are stored.");
	usage->setDescription("An interactive XGIS visualization client.");
	cliUsage += " [--start-gkr <x> <y> | --start-wgs84 <latitude> <longitude>] [--start-altitude <z>]";
#endif /* XGIS_UI_CONSOLE */
	usage->setCommandLineUsage(cliUsage);

	osg::ArgumentParser parser(&argc, argv);
	parser.setApplicationUsage(usage);

	if(parser.read("-h") || parser.read("--help")) {
		parser.getApplicationUsage()->write(std::cout);
		return 181;
	}

	std::string storageUrl = "http://app.gmajna.net/uniserver-0.2.0/raw";
	while(parser.read("--storage", storageUrl)) {}

	std::string gpsTrackFile;
	while(parser.read("--start-track", gpsTrackFile)) {}

	int fps = 30;
	while(parser.read("--fps", fps)) {}

	int viewWidth = 800, viewHeight = 600;
	while(parser.read("--view-size", viewWidth, viewHeight)) {}

	float rotSpeed = M_PI/10.0;
	while(parser.read("--rotation-speed", rotSpeed)) {}

	float trsSpeed = 10.0;
	while(parser.read("--translation-speed", trsSpeed)) {}

	float farPlane = 50000.0;
	while(parser.read("--far-plane", farPlane)) {}

	float fogSize = 0.0;
	while(parser.read("--fog-size", fogSize)) {}

	std::string floodPath = ".";
	while(parser.read("--flood-path", floodPath)) {}

	std::string startDate;
	time_t startTime = time(NULL);
	while(parser.read("--start-date", startDate)) {}
	if(startDate.length() > 0) {
		struct tm tm;
		tm.tm_sec = 0;
		tm.tm_wday = 0;
		tm.tm_yday = 0;
		tm.tm_isdst = -1;
		if(5 != sscanf(startDate.c_str(), "%04d%02d%02d-%02d%02d",
			&(tm.tm_year), &(tm.tm_mon), &(tm.tm_mday),
			&(tm.tm_hour), &(tm.tm_min))) {
			XGIS_LOG_WARNING("Wrong date string: using current time as start date.");
		}
		else {
			tm.tm_mon--;
			tm.tm_year -= 1900;
			startTime = mktime(&tm);
			if(-1 == startTime) {
				startTime = time(NULL);
			}
		}
	}

	std::string layers;
	while(parser.read("--layers", layers)) {}

	int dls = 0;
	while(parser.read("--detail-level-set", dls)) {}

	std::string imagePath;

	bool gkrSet = false;
	float gkrX = 0, gkrY = 0;

	bool wgsSet = false;
	float wgsLat = 0, wgsLon = 0;

	float alt = NAN;

	int startFrame = 0;

#ifdef XGIS_UI_CONSOLE
	while(parser.read("--image-dir", imagePath)) {}

	while(parser.read("--start-frame", startFrame)) {}
#else
	while(parser.read("--start-gkr", gkrX, gkrY)) {
		gkrSet = true;
	}

	while(parser.read("--start-wgs84", wgsLat, wgsLon)) {
		wgsSet = true;
	}

	while(parser.read("--start-altitude", alt)) {}
#endif /* XGIS_UI_CONSOLE */

	if(gkrSet && wgsSet) {
		parser.reportError("You may only set starting location in "
			"Gauss-Krueger or WGS-84 coordinates, not both!");
	}

	parser.reportRemainingOptionsAsUnrecognized();

    // report any errors if they have occured when parsing the program aguments.
	if(parser.errors()) {
		parser.writeErrorMessages(std::cout);
		return 42;
	}

#ifdef WIN32
		// set priority in order to be able to kill process if (when!) it hangs
		SetPriorityClass(GetCurrentProcess(), BELOW_NORMAL_PRIORITY_CLASS);
#else
		/* TBD: for Linux & Apple */
#endif /* WIN32 */

	// load xgis viewer configuration
	XGisConfig *cfg = XGisConfig::getInstance();

	// store default configuration passed from CLI pars
	cfg->setRotationSpeed(rotSpeed);
	cfg->setTranslationSpeed(trsSpeed);
	cfg->setFarPlane(farPlane);
	cfg->setFogSize(fogSize);
	cfg->setDefaultDetailLevelSet(dls);
	cfg->setStartTime(startTime);
	cfg->setFloodPath(floodPath);

	// add layers passed by CLI par
	int s = 0, e = -1;
	do {
		std::string layer;
		e = layers.find(',', s);
		if(e != -1 && e > s) {
			layer = std::string(layers, s, e - s);
			s = e + 1;
		}
		else {
			layer = std::string(layers, s);
		}
		XGIS_LOG_INFO("Adding layer: %s", layer.c_str());
		cfg->addLayer(layer);
	} while(s < (int)layers.length() && e != -1);

	// report configuration
	cout << "XGIS Viewer configuration:" << endl;
	cout << "  Quad loader loads ";
	if(cfg->useCompiledQuads()) {
		cout << "precompiled quad geodes." << endl;
	}
	else {
		cout << "quad height maps and textures." << endl;
	}
	cout << "  Default detail level set: " << cfg->getDefaultDetailLevelSet() << endl;
	cout << "  Available layers: " << endl;
	std::vector<std::string>::const_iterator i = cfg->getAvailableLayers().begin();
	while(i != cfg->getAvailableLayers().end()) {
		cout << "    - " << *i++ << endl;
	}
	cout << endl;
	cout << "  Visible layers: " << endl;
	i = cfg->getVisibleLayers().begin();
	while(i != cfg->getVisibleLayers().end()) {
		cout << "    - " << *i++ << endl;
	}
	cout << endl;

	// make us as thread-safe as possible
	osg::Referenced::setThreadSafeReferenceCounting(true);

	// set-up world storage
	std::string worldName = "a-brave-new-world";
	WorldStorage *storage =	WorldStorage::createFromURL(worldName, storageUrl);

	Coords::GaussKr defPos(519870, 80660);
	if(gkrSet) {
		defPos = Coords::GaussKr(gkrX, gkrY);
	}
	else if(wgsSet) {
		defPos = Coords::WGS84(wgsLon, wgsLat).toGaussKr();
	}
	osg::ref_ptr<GpsTrack> track = new GpsTrack;

	if(gpsTrackFile.length() == 0) {
		track->addPoint(osg::Vec3(defPos.x, defPos.y, alt));
	}
	else {
		if(!track->parseGpx(gpsTrackFile)) {
			XGIS_LOG_ERROR("Invalid GPS track!");
			track->clear();
			track->addPoint(osg::Vec3(defPos.x, defPos.y, alt));
		}
		else if(track->begin() != track->end()) {
			// if parsed and has >= 1 point,
			// move initial position to first point in GPS track
			defPos.x = (*track->begin()).x();
			defPos.y = (*track->begin()).y();
			alt = (*track->begin()).z();
		}
		else {
			// no points in track...
			track->addPoint(osg::Vec3(defPos.x, defPos.y, alt));
		}
	}
	cfg->setStartTrack(track.get());

	// create UI
	XGisUI *ui = XGisUIFactory::getInstance();
	// set it up with necessary data
	ui->setup(argc, argv, viewWidth, viewHeight, fps, startFrame,
		"XGIS Viewer " VERSION " by XLAB - Hydraulics by UL-FGG KMTE", imagePath,
		new StartupThread(worldName, osg::Vec3(defPos.x, defPos.y, alt),
			storage, ui));
	// finally, start main loop and pass control to the windowing subsystem
	ui->run();

	return 0;
}
