#include "config.h"

#include <assert.h>

#include "obscenemanipulator.h"

#include <xgislib/xputils.h>

#ifdef XGIS_REFCOUNT_DEBUG
#	include <xgislib/quad.h>
#	include <xgislib/quadsingle.h>
#	include <xgislib/layerset.h>
#endif /* XGIS_REFCOUNT_DEBUG */

using namespace osg;

class AnimationPathEx : public AnimationPath {
private:
	double sumTimeTasks, sumTime, lastTime, maxTime;
	int numFrames;

public:
	AnimationPathEx()
	: AnimationPath(), sumTimeTasks(0), sumTime(0), lastTime(0),  maxTime(0), numFrames(0)
	{}

	void noteFrame(float time, int pendTasks) {
		double thisTime = time - lastTime;
		if (thisTime > 1.0) {
			XGIS_LOG_DEBUG2("Frame lasted %lf sec with %d pending tasks", thisTime, pendTasks);
		}
		sumTimeTasks += thisTime * pendTasks;
		sumTime += thisTime;
		maxTime = osg::maximum(maxTime, thisTime);
		lastTime = time;
		numFrames++;
	}

	int getNumFrames() {
		return numFrames;
	}

	float getAvgPendTasks() {
		return (float) (sumTimeTasks/sumTime);
	}

	float getMaxTime() {
		return maxTime;
	}
};

const float
 	ObsceneManipulator::shiftFactor = 10.0,
	ObsceneManipulator::plusFactor = 1.5,
	ObsceneManipulator::minEyeHeight = 50.0,
	ObsceneManipulator::lockedHeightTolerance = 20.0;
	

ObsceneManipulator::ObsceneManipulator(World *_world, ObsViewer *_viewer,
	const osg::Matrixd& _homeM)	:
	osgGA::MatrixManipulator(), world(_world), viewer(_viewer) {
	speed = 0; //don't start flying immediately
	
	// animation speeds
	aSpeedMove = world->getConfiguration()->getTranslationSpeed();
	aSpeedRot = world->getConfiguration()->getRotationSpeed();
	
	moveStep = 10; // m
	rotStep = 0.4/180*M_PI; // rad
	speedStep = 1; // m/s

	homeM = _homeM;
	home();
	
	isHeightLocked = false;
	lockedHeight = -1;

	currAnim = NULL;

	isFirstFrame = true;
	tLastFrame = 0;
	
	updateOnAnimEnd = false;
}

ObsceneManipulator::~ObsceneManipulator() {
}

void ObsceneManipulator::rotate(float angle, Vec3 axis) {
	Vec3 eye = getEye();
	setByMatrix(getMatrix()
		* Matrix::translate(-eye)
		* Matrix::rotate(angle, axis)
		* Matrix::translate(eye));
}

void ObsceneManipulator::pitch(float angle) {
	rotate(angle, getStrafe());
}

void ObsceneManipulator::yaw(float angle) {
	// NOTE: we don't really want a proper yaw, we want rotation around up
	// vector
#if 0
	rotate(angle, getUp());
#else
	rotate(angle, osg::Vec3(0, 0, 1));
#endif /* 0/1 */
}

void ObsceneManipulator::roll(float angle) {
	rotate(angle, getForward());
}

void ObsceneManipulator::translate(Vec3 const &t) {
	setByMatrix(getMatrix() * Matrix::translate(t));
}

void ObsceneManipulator::beamTo(const osg::Vec3f &dest) {
	osg::Vec3d eye = getEye();
	setByMatrix(getMatrix() * Matrix::translate(dest - eye));
}

void ObsceneManipulator::left(float d) {
	// when we yaw, we need to adjust the strafe vector
	// accordingly (I suppose...)
	Vec3 forward = getForward();
	Vec3 strafe = forward^Vec3(0, 0, 1);
	strafe.normalize();
	strafe = -strafe*d;
	translate(strafe);	
}

void ObsceneManipulator::right(float d) {
	left(-d);
}

void ObsceneManipulator::forward(float d) {
	Vec3 forward = getForward();
	forward.z() = 0;
	forward.normalize();
	forward = forward*d;
	translate(forward);	
}

void ObsceneManipulator::backward(float d) {
	forward(-d);
}


void ObsceneManipulator::lookAt(float x, float y, float z) {
	Vec3 center(x, y, z);
	lookAt(center);
}

void ObsceneManipulator::lookAt(Vec3 const &center) {
	Vec3 up(0, 0, 1);
	Vec3 eye = getEye();
	Vec3 f = getForward();
	Vec3 nf(center[0] - eye[0], center[1] - eye[1], center[2] - eye[2]);
	Vec3 f2(f[0], f[1], 0);
	Vec3 nf2(nf[0], nf[1], 0);
	Vec3 axis;
	
	// in order to look at a certain point (and preserve the up vector
	// as it currently is), we need to:
	// 1. rotate about the axis defined by the up vector for the angle
	//    between the current forward vector and the desired forward vector,
	//    both projected onto the XY plane
	f2.normalize();
	nf2.normalize();
	float sp = f2*nf2;
	float angle = acos(sp);
	axis = f2^nf2;
	rotate(angle, axis);

	// 2. rotate about the axis defined by the strafe vector for the
	//    angle between the current forward vector and the desired
	//    forward vector, both projected onto the plane defined by
	//    XY-plane vector from eye to new center point and Z axis
	f = getForward();
	Vec3 strafe = getStrafe();
	float xyDist = sqrt(f[0]*f[0] + f[1]*f[1]);
	Vec3 fz(0, xyDist, f[2]);
	xyDist = sqrt(nf[0]*nf[0] + nf[1]*nf[1]);
	Vec3 nfz(0, xyDist, nf[2]);
	fz.normalize();
	nfz.normalize();
	sp = fz*nfz;
	angle = acos(sp);
	if(fz[2] > nfz[2])
		angle = -angle;
	rotate(angle, strafe);
	
	// fuck, I hate euclidean geometry!
	// would you rather have another geometry?
	// how about riemann?
}

void ObsceneManipulator::moveToLockedHeight() {
	if(lockedHeight != -1) {
		Vec3d eye = getEye();
		float height = world->getTerrainHeight(eye[0], eye[1]);
		if(height != -1) {
			height = eye[2] - height;
			height = lockedHeight - height;
			translate(0, 0, height);
		}
	}
}

void ObsceneManipulator::handleFrame(float time) {
#ifdef XGIS_UI_CONSOLE
		static int frameNum = 0;
		static float lastTime = time;
		frameNum++;
		if (time > lastTime + 5) {
			Coords::GaussKr ey(getEye().x(), getEye().y());
			Coords::WGS84 eyWgs(ey);
			float avgPendTasks = NAN;
			if (currAnim.valid()) {
				AnimationPathEx *animEx	= dynamic_cast<AnimationPathEx*>(currAnim.get());
				if (NULL != animEx)
					avgPendTasks = animEx->getAvgPendTasks();
			}
			XGIS_LOG_INFO("Frame %d, time %.1f, pendTasks curr/avg %d/%.2f, eye at GaussKr %.0f,%.0f,%.0f, WGS %.4f,%.4f",
				frameNum, time,
				world->getDataLoader()->getTaskCount(), avgPendTasks,
				ey.x, ey.y, getEye().z(),
				eyWgs.longit, eyWgs.latit);
			lastTime = time;
		}
#endif /* XGIS_UI_CONSOLE */

	if (isFirstFrame) {
		isFirstFrame = false;
#ifdef XGIS_MIPRO_GPX_FLIGHT
			static float fullLoadTime = NAN;
			if (isnan(fullLoadTime)) {
				isFirstFrame = true;
				if (world->getDataLoader()->getTaskCount() == 0) {
					fullLoadTime = time;
					XGIS_LOG_DEBUG4("fully loaded, waiting a bit...");
				}
			} else {
				if (world->getDataLoader()->getTaskCount() > 0) {
					fullLoadTime = NAN;
					isFirstFrame = true;
					XGIS_LOG_DEBUG4("...actually not fully loaded...");
				} else if (time < fullLoadTime + 5.0)
					isFirstFrame = true;
				else {
					ParDataLoader *parLd
						= dynamic_cast<ParDataLoader*>(world->getDataLoader());
					if (parLd != NULL)
						parLd->startupDone = true;
					XGIS_LOG_DEBUG4("...starting inital gpx flight!");
				}
			}
			if (currAnim.valid()) {
				tAnimStart = time;
			}
#endif /* XGIS_MIPRO_GPX_FLIGHT */
		tLastFrame = time;
		return;
	}

	if (currAnim.valid()) {
		osg::Matrixf animM;
		float tAnim = time - tAnimStart;
		if (tAnim <= currAnim->getLastTime()) {
			// animate
			currAnim->getMatrix(time - tAnimStart, animM);
			setByMatrix(animM);
			osg::Vec3f eye = getEye();
			XGIS_LOG_DEBUG5("New position: %f,%f,%f", eye.x(), eye.y(), eye.z());
			AnimationPathEx *animEx	= dynamic_cast<AnimationPathEx*>(currAnim.get());
			if (NULL != animEx)
				animEx->noteFrame(time - tAnimStart, world->getDataLoader()->getTaskCount());
		}
		else { 
			// animation finished
			AnimationPathEx *animEx	= dynamic_cast<AnimationPathEx*>(currAnim.get());
			if (NULL != animEx) {
				XGIS_LOG_INFO("anim done; avg %.2f pending tasks in %d frames",
					animEx->getAvgPendTasks(), animEx->getNumFrames());
				XGIS_LOG_INFO("          avg %f FPS, min %f FPS",
					animEx->getNumFrames() / animEx->getLastTime(),
					1 / animEx->getMaxTime());
			}

			stop();
#ifdef XGIS_UI_CONSOLE
			// stop console viewer after track is done
			viewer->setDone(true);
#endif /* XGIS_UI_CONSOLE */
		}
	}
	else {
		// no animation, just fly forward
		forward((time - tLastFrame)*getSpeed());
		tLastFrame = time;
	}
}

// stop any ongoing observer motion
void ObsceneManipulator::stop() {
	if (NULL != currAnim) {
		setAnim(NULL);
		if(updateOnAnimEnd) {
			world->updateDetailLevel(llUpdateQ, urUpdateQ);
			world->updateDetailLevel();
			world->setDoAutomaticUpdates(true);
		}
	}
}

void ObsceneManipulator::jumpTo(const osg::Vec3 &destination) {
	translate(destination - getEye());
}

void ObsceneManipulator::flyTo(const osg::Vec3 &destination) {
	AnimationPathEx *path = new AnimationPathEx();
	path->setLoopMode(osg::AnimationPath::NO_LOOPING);

	osg::Vec3f pos, scale;
	osg::Quat rotation, so;
	getMatrix().decompose(pos, rotation, scale, so);
	
	/* start point */
	path->insert(0, osg::AnimationPath::ControlPoint(pos, rotation));

	/* first, rotate towards the destination */
	osg::Vec3f fwd = getForward();
	fwd.z() = 0;
	osg::Vec3f dir = destination - pos;
	dir.z() = 0;
	osg::Quat incrot;
	incrot.makeRotate(fwd, dir);
	double angle;
	osg::Vec3f axis;
	incrot.getRotate(angle, axis);
	rotation = rotation*incrot;
	path->insert(angle/aSpeedRot, osg::AnimationPath::ControlPoint(pos, rotation));

	/* then, move towards destination */
	XGIS_LOG_INFO("Len: %f, speed %f, time %f", (float)dir.length(), (float)aSpeedMove, (float)(dir.length()/aSpeedMove));
	path->insert(dir.length()/aSpeedMove,
		osg::AnimationPath::ControlPoint(destination, rotation));
	setAnim(path);
	
	XGIS_LOG_DEBUG5("Starting flyTo(%f,%f,%f)", destination.x(), destination.y(), destination.z());
	
}

void ObsceneManipulator::flyTo(const osg::Vec3 &destination, double flightTime) {
	AnimationPathEx *path = new AnimationPathEx();
	path->setLoopMode(osg::AnimationPath::NO_LOOPING);

	world->setDoAutomaticUpdates(false);
	world->getUpdateFrame(world->getDetailRefPoint(), llUpdateQ, urUpdateQ);
	updateOnAnimEnd = true;
	
	osg::Vec3f pos, scale;
	osg::Quat rotation, so;
	getMatrix().decompose(pos, rotation, scale, so);
	
	/* start point */
	path->insert(0, osg::AnimationPath::ControlPoint(pos, rotation));

	/* compute rotation towards destination */
	osg::Vec3f fwd = getForward();
	fwd.z() = 0;
	osg::Vec3f dir = destination - pos;
	dir.z() = 0;
	osg::Quat incrot;
	incrot.makeRotate(fwd, dir);
	double angle;
	osg::Vec3f axis;
	incrot.getRotate(angle, axis);
	rotation = rotation*incrot;
	
	/* fly to destination over a parabole with a peak distance/10 high and turn,
	 * all in flightTime seconds */
	dir = destination - pos;
	int numSteps = 30;
	double stepTime = flightTime/numSteps;
	double len = dir.length();
	double stepLen = len/numSteps;
	double peak = len/10.0;
	double a = peak*2.0*2.0/len/len;
	int i;
	double x;
	for(i = 0, x = -len/2.0; i <= numSteps; i++, x += stepLen) {
		path->insert(i*stepTime, osg::AnimationPath::ControlPoint(pos +
			dir*((double)i/(double)numSteps) + osg::Vec3(0, 0, -a*x*x + peak), rotation));
		XGIS_LOG_DEBUG5("Height at step %d: %f, elev %f, x %lf", i, (pos +
			dir*((double)i/(double)numSteps) + osg::Vec3(0, 0, -a*x*x + peak)).z(), osg::Vec3(0, 0, -a*x*x + peak).z(), x);
	}
	XGIS_LOG_DEBUG5("Starts at %f, end at %f, expected z diff: %f", pos.z(), destination.z(), dir.z());		

	setAnim(path);
	
	XGIS_LOG_DEBUG3("Starting flyTo (%f,%f,%f) in %lfs",
		destination.x(), destination.y(), destination.z(), flightTime);
}

void ObsceneManipulator::flyGpsTrack(const GpsTrack &track, float alt) {
	GpsTrack realTrack;
	std::vector<osg::Vec3>::const_iterator
		i = track.begin(),
		end = track.end();
	while (i != end) {
		osg::Vec3 p = *i;
		p.z() = world->getTerrainHeight(p.x(), p.y()) + alt;
		realTrack.addPoint(p);
	}
	realTrack.setSpeed(track.getSpeed());
	flyGpsTrack(realTrack);
}

void ObsceneManipulator::flyGpsTrack(const GpsTrack &track) {
	XGIS_LOG_DEBUG5("Flying over track with %d points.", track.getNumPoints());

	float tSpeed = aSpeedMove;
	
	std::vector<osg::Vec3>::const_iterator
		i = track.begin(),
		end = track.end();

	if (i == end)
		return;

	AnimationPathEx *path = new AnimationPathEx();
	path->setLoopMode(osg::AnimationPath::NO_LOOPING);
	if (!isnan(track.getSpeed())) {
		tSpeed = track.getSpeed();
	}

	osg::Vec3f pos, scale;
	osg::Quat rotation, so;
	getMatrix().decompose(pos, rotation, scale, so);
	float time = 0, turnTime, travelTime;

	/* start point */
	pos = *i;
	float terrH = world->getTerrainHeight(pos.x(), pos.y());
	if (isnan(pos.z()) || pos.z() < terrH + minEyeHeight) {
		pos.z() = terrH + minEyeHeight;
	}
	path->insert(time, osg::AnimationPath::ControlPoint(pos, rotation));
	i++;

	osg::Vec3 fwd = getForward();
	fwd.z() = 0;
	osg::Vec3 dst = pos;
	osg::Vec3 turnDst;
	osg::Vec3 dir;
	osg::Vec3 zdir;
	osg::Vec3 axis;
	double angle;
	while(i != end) { 
		//find first point (dst) that's different from current pos
		while ((dst-pos).length() < 1 && i != end) {
			dst = *i;
			float terrH = world->getTerrainHeight(dst.x(), dst.y());
			if (isnan(dst.z()) || dst.z() < terrH + minEyeHeight) {
				dst.z() = terrH + minEyeHeight;
			}
			i++;
		}
		// construct & insert ControlPoint
		dir = dst - pos;
		dir.z() = 0;
		osg::Quat incrot;
		incrot.makeRotate(fwd, dir);
		incrot.getRotate(angle, axis);
		rotation = rotation*incrot;
		turnTime = angle/aSpeedRot;
		travelTime = dir.length()/tSpeed;
		if(turnTime < travelTime) {
			turnDst = dir;
			turnDst.normalize();
			turnDst = turnDst*tSpeed*turnTime;
			assert(turnDst.length() < dir.length());
			turnDst = pos + turnDst;
			path->insert(time + turnTime,
				osg::AnimationPath::ControlPoint(turnDst, rotation));
		}
		time += travelTime;
		path->insert(time,
			osg::AnimationPath::ControlPoint(dst, rotation));
		fwd = dir;
		pos = dst;
	}
	setAnim(path);	
}

void ObsceneManipulator::setByMatrix(const osg::Matrixd& _matrix) {
	matrix = _matrix;

	osg::Vec3d newEye = getEye();
	double terrH = world->getTerrainHeight(newEye.x(), newEye.y());
	double newEyeH = newEye.z() - terrH;

	// if inappropriate, correct t.z()
	if (isHeightLocked) {
		if (newEyeH < lockedHeight - lockedHeightTolerance) {
			newEye.z() += lockedHeight - lockedHeightTolerance - newEyeH;
			matrix(3, 2) = newEye.z();
		}
		else if (newEyeH > lockedHeight + lockedHeightTolerance) {
			newEye.z() -= newEyeH - lockedHeight - lockedHeightTolerance;
			matrix(3, 2) = newEye.z();
		}
	} else {
		if(newEyeH < minEyeHeight) {
			newEye.z() += minEyeHeight - newEyeH;
			matrix(3, 2) = newEye.z();
		}
	}
	viewer->syncSlaveCameras();
}

bool ObsceneManipulator::handle(const osgGA::GUIEventAdapter& ea,
	osgGA::GUIActionAdapter& us) {

	if (ea.getHandled())
		return true;

	// FIXME: use us.requestRedraw() and us.requestContinuousUpdate() to
	// stop rendering when stationary and restart when not

	switch(ea.getEventType()) {
	case(osgGA::GUIEventAdapter::RESIZE):
		viewer->resizeViewer(ea.getWindowX(), ea.getWindowY(),
			ea.getWindowWidth(), ea.getWindowHeight());
		return false;
	case(osgGA::GUIEventAdapter::FRAME):
		handleFrame(viewer->getFrameStamp()->getSimulationTime());
		return false;

	default:
		break;
	}
	
	return osgGA::MatrixManipulator::handle(ea, us);
}
