#ifndef __OBSCENE_OBSCENEMANIPULATOR_H__
#  define __OBSCENE_OBSCENEMANIPULATOR_H__

#include <osgGA/MatrixManipulator>
#include <osg/AnimationPath>

#include <xgislib/gpstrack.h>
#include <xgislib/world.h>

#include "obsviewer.h"

class ObsceneManipulator : public osgGA::MatrixManipulator {
friend class ObsViewer;

protected:
	osg::ref_ptr<World> world;       // the World
	osg::ref_ptr<ObsViewer> viewer;  // the Viewer
	
	bool isFirstFrame;
	double tLastFrame;
	
	osg::Matrixf matrix,      //current camera position
				homeM;        // home position

	// current observer speed (m/s)
	float speed;

	// animation speeds
	float aSpeedMove, aSpeedRot;
	
	// step-by-step motion and rotation, and speed control
	float moveStep, rotStep, speedStep;
								// how many meters a single keypress moves;
	                            // how many radians it rotates;
								// how fast does it accelerate

	static const float
				shiftFactor,	// step factor when shift is pressed
				plusFactor,		//multiply/divide step by this on +/- key
				minEyeHeight,	//eye can't get lower than this many metres above land
				lockedHeightTolerance; //locked height is followed this closely

	bool isHeightLocked;		// do we maintain a constant height-over-terrain?
	float lockedHeight;			// what is the locked height?

	osg::ref_ptr<osg::AnimationPath> currAnim;	// current motion data
	float tAnimStart;							// time when animation was started
	bool updateOnAnimEnd;
	QuadC llUpdateQ, urUpdateQ;
	
	void setAnim(osg::AnimationPath *anim) {
		currAnim = anim;
		if (currAnim.valid()) {
			tAnimStart = viewer->getFrameStamp()->getSimulationTime(); // FIXME: store real start time
		}
	}

	void handleFrame(float time);

	float _mX, _mY;
	float shift, ctrl;

public:

	ObsceneManipulator(World *_world, ObsViewer *_viewer, const osg::Matrixd& _homeM);
	virtual ~ObsceneManipulator();

	float getSpeed() { return speed; }
	void accelerate(float dspeed) {
		speed += dspeed;
	}
	
	//matrix access functions
	virtual void setByMatrix(const osg::Matrixd& _matrix);
	virtual void setByInverseMatrix(const osg::Matrixd& invM) { 
		setByMatrix(osg::Matrixd::inverse(invM));
	}
	virtual osg::Matrixd getMatrix() const { return matrix; } 
	virtual osg::Matrixd getInverseMatrix() const {
		return osg::Matrixd::inverse(matrix);
	}

	// general rotation
	void rotate(float angle, osg::Vec3 axis);

	// rotation about the viewer's principal axes
	void pitch(float angle);
	void yaw(float angle);
	void roll(float angle);

	// translation (relative to world coordinate system)
	void translate(osg::Vec3 const &t);
	void translate(float x, float y, float z) { translate(osg::Vec3(x, y, z)); }

	// translation (relative to viewer's coordinate system axes)
	void left(float d);
	void right(float d);
	void forward(float d);
	void backward(float d);

	// jump to the given location (beam me up, scotty!)
	void beamTo(const osg::Vec3f &dest);

	// direct the forward vector towards a location
	void lookAt(float x, float y, float z);
	void lookAt(osg::Vec3 const &center);

	void home() { setByMatrix(homeM); }

	void moveToLockedHeight(void);

	osg::Vec3 getEye() { return osg::Vec3(matrix(3, 0), matrix(3, 1), matrix(3, 2)); }
	osg::Vec3 getForward() { return osg::Vec3(-matrix(2, 0), -matrix(2, 1), -matrix(2, 2)); }
	osg::Vec3 getStrafe() { return osg::Vec3(matrix(0,0), matrix(0, 1), matrix(0, 2)); }
	osg::Vec3 getUp() { return getForward()^getStrafe(); }
	float getShiftFactor() { return shiftFactor; } 

	void jumpTo(const osg::Vec3 &destination);
	void flyTo(const osg::Vec3 &destination);
	void flyTo(const osg::Vec3 &destination, double flightTime);

	//fly following a gps track at specified altitued over terrain
	void flyGpsTrack(const GpsTrack &track, float alt);

	//fly following the gps track; if track's altitude invalid, fly at minEyeHeight
	void flyGpsTrack(const GpsTrack &track);
	
	// stop any ongoing observer motion
	void stop();
	// is there any motion going on
	bool isAnimating() {
		return currAnim.valid();
	}

	// take care of generic events (RESIZE, FRAME)
	virtual bool handle(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter& us);

};

#endif /* __OBSCENE_OBSCENEMANIPULATOR_H__ */
