#include "config.h"

#include <osg/Camera>

#include <xgislib/xputils.h>

#include "obsviewer.h"

#include "hud.h"

#undef near //this is defined somewhere in windows headers, and we DON'T WANT IT!
#undef far

ObsViewer::ObsViewer(float _farPlane) : osgViewer::Viewer(),
	flNow(0), flLast(0), maxFps(25), farPlane(_farPlane) {

	double fovy, ar, mnear, mfar;

	// the main camera; renders the world
	getCamera()->getProjectionMatrixAsPerspective(fovy, ar, mnear, mfar);
	getCamera()->setComputeNearFarMode(osg::CullSettings::COMPUTE_NEAR_FAR_USING_PRIMITIVES);
	getCamera()->setClearMask(GL_DEPTH_BUFFER_BIT);

	// the HUD camera; post-rendered
	hudCam = createSlaveCamera(HUD_CAMERA_NAME, osg::Camera::POST_RENDER, 100, 0, 0);
    hudCam->setViewMatrix(osg::Matrix::identity());
	hudCam->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
	hudCam->setClearMask(GL_DEPTH_BUFFER_BIT);

	// the environment (=sky) camera; pre-rendered,
	// thus the only one that clear the color buffer
    envCam = createSlaveCamera(ENV_CAMERA_NAME, osg::Camera::PRE_RENDER, 0, 0, 0);
    envCam->setClearMask(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	envCam->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
	envCam->setComputeNearFarMode(osg::CullSettings::COMPUTE_NEAR_FAR_USING_BOUNDING_VOLUMES);
	double skyDist = 1e+7; // FIXME: should approximately equal the skydome radius
    envCam->setProjectionMatrixAsPerspective(fovy, ar, skyDist/10, skyDist*10);    
}

osg::Camera *ObsViewer::createSlaveCamera(const char *name,
	osg::Camera::RenderOrder order, int pos, double mnear, double mfar) {

	osg::Camera *cam = new osg::Camera;
	cam->setRenderOrder(order, pos);
	cam->setAllowEventFocus(false);
	cam->setName(name);
	
	return cam;
}

void ObsViewer::setupSlaveCamera(osg::Camera *cam) {
	osg::Camera *mainCam = getCamera();

	cam->setViewport(mainCam->getViewport());
    cam->setGraphicsContext(mainCam->getGraphicsContext());

	/* insert the camera into the scene graph
		(if we add it using addSlave() then we cannot control 
		the rendering order) */
	osg::Group *scene = dynamic_cast<osg::Group*>(getSceneData());
	if (NULL == scene) {
		XGIS_LOG_ERROR("Scene is NULL or not a group -- cannot add a slave camera to it!");
		return;
	}
	scene->addChild(cam);
}

void ObsViewer::setupSlaveCameras() {
	setupSlaveCamera(hudCam.get());
	setupSlaveCamera(envCam.get());
}

void ObsViewer::syncSlaveCameras() {
	osg::Matrix &vm = getCamera()->getViewMatrix();
	envCam->setViewMatrix(vm);
}

void ObsViewer::resizeViewer(int x, int y, int w, int h) {
   	getCamera()->setViewport(x, y, w, h);
   	hudCam->setViewport(getCamera()->getViewport());
	hudCam->setProjectionMatrix(osg::Matrix::ortho2D(x, w, y, h));
	if(hud.valid()) {
		hud->setViewport(getCamera()->getViewport());
	}
	envCam->setViewport(getCamera()->getViewport());

	updateSlaves();
}

const float ObsViewer::maxFieldOfView = 170.0;
const float ObsViewer::minFieldOfView = 0.5;

void ObsViewer::zoom(float zoomFactor) {
	osg::Camera *camera = getCamera();
	double fovy, ar, dnear, dfar;
	
	XGIS_LOG_DEBUG3("Zooming for factor: %f", zoomFactor);
	camera->getProjectionMatrixAsPerspective(fovy, ar, dnear, dfar);
	XGIS_LOG_DEBUG3("Old FOV: %lf", fovy);
	XGIS_LOG_DEBUG3("ar=%lf, near/far=%lf,%lf", ar, dnear, dfar);
	fovy /= zoomFactor;
	fovy = CLAMP(fovy, minFieldOfView, maxFieldOfView);
	camera->setProjectionMatrixAsPerspective(fovy, ar, dnear, dfar);
#ifdef XGIS_LOG_LEVEL_INTERNAL_DEBUG3
	camera->getProjectionMatrixAsPerspective(fovy, ar, dnear, dfar);
	XGIS_LOG_DEBUG3("New FOV: %lf\n", fovy);
#endif 
}

void ObsViewer::frameLimiter() {
	flNow = flTimer.tick();
	int ms_rem = (int)(1000*((1.0/(float)maxFps) - flTimer.delta_s(flLast, flNow))) ;
	if (ms_rem > 0)
		usleep(1000*(ms_rem));

	flLast = flTimer.tick();
}
