#include "config.h"

#include <osgParticle/ExplosionEffect>
#include <osgParticle/ExplosionDebrisEffect>

#include <xgislib/xputils.h>
#include <xgislib/pickaction.h>

#include "worldloader.h"
#include "obscenemanipulator.h"
#include "pickhandler.h"

// PickHandler -- A GUIEventHandler that implements picking.

PickHandler::PickHandler(World *_world, HUD *_hud) : world(_world), hud(_hud) {
	_mX = 0;
	_mY = 0;
	_usePolytopeIntersector = false;
}

bool PickHandler::pickOnHUDMap(
	osgUtil::IntersectionVisitor &iv,
	ObsViewer *viewer, unsigned int button, bool doubleClick)
	//check if the pick was on the overlay map;
	//if yes, process it & return true
	//otherwise return false
{
	viewer->getHUDCamera()->accept(iv);
	if (!iv.getIntersector()->containsIntersections())
		return false;

	if (doubleClick) {
		osgUtil::LineSegmentIntersector::Intersection intersection =
			dynamic_cast<osgUtil::LineSegmentIntersector*>(iv.getIntersector())
			->getFirstIntersection();
		Coords::GaussKr clickLoc = hud->map2world(
			intersection.getLocalIntersectPoint().x(),
			intersection.getLocalIntersectPoint().y());
		ObsceneManipulator *manip =
			dynamic_cast<ObsceneManipulator*>
				(viewer->getCameraManipulator());
		switch (button) {
			case osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON:
				manip->jumpTo(osg::Vec3(clickLoc.x, clickLoc.y,
					manip->getEye().z()));
				break;
			case osgGA::GUIEventAdapter::RIGHT_MOUSE_BUTTON:
				manip->flyTo(osg::Vec3(clickLoc.x, clickLoc.y,
					manip->getEye().z()), 5.0);
				break;
			default:
				break;
		}
	}
	return true;
}

void PickHandler::pick(const osgGA::GUIEventAdapter &ea,
	ObsViewer *viewer, osgGA::GUIActionAdapter &aa,
	unsigned int button, bool doubleClick) {

	osg::Node* node = 0;
	osg::Group* parent = 0;

	/* camera manipulator */
	ObsceneManipulator *manip =
		dynamic_cast<ObsceneManipulator*>
			(viewer->getCameraManipulator());

	// any pick will immediately cancel each and every on-going
	// motion!
	manip->stop();
	
	//setup the Intersector object
	osgUtil::LineSegmentIntersector* picker;
	if (!_useWindowCoordinates) {
		// use non dimensional coordinates - in projection/clip space
		picker = new osgUtil::LineSegmentIntersector
			(osgUtil::Intersector::PROJECTION,
			ea.getXnormalized(), ea.getYnormalized());
	}
	else {
		// use window coordinates
		// remap the mouse x,y into viewport coordinates.
		osg::Viewport* viewport = viewer->getCamera()->getViewport();
		float mx = viewport->x() +
			(int)((float)viewport->width()*(ea.getXnormalized()*0.5f + 
			0.5f));
		float my = viewport->y() +
			(int)((float)viewport->height()*(ea.getYnormalized()*0.5f +
			0.5f));
		picker = new osgUtil::LineSegmentIntersector
			(osgUtil::Intersector::WINDOW, mx, my);
    }
	osgUtil::IntersectionVisitor iv(picker);

	// first, see if a point on the overlay mapped was picked
	bool mapPicked = pickOnHUDMap(iv, viewer, button, doubleClick);
	if (mapPicked)
		return;

#ifdef XGIS_MULTI_PASS_RENDERING	
	// otherwise, process the pick of the terrain/object
	// first, check intersection against fine model
	viewer->getFineCamera()->accept(iv);
	if(!picker->containsIntersections()) {
		// if nothing, check intersection against coarse model
		viewer->getCoarseCamera()->accept(iv);
	}
#else
	viewer->getCamera()->accept(iv);
#endif /* XGIS_MULTI_PASS_RENDERING */

	if (picker->containsIntersections()) {
		osgUtil::LineSegmentIntersector::Intersection intersection =
			picker->getFirstIntersection();

		XGIS_LOG_DEBUG5("Picked %f %f %f\n",
			intersection.localIntersectionPoint.x(),
			intersection.localIntersectionPoint.y(),
			intersection.localIntersectionPoint.z());

		if(doubleClick) {
			if(button == osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON) {
				// double-click: move to
				if (manip) {
					// animate move to location
					intersection.localIntersectionPoint.z() +=
						(manip->getEye().z() - world->getTerrainHeight
							(manip->getEye().x(), manip->getEye().y()));
					if(ea.getModKeyMask() &
						osgGA::GUIEventAdapter::MODKEY_SHIFT) {
						// slow, straight flight to location
						manip->flyTo(intersection.localIntersectionPoint);
					}
					else {
						// 5 second jump over a parabole to location
						manip->flyTo(intersection.localIntersectionPoint,
							5.0);
					}
				}
			}
			else if(button == osgGA::GUIEventAdapter::MIDDLE_MOUSE_BUTTON) {
				char locmsg[128 + 1];
				snprintf(locmsg, 128, "Clicked on location (%f,%f,%f)",
					intersection.localIntersectionPoint.x(),
					intersection.localIntersectionPoint.y(),
					intersection.localIntersectionPoint.z());
				XGIS_LOG_INFO(locmsg);
				hud->setStatus(locmsg);
			}
			else if(button == osgGA::GUIEventAdapter::RIGHT_MOUSE_BUTTON) {
				// double-click: put an explosion at intersection
				osg::NodePath& nodePath = intersection.nodePath;
				node = (nodePath.size()>=1)?nodePath[nodePath.size()-1]:0;
				parent = (nodePath.size()>=2)?
					dynamic_cast<osg::Group*>(nodePath[nodePath.size()-2]):0;

				XGIS_LOG_DEBUG4("Blowing up object: ");
				// if any object on the node path has PickAction data, execute
				// its pick() method
				for (unsigned int i = 1; i < nodePath.size(); i++) {
					osg::Node::DescriptionList dscs = nodePath[i]->getDescriptions();
					XGIS_LOG_DEBUG4("  object %s (%s), descriptions: %d", nodePath[i]->className(), nodePath[i]->getName().c_str(), dscs.size());
					osg::Node::DescriptionList::iterator it = dscs.begin();
					while(it != dscs.end()) {
						XGIS_LOG_DEBUG4("Found description '%s' on %p", (*it).c_str(), nodePath[i]);
						if(*it == "explodable") {
							world->getDataLoader()->pushTask
								(new RemoveNode(world.get(), nodePath[i], (osg::Group *)(nodePath[i - 1])));
							break;
						}
						it++;
					}
				}

				osg::Group *explosionGroup = new osg::Group;
				explosionGroup->addChild(new osgParticle::ExplosionDebrisEffect
					(intersection.localIntersectionPoint, 10.0f, 1.0f));
				explosionGroup->addChild(new osgParticle::ExplosionEffect
					(intersection.localIntersectionPoint, 20.0f, 1.0f));
				world->getDataLoader()->pushTask
					(new AddNode(world.get(),	explosionGroup));
			}					
		}
		else {
			// single-click: perform action on the picked object
			osg::NodePath& nodePath = intersection.nodePath;
			node = (nodePath.size()>=1)?nodePath[nodePath.size()-1]:0;
			parent = (nodePath.size()>=2)?
				dynamic_cast<osg::Group*>(nodePath[nodePath.size()-2]):0;

			// if any object on the node path has PickAction data, execute
			// its pick() method
			XGIS_LOG_DEBUG4("Picking object at (%.2f,%.2f,%.2f): ",
				intersection.localIntersectionPoint.x(),
				intersection.localIntersectionPoint.y(),
				intersection.localIntersectionPoint.z() );
			for (unsigned int i = 0; i < nodePath.size(); i++) {
				SceneUserData *udata = (SceneUserData *)nodePath[i]->getUserData();
				XGIS_LOG_DEBUG4("  object %s (%s), user data: %p", nodePath[i]->className(), nodePath[i]->getName().c_str(), udata);
				if(NULL != udata) {
					if(0 == strcmp(udata->getTypeName(), PickAction::TYPE_NAME)) {
						XGIS_LOG_DEBUG4("  executing pick action.");
						((PickAction *)udata)->pick(nodePath[i]);
					}
				}
			}

		}
	}
}

bool PickHandler::handle(const osgGA::GUIEventAdapter& ea,
	osgGA::GUIActionAdapter& aa) {
	ObsViewer* viewer = dynamic_cast<ObsViewer*>(&aa);
	if (!viewer) return false;

	ObsceneManipulator* manip =  (ObsceneManipulator*)viewer->getCameraManipulator();

	int et = ea.getEventType();

	switch(et) {
	case(osgGA::GUIEventAdapter::PUSH):
		_mX = ea.getX();
		_mY = ea.getY();
		pick(ea, viewer, aa, ea.getButton(), false);
		return false;
		
	case(osgGA::GUIEventAdapter::DOUBLECLICK):
		_mX = ea.getX();
		_mY = ea.getY();
		pick(ea, viewer, aa, ea.getButton(), true);
		return false;

	case(osgGA::GUIEventAdapter::DRAG): {
		float dx = _mX - ea.getX();
		_mX = ea.getX();
		float dy = _mY - ea.getY();
		_mY = ea.getY();
		if(ea.getButtonMask() == osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON) {
			if(dx != 0) {
				manip->rotate(0.01*dx, osg::Vec3f(0, 0, 1));
			}
			if(dy != 0) {
				manip->rotate(0.01*dy, manip->getStrafe());
			}
		}
		else if(ea.getButtonMask() == osgGA::GUIEventAdapter::RIGHT_MOUSE_BUTTON) {
			if(dx != 0) {
				manip->translate(manip->getStrafe()*dx);
			}
			if(dy != 0) {
				osg::Vec3 horizFwd = manip->getForward();
				horizFwd.z() = 0.0;
				horizFwd *= 4.0;
				manip->translate(horizFwd*dy);
			}				
		}
		return true;
	}

	case (osgGA::GUIEventAdapter::SCROLL): {
		double shiftFactor = 1;
		if (ea.getModKeyMask() & osgGA::GUIEventAdapter::MODKEY_SHIFT)
			shiftFactor *= manip->getShiftFactor();
		if(ea.getScrollingMotion() == osgGA::GUIEventAdapter::SCROLL_DOWN) {
			manip->translate(manip->getForward()*shiftFactor*-20.0);
		}
		else if(ea.getScrollingMotion() == osgGA::GUIEventAdapter::SCROLL_UP) {
			manip->translate(manip->getForward()*shiftFactor*20.0);
		}
		return true;
	}									   
	default:
		// XGIS_LOG_DEBUG4("%d\n",et);
		return false;
	}

}
