#include "config.h"

#include "skydome.h"

#include <osg/Geode>
#include <osg/Geometry>
#include <osg/ShapeDrawable>
#include <osg/Texture2D>
#include <osg/MatrixTransform>

#include <xgislib/xputils.h>
#include <xgislib/pointdata.h>

float SCALING_FACTOR = 10.0;

using namespace osg;



float angleBetween(float thetaVertex, float phiVertex, float thetaSun, float phiSun) {
  float cospsi = sin(thetaVertex) * sin(thetaSun) * cos(phiSun-phiVertex) + cos(thetaVertex) * cos(thetaSun);
  /*if (cospsi > 1)  return 0;
  if (cospsi < -1) return M_PI;*/
  float angle = acos(cospsi);

/*  printf("%f %f %f %f %f\n",osg::RadiansToDegrees(thetaVertex),osg::RadiansToDegrees(phiVertex)
	  ,osg::RadiansToDegrees(thetaSun),osg::RadiansToDegrees(phiSun),osg::RadiansToDegrees(angle));*/
  return angle;
}

void SkyDome::setSkyColour(time_t rawtime) {

	double azimuth, altitude;
	osg::Vec4Array *colours = new osg::Vec4Array;

	Coords::GaussKr c(center.x(), center.y());


	Pointdata::FeatureInfo::getSunPosition(
		c, rawtime, azimuth, altitude);

	float compAlt = altitude;


	float startFlat = 5.0;
	float startDrop = -6.0;
	float finalAlt = -9.0;

	if (altitude > finalAlt) {
		SCALING_FACTOR = 3.0;

#define USE_COMP_ALT
#ifdef USE_COMP_ALT
		if (altitude < startFlat) {
			compAlt = startFlat;
		}

		if (altitude < startDrop) {
			float k = ((startFlat - finalAlt)/(startDrop - finalAlt));
			float n = startFlat - k*startDrop;
			compAlt = k*altitude + n;
		}
#endif

		float add = MIN(2.0,2.0*abs(compAlt)/7.0);
		SCALING_FACTOR -= add;

		if (compAlt < 0) {
			SCALING_FACTOR = -compAlt + 3;
		}
		
		
	} else {
		SCALING_FACTOR = 5 + 8*abs(altitude)/10.0;
	}


	//if (altitude<-10.0) altitude = -10.0;

	double suZ = sin(osg::DegreesToRadians(compAlt));
	double sproj = cos(osg::DegreesToRadians(compAlt));
	double xangle = osg::DegreesToRadians(270.0-azimuth);
	double suX = sproj*cos(xangle);
	double suY = sproj*sin(xangle);
	//printf("sun unit %f %f %f\n",suX,suY,suZ);
	osg::Vec3d sunUnit = osg::Vec3d(suX,suY,suZ);

	thetaSun = osg::DegreesToRadians(90.0 - compAlt);

	/*TODO: check if negative values are OK here!*/
	phiSun = osg::DegreesToRadians(azimuth);

	//double zenithY, zenith_x, zenith_y;
	zenithY = getZenithY(thetaSun);
	zenithY = 1 - exp(-(1.0/(SCALING_FACTOR)) * zenithY);
	zenith_x = getZenith_x(thetaSun);
	zenith_y = getZenith_y(thetaSun);
	double r,g,b;
	YxyToRGB(zenithY, zenith_x, zenith_y, r, g, b);

	/*printf("Y %f x %f y %f\n",zenithY,zenith_x,zenith_y);
	printf("rgb %f %f %f\n",r*255,g*255,b*255);*/

	double minz = 0;
	for (unsigned int i = 0; i < vertices->getNumElements(); i++) {
		Vec3 v = (*vertices)[i];
		if (v.z()<minz) {
			minz = v.z();
		}
	}

	for (unsigned int i = 0; i < vertices->getNumElements(); i++) {
		Vec3 v = (*vertices)[i];

		double zCoord = v.z()+abs(minz);

		/*theta is 90 - point altitude, phi is easting from south*/
		double theta, phi;
		theta = M_PI/2.0 - atan(zCoord/sqrt(v.x()*v.x() + v.y()*v.y()));

		phi = (v.x() / v.y());

		if (v.x() > 0) {
			if (v.y() != 0) {
				/*phi is positive angle over the x axis or negative angle under the x axis here*/
				phi = M_PI + phi;
			} else {
				phi = M_PI;
			}
		} else {
			if (v.y() == 0) {
				phi = -M_PI / 2.0;
			} else {
				/*nothing*/
			}
		}

		v.normalize();

		//float gamma = angleBetween(theta,phi,thetaSun,phiSun);
		float gamma2 = acos(v.x()*sunUnit.x() + v.y()*sunUnit.y() + v.z()*sunUnit.z());
		//printf("%f\n",gamma2);

		if (gamma2 > M_PI)
			gamma2 = 2*M_PI - gamma2;

		//printf("gamma %f gamma2 %f\n",osg::RadiansToDegrees(gamma),osg::RadiansToDegrees(gamma2));


		double distribution_x = fPerez(turbidity,'x',theta,gamma2)/fPerez(turbidity,'x',0,thetaSun);
		double distribution_y = fPerez(turbidity,'y',theta,gamma2)/fPerez(turbidity,'y',0,thetaSun);
		double distributionY = fPerez(turbidity,'Y',theta,gamma2)/fPerez(turbidity,'Y',0,thetaSun);

		//printf("%f %f %f\n",distributionY,distribution_x,distribution_y);

		double vY, vy, vx;
		vY = zenithY*distributionY;
		vy = zenith_y*distribution_y;
		vx = zenith_x*distribution_x;

		double r,g,b;

		/*returns rgb colours in range 0..255*/
		YxyToRGB(vY,vx,vy,r,g,b);

		//printf("x %f y %f z %f\n",v.x(),v.y(),v.z());
		//printf("rgb %f %f %f\n",r,g,b);


		osg::Vec4f colour = osg::Vec4f(r,g,b,1.0);

		colours->push_back(colour);

	}

	skyGeometry->setColorArray(colours);
}

osg::PositionAttitudeTransform *SkyDome::createSphere(osg::Vec3 center, float radius,
	WorldStorage *storage, time_t rawtime) {

	skydome = storage->getModel("skydome2.osg");

		// osgDB::readNodeFile("models/skydome2.osg");
	osg::PositionAttitudeTransform *skyPos = new osg::PositionAttitudeTransform();

	if (skydome != NULL) {

		osg::Group *skyGroup = skydome.get()->asGroup();

		skyGeometry = (osg::Geometry*)(((osg::Geode*)(skyGroup->getChild(0)))->getDrawable(0));

		/*skydome geometry, vec3d array*/
		vertices = (osg::Vec3Array*)(skyGeometry->getVertexArray());

		setSkyColour(rawtime);

		skyGeometry->setColorBinding(osg::Geometry::BIND_PER_VERTEX);

		osg::ref_ptr<osg::StateSet> skyState = new osg::StateSet();
		skyState->setMode(GL_LIGHTING,  osg::StateAttribute::OFF);
		skyState->setMode(GL_CULL_FACE, osg::StateAttribute::OFF);
		skyState->setRenderBinDetails(-10, "RenderBin");
		skydome->setStateSet(skyState.get());
		skyPos->addChild(skydome.get());
		float sf = 1000;
		skyPos->setScale(osg::Vec3d(sf, sf, sf));
		skyPos->setPosition(center);
	}

	return skyPos;
}

void getPerezQuotients(int turbidity, char type, double &A, double &B, double &C, double &D, double&E) {
	switch(type) {
		case 'x':  {
			A = -0.0193*turbidity - 0.2592;
			B = -0.0665*turbidity + 0.0008;
			C = -0.0004*turbidity + 0.2125;
			D = 0.0641*turbidity - 0.8989;
			E = -0.0033*turbidity + 0.0452;
			break;
				   }
		case 'y': {
			A = -0.0167*turbidity - 0.2608;
			B = -0.0950*turbidity + 0.0092;
			C = -0.0079*turbidity + 0.2102;
			D = -0.0441*turbidity - 1.6537;
			E = -0.0109*turbidity + 0.0529;
			break;
					}
		case 'Y': {
			A = 0.1787*turbidity - 1.463;
			B = -0.3554*turbidity + 0.4275;
			C = -0.0227*turbidity + 5.3251;
			D = 0.1206*turbidity - 2.5771;
			E = -0.0670*turbidity + 0.3703;
			break;
				  }
		default: break;
	}
}

/*Perez distribution; I presume theta and gamma have to be in radians!!!*/
double SkyDome::fPerez(int turbididty, char type, double theta, double gamma) {
	double A, B, C, D, E;
	getPerezQuotients(turbididty, type,A,B,C,D,E);
	double ret = 0;

	double t1 = 1 + A*exp(B /cos(theta));
	double t2 = 1 + C*exp(D*gamma) + E*(cos(gamma)*cos(gamma));

	ret = t1*t2;
	//if (abs(ret)>10)
		//printf("th %f gam %f\n",theta,gamma);

	return ret;
}

/*conversion between colour systems:
Yxy -> XYZ -> RGB*/
void SkyDome::YxyToRGB(double Y, double x, double y, double& r, double& g, double& b) {

	//printf("Yxy %f %f %f\n",Y,x,y);
	//Y = 1 - exp(-1.0/25.0)*Y;

	Y = 1 - exp(-(1.0/SCALING_FACTOR) * Y);

	double X1 = x * (Y/y);
	double Y1 = Y;
	double Z1 = (1.0 - x - y)*(Y/y);

	r = 3.240479*X1 - 1.53715*Y1 - 0.49853*Z1;
	g = -0.969256*X1 + 1.875991*Y1 + 0.041556*Z1;
	b = 0.055648*X1 - 0.204043*Y1 + 1.057311*Z1;

/*	r = 255*r;
	g = 255*g;
	b = 255*b;*/

}

/*return zenith luminance in Yxy colour system
theta is 90 degrees - sun elevation, but presumably in radians here;
phi is azimuth, east of south. also presumably in radians*/
double SkyDome::getZenithY(double theta) {
	double hi = (4.0/9.0 - turbidity / 120.0) * (M_PI - 2*theta);
	double ret = (4.0453*turbidity - 4.971)*tan(hi) - 0.2155*turbidity + 2.4192;

	return abs(ret);
}

/*zenith x color value, theta same as above*/
double SkyDome::getZenith_x(double theta) {
	double ret = 0;
	double T = turbidity;
	double T2 = turbidity*turbidity;

	double x1 = T2*0.00166 - T*0.02903 + 0.11693;
	double x2 = -T2*0.00375 + T*0.06377 - 0.21196;
	double x3 = T2*0.00209 - T*0.03203 + 0.06052;
	double x4 = T*0.00394 + 0.25886;

	double theta2 = theta*theta;
	double theta3 = theta2*theta;

	ret = x1*theta3 + x2*theta2 + x3*theta + x4;

	return ret;

}

/*zenith y color value, theta same as above*/
double SkyDome::getZenith_y(double theta) {
	double ret = 0;
	double T = turbidity;
	double T2 = turbidity*turbidity;

	double x1 =  T2*0.00275 - T*0.04214 + 0.15346;
	double x2 = -T2*0.0061 + T*0.0897 - 0.26756;
	double x3 =  T2*0.00317 - T*0.04153 + 0.0667;
	double x4 =  T*0.00516 + 0.26688;

	double theta2 = theta*theta;
	double theta3 = theta2*theta;

	ret = x1*theta3 + x2*theta2 + x3*theta + x4;

	return ret;
}

SkyDome::SkyDome(osg::Vec3 _center, float radius, WorldStorage *storage, World* _world,
	int _turbidity, time_t rawtime) {

	this->turbidity = _turbidity;
	this->world = _world;
	this->center = _center;

	transform = createSphere(center, radius, storage, rawtime);
	addChild(transform.get());
}

SkyDome::~SkyDome(void) {
}

void SkyDome::setCenter(const osg::Vec3 &_center) {
	center = _center;
	transform->setPosition(center);
}
