#include "config.h"

#include <xgislib/xputils.h>
#include <xgislib/relativealtitudegroup.h>
#include <xgislib/osgutils.h>

#include <loaders/kml/kmlloader.h>

#include <osgDB/ReadFile>
#include <osgDB/FileNameUtils>
#include <osg/Vec3f>

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"
#include <wx/arrstr.h>

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include "wx/sizer.h"
#include "wx/checklst.h"
#include "wx/listctrl.h"
#include "wx/listbase.h"
#include "wx/filedlg.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#include "wxui.h"

// first, a suitable derivative of wxGLCanvas to serve as a GraphicsWindow for
// our viewer: WxUIGraphicsWindow

class WxUIGraphicsWindow: public wxGLCanvas, public osgViewer::GraphicsWindow
{
	static int glAttributes[];

public:
    WxUIGraphicsWindow(wxWindow *parent, wxWindowID id = wxID_ANY,
        const wxPoint& pos = wxDefaultPosition,
        const wxSize& size = wxDefaultSize, long style = 0,
        const wxString& name = wxT("TestGLCanvas"));

    ~WxUIGraphicsWindow();

    void OnPaint(wxPaintEvent& event);
    void OnSize(wxSizeEvent& event);
    void OnEraseBackground(wxEraseEvent& event);
    void OnKeyDown(wxKeyEvent &event);
    void OnKeyUp(wxKeyEvent &event);
    void OnMouse(wxMouseEvent &event);

    //
    // GraphicsWindow interface
    //

    void grabFocus();
    void grabFocusIfPointerInWindow();
    void useCursor(bool cursorOn);

    bool makeCurrentImplementation();
    void swapBuffersImplementation();

    // note implemented yet...just use dummy implementation to get working.
    virtual bool valid() const { return true; }
    virtual bool realizeImplementation() { return true; }
    virtual bool isRealizedImplementation() const  { return true; }
    virtual void closeImplementation() {}
    virtual bool releaseContextImplementation() { return true; }

private:
	int keyCodeWx2Osg(int key);

    wxCursor _oldCursor;

    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(WxUIGraphicsWindow, wxGLCanvas)
    EVT_SIZE(WxUIGraphicsWindow::OnSize)
    EVT_PAINT(WxUIGraphicsWindow::OnPaint)
    EVT_ERASE_BACKGROUND(WxUIGraphicsWindow::OnEraseBackground)
    EVT_KEY_DOWN(WxUIGraphicsWindow::OnKeyDown)
    EVT_KEY_UP(WxUIGraphicsWindow::OnKeyUp)
    EVT_MOUSE_EVENTS(WxUIGraphicsWindow::OnMouse)
END_EVENT_TABLE()

int WxUIGraphicsWindow::glAttributes[] = {
	WX_GL_RGBA,
	WX_GL_DOUBLEBUFFER,
	WX_GL_DEPTH_SIZE, 16,
	0
};

WxUIGraphicsWindow::WxUIGraphicsWindow(wxWindow *parent, wxWindowID id,
	const wxPoint& pos, const wxSize& size, long style, const wxString& name) :
	wxGLCanvas(parent, id, pos, size, style|wxFULL_REPAINT_ON_RESIZE, name,
		WxUIGraphicsWindow::glAttributes)
{
    // default cursor to standard
    _oldCursor = *wxSTANDARD_CURSOR;

    _traits = new GraphicsContext::Traits;
    _traits->x = pos.x;
    _traits->y = pos.y;
    _traits->width = size.x;
    _traits->height = size.y;

    if (valid())
    {
        setState( new osg::State );
        getState()->setGraphicsContext(this);

        if (_traits.valid() && _traits->sharedContext)
        {
            getState()->setContextID( _traits->sharedContext->getState()->getContextID() );
            incrementContextIDUsageCount( getState()->getContextID() );
        }
        else
        {
            getState()->setContextID( osg::GraphicsContext::createNewContextID() );
        }
    }

    grabFocus();
}

WxUIGraphicsWindow::~WxUIGraphicsWindow() {
}

void WxUIGraphicsWindow::OnPaint(wxPaintEvent& WXUNUSED(event)) {
	/* must always be here */
	wxPaintDC dc(this);
}

void WxUIGraphicsWindow::OnSize(wxSizeEvent& event) {
    // this is also necessary to update the context on some platforms
    wxGLCanvas::OnSize(event);

    // set GL viewport (not called by wxGLCanvas::OnSize on all platforms...)
    int width, height;
    GetClientSize(&width, &height);

    // update the window dimensions, in case the window has been resized.
    getEventQueue()->windowResize(0, 0, width, height);
    resized(0, 0, width, height);
}

void WxUIGraphicsWindow::OnEraseBackground(wxEraseEvent& WXUNUSED(event))
{
    /* Do nothing, to avoid flashing on MSW */
}

int WxUIGraphicsWindow::keyCodeWx2Osg(int key) {

	switch(key) {
	case WXK_END: key = osgGA::GUIEventAdapter::KEY_End; break;
	case WXK_HOME: key = osgGA::GUIEventAdapter::KEY_Home; break;
	case WXK_LEFT: key = osgGA::GUIEventAdapter::KEY_Left; break;
	case WXK_UP: key = osgGA::GUIEventAdapter::KEY_Up; break;
	case WXK_RIGHT: key = osgGA::GUIEventAdapter::KEY_Right; break;
	case WXK_DOWN: key = osgGA::GUIEventAdapter::KEY_Down; break;
	case WXK_PAGEUP: key = osgGA::GUIEventAdapter::KEY_Page_Up; break;
	case WXK_PAGEDOWN: key = osgGA::GUIEventAdapter::KEY_Page_Down; break;
	case WXK_INSERT: key = osgGA::GUIEventAdapter::KEY_Insert; break;
	case WXK_DELETE: key = osgGA::GUIEventAdapter::KEY_Delete; break;
	case WXK_SHIFT: key = osgGA::GUIEventAdapter::KEY_Shift_L; break;
	case WXK_ALT: key = osgGA::GUIEventAdapter::KEY_Alt_L; break;
	case WXK_CONTROL: key = osgGA::GUIEventAdapter::KEY_Control_L; break;
	case WXK_NUMPAD_ADD: key = '+'; break;
	case WXK_NUMPAD_SUBTRACT: key = '-'; break;
	default: break;
	}
	return key;
}

void WxUIGraphicsWindow::OnKeyDown(wxKeyEvent &event)
{
#ifdef WIN32
    int key = event.GetKeyCode();
	if ((key >= WXK_START  &&  key <= WXK_COMMAND)
		|| (key < 'A')) {

	} else {
		key = event.GetUnicodeKey();
	}
#else
	int key = event.GetUnicodeKey();
#endif

    getEventQueue()->keyPress(keyCodeWx2Osg(key));

    // propagate event
    event.Skip();
}

void WxUIGraphicsWindow::OnKeyUp(wxKeyEvent &event)
{
#ifdef WIN32
    int key = event.GetKeyCode();
	if ((key >= WXK_START  &&  key <= WXK_COMMAND)
		|| (key < 'A')) {

	} else {
		key = event.GetUnicodeKey();
	}
#else
    int key = event.GetKeyCode();
#endif
    getEventQueue()->keyRelease(keyCodeWx2Osg(key));

    // propagate event
    event.Skip();
}

void WxUIGraphicsWindow::OnMouse(wxMouseEvent& event)
{
	if (event.GetWheelRotation() != 0) {
		if(event.GetWheelRotation() > 0) {
			getEventQueue()->mouseScroll(osgGA::GUIEventAdapter::SCROLL_UP);
		}
		else {
			getEventQueue()->mouseScroll(osgGA::GUIEventAdapter::SCROLL_DOWN);
		}
	}
	else if (event.ButtonDown()) {
        int button = event.GetButton();
        getEventQueue()->mouseButtonPress(event.GetX(), event.GetY(), button);
    }
    else if (event.ButtonUp()) {
        int button = event.GetButton();
        getEventQueue()->mouseButtonRelease(event.GetX(), event.GetY(), button);
    }
    else if (event.ButtonDClick()) {
        int button = event.GetButton();
    	getEventQueue()->mouseDoubleButtonPress(event.GetX(), event.GetY(), button);
    }
    else if (event.Dragging()) {
        getEventQueue()->mouseMotion(event.GetX(), event.GetY());
    }

	grabFocus();
}

void WxUIGraphicsWindow::grabFocus()
{
    // focus this window
    SetFocus();
}

void WxUIGraphicsWindow::grabFocusIfPointerInWindow()
{
    // focus this window, if the pointer is in the window
    wxPoint pos = wxGetMousePosition();
    if (this == wxFindWindowAtPoint(pos)) {
        SetFocus();
    }
}

void WxUIGraphicsWindow::useCursor(bool cursorOn)
{
    if (cursorOn) {

        // show the old cursor
        SetCursor(_oldCursor);
    }
    else {

        // remember the old cursor
        _oldCursor = GetCursor();

        // hide the cursor
        //    - can't find a way to do this neatly, so create a 1x1, transparent image
        wxImage image(1,1);
        image.SetMask(true);
        image.SetMaskColour(0, 0, 0);
        wxCursor cursor(image);
        SetCursor(cursor);
    }
}

bool WxUIGraphicsWindow::makeCurrentImplementation()
{
    SetCurrent();
    return true;
}

void WxUIGraphicsWindow::swapBuffersImplementation()
{
    SwapBuffers();
}


// A main frame (top-level window): WxUIMainFrame

class WxUIMainFrame : public wxFrame
{
public:
    WxUIMainFrame(wxFrame *frame, const wxString& title, const wxPoint& pos,
        const wxSize& size, long style = wxDEFAULT_FRAME_STYLE);

    void SetViewer(ObsViewer *viewer);
    void OnIdle(wxIdleEvent &event);
    void OnSetScene(wxCommandEvent &event);

private:
    osg::ref_ptr<ObsViewer> _viewer;

    DECLARE_EVENT_TABLE()
};

DECLARE_EVENT_TYPE(wxEVT_SET_SCENE, -1)
DEFINE_EVENT_TYPE(wxEVT_SET_SCENE)

BEGIN_EVENT_TABLE(WxUIMainFrame, wxFrame)
    EVT_IDLE(WxUIMainFrame::OnIdle)
    EVT_COMMAND(wxID_ANY, wxEVT_SET_SCENE, WxUIMainFrame::OnSetScene)
END_EVENT_TABLE()

/* My frame constructor */
WxUIMainFrame::WxUIMainFrame(wxFrame *frame, const wxString& title, const wxPoint& pos,
    const wxSize& size, long style)
    : wxFrame(frame, wxID_ANY, title, pos, size, style)
{
}

void WxUIMainFrame::SetViewer(ObsViewer *viewer)
{
    _viewer = viewer;
}

void WxUIMainFrame::OnIdle(wxIdleEvent &event)
{

	// update and render the scene graph
	if (_viewer.valid()) {
		// limit FPS to max
		_viewer->frameLimiter();
		_viewer->frame();
	    event.RequestMore();
	}

}

void WxUIMainFrame::OnSetScene(wxCommandEvent &event)
{
	WxUI *ui = (WxUI *)XGisUIFactory::getInstance();

    ObsViewer *viewer = ui->getViewer();
	int w, h;
	((WxUIGraphicsWindow *)ui->gw)->GetClientSize(&w, &h);
    viewer->getCamera()->setViewport(0, 0, w, h);
	viewer->getCamera()->setGraphicsContext(ui->gw);
    viewer->setThreadingModel(osgViewer::Viewer::SingleThreaded);

    ui->setupUI();

    SetViewer(viewer);

    wxIdleEvent ievent;
    ievent.SetEventObject((WxUIGraphicsWindow *)ui->gw);
    ((WxUIGraphicsWindow *)ui->gw)->GetEventHandler()->AddPendingEvent(ievent);
}



// MultiData
/*
void WxMultiData::OnCheckBox(wxCommandEvent& event)
{
    DataChanged(1);
}
*/
void WxMultiData::DataChanged(int id=1)
{
    WxUI *ui = (WxUI *)XGisUIFactory::getInstance();

    if(!ui)
    {
        XGIS_LOG_DEBUG2("Could not bountain World Instance");
    }

    // Area changed
    if(id==0)
    {
        level->Clear();
        if(! area->GetStringSelection().CmpNoCase(wxT("Kokra-cel-1D")))
        {
            level->Append( wxT("0-5") );
            level->Append( wxT("5-10") );
            level->Append( wxT("10-20") );
            level->Append( wxT("20-30") );
            level->Append( wxT("30-40") );
            level->Append( wxT("40-50") );
            level->Append( wxT("50-60") );
            level->Append( wxT("60-70") );
            level->Append( wxT("70-188") );
            level->Append( wxT("188-201") );
            level->Select(0);
        }

        if(! area->GetStringSelection().CmpNoCase(wxT("Mali_graben-cel-1D")))
        {
            level->Append( wxT("0-95") );
            level->Append( wxT("95-100") );
            level->Append( wxT("100-110") );
            level->Append( wxT("110-115") );
            level->Append( wxT("115-120") );
            level->Append( wxT("120-125") );
            level->Append( wxT("150-175") );
            level->Append( wxT("175-200") );
            level->Append( wxT("200-225") );
            level->Select(0);
        }
        if(! area->GetStringSelection().CmpNoCase(wxT("Mali_graben-del1-2D")))
        {
            level->Append( wxT("0-150") );
            level->Append( wxT("150-190") );
            level->Append( wxT("190-250") );
            level->Select(0);
        }
        if(! area->GetStringSelection().CmpNoCase(wxT("Mali_graben-del2-2D")))
        {
            level->Append( wxT("0-150") );
            level->Append( wxT("150-190") );
            level->Append( wxT("190-250") );
            level->Select(0);
        }
        if(! area->GetStringSelection().CmpNoCase(wxT("Mali_graben-del3-2D")))
        {
            level->Append( wxT("0-130") );
            level->Append( wxT("130-150") );
            level->Append( wxT("150-190") );
            level->Append( wxT("190-250") );
            level->Select(0);
        }
        if(! area->GetStringSelection().CmpNoCase(wxT("Mura-cel-1D")))
        {
            level->Append( wxT("0-410") );
            level->Append( wxT("410-500") );
            level->Append( wxT("500-600") );
            level->Append( wxT("600-700") );
            level->Append( wxT("700-800") );
            level->Append( wxT("800-1000") );
            level->Append( wxT("1000-1500") );
            level->Append( wxT("1500-2000") );
            level->Select(0);
        }
        if(! area->GetStringSelection().CmpNoCase(wxT("Reka-cel-1D")))
        {
            level->Append( wxT("0-200") );
            level->Append( wxT("200-250") );
            level->Append( wxT("250-300") );
            level->Select(0);
        }
        if(! area->GetStringSelection().CmpNoCase(wxT("Sava-cel-1D")))
        {
            level->Append( wxT("0-3005") );
            level->Append( wxT("3005-3050") );
            level->Append( wxT("3050-3290") );
            level->Append( wxT("3290-3350") );
            level->Append( wxT("3350-3830") );
            level->Select(0);
        }
        if(! area->GetStringSelection().CmpNoCase(wxT("Vipava-cel-1D")))
        {
            level->Append( wxT("0-296") );
            level->Append( wxT("296-369") );
            level->Append( wxT("369-386") );
            level->Append( wxT("386-422") );
            level->Select(0);
        }
    }

	std::string path = XGisConfig::getInstance()->getFloodPath() + "/";
	wxString filename(path.c_str(), wxConvUTF8);
	filename.append(area->GetStringSelection());
	filename.append(wxT("/"));
	filename.append(level->GetStringSelection());


	XGIS_LOG_DEBUG2("NewFile: %ls\t", filename.c_str());

	if(current )
	{
		ui->getWorld()->removeLocalNode(current);
		current=NULL;
	}


	osgDB::ReaderWriter::Options *opt = new osgDB::ReaderWriter::Options();
	opt->setObjectCacheHint(opt->CACHE_ALL);

    osg::ref_ptr<osg::Node> srf = NULL;
	if( relative->IsChecked() )
	{
        filename.append(wxT("/rel.grd"));

        //printf("Loading relative file %s", )
        osg::ref_ptr<osg::HeightField> hf = osgDB::readHeightFieldFile(std::string(filename.mb_str()));
        XGIS_LOG_DEBUG4("Read GRD file to hfield.");
        if(hf.valid()) {
            srf = osgCreateTransparentSurface(hf.get(), true,
                osg::Vec4(0, 0, 0.75, 0.2 ), 5.0, ui->getWorld());
            XGIS_LOG_DEBUG4("created surface");
        }
	}
	else
	{
        filename.append(wxT("/abs.ive"));
        srf = osgDB::readNodeFile(std::string(filename.mb_str()), opt);
	}

	if( srf.valid() ) {
		XGIS_LOG_DEBUG2("Successfully imported IVE.");
		ui->getWorld()->addLocalNode(srf.get());
		current = srf.get();
        if(id==0)
        {
            ui->getManipulator()->flyTo(srf->getBound().center(), 5.0);
        }
	} else
		XGIS_LOG_DEBUG2("Reading IVE failed.");
}

WxMultiData::WxMultiData(wxWindow *p_parent) : wxBoxSizer(wxVERTICAL)
						 //wxWindow *parent, wxWindowID id) :wxListCtrl(parent, id, wxDefaultPosition, wxDefaultSize, wxLC_LIST, wxDefaultValidator, wxString("local_data_list", wxConvUTF8))
{

	this->parent = p_parent;
	this->current = NULL;

	//wxStaticText* label = new wxStaticText(parent, wxID_ANY, wxString("Multilevel data", wxConvUTF8));
	//this->Add(label, 0, wxEXPAND);

	wxArrayString choices;// = new wxArrayString();
	choices.Add( wxT("") );
	choices.Add( wxT("Kokra-cel-1D") );
	choices.Add( wxT("Mali_graben-cel-1D") );
	choices.Add( wxT("Mali_graben-del1-2D") );
	choices.Add( wxT("Mali_graben-del2-2D") );
	choices.Add( wxT("Mali_graben-del3-2D") );
	choices.Add( wxT("Mura-cel-1D") );
	choices.Add( wxT("Reka-cel-1D") );
	choices.Add( wxT("Sava-cel-1D") );
	choices.Add( wxT("Vipava-cel-1D") );
	wxPoint p;
	wxSize s;

	area = new WxMultiChoice(this, choices);
	area->Select(0);
	area->SetID(0);
	this->Add(area, 1, wxEXPAND);

	choices.Clear();
	choices.Add( wxT("") );

	wxStaticText* label = new wxStaticText(parent, wxID_ANY, wxString("Razpon pretoka [m3/s]", wxConvUTF8));
	this->Add(label, 0, wxEXPAND);

	level = new WxMultiChoice(this, choices);
	level->Select(0);
	level->SetID(1);
	this->Add(level, 1, wxEXPAND);

	relative = new WxCheckRelative(this, wxID_ANY, wxT("Relativne visine") );
    this->Add(relative, 1, wxEXPAND);
}


void WxCheckRelative::OnClick(wxCommandEvent& WXUNUSED(event))
{
	if(parent)
		parent->DataChanged(this->id);
}

BEGIN_EVENT_TABLE(WxCheckRelative, wxCheckBox)
    EVT_CHECKBOX(wxID_ANY, WxCheckRelative::OnClick)
END_EVENT_TABLE()


void WxMultiChoice::OnChoice(wxCommandEvent& WXUNUSED(event))
{
	if(parent)
		parent->DataChanged(this->id);
}

BEGIN_EVENT_TABLE(WxMultiChoice, wxChoice)
    EVT_CHOICE(wxID_ANY, WxMultiChoice::OnChoice)
END_EVENT_TABLE()


// A local data list control

class WxUILocalDataList : public wxListCtrl
{
public:
	WxUILocalDataList(wxWindow *parent, wxWindowID id);

	void OnActivate(wxListEvent &event);
	void OnKeyDown(wxListEvent &event);

private:
    DECLARE_EVENT_TABLE()

	void AddLocalDataItem(const wxString &name, void *data);
};

BEGIN_EVENT_TABLE(WxUILocalDataList, wxListCtrl)
	EVT_LIST_ITEM_ACTIVATED(wxID_ANY, WxUILocalDataList::OnActivate)
	EVT_LIST_KEY_DOWN(wxID_ANY, WxUILocalDataList::OnKeyDown)
END_EVENT_TABLE()

static const char *supportedLocalFileTypes =
	"KML files (*.kml)|*.kml|"
	"SHP files (*.shp)|*.shp|"
	"GRD files (*.grd)|*.grd|"
	"IVE files (*.ive)|*.ive|"
	"other files (*.*)|*.*";

WxUILocalDataList::WxUILocalDataList(wxWindow *parent, wxWindowID id) :
	wxListCtrl(parent, id, wxDefaultPosition, wxDefaultSize, wxLC_LIST,
		wxDefaultValidator, wxString("local_data_list", wxConvUTF8)) {
	wxListItem item;
	item.SetText(wxString("Add new item...", wxConvUTF8));
	item.SetData(0L);
	InsertItem(item);
}

void WxUILocalDataList::AddLocalDataItem(const wxString &name, void *data) {
	wxListItem item;
	item.SetText(name);
	item.SetData((long)data);
	InsertItem(item);
}

void WxUILocalDataList::OnKeyDown(wxListEvent &event) {
	if(event.GetKeyCode() != WXK_DELETE) {
		return;
	}

	long data = event.GetData();
	if(0 != data) {
		WxUI *ui = (WxUI *)XGisUIFactory::getInstance();

		DeleteItem(event.GetIndex());
		ui->getWorld()->removeLocalNode((osg::Node *)data);
	}
}

void WxUILocalDataList::OnActivate(wxListEvent &event) {
	XGIS_LOG_INFO("Clicked (%ld): %ld; mask %ld", event.GetIndex(), event.GetData(), event.GetMask());

	long data = event.GetData();

	WxUI *ui = (WxUI *)XGisUIFactory::getInstance();

	if(0 == data) {
		wxFileDialog *fd = new wxFileDialog(GetParent(), wxString("Choose a local data file...", wxConvUTF8),
			wxString("", wxConvUTF8), wxString("", wxConvUTF8), wxString(supportedLocalFileTypes, wxConvUTF8),
			wxFD_FILE_MUST_EXIST, wxDefaultPosition, wxDefaultSize, wxString("local_data_file_dlg", wxConvUTF8));
		fd->ShowModal();
		wxString spath = fd->GetPath();
		wxString sfname = fd->GetFilename();
		char *path = strdup((char *)spath.char_str());
		char *fname = strdup((char *)sfname.char_str());
		XGIS_LOG_INFO("Got file: %s, %s", path, fname);

		if(strlen(fname) <= 4) {
			// TODO report error
		}
		else if(0 == stricmp(fname + strlen(fname) - 4, ".shp")) {
			XGIS_LOG_INFO("Importing SHP file.");
			osg::ref_ptr<osg::Node> shpNode = osgDB::readNodeFile(path);
			if(shpNode.valid()) {
				XGIS_LOG_INFO("Successfully imported SHP.");
				osg::BoundingSphere s = shpNode->getBound();
				osg::ref_ptr<RelativeAltitudeGroup> grp = new RelativeAltitudeGroup
					(osg::Vec2f(s.center().x(), s.center().y()), 0.0,
						ui->getWorld());
				grp->addChild(shpNode.get());
				ui->getWorld()->addLocalNode(grp.get());
				AddLocalDataItem(sfname, grp.get());
			}
		}
		else if(0 == stricmp(fname + strlen(fname) - 4, ".kml")) {
			XGIS_LOG_INFO("Importing KML file.");
			KMLLoader loader;
			osg::ref_ptr<osg::Node> kmlNode = loader.loadNode(path, ui->getWorld());
			if(kmlNode.valid()) {
				XGIS_LOG_INFO("Successfully imported KML.");
				ui->getWorld()->addLocalNode(kmlNode.get());
				AddLocalDataItem(sfname, kmlNode.get());
			}
		}
		else if(0 == stricmp(fname + strlen(fname) - 4, ".grd")) {
			XGIS_LOG_INFO("Importing GRD file.");
			osg::ref_ptr<osg::HeightField> hf = osgDB::readHeightFieldFile(path);
			XGIS_LOG_DEBUG4("Read GRD file to hfield.");
			if(hf.valid()) {
				bool relative = toupper(osgDB::getSimpleFileName(std::string(path))[0]) == 'R';
				osg::ref_ptr<osg::Node> srf =
					osgCreateTransparentSurface(hf.get(), relative,
					osg::Vec4(0, 0, 0.75, relative ? 0.2 : 0.3), 5.0, ui->getWorld());
				XGIS_LOG_DEBUG4("created surface");
				if(srf.valid()) {
					ui->getWorld()->addLocalNode(srf.get());
					XGIS_LOG_INFO("Successfully imported GRD.");
					AddLocalDataItem(sfname, srf.get());
				}
			}
		} else {
			XGIS_LOG_INFO("Importing other file.");
			osg::ref_ptr<osg::Node> node = osgDB::readNodeFile(path);
			XGIS_LOG_DEBUG4("Read file.");
			if(node.valid()) {
				ui->getWorld()->addLocalNode(node.get());
				XGIS_LOG_INFO("Successfully imported.");
				AddLocalDataItem(sfname, node.get());
			}
		}

		delete path;
		delete fname;
		delete fd;

	}
	else {
		osg::Node *node = (osg::Node *)data;
		osg::Vec3f above(0.0, 0.0, 500.0);
		ui->getManipulator()->flyTo(node->getBound().center()+above, 5.0);
	}
}

// A layer check list control
class WxUILayerList : public wxCheckListBox
{
public:
    WxUILayerList(wxWindow *parent, wxWindowID id);

    void OnCheck(wxCommandEvent &event);

private:
    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(WxUILayerList, wxCheckListBox)
    EVT_CHECKLISTBOX(wxID_ANY, WxUILayerList::OnCheck)
END_EVENT_TABLE()

WxUILayerList::WxUILayerList(wxWindow *parent, wxWindowID id) :
	wxCheckListBox(parent, id, wxDefaultPosition, wxDefaultSize, 0,
		NULL, 0, wxDefaultValidator, wxString("layer_check_list", wxConvUTF8)) {
	XGisConfig *cfg = XGisConfig::getInstance();

	const std::vector<std::string> &layers = cfg->getAvailableLayers();
	int layerCount = layers.size();
	if(layerCount > 0) {
		wxString *layerStrings = new wxString[layers.size()];
		std::vector<std::string>::const_iterator i = layers.begin();
		int n = 0;
		while(i != layers.end()) {
			layerStrings[n++] = wxString((*i).c_str(), wxConvUTF8);
			i++;
		}
		InsertItems(layers.size(), layerStrings, 0);
		const std::vector<std::string> vlayers = cfg->getVisibleLayers();
		if(vlayers.size() > 0) {
			i = vlayers.begin();
			while(i != vlayers.end()) {
				for(n = 0; n < layerCount; n++) {
					if((*i) == layers[n]) {
						Check(n);
						break;
					}
				}
				i++;
			}
		}
	}

}

void WxUILayerList::OnCheck(wxCommandEvent &event) {
	int n = event.GetInt();
	XGisConfig *cfg = XGisConfig::getInstance();
	const std::string &s = cfg->getAvailableLayers()[n];
	if(IsChecked(n)) {
		cfg->addLayer(s);
	}
	else {
		cfg->removeLayer(s);
	}
}

// We need a WX application object: WxUIApp

class WxUIApp : public wxApp {
public:
    bool OnInit();
};

// `Main program' equivalent, creating windows and returning main app frame
bool WxUIApp::OnInit()
{
	WxUI *ui = (WxUI *)XGisUIFactory::getInstance();

#ifdef XGIS_ASYNC_STARTUP
	ui->startup();
#endif /* XGIS_ASYNC_STARTUP */

	// this is the only conversion I could discover...
	wxString title(ui->title.c_str(), wxConvUTF8);;
    // Create the main frame window
	WxUIMainFrame *frame = new WxUIMainFrame(NULL, title,
        wxDefaultPosition, wxDefaultSize);

	wxBoxSizer *hbox = new wxBoxSizer(wxHORIZONTAL);
	frame->SetSizer(hbox);

	wxBoxSizer *vbox = new wxBoxSizer(wxVERTICAL);
	hbox->Add(vbox, 0, wxEXPAND);

	// layers
	wxStaticText *label = new wxStaticText(frame, wxID_ANY, wxString("Layers", wxConvUTF8));
	vbox->Add(label, 0, wxEXPAND);

	// create listbox
	wxCheckListBox *listBox = new WxUILayerList(frame, wxID_ANY);
	listBox->SetInitialSize(wxSize(192, wxDefaultCoord));
	vbox->Add(listBox, 1, wxEXPAND);

	// local data
	label = new wxStaticText(frame, wxID_ANY, wxString("Local data", wxConvUTF8));
	vbox->Add(label, 0, wxEXPAND);

	// create listview
	wxListCtrl *listCtrl = new WxUILocalDataList(frame, wxID_ANY);
	vbox->Add(listCtrl, 1, wxEXPAND);


	// local multidata
	label = new wxStaticText(frame, wxID_ANY, wxString("Odsek vodotoka", wxConvUTF8));
	vbox->Add(label, 0, wxEXPAND);


	/////  MARIANO
	//wxListCtrl* listChoiceCtrl = new WxUILocalMultiDataList(frame, wxID_ANY);
	//vbox->Add(listChoiceCtrl, 1, wxEXPAND);
	// create Choice selector

	WxMultiData* multidataCtrl = new WxMultiData(frame);
	vbox->Add(multidataCtrl, 1, wxEXPAND);
/*
	wxArrayString choices;// = new wxArrayString();
	choices.Add( wxT("") );
	choices.Add( wxT("Vipava") );
	choices.Add( wxT("Mali Graben") );
	wxPoint p;
	wxSize s;
	wxChoice *choiceCtrl = new wxChoice(frame, wxID_ANY, p, s, choices);
	choiceCtrl->Select(0);
	vbox->Add(choiceCtrl, 1, wxEXPAND);

	choices.Clear();
	choices.Add( wxT("0") );
	choices.Add( wxT("1") );
	choices.Add( wxT("2") );
	choices.Add( wxT("3") );
	choices.Add( wxT("4") );
	choices.Add( wxT("5") );
	choices.Add( wxT("6") );
	choices.Add( wxT("7") );
	choices.Add( wxT("8") );
	choices.Add( wxT("9") );
	choices.Add( wxT("10") );

	choiceCtrl = new wxChoice(frame, wxID_ANY, p, s, choices);
	choiceCtrl->Select(0);
	vbox->Add(choiceCtrl, 1, wxEXPAND);
*/
	// hbox->AddStretchSpacer();

	vbox = new wxBoxSizer(wxVERTICAL);
	hbox->Add(vbox, 1, wxEXPAND);

	// create osg canvas
    //    - initialize
	WxUIGraphicsWindow* gw = new WxUIGraphicsWindow(frame, wxID_ANY,
		wxDefaultPosition, wxSize(ui->width, ui->height), wxSUNKEN_BORDER);
	// gw->SetBestSize(wxSize(ui->width, ui->height));
	vbox->Add(gw, 1, wxEXPAND);

	ui->frame = frame;
	ui->gw = gw;

	hbox->Fit(frame);

    // Show the frame
    frame->Show(true);

    ui->postInitialized();

#ifndef XGIS_ASYNC_STARTUP
	ui->startup();
#endif /* XGIS_ASYNC_STARTUP */

    return true;
}

// implement the application but omit main()
IMPLEMENT_APP_NO_MAIN(WxUIApp)


// Finally, our WX XGisUI implementation: WxUI

void WxUI::setupViewer() {
    wxCommandEvent event( wxEVT_SET_SCENE, wxID_ANY );
    event.SetEventObject( frame );
    frame->GetEventHandler()->ProcessEvent( event );
}

void WxUI::run() {
	wxEntry(argc, argv);
}
