#include "config.h"

#include <osgViewer/ViewerEventHandlers>

#include "xgisui.h"
#include "interactivemanipulator.h"
#include "hud.h"
#include "handlers.h"

#if defined(XGIS_UI_GLUT)
#	ifdef WIN32
#		pragma message ("WARNING: GLUT user interface does not fully support all user commands (keyboard, mouse).")
#	else
#		warning "GLUT user interface does not fully support all user commands (keyboard, mouse)."
#	endif
#	include "glutui.h"
#elif defined(XGIS_UI_WX)
#	include "wxui.h"
#elif defined(XGIS_UI_CONSOLE)
#	include "consoleui.h"
#else
#	error "No UI flavour was defined. Currently console, GLUT or wxWindows are supported."
#endif

#include <osg/ShapeDrawable>
#include <osg/Geode>

/* distance of the far plane */
#define FAR_PLANE_DIST 100000
/* interval for periodic detail level updates (in ms) */
#define UPDATE_INTERVAL 2500

XGisUI *XGisUIFactory::instance = NULL;

XGisUIFactory::XGisUIFactory() {}

XGisUI *XGisUIFactory::getInstance() {
	if(NULL == instance) {
#if defined(XGIS_UI_GLUT)
		instance = new GlutUI;
#elif defined(XGIS_UI_WX)
		instance = new WxUI;
#elif defined(XGIS_UI_CONSOLE)
		instance = new ConsoleUI;
#endif
	}
	
	return instance;
}

void XGisUI::setup(int _argc, char **_argv,
	int _width, int _height, int _fps, int _startFrame,
	const std::string &_title, const std::string &_imageDir,
	OpenThreads::Thread *_startupThread)
{
	argc = _argc;
	argv = _argv;
	width = _width;
	height = _height;
	fps = _fps;
	startFrame = _startFrame;
	title = _title;
	imageDir = _imageDir;
	startupThread = _startupThread;

	root = NULL;
	world = NULL;
	viewer = NULL;
}

void XGisUI::startup() {
#ifdef XGIS_ASYNC_STARTUP
	startupThread->start();
#else
	startupThread->run();
#endif /* XGIS_ASYNC_STARTUP */
}
void XGisUI::setScene(osg::Group *_root, World *_world, HUD *_hud) {	
	uiInitSemaphore.wait();

	root = _root;
	world = _world;
	hud = _hud;
	
	setupViewer();
}

ObsViewer *XGisUI::getViewer() {
	if(NULL != viewer.get()) {
		return viewer.get();
	}
	
	// create viewer
	viewer = new ObsViewer(XGisConfig::getInstance()->getFarPlane());
	osg::DisplaySettings * ds = osg::DisplaySettings::instance();
	ds->setNumMultiSamples(16); 
	viewer->setDisplaySettings(ds);
	viewer->setMaxFPS(fps);
	
	viewer->resizeViewer(0, 0,width, height);

	return viewer.get();
}

void XGisUI::setupUI() {
	getViewer();

	// add scene to the viewer
	viewer->setSceneData(world.get()); 

	// add enviromnet, HUD & respective cameras
	viewer->setEnvironment(root.get());
	viewer->addHUD(hud.get());
	viewer->setupSlaveCameras(); //must be called after setSceneData!

#if 0
	osg::Sphere *sphere = new osg::Sphere(world->getDetailRefPoint() + osg::Vec3(1000, 0, 300), 100);		
	osg::ShapeDrawable *drawable = new osg::ShapeDrawable(sphere);
	drawable->setColor(osg::Vec4(1.0f, .25f, .25f, 1.0f));
	osg::Geode *model = new osg::Geode;
	model->addDrawable(drawable);
	osg::Group *grp = new osg::Group;
	grp->addChild(model);
	viewer->setEnvironment(grp);
#endif /* 0/1 */

	// create camera manipulator and add it to the viewer
	manipulator =
		new InteractiveManipulator(world.get(), viewer.get(), hud.get(),
			world->getDefaultCameraPos());
	viewer->setCameraManipulator(manipulator.get());
	viewer->getCameraManipulator()->setByMatrix(world->getDefaultCameraPos());
	
	GpsTrack *track = world->getConfiguration()->getStartTrack();
	if(NULL != track) {
		manipulator->flyGpsTrack(*track);
	}

	// add a handler for displaying statistics
	osgViewer::StatsHandler *sh = new osgViewer::StatsHandler();
	viewer->addEventHandler(sh);
	sh->setKeyEventTogglesOnScreenStats('X');
	sh->setKeyEventPrintsOutStats(0);
	//note: 1. StatsHandler treats Shift incorrectly (thus we use upper-case key)
	//		2. keyEventPrintsOutStats can't be the same (thus we don't use it)

	// create a callback for performing updates to the world
	WorldUpdateCallback *wu = world->getUpdateCallback();
	wu->addUpdateHandler(new WorldLoader(world.get(), manipulator.get()));
	wu->addUpdateHandler(new HUDUpdater(hud.get(), manipulator.get()));
#ifndef XGIS_UI_CONSOLE
	// only do periodic updating in interactive viewers
	wu->addUpdateHandler(new PeriodicUpdateHandler(UPDATE_INTERVAL));
#endif /* !XGIS_UI_CONSOLE */
	
	// initialize HUD camera display
	hud->updateCompass(0);
	hud->updatePosition(0, manipulator->getEye());
	hud->updateTasks(0);
	hud->setStatus("This is a status message that will fade away in 10 seconds.", 10);
	
}
