#include "config.h"

#include <math.h>
#include <stdio.h>
#include <iostream>
#include <fcntl.h>

#include <osg/Vec3f>
#include <xgislib/xputils.h>
#include <xgislib/osgutils.h>
#include <xgislib/coords.h>
#include <xgislib/readerwritergrd6.h>
#include <osgDB/ReadFile>
#include <osgDB/WriteFile>

osgDB::RegisterReaderWriterProxy<ReaderWriterGrd6> g_readerWriter_Grd6_Proxy;

bool readVec3(FILE *f, osg::Vec3f &v) {
	float x, y, z;
	int floatsRead = fscanf(f, "%f %f %f", &x, &y, &z);
	if (floatsRead != 3) {
		v.x() = NAN;
		v.y() = NAN;
		v.z() = NAN;
		return false;
	} else {
		v.x() = x;
		v.y() = y;
		v.z() = z;
		return true;
	}
}

FILE *myfopen(const char* fname, const char *mode) {
	FILE *rv = NULL;
#	ifndef WIN32
		// necessary to facilitate opening files >2GB
		int fd = open(fname, O_LARGEFILE|O_RDONLY);
		if(-1 != fd)
			rv = fdopen(fd, mode);
#	else
		rv = fopen(fname, mode);
#	endif // !WIN32 
	if (rv == NULL) {
		std::cerr << "Error opening file " << fname << std::endl;
		exit(-4);
	}
	return rv;
}

int fround(float x) {
	return (int)floor(x+0.5);
}

//converts a short (encoded the litte-endian way) to float
float toFloat(short s, bool bigEndianMachine) {
	if (bigEndianMachine)
		s = (s >> 8) | (s << 8);
	return (float)s;
}

#define EPS (1e-5f)

class BagOfPoints {
private:
	float res;
	float minx, maxx, miny, maxy;

	float *data;
	int sx, sy;

	float &val(int i, int j) {
		assert(0 <= i && i < sx);
		assert(0 <= j && j < sy);
		return data[i + sx*(sy-j-1)];
	}
	void setAt(float x, float y, float z) {
		assert(minx - EPS <= x && x <= maxx + EPS);
		assert(miny - EPS <= y && y <= maxy + EPS);
		float xx = (x-minx)/res;
		float yy = (y-miny)/res;
		assert( fabs(xx - fround(xx)) < EPS );
		assert( fabs(yy - fround(yy)) < EPS );
		val(fround(xx), fround(yy)) = z;
	}

public:
	float getAt(float x, float y) {
		assert(minx - EPS <= x && x <= maxx + EPS);
		assert(miny - EPS <= y && y <= maxy + EPS);
		float ll, lr, ul, ur;

		float i = (x-minx)/res, j = (y-miny)/res;
		int ifl = (int)floor(i), icl = (int)ceil(i);
		int jfl = (int)floor(j), jcl = (int)ceil(j);
		int is = -1;
		int js = -1;
		if (ifl < 0)
			is = icl;
		else if (icl >= sx)
			is = ifl;
		if (jfl < 0)
			js = jcl;
		else if (jcl >= sy)
			js = jfl;
		
		if (is != -1 && js != -1)
			return val(is, js);
		else if (is != -1) {
			ll = val(is, jfl);
			ul = val(is, jcl);
			return ll + (ul-ll)*(j-jfl);
		} else if (js != -1) {
			ll = val(ifl, js);
			lr = val(icl, js);
			return ll + (lr-ll)*(i-ifl);
		} else {
			ll = val(ifl, jfl);
			lr = val(icl, jfl);
			float lower = ll + (lr-ll)*(i-ifl);
			ul = val(ifl, jcl);
			ur = val(icl, jcl);
			float upper = ul + (ur-ul)*(i-ifl);
			return lower + (upper-lower)*(j-jfl);
		}
	}

	BagOfPoints(float _res, char *bboxfile, char *file) {
		//read bbox
		FILE *bbf = fopen(bboxfile, "r");
		minx = miny = maxx = maxy = NAN;
		int numread = fscanf(bbf, "%f %f %f %f\n", &minx, &miny, &maxx, &maxy);
		fclose(bbf);
		if (numread != 4) {
			std::cerr << "Invalid bbox file.\n";
			exit(-2);
		}

		res = _res;
		if (fabs((maxx-minx)/res - fround((maxx-minx)/res)) > EPS) {
			std::cerr << "Bounding box size not divisible by resolution.\n";
			exit(-3);
		}
		sx = fround((maxx-minx)/res) + 1;
		sy = fround((maxy-miny)/res) + 1;
		data = new float[sx*sy];

		for (int x = 0; x < sx; x++)
			for (int y = 0; y < sy; y++)
				val(x,y) = NAN;

		FILE *fRead = myfopen(file, "rb");
		osg::Vec3f v;
		bool readOK;
		int numRead = 0;

		while (true) {
			readOK = readVec3(fRead, v);
			if (!readOK)
				break;
			if (v.x() < minx || v.y() < miny || v.x() > maxx || v.y() > maxy) {
				std::cerr << "Point " << v.x() << "," << v.y() << " outside of BBOX!\n";
				exit(-4);
			}
			setAt(v.x(), v.y(), v.z());
			numRead ++;
		} 
		fclose(fRead);
		std::cerr << numRead << " points read.\n";
		
		for (float x = minx; x < maxx; x += (maxx-minx)/10)
			for (float y = miny; y < maxy; y += (maxy-miny)/10)
				printf("%f, %f : %f\n", x, y, getAt(x, y));
	}
	BagOfPoints(char *file) {
		unsigned short magic = 258;
		char *pm = (char*)(&magic);
		bool bigEndian = false;
		if (*pm != 2 || *(pm+1) != 1) {
			if (*pm == 1 && *(pm+1) == 2) {
				std::cerr << "WARNING: this is a big-endian machine\n and this program isn't tested on those.\n";
				bigEndian = true;
			} else  {
				std::cerr << "FATAL: this machine appears to be neither little nor big-endian!\n";
				exit(-6);
			}
		}

		int fsize = xpGetFileSize(file);
		float sqrtfsize = sqrt((float)fsize/(float)sizeof(short));
		sx = sy = fround(sqrtfsize);
		res = 1;

		if (fsize % sizeof(short) != 0 || fabs(sx - sqrtfsize) > EPS) {
			std::cerr << "File size invalid (must contain a square number of shorts)\n";
			exit(-5);
		}
		FILE *f = fopen(file, "rb");
		short *buf = new short[sx*sy];
		size_t numRead = fread(buf, sizeof(short), sx*sy, f);
		fclose(f);
		if (numRead != sx*sy) {
			std::cerr << "Error reading source file.\n";
			exit(-8);
		}

		data = new float[sx*sy];
		for (int x = 0; x < sx; x++)
			for (int y = 0; y < sy; y++)
				val(x,sy-y-1) = toFloat(buf[x + sx*y], bigEndian);
	}

	void dumpRaw(const char *file) {
		FILE *f = myfopen(file, "wb");
		fwrite(data, sizeof(float), sx*sy, f);
		fclose(f);
	}
};

bool writeMtdAndHdr(const osg::HeightField &hf, char *basename) {
	std::string hdrFormat =	"ENVI\n"
							"samples = %d\n"
							"lines = %d\n"
							"bands = 1\n"
							"header offset = 0\n"
							"file type = ENVI Standard\n"
							"data type = 4\n"
							"interleave = bsq\n"
							"byte order = 0\n";
	std::string mtdFormat = "Geotiff_Information:\n"
							"	Version: 1\n"
							"	Key_Revision: 1.0\n"
							"	Tagged_Information:\n"
							"		ModelTiepointTag (2,3):\n"
							"			0      0     0\n"
							"			%.3lf %.3lf 0\n"
							"		ModelPixelScaleTag (1,3):\n"
							"			%.3lf %.3lf 0\n"
							"		End_Of_Tags.\n"
							"	Keyed_Information:\n"
							"	End_Of_Keys.\n"
							"End_Of_Geotiff.\n";
		
	char *fname = new char[strlen(basename) + 10];

	//write hdr
	sprintf(fname, "%s.hdr", basename);
	FILE *f = fopen(fname, "w");
	if (NULL == f) {
		std::cerr << "Error writing to file " << fname << std::endl;
		return false;
	}
	fprintf(f, hdrFormat.c_str(), hf.getNumColumns(), hf.getNumRows());
	fclose(f);

	//write mtd (geo-info)
	sprintf(fname, "%s.mtd", basename);
	f = fopen(fname, "w");
	if (NULL == f) {
		std::cerr << "Error writing to file " << fname << std::endl;
		return false;
	}
	fprintf(f, mtdFormat.c_str(), hf.getOrigin().x(), hf.getOrigin().y() + hf.getYInterval()*hf.getNumRows(),
		hf.getXInterval(), - hf.getYInterval());
	fclose(f);

	return true;
}

int main(int argc, char **argv) {
	if (argc < 2) {
		std::cerr << "Missing command.\n";
		return -1;
	}

	if (strcmp(argv[1], "wgs2gkr") == 0) {
		//wgs2gkr <lat> <lon> - prints GaussKr 

		Coords::WGS84 wgs(atof(argv[3]), atof(argv[2]));
		Coords::GaussKr gk = wgs.toGaussKr();
		printf("%.2lf %.2lf\n", gk.x, gk.y);

	} else if (strcmp(argv[1], "gkr2wgs") == 0) {
		//gkr2wgs <x> <y> - prints WGS84 as <lat>,<lon>

		Coords::WGS84 wgs(Coords::GaussKr(atof(argv[2]), atof(argv[3])));
		printf("%.8lf %.8lf\n", wgs.latit, wgs.longit);

	} else if (strcmp(argv[1], "findbbox") == 0) {
		//findbbox <file> - prints bounding box of given .xyz file

		FILE *fRead = myfopen(argv[2], "rb");
		osg::Vec3f v;
		bool readOK;

		float minx = 1e+30f;
		float miny = 1e+30f;
		float maxx = -minx;
		float maxy = -miny;

		while (true) {
			readOK = readVec3(fRead, v);
			if (!readOK)
				break;
			if (v.x() < minx)
				minx = v.x();
			if (v.y() < miny)
				miny = v.y();
			if (v.x() > maxx)
				maxx = v.x();
			if (v.y() > maxy)
				maxy = v.y();
		} 

		fclose(fRead);
		printf("%f %f %f %f\n", minx, miny, maxx, maxy);
	
	} else if (strcmp(argv[1], "point2raw") == 0) {
		//point2raw <infile> <bbox-file> <resol [m]> <outfile> - reads all points and writes them as raw float table

		float res = (float)atof(argv[4]);
		assert(res > 0);
		BagOfPoints points(res, argv[3], argv[2]);
		points.dumpRaw(argv[5]);

	} else if (strcmp(argv[1], "getelev") == 0) {
		//getelev <infile> <bbox-file> <resol [m]> <lon> <lat> - returns elevation at given WGS84 coordinates

		float res = (float)atof(argv[4]);
		assert(res > 0);
		Coords::WGS84 wgs(atof(argv[5]), atof(argv[6]));
		Coords::GaussKr gk = wgs.toGaussKr();
		printf("gk: %.1lf, %.1lf\n", gk.x, gk.y);
		BagOfPoints points(res, argv[3], argv[2]);
		printf("elev: %.3lf\n", points.getAt(gk.x, gk.y));

	} else if (strcmp(argv[1], "bil2float") == 0) {
		//bil2float <infile> <outfile> - reads a WWJ bil file and converts it to bil32

		BagOfPoints points(argv[2]);
		points.dumpRaw(argv[3]);

	} else if (strcmp(argv[1], "grd2bil") == 0) {
		//grd2bil <infile> <outfilename-without-extension> - converts a GRD6 file to GDAL-readable bil+hdr+mhd

#		define MAX_ALLOWED_HEIGHT 100000.0 //heights above that are replaced by NaN

		osg::ref_ptr<osg::HeightField> hf = osgDB::readHeightFieldFile(argv[2]);
		if (!hf.valid()) {
			std::cerr << "Error reading file " << argv[2] << std::endl;
			return -4;
		}
		Coords::GaussKr gk(hf->getOrigin().x(), hf->getOrigin().y());
		Coords::WGS84 wgs(gk);
		printf("Origin: %lf, %lf\n", wgs.latit, wgs.longit);

		if (!writeMtdAndHdr(*hf, argv[3]))
			return -4;

		//write raw
		char *fname = new char[strlen(argv[3]) + 10];
		sprintf(fname, "%s.bil", argv[3]);
		FILE *f = fopen(fname, "wb");
		if (NULL == f) {
			std::cerr << "Error writing to file " << fname << std::endl;
			return -4;
		}
		for (int y = hf->getNumRows()-1; y >= 0; y--)
			for (int x = 0; x < (int)hf->getNumColumns(); x++) {
				float v = hf->getHeight(x, y);
				if (fabs(v) > MAX_ALLOWED_HEIGHT)
					v = (float)NAN;
				fwrite(&v, sizeof(float), 1, f);
			}
		fclose(f);

	} else if (strcmp(argv[1], "grd2tiff") == 0) {
		//grd2tiff <infile> <outfilename-without-extension> - converts a GRD6 file to GDAL-readable tiff+mtd

#		define MAX_ALLOWED_HEIGHT 100000.0 //heights above that are replaced by NaN

		osg::ref_ptr<osg::HeightField> hf = osgDB::readHeightFieldFile(argv[2]);
		if (!hf.valid()) {
			std::cerr << "Error reading file " << argv[2] << std::endl;
			return -4;
		}
		Coords::GaussKr gk(hf->getOrigin().x(), hf->getOrigin().y());
		Coords::WGS84 wgs(gk);
		printf("Origin: %lf, %lf\n", wgs.latit, wgs.longit);

		if (!writeMtdAndHdr(*hf, argv[3]))
			return -4;

		//write tiff
		char *fname = new char[strlen(argv[3]) + 10];
		sprintf(fname, "%s.tiff", argv[3]);
		osg::ref_ptr<osg::Image> img = new osg::Image;
		img->allocateImage(hf->getNumColumns(), hf->getNumRows(), 1, GL_RGBA, GL_UNSIGNED_BYTE);
		unsigned char val[4] = {0,0,255,0};
		for (int y = hf->getNumRows()-1; y >= 0; y--)
			for (int x = 0; x < (int)hf->getNumColumns(); x++) {
				float v = hf->getHeight(x, y);
				if (fabs(v) > MAX_ALLOWED_HEIGHT)
					v = (float)NAN;
				val[3] = isnan(v) ? 0 : (unsigned char)osg::clampBetween(osg::round(v*30.0+80), 90.0, 210.0);
				memcpy(img->data(x, y), val, 4);
			}
		osgDB::writeImageFile(*img, fname);

	} else if (strcmp(argv[1], "bil2grd") == 0) {
		//bil2grd <inrawfile> <ingeoinfofile> <outfile> - converts RAW+geoinfo file to GRD6

		//read geo info
		int w, h;
		double lat, lon, latDelta, lonDelta;
		FILE *f = fopen(argv[3], "r");
		if (NULL == f) {
			std::cerr << "Error reading file " << argv[3] << std::endl;
			return -4;
		}
		int numParsed = fscanf(f, "Size is %d, %d\nOrigin = (%lf,%lf)\nPixel Size = (%lf,%lf)",
			&w, &h, &lon, &lat, &lonDelta, &latDelta);
		fclose(f);
		if (numParsed != 6) {
			std::cerr << "Invalid geo info file " << argv[3] << std::endl;
			return -45;
		}

		//create heightField
		osg::ref_ptr<osg::HeightField> hf = new osg::HeightField();
		hf->allocate(w, h);
		hf->setOrigin(osg::Vec3(lon, lat+h*latDelta, 0));
		hf->setXInterval(lonDelta);
		hf->setYInterval(-latDelta);

		//read raw data
		f = fopen(argv[2], "rb");
		if (NULL == f) {
			std::cerr << "Error opening file " << argv[2] << std::endl;
			return -4;
		}
		for (int y = hf->getNumRows()-1; y >= 0; y--)
			for (int x = 0; x < (int)hf->getNumColumns(); x++) {
				float v;
				size_t numRead = fread(&v, sizeof(float), 1, f);
				if (numRead != 1) {
					std::cerr << "Error reading file " << argv[2] << std::endl;
					return -4;
				}
				hf->setHeight(x, y, v);
			}
		fclose(f);

		//write
		if (!osgDB::writeHeightFieldFile(*hf, argv[4])) {
			std::cerr << "Error writing to file " << argv[4] << std::endl;
			return -4;
		}

	} else if (strcmp(argv[1], "grd2ive") == 0) {
		//grd2ive <ingrdfile> <outivefile> [alpha (default=0.3)] - compiles GRD flood surface with absolute heights

		osg::ref_ptr<osg::HeightField> hf = osgDB::readHeightFieldFile(argv[2]);
		if(!hf.valid()) {
			std::cerr << "Error reading file " << argv[2] << '\n';
			return -89;
		}

		float transp = osg::clampBetween(argc >= 5 ? atof(argv[4]) : 0.3, 0.0, 1.0);
		osg::ref_ptr<osg::Node> srf = osgCreateTransparentSurface(hf.get(), false,
				osg::Vec4(0, 0, 0.75, transp), NAN, NULL);
		osgDB::writeNodeFile(*(srf.get()), argv[3]);

	} else {
		std::cerr << "Invalid command.\n";
		return -2;
	}
}
